// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/* 
 * File:   data_set.h
 * Author: gleseach
 *
 * Created on 26 août 2015, 15:17
 */

#ifndef DATA_SET_H
#define	DATA_SET_H

#include "core/structs.h"
#include "3d/pickable.h"

namespace OGL
{
    class StaticMesh;
}

class Image;
class BBox;
class StreamlinesData;
class ParticulesData;

class DataSet final : public Pickable
{
    int _width;
    int _height;
	float _resolution;

    unsigned int _dimensionCount;
    Image* _data;
    Image* _dataMask;
    float * _offset;
    float * _scaleFactor;
    float _scaleFactorGL;        // 1.0 -> 255.0 for unsigned byte raw data

    float * lons;
    float * lats;
    unsigned int lat_size;
    unsigned int lon_size;
    float * min;
    float * max;

    glm::vec4 corners[5];
    glm::vec3 normal;
    glm::vec4 centerPlane;

    unsigned short * gcpx;
    unsigned short * gcpy;

    OGL::StaticMesh * mesh_spherical;
    OGL::StaticMesh * mesh_planar;

    std::vector<BBox*> bboxArray;
    glm::vec4 bbox; // simple bbox in degrees, does not handle poles.
	bool _xInvertedBbox;
    bool _yInvertedBbox;

    StreamlinesData* streamlinesData;
    ParticulesData* particulesData;

    std::vector<glm::vec3> _gcps;

    bool _lonLatAligned;
    glm::bvec2 _seamLess;
    std::vector<glm::ivec4> _areas;

public:
    DataSet(const SEAScope::RenderableId& rId);
    ~DataSet() override;

    // Prevent copies
    DataSet(const DataSet&) = delete;
    DataSet& operator=(const DataSet&) = delete;

    glm::vec4 getBbox(unsigned int id) const;

    // Allocate buffers and send data to the gpu
    void generateGPUData(OGL::Context* ctx);

    //! Return the offset and scale factor to uncompress the physical data.
    glm::vec4 uncompressVector() const;
    glm::vec3 uncompressVector3() const;

    unsigned int uncompressVectorSize() const;
    float* offsetArray() const;
    float* scaleFactorArray() const;

    void draw(OGL::Context* ctx, const glm::mat4& mvp) override;
    void drawPicking(OGL::Context* ctx, const glm::mat4& mvp) override;

    // Accessors
    void setStreamlinesData(StreamlinesData* streamlinesData);
    StreamlinesData* getStreamlinesData() const;
    void setBbox(glm::vec4 bbox, bool xInverted, bool yInverted);
    glm::vec4 getBbox() const;
    void addToBboxArray(BBox* bbox);
    void addToBboxArray(const glm::vec4&);
    void setMesh_planar(OGL::StaticMesh* mesh_planar);
    OGL::StaticMesh* getMesh_planar() const;
    void setMesh_spherical(OGL::StaticMesh* mesh_spherical);
    OGL::StaticMesh* getMesh_spherical() const;
    void setGcpy(unsigned short* gcpy);
    unsigned short* getGcpy() const;
    void setGcpx(unsigned short* gcpx);
    unsigned short* getGcpx() const;
    void setCenterPlane(glm::vec4 centerPlane);
    glm::vec4 getCenterPlane() const;
    void setNormal(glm::vec3 normal);
    glm::vec3 getNormal() const;
    void setCorner(int i, const glm::vec4& cornersIn);
    const glm::vec4& getCorner(int i) const;
    void setMax(float* max);
    float* getMax() const;
    void setMin(float* min);
    float* getMin() const;
    void setLon_size(unsigned int lon_size);
    const unsigned int& getLon_size() const;
    void setLat_size(unsigned int lat_size);
    const unsigned int& getLat_size() const;
    void setLats(float* lats);
    float* getLats() const;
    void setLons(float* lons);
    float* getLons() const;
    void setScaleFactorGL(float _scaleFactorGL);
    float getScaleFactorGL() const;
    void setScaleFactor(float* _scaleFactor);
    float* getScaleFactor() const;
    void setOffset(float* _offset);
    float* getOffset() const;
    void setDataMask(unsigned int width, unsigned int height, unsigned int componentCount, OGL::DataType dataType, void * data);
    Image* getDataMask() const;
    void setData(unsigned int width, unsigned int height, unsigned int componentCount, OGL::DataType dataType, void * data);
    Image* getData() const;
    void setDimensionCount(unsigned int _dimensionCount);
    const unsigned int& getDimensionCount() const;
    void setResolution(float _resolution);
    float getResolution() const;
    void setHeight(int _height);
    const int& getHeight() const;
    void setWidth(int _width);
    const int& getWidth() const;
    ParticulesData* getParticulesData() const;
    void setParticulesData(ParticulesData* pData);
    void setFilterMode(OGL::FilterMode filterMode);
    const std::vector<glm::vec3>& getGcps() const;
    void setSeamLess(bool x, bool y);
    const glm::bvec2& getSeamLess() const;
    void setNEWSAligned(bool status);
    bool isNEWSAligned() const;
    void setVisibleAreas(std::vector<glm::ivec4>&& areas);
    const std::vector<glm::ivec4>& getVisibleArea() const;
    bool isBboxYInverted() const;
	bool isBboxXInverted() const;
};

#endif	/* DATA_SET_H */

