// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
/*
 * File:   streamlines_data.h
 * Author: gleseach
 *
 * Created on 9 janvier 2017, 12:32
 */

#ifndef STREAMLINES_DATA_H
#define STREAMLINES_DATA_H

#include "3d/chrono.h"
#include "data/image.h"

namespace OGL
{
    class VertexArrayObject;
    class FrameBuffer;
    class Buffer;
    class Context;
    class Texture;
}

class DataSet;

class StreamlinesData
{
public:
    StreamlinesData();
    virtual ~StreamlinesData();

    void create(OGL::Context* ctx,
        unsigned int streamlineCount,
        unsigned int streamlineLength,
        float speed,
        DataSet* _dataSet);

    OGL::Texture* getGcpTexture() const { return _gcp; }
    OGL::Texture* getGcpNorthEastTexture() const { return _gcpNorthEast; }
    glm::vec2  getGcpUVMin() { return _uvMin; }
    glm::vec2  getGcpUVRange() { return _uvRange; }

    unsigned int getIndiceCount() const { return _indiceCount; }
    void setIndiceCount(unsigned int val) { _indiceCount = val; }
    unsigned int getStreamlineLength() const { return _streamlineLength; }
    void setStreamlineLength(unsigned int val);
    OGL::VertexArrayObject* getTargetVao() const { return this->_vaoRendering[_targetId]; }
    unsigned int getParticleCount() const { return _particleCount; }
    void setParticleCount(unsigned int val) { _particleCount = val; }
    unsigned int getCurrentStreamline() const { return _currentStreamline; }
    void setCurrentStreamline(unsigned int val) { _currentStreamline = val; }
    float getArea() const { return _area; }
    void setArea(float val) { _area = val; }
    unsigned int getStreamlineCount() const { return _streamlineCount; }
    void setStreamlineCount(unsigned int val);
    float getSpeed() const { return _speed; }
    void setSpeed(float val) { _speed = val; }

    OGL::Buffer* getVboSrc() const { return _vbo[_sourceId]; }
    OGL::Buffer* getVboDst() const { return _vbo[_targetId]; }

    OGL::VertexArrayObject* getVaoPhysicSrc() const { return _vaoPhysic[_sourceId]; }
    OGL::VertexArrayObject* getVaoPhysicDst() const { return _vaoPhysic[_targetId]; }

    // Chrono
    Chrono& getLastTime() { return _lastTime; }
    Chrono& getLastUpdateTime() { return _lastUpdateTime; }

    void swap();
    void update(OGL::Context* ctx);
    void setSaveConfig(bool);
    bool getSaveConfig() const;

private:
    void clearBuffers();
    void initBuffers(OGL::Context* ctx);

private:
    unsigned int _currentStreamline; //< Streamline id to update
    unsigned int _particleCount; //< Number of particles
    unsigned int _streamlineCount;
    unsigned int _streamlineCountTmp;
    bool _update;
    unsigned int _streamlineLength;
    unsigned int _streamlineLengthTmp;
    //std::uint8_t _particleTTL;
    float _streamlinePerSec;
    float _speed;

    unsigned int _indiceCount;
    OGL::Buffer* _vertices_local_id_buffer;
    OGL::Buffer* _indices_buffer;
    OGL::Buffer* _vbo[2];
    OGL::VertexArrayObject* _vaoPhysic[2];
    OGL::VertexArrayObject* _vaoRendering[2];
    unsigned int _sourceId;
    unsigned int _targetId;
    Chrono _lastTime;
    Chrono _lastUpdateTime;
    Chrono _origin;

    OGL::Texture* _gcp;
    OGL::Texture* _gcpNorthEast;

    float _area;
    //float _minX;
    glm::vec2 _uvMin;
    glm::vec2 _uvRange;
    bool _saveConfig;
};

#endif /* STREAMLINES_DATA_H */

