// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef ENGINE_H
#define ENGINE_H

#include <map>
#include <set>
#include <chrono>
#include <vector>
#include <string>
#include <unordered_map>

#include "core/types.h"
#include "core/structs.h"
#include "core/interfaces/imetadata_store.h"
#include "core/interfaces/idata_store.h"
#include "core/interfaces/iplugins_loader.h"
#include "core/plugins_manager.h"
#include "core/metadata_mediator.h"
#include "core/data_mediator.h"
#include "core/rendering_control.h"
#include "core/rendering_cfg.h"
#include "core/sources_control.h"
#include "core/collections_control.h"
#include "core/timeline.h"
#include "core/colormap_manager.h"
#include "core/sockets_manager.h"
#include "core/settings.h"
#include "core/state.h"
#include "data/events_coverage.h"

#include "ui/iadvanced_rendering_control_view.h"
#include "ui/icollections_control_view.h"
#include "ui/irendering_control_view.h"
#include "ui/isources_control_view.h"
#include "ui/itimeline_view.h"

#include "core/display_queue.h"
#include "data/gpu_data_builder.h"

#include "core/tools.h"
#include "core/interfaces/iapp_events_subscriber.h"
#include "local/user_generated.h"

namespace SEAScope
{
    // Forward declarations
    class AppEvent;
    class AppEventsDispatcher;

	class Engine: public virtual IAppEventsSubscriber
	{
		public:
			Engine(AppEventsDispatcher&);
			virtual ~Engine() override;

            bool handleAppEvent(std::type_index, const AppEvent&) override;

			bool init(SceneView*, IMetadataStore*, IDataStore*, IPluginsLoader*, IBlueprintFactory*);
			bool initColormaps();

			bool setCollectionsUI(ICollectionsControlView*);
			bool setRenderingUI(IRenderingControlView*);
			bool setAdvancedRenderingUI(IAdvancedRenderingControlView*);
			bool setTimelineUI(ITimelineView*);
			bool setSourcesUI(ISourcesControlView*);

			bool shutdown();

			bool setRenderables(const std::map<CollectionId, std::vector<GranuleMetadata> >&);

			void updateRendering(const RenderingCfg*, const RenderingParam&);
			bool updateRenderSet(const CollectionId&, const std::vector<GranuleMetadata>&, std::set<RenderableId>&);
			void renderablesChanged();
			bool loadVariableData(const RenderableId&, const DisplayBatchId&);
			bool loadOneGranule( const RenderableId&
							   , const VariableCfg&
							   , const GranuleMetadata&
							   , float, int, int, const DisplayBatchId&);
			bool loadGranuleData( const IDFDescriptor&
								, const std::set<std::string>&
								, const GranuleId&
								, const DataModel&
								, const RenderingCfg*
								, const RenderableId&
                                , const DisplayBatchId&);
			bool getRenderableLabel(const RenderableId&, std::string&);
			bool getRenderableLabel(const RenderableId&, std::string&, const std::string&);
			bool getRenderableUnits(const RenderableId&, std::string&);
			bool selectResolution(const GranuleMetadata&, float, std::int32_t, std::int32_t, std::string&, float&, unsigned int&) const;

			bool clearExtractedData();
			bool initExtractedData();
            const std::map<RenderableId, Extraction>& getExtractedData() const;
            const std::map<RenderableId, std::string>& getExtractedLabels() const;

            bool updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel);
            void update();

			bool getRenderableInfo(const RenderableId&, GranuleInfo&);

            bool sendProcessingOpcode(std::uint64_t);

            Tools* getTools();
            UserGenerated* getUserGenerated();

            AppEventsDispatcher& _appEventsDispatcher;
			std::map<CollectionId, std::vector<GranuleMetadata> > renderables;
			MetadataMediator metadataMediator;
			DataMediator dataMediator;
			Timeline timeline;
			RenderingControl renderingControl;
			SourcesControl sourcesControl;
			CollectionsControl collectionsControl;
			PluginsManager plugins;
			ColormapManager cmaps;
			Settings settings;
			State state;
			SocketsManager sockets;
			DisplayQueue displayQueue;
			GPUDataBuilder gpuData;
			std::map<RenderableId, std::string> uris;
            int maxTextureSize;
            bool updateRequested;
            std::chrono::high_resolution_clock::time_point lastUpdate;

        protected:
            Tools _tools;
            UserGenerated _userGenerated;

            std::map<RenderableId, Extraction> _extractedData;
            std::map<RenderableId, std::string> _extractedLabels;
	private:
		void saveConfig(const std::string& path, SEAScope::RenderingCfg* cfg, const SEAScope::CollectionCfg& cCfg);
	};
};

#endif
