// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef RENDERING_CFG_H
#define RENDERING_CFG_H

#include <cstdint>
#include <functional>
#include "core/types.h"
#include "core/structs.h"

#undef min
#undef max

namespace SEAScope
{
	class RenderingCfg
	{
		public:
			RenderingCfg();
            RenderingCfg(const RenderingCfg& src)
            : _rendered(src._rendered), _logScale(src._logScale), _min(src._min),
              _max(src._max), _opacity(src._opacity), _zindex(src._zindex),
              _color({src._color[0], src._color[1], src._color[2]}),
              _colormap(src._colormap), _renderMethod(src._renderMethod),
              _particlesCount(src._particlesCount), _particleTTL(src._particleTTL),
              _streamlineLength(src._streamlineLength), _streamlineSpeed(src._streamlineSpeed),
              _filterMode(src._filterMode), target(src.target), callback(src.callback)
            {
            }

            RenderingCfg& operator=(const RenderingCfg& src)
            {
			    this->_rendered = src._rendered;
			    this->_logScale = src._logScale;
			    this->_min = src._min;
			    this->_max = src._max;
			    this->_opacity = src._opacity;
			    this->_zindex = src._zindex;
			    this->_color = src._color;
			    this->_colormap = src._colormap;
			    this->_renderMethod = src._renderMethod;
			    this->_particlesCount = src._particlesCount;
			    this->_particleTTL = src._particleTTL;
			    this->_streamlineLength = src._streamlineLength;
			    this->_streamlineSpeed = src._streamlineSpeed;
                this->_filterMode = src._filterMode;
			    this->target = src.target;
			    this->callback = src.callback;

                return *this;
            }

			virtual ~RenderingCfg();

			bool setRendered(bool);
			bool rendered() const;

			bool setLogscale(bool);
			bool logscale() const;

			bool setMin(float);
			float min() const;

			bool setMax(float);
			float max() const;

			bool setOpacity(float);
			float opacity() const;

			bool setZindex(float);
			float zindex() const;

			bool setColormap(const std::string&);
			const std::string& colormap() const;

			bool setColor(const std::array<std::uint8_t, 3>&);
			const std::array<std::uint8_t, 3>& color() const;

			bool setRenderMethod(const RenderMethod&);
			const RenderMethod& renderMethod() const;

			bool setParticlesCount(uint32_t);
			uint32_t particlesCount() const;

			bool setParticleTTL(uint8_t);
			uint8_t particleTTL() const;

			bool setStreamlineLength(uint32_t);
			uint32_t streamlineLength() const;

			bool setStreamlineSpeed(float);
			float streamlineSpeed() const;

			bool setScale(float, float, bool);

            bool setFilterMode(FilterMode);
            FilterMode filterMode() const;

			bool applyTo(RenderingCfg*) const;

			bool setCallback(const std::function<void(const RenderingCfg*, const RenderingParam&)>&);

			bool scrollExtrema(float, bool);

			bool _rendered;
			bool _logScale;
			float _min;
			float _max;
			float _opacity;
			float _zindex;
			std::array<std::uint8_t, 3> _color;
			std::string _colormap;
			RenderMethod _renderMethod;
			uint32_t _particlesCount;
			uint8_t _particleTTL;
			uint32_t _streamlineLength;
			float _streamlineSpeed;
            FilterMode _filterMode;

			RenderableId target;
			std::function<void(const RenderingCfg*, const RenderingParam&)> callback;
	};
};

#endif
