// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#ifndef TIMELINE_CFG_H
#define TIMELINE_CFG_H

#include <cstdint>
#include <string>
#include <array>
#include <vector>
#include <map>

#include "core/types.h"


namespace SEAScope
{
	class Timeframe final
	{
		public:
			Timeframe(const std::string& label, const std::string& description, timestamp_t pastOffset, timestamp_t futureOffset);
			Timeframe();
            Timeframe(const Timeframe& src)
            : label(src.label), description(src.description), pastOffset(src.pastOffset), futureOffset(src.futureOffset)
            {
            }
            Timeframe& operator=(const Timeframe& src)
            {
                this->label = src.label;
                this->description = src.description;
                this->pastOffset = src.pastOffset;
                this->futureOffset = src.futureOffset;

                return *this;
            }
			~Timeframe();

			std::string label;
			std::string description;
			timestamp_t pastOffset;
			timestamp_t futureOffset;
	};

	class Timestep final
	{
		public:
			Timestep(const std::string& label, const std::string& description, timestamp_t value);
			Timestep();
            Timestep(const Timestep& src)
            : label(src.label), description(src.description), value(src.value)
            {
            }
            Timestep& operator=(const Timestep& src)
            {
                this->label = src.label;
                this->description = src.description;
                this->value = src.value;

                return *this;
            }
			~Timestep();

			std::string label;
			std::string description;
			timestamp_t value;
	};

	class TimelineCfg final
	{
		public:
			TimelineCfg();
            TimelineCfg(const TimelineCfg& src)
			: minYear(src.minYear), maxYear(src.maxYear),
              defaultTimespan(src.defaultTimespan), defaultTimestep(src.defaultTimestep),
              useNearestMode(src.useNearestMode), useNearestInDayMode(src.useNearestInDayMode),
              advancedScrolling(src.advancedScrolling)
            {
			    this->timeframes.assign(src.timeframes.begin(), src.timeframes.end());
			    this->timesteps.assign(src.timesteps.begin(), src.timesteps.end());
            }
            TimelineCfg& operator=(const TimelineCfg& src)
            {
			    this->timeframes.assign(src.timeframes.begin(), src.timeframes.end());
			    this->timesteps.assign(src.timesteps.begin(), src.timesteps.end());
			    this->minYear = src.minYear;
			    this->maxYear = src.maxYear;
			    this->defaultTimespan = src.defaultTimespan;
			    this->defaultTimestep = src.defaultTimestep;
                this->useNearestMode = src.useNearestMode;
                this->useNearestInDayMode = src.useNearestInDayMode;
                this->advancedScrolling = src.advancedScrolling;

                return *this;
            }
			~TimelineCfg();

			bool addTimeframe(const std::string&, const std::string&, timestamp_t, timestamp_t);
			bool getTimeframe(const std::string& label, Timeframe& result) const;
			bool getTimeframe(timestamp_t pastOffset, timestamp_t futureOffset, Timeframe& result) const;
			bool getTimeframes( std::vector<std::string>& labels
							  , std::vector<std::string>& descriptions) const;
			bool getTimeframes(std::map<std::string, std::array<timestamp_t, 2>>& result) const;

			bool addTimestep(const std::string&, const std::string&, timestamp_t);
			bool getTimestep(const std::string& label, Timestep& result) const;
			bool getTimestep(timestamp_t value, Timestep& result) const;
			bool getTimesteps( std::vector<std::string>& labels
							 , std::vector<std::string>& descriptions) const;
			bool getTimesteps( std::map<std::string, timestamp_t>& result) const;

            bool setInitialDatetime(timestamp_t);
            timestamp_t getInitialDatetime() const;

			std::vector<Timeframe> timeframes;
			std::vector<Timestep> timesteps;
			std::uint16_t minYear;
			std::uint16_t maxYear;
			std::string defaultTimespan;
			std::string defaultTimestep;
            bool useNearestMode;
            bool useNearestInDayMode;
            bool advancedScrolling;

        protected:
            timestamp_t _initialDatetime;
	};
};

#endif
