// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "core/opengl_enum.h"
#include "3d/data_set.h"

#include "data/image.h"
#include "3d/static_mesh.h"
#include "3d/bbox.h"
#include "3d/mesh_creator.h"
#include "3d/grid.h"
#include "3d/core_objects.h"
#include "3d/streamlines_data.h"
#include "3d/particules_data.h"

#include "core/log.h"

using SEAScope::Log;

DataSet::DataSet(const SEAScope::RenderableId& rId)
: Pickable(rId)
{
    _width = 0;
    _height = 0;
	_resolution = 0.f;
    _dimensionCount = 0;
    _data = nullptr;
    _dataMask = nullptr;
    _offset = nullptr;
    _scaleFactor = nullptr;
    _scaleFactorGL = 0.f;
    lons = nullptr;
    lats = nullptr;
    lat_size = 0;
    lon_size = 0;
    min = nullptr;
    max = nullptr;
    gcpx = nullptr;
    gcpy = nullptr;
    mesh_spherical = nullptr;
    mesh_planar = nullptr;
	_xInvertedBbox = false;
    _yInvertedBbox = false;
    streamlinesData = nullptr;
    particulesData = nullptr;
}

void DataSet::setStreamlinesData(StreamlinesData* streamlinesDataIn)
{
    this->streamlinesData = streamlinesDataIn;
}

StreamlinesData* DataSet::getStreamlinesData() const
{
    return streamlinesData;
}

void DataSet::setBbox(glm::vec4 bboxIn, bool xInverted, bool yInverted)
{
    this->bbox = bboxIn;
	this->_xInvertedBbox = xInverted;
	this->_yInvertedBbox = yInverted;
}

glm::vec4 DataSet::getBbox() const
{
    return bbox;
}

void DataSet::addToBboxArray(BBox* bboxIn)
{
    bboxArray.push_back(bboxIn);
}

void DataSet::addToBboxArray(const glm::vec4& bboxIn)
{
    bboxArray.push_back(new BBox(bboxIn));
}

void DataSet::setMesh_planar(OGL::StaticMesh* mesh_planarIn)
{
    this->mesh_planar = mesh_planarIn;
}

OGL::StaticMesh* DataSet::getMesh_planar() const
{
    return mesh_planar;
}

void DataSet::setMesh_spherical(OGL::StaticMesh* mesh_sphericalIn)
{
    this->mesh_spherical = mesh_sphericalIn;
}

OGL::StaticMesh* DataSet::getMesh_spherical() const
{
    return mesh_spherical;
}

void DataSet::setGcpy(unsigned short* gcpyIn)
{
    this->gcpy = gcpyIn;
}

unsigned short* DataSet::getGcpy() const
{
    return gcpy;
}

void DataSet::setGcpx(unsigned short* gcpxIn)
{
    this->gcpx = gcpxIn;
}

unsigned short* DataSet::getGcpx() const
{
    return gcpx;
}

void DataSet::setCenterPlane(glm::vec4 centerPlaneIn)
{
    this->centerPlane = centerPlaneIn;
}

glm::vec4 DataSet::getCenterPlane() const
{
    return centerPlane;
}

void DataSet::setNormal(glm::vec3 normalIn)
{
    this->normal = normalIn;
}

glm::vec3 DataSet::getNormal() const
{
    return normal;
}

void DataSet::setCorner(int i, const glm::vec4& cornersIn)
{
    if (i >= 0 && i < 5)
    {
        corners[i] = cornersIn;
    }
}

const glm::vec4& DataSet::getCorner(int i) const{return corners[i];}

void DataSet::setMax(float* maxIn)
{
    this->max = maxIn;
}

float* DataSet::getMax() const
{
    return max;
}

void DataSet::setMin(float* minIn)
{
    this->min = minIn;
}

float* DataSet::getMin() const
{
    return min;
}

void DataSet::setLon_size(unsigned int lon_sizeIn)
{
    this->lon_size = lon_sizeIn;
}

const unsigned int& DataSet::getLon_size() const
{
    return lon_size;
}

void DataSet::setLat_size(unsigned int lat_sizeIn)
{
    this->lat_size = lat_sizeIn;
}

const unsigned int& DataSet::getLat_size() const
{
    return lat_size;
}

void DataSet::setLats(float* latsIn)
{
    this->lats = latsIn;
}

float* DataSet::getLats() const
{
    return lats;
}

void DataSet::setLons(float* lonsIn)
{
    this->lons = lonsIn;
}

float* DataSet::getLons() const
{
    return lons;
}

void DataSet::setScaleFactorGL(float _scaleFactorGLIn)
{
    this->_scaleFactorGL = _scaleFactorGLIn;
}

float DataSet::getScaleFactorGL() const
{
    return _scaleFactorGL;
}

void DataSet::setScaleFactor(float* _scaleFactorIn)
{
    this->_scaleFactor = _scaleFactorIn;
}

float* DataSet::getScaleFactor() const
{
    return _scaleFactor;
}

void DataSet::setOffset(float* _offsetIn)
{
    this->_offset = _offsetIn;
}

float* DataSet::getOffset() const
{
    return _offset;
}

void DataSet::setDataMask(unsigned int width, unsigned int height, unsigned int componentCount, OGL::DataType dataType, void * data)
{
    if
        (nullptr != this->_dataMask)
    {
        delete this->_dataMask;
        this->_data = nullptr;
    }
    this->_dataMask = new Image();
    this->_dataMask->allocate(width, height, componentCount, dataType, data);
}

Image* DataSet::getDataMask() const
{
    return _dataMask;
}

void DataSet::setData(unsigned int width, unsigned int height, unsigned int componentCount, OGL::DataType dataType, void * data)
{
    if
        (nullptr != this->_data)
    {
        delete this->_data;
        this->_data = nullptr;
    }
    this->_data = new Image();
    this->_data->allocate(width, height, componentCount, dataType, data);
}

Image* DataSet::getData() const
{
    return _data;
}

void DataSet::setDimensionCount(unsigned int _dimensionCountIn)
{
    this->_dimensionCount = _dimensionCountIn;
}

const unsigned int& DataSet::getDimensionCount() const
{
    return _dimensionCount;
}

void DataSet::setResolution(float _resolutionIn)
{
    this->_resolution = _resolutionIn;
}

float DataSet::getResolution() const
{
    return _resolution;
}

void DataSet::setHeight(int _heightIn)
{
    this->_height = _heightIn;
}

const int& DataSet::getHeight() const
{
    return _height;
}

void DataSet::setWidth(int _widthIn)
{
    this->_width = _widthIn;
}

const int& DataSet::getWidth() const
{
    return _width;
}

DataSet::~DataSet()
{
    if
        (nullptr != this->_data)
    {
        delete this->_data;
        this->_data = nullptr;
    }

    if
        (nullptr != this->_dataMask)
    {
        delete this->_dataMask;
	    this->_dataMask = nullptr;
    }

    if
        (nullptr != this->_offset)
    {
        delete [] this->_offset;
	    this->_offset = nullptr;
    }

    if
        (nullptr != this->_scaleFactor)
    {
        delete [] this->_scaleFactor;
	    this->_scaleFactor = nullptr;
    }

    if
        (nullptr != this->lons)
    {
        delete [] this->lons;
	    this->lons = nullptr;
    }

    if
        (nullptr != this->lats)
    {
        delete [] this->lats;
	    this->lats = nullptr;
    }

    if
        (nullptr != this->min)
    {
        delete [] this->min;
	    this->min = nullptr;
    }

    if
        (nullptr != this->max)
    {
        delete [] this->max;
	    this->max = nullptr;
    }

    if
        (nullptr != this->gcpx)
    {
        delete [] this->gcpx;
	    this->gcpx = nullptr;
    }

    if
        (nullptr != this->gcpy)
    {
        delete [] this->gcpy;
	    this->gcpy = nullptr;
    }

    for
        (auto bboxTmp : this->bboxArray)
    {
        delete bboxTmp;
        bboxTmp = nullptr;
    }
    this->bboxArray.clear();


    if
        (nullptr != this->mesh_spherical)
    {
        // mesh spherical and planar have the same texture objects shared.
        // delete only the pointers of one object.
        auto textureMap = mesh_spherical->getTextureMap();
        for (auto& texPair : textureMap)
        {
            delete texPair.second;
            texPair.second = nullptr;
        }

        delete this->mesh_spherical;
	    this->mesh_spherical = nullptr;
    }

    if
        (nullptr != this->mesh_planar)
    {
        delete this->mesh_planar;
	    this->mesh_planar = nullptr;
    }

    if
        (nullptr != this->streamlinesData)
    {
        delete this->streamlinesData;
	    this->streamlinesData = nullptr;
    }

    if
        (nullptr != this->particulesData)
    {
        delete this->particulesData;
	    this->particulesData = nullptr;
    }
}

glm::vec4 DataSet::getBbox(unsigned int id) const
{
    return bboxArray[id]->getMinMax();
}

void DataSet::generateGPUData(OGL::Context* ctx)
{
    // Geometry
    std::vector<glm::vec2> uv;
    std::vector<unsigned int> indices;
    
    MeshCreator mesh_creator_globe;

    createGridFromGCP(lons, lon_size, lats, lat_size, gcpx, gcpy, static_cast<unsigned int>(_width), static_cast<unsigned int>(_height), _gcps, uv, indices);

    mesh_creator_globe.setDrawType(OGL::draw_elements, OGL::triangles);
    mesh_creator_globe.setVertices(makeArray(_gcps));
    mesh_creator_globe.setTextureCoordinates(makeArray(uv));
    mesh_creator_globe.setIndices(makeArray(indices));
    
    Image* finalImage = _data;
    Image* finalImageMask = _dataMask;
    
    OGL::Texture * colors = OGL::Helper::createTexture(ctx, *finalImage);
    colors->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::linear);
    mesh_creator_globe.setTexture(colors, 0);
    
    OGL::Texture * mask = OGL::Helper::createTexture(ctx, *finalImageMask);
    mask->setWrapAndFilterMode(OGL::clamp_to_edge, OGL::nearest);
    mesh_creator_globe.setTexture(mask, 1);
    
    mesh_spherical = mesh_creator_globe.createStaticMesh(ctx);
}

void DataSet::draw(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
    mesh_spherical->bindTextures(ctx);
    mesh_spherical->draw();
}

void DataSet::drawPicking(OGL::Context* ctx, const glm::mat4& /*mvp*/)
{
    // Bind mask data to texture unit 0
    ctx->activeTextureUnit(0, mesh_spherical->getTexture(1));
    
    mesh_spherical->draw();
}

glm::vec4 DataSet::uncompressVector() const
{
    return glm::vec4(_scaleFactorGL, _offset[0], _scaleFactor[0], static_cast<float>(_dimensionCount));
}

glm::vec3 DataSet::uncompressVector3() const
{
    return glm::vec3(_scaleFactorGL, _offset[0], _scaleFactor[0]);
}

float* DataSet::offsetArray() const
{
    return _offset;
}

float* DataSet::scaleFactorArray() const
{
    return _scaleFactor;
}

ParticulesData* DataSet::getParticulesData() const
{
    return particulesData;
}

void DataSet::setParticulesData(ParticulesData* pData)
{
    particulesData = pData;
}

void DataSet::setFilterMode(OGL::FilterMode filterMode)
{
    if (nullptr != mesh_spherical)
    {
        // Get texture which store physical data
        OGL::Texture* texture = mesh_spherical->getTexture(0);
        
        if (nullptr != texture)
        {
            texture->setFilterMode(filterMode);
        }
    }
}

const std::vector<glm::vec3>& DataSet::getGcps() const
{
    return _gcps;
}

void DataSet::setSeamLess(bool x, bool y)
{
    this->_seamLess = glm::bvec2(x, y);
}

const glm::bvec2& DataSet::getSeamLess() const
{
    return _seamLess;
}

void DataSet::setNEWSAligned(bool status)
{
    this->_lonLatAligned = status;
}

bool DataSet::isNEWSAligned() const
{
    return _lonLatAligned;
}

void DataSet::setVisibleAreas(std::vector<glm::ivec4>&& areas)
{
    _areas = std::move(areas);
}

const std::vector<glm::ivec4>& DataSet::getVisibleArea() const
{
    return _areas;
}

bool DataSet::isBboxYInverted() const
{
	return this->_yInvertedBbox;
}

bool DataSet::isBboxXInverted() const
{
	return this->_xInvertedBbox;
}
