// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/particules_data.h"

#include "3d/core_objects.h"
#include "3d/data_set.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"

ParticulesData::ParticulesData()
{
    _displacement_buffer = nullptr;
    _drawMode = nullptr;
    _texcoords_buffer = nullptr;
    _vao = nullptr;
    _vertices_buffer = nullptr;
}

ParticulesData::~ParticulesData()
{
    delete _displacement_buffer;
    _displacement_buffer = nullptr;

    delete _drawMode;
    _drawMode = nullptr;

    delete _texcoords_buffer;
    _texcoords_buffer = nullptr;

    delete _vao;
    _vao = nullptr;

    delete _vertices_buffer;
    _vertices_buffer = nullptr;
}

void ParticulesData::create(OGL::Context* ctx, DataSet* dataSet)
{
    glm::vec2 texcoord;
    glm::vec3 position;
    
    unsigned int id = 0;
    unsigned int lonSize = dataSet->getLon_size();
    unsigned int latSize = dataSet->getLat_size();
    unsigned int dataXSize = static_cast<unsigned int>(dataSet->getWidth());
    unsigned int dataYSize = static_cast<unsigned int>(dataSet->getHeight());
    unsigned short* gcpx = dataSet->getGcpx();
    unsigned short* gcpy = dataSet->getGcpy();
    float* lons = dataSet->getLons();
    float* lats = dataSet->getLats();
    
    std::vector<glm::vec3> vertices;
    std::vector<glm::vec2> texcoords;
    std::vector<glm::vec2> displacement;
    
    for (unsigned int lon = 0; lon < lonSize; ++lon)
    {
        for (unsigned int lat = 0; lat < latSize; ++lat)
        {
            id = lon * latSize + lat;
            
            texcoord.x = static_cast<float>(gcpx[id]) / static_cast<float>(dataXSize);
            texcoord.y = static_cast<float>(gcpy[id]) / static_cast<float>(dataYSize);
            
            position = UnitSphere::ToVector3D(Geodetic2D(static_cast<double>(glm::radians(lons[id])),
                                                                   static_cast<double>(glm::radians(lats[id]))));
            
            // Dupplicate to make a square with two triangles
            vertices.push_back(position);
            vertices.push_back(position);
            vertices.push_back(position);
            vertices.push_back(position);
            vertices.push_back(position);
            vertices.push_back(position);
            
            texcoords.push_back(texcoord);
            texcoords.push_back(texcoord);
            texcoords.push_back(texcoord);
            texcoords.push_back(texcoord);
            texcoords.push_back(texcoord);
            texcoords.push_back(texcoord);
            
            displacement.push_back(glm::vec2(-1.f,  1.f));
            displacement.push_back(glm::vec2( 1.f,  1.f));
            displacement.push_back(glm::vec2( 1.f, -1.f));
            displacement.push_back(glm::vec2( 1.f, -1.f));
            displacement.push_back(glm::vec2(-1.f, -1.f));
            displacement.push_back(glm::vec2(-1.f,  1.f));
        }    
    }
    
    _vertices_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(vertices));
    _texcoords_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(texcoords));
    _displacement_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(displacement));
        
    OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(0, 3, _vertices_buffer);
    OGL::VertexAttribArray* texcoords_attrib = ctx->createVertexAttribArray(1, 2, _texcoords_buffer);
    OGL::VertexAttribArray* displacements_attrib = ctx->createVertexAttribArray(2, 2, _displacement_buffer);
        
    _vao = ctx->createVertexArrayObject();
    _vao->create(ctx);
    _vao->addObject(vertices_attrib);
    _vao->addObject(texcoords_attrib);
    _vao->addObject(displacements_attrib);
    
    _vao->fill();
    
    _drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::triangles, static_cast<unsigned int>(vertices.size()));
}

void ParticulesData::bind(OGL::Context* /*ctx*/)
{
}

void ParticulesData::draw(OGL::Context* ctx)
{
    ctx->setFlag(OGL::cull_face, false);
    
    _vao->bind();
    _drawMode->draw();
    _vao->unbind();

    ctx->setFlag(OGL::cull_face, true);
}
