// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <png.h>

#include "3d/stdafx.h"
#include "3d/png_reader.h"

#include "data/image.h"
#include "core/log.h"

// temporarily disable the warning C4611 (interaction between '_setjmp' and
// C++ object destruction is non-portable)
#ifdef _WIN32
    #pragma warning(disable:4611)
#endif /* VC++ */

bool readPNG(const char* file_name, int& width, int& height, int& channels, unsigned char*& pixels)
{
    unsigned int y;
    png_byte color_type;
    png_byte bit_depth;
    png_structp png_ptr;
    png_infop info_ptr;
    png_bytep * row_pointers;
    unsigned char header[8];    // 8 is the maximum size that can be checked

    /* open file and test for it being a png */
    FILE *fp = fopen(file_name, "rb");

    if (!fp)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "File {} could not be opened for reading", file_name);
        return false;
    }

    fread(header, 1, 8, fp);

    if (png_sig_cmp(static_cast<png_const_bytep>(header), 0, 8))
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "File {} is not recognized as a PNG file", file_name);
        return false;
    }


    /* initialize stuff */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);

    if (!png_ptr)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "png_create_read_struct failed");
        return false;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (!info_ptr)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "png_create_info_struct failed");
        return false;
    }

#ifdef __APPLE__
    #pragma clang diagnostic push
    #pragma clang diagnostic ignored "-Wdisabled-macro-expansion"
    if (setjmp(png_jmpbuf(png_ptr)))
    #pragma clang diagnostic pop
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "Error during init_io");
        return false;
    }
#else
    if (setjmp(png_jmpbuf(png_ptr)))
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "Error during init_io");
        return false;
    }
#endif

    png_init_io(png_ptr, fp);
    png_set_sig_bytes(png_ptr, 8);

    png_read_info(png_ptr, info_ptr);

    width               = static_cast<int>(png_get_image_width(png_ptr, info_ptr));
    height              = static_cast<int>(png_get_image_height(png_ptr, info_ptr));
    color_type          = png_get_color_type(png_ptr, info_ptr);
    bit_depth           = png_get_bit_depth(png_ptr, info_ptr);
    channels            = png_get_channels(png_ptr, info_ptr);

    switch
        (color_type)
    {
        case PNG_COLOR_TYPE_PALETTE:
            png_set_palette_to_rgb(png_ptr);
            //Don't forget to update the channel info (thanks Tom!)
            //It's used later to know how big a buffer we need for the image
            channels = 3;
            break;
        case PNG_COLOR_TYPE_GRAY:
            if (bit_depth < 8)
            png_set_expand_gray_1_2_4_to_8(png_ptr);
            //And the bitdepth info
            bit_depth = 8;
            break;
    }

    /* if the image has a transparency set.. convert it to a full Alpha
       channel..*/
    if
        (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
    {
        png_set_tRNS_to_alpha(png_ptr);
        channels = channels + 1;
    }

    /* We don't support 16 bit precision.. so if the image Has 16 bits per
       channel precision... round it down to 8. */
    if
        (16 == bit_depth)
    {
        png_set_strip_16(png_ptr);
    }

    png_read_update_info(png_ptr, info_ptr);

    /* read file */
#ifdef __APPLE__
    #pragma clang diagnostic push
    #pragma clang diagnostic ignored "-Wdisabled-macro-expansion"
    if (setjmp(png_jmpbuf(png_ptr)))
    #pragma clang diagnostic pop
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "Error during read_image");
        return false;
    }
#else
    if (setjmp(png_jmpbuf(png_ptr)))
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "read_png_file",
                           "Error during read_image");
        return false;
    }
#endif

    row_pointers = static_cast<png_bytep*>(malloc(sizeof(png_bytep) * static_cast<std::size_t>(height)));

    pixels = static_cast<unsigned char*>(malloc(static_cast<std::size_t>(width * height * channels)));

    png_size_t size = png_get_rowbytes(png_ptr, info_ptr);

    for (y = 0; y < static_cast<std::size_t>(height); y++)
    {
        row_pointers[y] = static_cast<png_byte*>(pixels) + size * y;
    }

    png_read_image(png_ptr, row_pointers);

    png_destroy_read_struct(&png_ptr, &info_ptr, nullptr);

    free(row_pointers);

    fclose(fp);

    return true;
}

#ifdef _WIN32
    #pragma warning(default:4611)
#endif /* VC++ */

bool readPNG(const char* file_name, Image& img)
{
    int width , height, channels;
    unsigned char * pixels;

    readPNG(file_name, width , height, channels, pixels);

    img.allocate(static_cast<unsigned int>(width), static_cast<unsigned int>(height), static_cast<unsigned int>(channels), OGL::unsigned_byte, pixels);

    return true;
}
