// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/polyline_editor.h"

#include "3d/polyline_builder.h"

#include "3d/core_objects.h"
#include "3d/polyline_mesh.h"
#include "3d/geodetic.h"
#include "3d/ellipsoid.h"
#include "3d/geodetic.h"
#include "3d/spherical_math.h"
#include "3d/math_func.h"
#include "3d/geometry.h"
#include "3d/polyline_base.h"
#include "3d/scene_graph_node.h"
#include "3d/shape_anchor.h"
#include "3d/anchor_pool_node.h"

PolylineEditor::PolylineEditor()
: ShapeEditor(), _program(nullptr), _buffer(nullptr), _vao(nullptr), _drawMode(nullptr)
{
}

PolylineEditor::~PolylineEditor()
{
	delete _buffer;
	_buffer = nullptr;
	delete _program;
	_program = nullptr;
	delete _vao;
	_vao = nullptr;
	delete _drawMode;
	_drawMode = nullptr;
}

void PolylineEditor::initialize(OGL::Context* ctx, Pickable* obj)
{
	_shapeToEdit = obj;

	_program = OGL::Helper::createProgram(ctx, vertex_line_tesselate, fragment_uni_color);

	_program->bind();
	_program->setUniformByName("color", glm::vec4(1.f, 0.f, 0.f, 1.f));
	_program->unbind();

	const int segmentCount = 20;
	std::vector<float> percents;
	percents.reserve(segmentCount);

	for (int i = 0; i < segmentCount; ++i)
	{
		percents.push_back(static_cast<float>(i) / (segmentCount - 1));
	}

	_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(percents));
	OGL::VertexAttribArray* vaa = ctx->createVertexAttribArray(0, 1, _buffer, OGL::float32);

	_vao = ctx->createVertexArrayObject();
	_vao->create(ctx);
	_vao->addObject(vaa);
	_vao->fill();

	_drawMode = OGL::Helper::createDrawModeArrays(ctx, OGL::line_strip, segmentCount);
}

void PolylineEditor::draw(OGL::Context* /*ctx*/, const glm::mat4& mvp)
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	if (p->getPointCount() > 0)
	{
		_program->bind();

		_program->setUniformByName("mvp", mvp);
		_program->setUniformByName("color", green4f);

		_vao->bind();

		for (unsigned int i = 0; i < p->getPointCount() - 1; ++i)
		{
			_program->setUniformByName("p0", p->getPoint(i));
			_program->setUniformByName("p1", p->getPoint(i + 1));

			_drawMode->draw();
		}

		_vao->unbind();
	}
}

OGL::Shape* PolylineEditor::createShape()
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	std::vector<glm::vec3> points = p->getPoints();
	static_cast<PolylineBuilder*>(_builder)->setPoints(std::move(points));
	return _builder->build();
}

void PolylineEditor::setBuilder(ShapeBuilder* builder)
{
    // takes ownership
	this->_builder = builder;
}

unsigned int PolylineEditor::getAnchorCount() const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
	return p->getPointCount();
}

glm::vec3 PolylineEditor::getAnchorPos(AnchorType type, unsigned int vertexId) const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	glm::vec3 pos;

	if (AnchorType::PointMoveable == type)
	{
		pos = p->getPoint(vertexId);
	}
	else if (AnchorType::PointAdd == type)
	{
		pos = getMiddlePoint(vertexId);
	}

	return pos;
}

glm::vec3 PolylineEditor::getMiddlePoint(unsigned int id) const
{
	PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());

	glm::vec3 point;

	if ((p->getPointCount() - 1u) > id)
	{
		point = glm::normalize(glm::mix(p->getPoint(id), p->getPoint(id + 1), 0.5f));
	}

	return point;
}

void PolylineEditor::move(AnchorType type, unsigned int vertexId, const glm::vec3& pos)
{
	if (AnchorType::PointMoveable == type)
	{
		PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
		p->setPoint(vertexId, pos);
	}
}

void PolylineEditor::move(AnchorType type, unsigned int vertexId)
{
    this->move(type, vertexId, this->_currentPos);
}

void PolylineEditor::pressed(AnchorType type, unsigned int id)
{
	if (AnchorType::PointAdd == type)
	{
        glm::vec3 addedPos = this->getAnchorPos(AnchorType::PointAdd, id);

        unsigned int i = this->getAnchorCount();

		PolylineBase* p = static_cast<PolylineBase*>(this->_shapeToEdit->getShape());
        p->addPoint(addedPos);
        // now i is the index of the added point

        this->_anchorPool->addAnchor(this, AnchorType::PointMoveable, i);
        this->_anchorPool->addAnchor(this, AnchorType::PointAdd, i - 1);

        while
            (i > id + 1)
        {
            p->setPoint(i, p->getPoint(i - 1));
            i = i - 1;
        }

        p->setPoint(id + 1, addedPos);

	}
}

void PolylineEditor::setAnchorPool(AnchorPoolNode* val)
{
	ShapeEditor::setAnchorPool(val);

	this->_anchorPool->clear();

	unsigned int count = getAnchorCount();

	for (unsigned int i = 0; i < count; ++i)
	{
		this->_anchorPool->addAnchor(this, AnchorType::PointMoveable, i);
	}

	for (unsigned int i = 0; i < count-1; ++i)
	{
		this->_anchorPool->addAnchor(this, AnchorType::PointAdd, i);
	}
}

/**
 *
 */
void PolylineEditor::updateCursorPosition(const glm::vec3& position)
{
    this->_currentPos = position;
}

/**
 *
 */
void PolylineEditor::remove(AnchorType type, unsigned int id)
{
    if
        (AnchorType::PointMoveable != type)
    {
        return;
    }

    PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
    if
        (2 >= p->getPointCount())
    {
        // it cannot be a polyline unless there are at least two vertices
        return;
    }

    if
        (p->getPointCount() - 1 == id)
    {
        // the removed vertex was at the last position
        // => remove the corresponding moveable anchor and the preceeding add
        //    anchor
        this->_anchorPool->removeAnchor(AnchorType::PointMoveable, id);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, id - 1);
    }
    else if
        (0 == id)
    {
        // the removed vertex was at the first position
        // => remove the corresponding moveable anchor and the following add
        //    anchor
        this->_anchorPool->removeAnchor(AnchorType::PointMoveable, p->getPointCount() - 1);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, p->getPointCount() - 2);
    }
    else
    {
        // the removed vertex was at an intermediate position
        // => remove the corresponding moveable anchor, the following and the
        //    preceeding add anchors
        this->_anchorPool->removeAnchor(AnchorType::PointMoveable, p->getPointCount() - 1);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, p->getPointCount() - 2);
        this->_anchorPool->removeAnchor(AnchorType::PointAdd, p->getPointCount() - 3);
    }

    p->deletePoint(id);
}

//bool PolylineEditor::handleKeyEvent(int key, int /*scancode*/, int action, int /*mods*/)
//{
//	bool result = false;
//
//	if (GLFW_PRESS == action)
//	{
//		switch (key)
//		{
//		case GLFW_KEY_DELETE:
//		{
//			//PolylineBase* p = static_cast<PolylineBase*>(_shapeToEdit->getShape());
//
//			//if (2 < p->getPointCount())
//			//{
//			//	auto& anchors = this->_anchorPool->getAnchors();
//			//	int anchorId = -1;
//
//			//	for (ShapeAnchor* anchor : anchors)
//			//	{
//			//		if (true == anchor->getStyle().isHighlighted())
//			//		{
//			//			anchorId = anchor->getId();
//			//			break;
//			//		}
//			//	}
//
//			//	if (-1 != anchorId)
//			//	{
//			//		p->deletePoint(anchorId);
//			//		_anchorPool->removeAnchor(anchorId);
//			//	}
//			//}
//
//			result = true;
//		}
//		break;
//		}
//	}
//
//	return result;
//}
