// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/ppm.h"

#include "data/image.h"
#include "core/log.h"

bool loadPPM(const char* file_name,int & width, int & height, unsigned char *& pixels)
{
    const int PPMREADBUFLEN = 256;

    char buf[PPMREADBUFLEN], *t;
    unsigned int d;
    int r;
    
    FILE *pf;
    
    SEAScope::Log::log(SEAScope::LogLevel::Debug, "PPM",
                       "Loading {} file", file_name);
    
    pf = fopen(file_name, "rb");

    if (nullptr == pf)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "PPM",
                           "File {} not found, fopen failed", file_name);
        return false;
    }
    
    t = fgets(buf, PPMREADBUFLEN, pf);
    
    /* the code fails if the white space following "P6" is not '\n' */
    if ( (t == nullptr) || ( strncmp(buf, "P6\n", 3) != 0 ) )
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "PPM",
                           "Wrong PPM header");
        return false;
    }
    
    do
    {
        /* Px formats can have # comments after first line */
        t = fgets(buf, PPMREADBUFLEN, pf);
       
        if ( nullptr == t )
        {
            return false;
        }
    }
    while (0 == strncmp(buf, "#", 1));
    
    r = sscanf(buf, "%u %u", &width, &height);
    
    SEAScope::Log::log(SEAScope::LogLevel::Debug, "PPM",
                       "width={} height={}", width, height);
    
    if ( r < 2 )
    {
        return false;
    }

    r = fscanf(pf, "%u", &d);
    
    if ( (r < 1) || ( d != 255 ) )
    {
        return false;
    }
    
    fseek(pf, 1, SEEK_CUR); /* skip one byte, should be whitespace */

    std::size_t pixelsBytes = static_cast<size_t>(width * height * 3);
    pixels = static_cast<unsigned char*>(malloc(pixelsBytes));
    
    if ( nullptr != pixels )
    {
        size_t rd = static_cast<size_t>(fread(pixels, 1, pixelsBytes, pf));
        
        SEAScope::Log::log(SEAScope::LogLevel::Debug, "PPM",
                           "Read {} bytes", static_cast<int>(rd));
        
        if ( pixelsBytes > rd )
        {
            SEAScope::Log::log(SEAScope::LogLevel::Error, "PPM",
                               "Could not fully fill pixels buffer from file");

            free(pixels);
            return false;
        }
    }
    
    return true;
}

bool loadPGMBinary(const char* file_name, int & width, int & height, char *& pixels)
{
    FILE *pf;
    
    pf = fopen(file_name, "rb");

    if (nullptr == pf)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "PGMBinary",
                           "File {} not found, fopen failed", file_name);
        return false;
    }
    
    char magic[5];
    int precision;
    
    fscanf(pf, "%s\n", magic);
    
	char buffer[255];
	fgets(buffer, 255, pf);
    
    fscanf(pf, "%d %d\n%d\n", &width, &height, &precision);
    
    std::size_t pixelsBytes = static_cast<std::size_t>(width * height);
    pixels = static_cast<char*>(malloc(pixelsBytes));
    
    fread(pixels, pixelsBytes, 1, pf);
    
    fclose(pf);
    
    return true;
}

bool loadPGM(const char* file_name,int & width, int & height, char *& pixels)
{
    const int PPMREADBUFLEN = 256;

    char buf[PPMREADBUFLEN], *t;
    unsigned int d;
    int r;
    
    FILE *pf;
    
    SEAScope::Log::log(SEAScope::LogLevel::Debug, "PGM",
                       "Loading {} file", file_name);
    
    pf = fopen(file_name, "rb");

    if (nullptr == pf)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "PGM",
                           "File {} not found, fopen failed", file_name);
        return false;
    }
    
    t = fgets(buf, PPMREADBUFLEN, pf);
    
    /* the code fails if the white space following "P3" is not '\n' */
    if ( (t == nullptr) || ( strncmp(buf, "P3\n", 3) != 0 ) )
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "PGM",
                           "Wrong PGM header");
        return false;
    }
    
    do
    {
        /* Px formats can have # comments after first line */
        t = fgets(buf, PPMREADBUFLEN, pf);
       
        if ( nullptr == t )
        {
            return false;
        }
    }
    while (0 == strncmp(buf, "#", 1));
    
    r = sscanf(buf, "%u %u", &width, &height);
    
    SEAScope::Log::log(SEAScope::LogLevel::Debug, "PGM",
                       "width={} height={}", width, height);
    
    if ( r < 2 )
    {
        return false;
    }

    r = fscanf(pf, "%u", &d);
    
    if ( (r < 1) || ( d != 255 ) )
    {
        return false;
    }
    
    fseek(pf, 1, SEEK_CUR); /* skip one byte, should be whitespace */

    std::size_t pixelsBytes = static_cast<std::size_t>(width * height);
    pixels = static_cast<char*>(malloc(pixelsBytes));
    
    if ( nullptr != pixels )
    {
        size_t rd = fread(pixels, 1, pixelsBytes, pf);
        
        SEAScope::Log::log(SEAScope::LogLevel::Debug, "PGM",
                           "Read {} bytes", static_cast<int>(rd));
        
        if ( pixelsBytes > rd )
        {
            SEAScope::Log::log(SEAScope::LogLevel::Error, "PGM",
                               "Could not fully fill pixels buffer from file");


            free(pixels);
            return false;
        }
    }
    
    return true;
}
bool savePGM(const char* file_name, const Image& img)
{
    int width(0);    //
    int height(0);  //
    char* data(nullptr); //
    bool result(false); //

    width = static_cast<int>(img.width());
    height = static_cast<int>(img.height());
    data = static_cast<char*>(img.data());
    result = savePGM(file_name, width, height, data);
    data = nullptr;

    return result;
}

bool savePGM(const char* file_name, int width, int height, char * pixels)
{
    FILE *pf;
    
    pf = fopen(file_name, "wb");
    
    if (!pf)
    {
        return false;
    }
    
    fprintf(pf, "P5\n# debug export\n%d %d\n%d\n", width, height, 255);
    std::size_t pixelsBytes = static_cast<std::size_t>(width * height);
    fwrite(pixels, 1, pixelsBytes, pf);
    
    fclose(pf);
    
    return true;
}

bool savePPM(const char* file_name, const Image& img)
{
    int width(0);    //
    int height(0);  //
    char* data(nullptr); //
    bool result(false); //

    width = static_cast<int>(img.width());
    height = static_cast<int>(img.height());
    data = static_cast<char*>(img.data());
    result = savePPM(file_name, width, height, data);
    data = nullptr;

    return result;
}

bool savePPM(const char* file_name, int width, int height, char * pixels)
{
    FILE *pf;
    
    pf = fopen(file_name, "wb");
    
    if (!pf)
    {
        return false;
    }
    
    fprintf(pf, "P6\n# debug export\n%d %d\n%d\n", width, height, 255);
    
    unsigned char* p = reinterpret_cast<unsigned char*>(pixels);
    
    for (int y = 0; y < height; ++y)
    {
        for (int x = 0; x < width; ++x)
        {
            unsigned char* f = p + (((height-1) - y) * width + x) * 3;
            fwrite(f, 3, 1, pf);
        }
    }
    
    fclose(pf);
    
    return true;
}

// RGB file format is a text file with one pixel RGB values on each line.
// R G B\n
// R G B\n
// ...
bool saveRGB(const char* file_name, int width, int height, unsigned char * pixels)
{
    FILE *pf;
    unsigned int widthTmp(0);   //
    unsigned int heightTmp(0);  //
    
    pf = fopen(file_name, "wb");
    
    if (!pf)
    {
        return false;
    }

    unsigned int pos(0);    //

    widthTmp = static_cast<unsigned int>(width);
    heightTmp = static_cast<unsigned int>(height);

    for (unsigned int y = 0; y < heightTmp; ++y)
    {
        for (unsigned int x = 0; x < widthTmp; ++x)
        {
            pos = (y * widthTmp + x) * 3;
            
            fprintf(pf, "%d %d %d\n", *(pixels + pos), *(pixels + pos + 1), *(pixels + pos + 2));
        }
    }
    
    fclose(pf);
    
    return true;
}

bool savePFM(const char* file_name, const Image& img)
{
    int width(0);    //
    int height(0);  //
    float* data(nullptr); //
    bool result(false); //

    width = static_cast<int>(img.width());
    height = static_cast<int>(img.height());
    data = static_cast<float*>(img.data());
    result = savePFM(file_name, width, height, data);
    data = nullptr;

    return result;
}

bool savePFM(const char* file_name, int width, int height, float * pixels)
{
    FILE *pf;
    
    pf = fopen(file_name, "wb");
    
    if (!pf)
    {
        return false;
    }
    
    fprintf(pf, "PF\n%d %d\n%d %f\n", width, height, 255, 1.0);
    
    for (int y = 0; y < height; ++y)
    {
        for (int x = 0; x < width; ++x)
        {
            float* f = pixels + (((height-1) - y) * width + x) * 4;
            
            fwrite(f, 3, sizeof(float), pf);
        }
    }
    
    fclose(pf);
    
    return true;
}

bool saveTXT4f(const char* file_name, const Image& img)
{
    FILE *pf;
    
    pf = fopen(file_name, "wb");
    
    if (!pf)
    {
        return false;
    }
    
    fprintf(pf, "%d %d\n", img.width(), img.height());

    float* pixels = static_cast<float*>(img.data());
    
    for (unsigned int y = 0; y < img.height(); ++y)
    {
        for (unsigned int x = 0; x < img.width(); ++x)
        {
            float* f = pixels + (y * img.width() + x) * 4;
            
            fprintf(pf,
                    "%f %f %f %f\n",
                    static_cast<double>(f[0]),
                    static_cast<double>(f[1]),
                    static_cast<double>(f[2]),
                    static_cast<double>(f[3]));
        }
    }
    
    fclose(pf);

    return true;
}
