// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/scene_graph_controller.h"

#include "3d/scene_dictionnary.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_node.h"
#include "3d/data_set.h"
#include "3d/colormap_manager_gpu.h"
#include "3d/polyline_mesh.h"
#include "3d/streamlines_data.h"
#include "3d/globe.h"
#include "3d/particle_grid.h"
#include "3d/enum_convertor.h"

#include "core/types.h"
#include "core/rendering_cfg.h"
#include "3d/shape_mesh.h"
#include "3d/shapes_allocator.h"
#include "3d/datasets_allocator.h"
#include "3d/polyline.h"

#include "num_utils.h"

#include "core/app_events_dispatcher.h"
#include "events/renderable_added_to_scene_graph_event.h"
#include "events/renderable_removed_from_scene_graph_event.h"
#include "events/annotation_removed_event.h"

#include "core/log.h"

namespace
{
    RendererType toRendererType(SEAScope::RenderMethod rMethod, Pickable* pickable, bool& ok)
    {
        RendererType type = RendererType::renderer_raster;
        DataSet* dataSet(nullptr);
        bool isDataSet(false);

        dataSet = dynamic_cast<DataSet*>(pickable);
        isDataSet = (nullptr != dataSet);

        switch(rMethod)
        {
        case SEAScope::RenderMethod::RASTER:
            type = RendererType::renderer_raster;
            ok = isDataSet;
            break;
        case SEAScope::RenderMethod::BARBS:
            type = RendererType::renderer_barb;
            ok = isDataSet;
            break;
        case SEAScope::RenderMethod::ARROWS:
            type = RendererType::renderer_arrow;
            ok = isDataSet;
            break;
        case SEAScope::RenderMethod::DOTSCLOUD:
            type = RendererType::renderer_dot_cloud;
            ok = isDataSet;
            break;
        case SEAScope::RenderMethod::STREAMLINES:
        {
            type = RendererType::renderer_uv_streamline;
        }
            ok = isDataSet;
            break;
        case SEAScope::RenderMethod::TRAJECTORIES:
            type = RendererType::renderer_trajectory;
            ok = isDataSet;
            break;
		case SEAScope::RenderMethod::RAWRGB:
			type = RendererType::renderer_rawrgb;
            ok = isDataSet;
			break;
        case SEAScope::RenderMethod::POLYLINE:
			type = RendererType::renderer_polyline;
            ok = !isDataSet;
            break;
        case SEAScope::RenderMethod::POLYGON:
			type = RendererType::renderer_polygon;
            ok = !isDataSet;
            break;
        case SEAScope::RenderMethod::ELLIPSE:
            type = RendererType::renderer_polyline;
            ok = !isDataSet;
            break;
        case SEAScope::RenderMethod::POINT:
            type = RendererType::renderer_quadscreen_textured;
            ok = !isDataSet;
            break;
        }

        return type;
    }
}

SceneGraphController::SceneGraphController(SEAScope::AppEventsDispatcher& appEventsDispatcher)
:   _sceneGraph(nullptr),
    _colormapMngr(nullptr),
    _world(nullptr),
    _appEventsDispatcher(appEventsDispatcher)
{
    this->_appEventsDispatcher.subscribe<SEAScope::AnnotationRemovedEvent>(this);
}

SceneGraphController::~SceneGraphController()
{
    this->_appEventsDispatcher.unsubscribe<SEAScope::AnnotationRemovedEvent>(this);

    this->_world = nullptr;  // owned and deleted by scene graph root
    this->_colormapMngr = nullptr;  // owned and deleted by map_main
    this->_sceneGraph = nullptr;    // owned and deleted by map_main
}

bool SceneGraphController::addGranule(OGL::Context* ctx, DataSet* dataSet, const SEAScope::RenderingCfg& rCfg)
{
    bool ok(false);
    RendererType rType = toRendererType(rCfg._renderMethod, dataSet, ok);
    if
        (false == ok)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "SceneGraphController",
                           "Cannot render granule with the selected render method");
        return false;
    }

    Renderable2* renderable = nullptr;

    // Check if the granule already exist
    if (_dictionnary.exist(dataSet->renderableId(), renderable))
    {
        DataSet* dataSetFound = static_cast<DataSet*>(renderable);

        // The granule exist, check its resolution
        if (SEAScope::floating_point_equality(dataSetFound->getResolution(), dataSet->getResolution()))
        {
            // It's the same, nothing to do
            return false;
        }
        else
        {
            // The resolution is different, remove the current granule object
            remove(dataSet->renderableId());
        }
    }

    // Send the granule data to the GPU
    dataSet->generateGPUData(ctx);

    OGL::FilterMode filterMode = seascopeFilterModeEnumToGLWrapperEnum(rCfg._filterMode);

    dataSet->setFilterMode(filterMode);

    if (SEAScope::RenderMethod::STREAMLINES == rCfg._renderMethod)
    {
        dataSet->setStreamlinesData(new StreamlinesData);
        dataSet->getStreamlinesData()->create(ctx, rCfg.particlesCount(), rCfg.streamlineLength(), rCfg.streamlineSpeed(), dataSet);
    }
	else if (  SEAScope::RenderMethod::BARBS == rCfg._renderMethod
			|| SEAScope::RenderMethod::ARROWS == rCfg._renderMethod)
    {
        ParticleGrid::generateParticulesData(ctx, dataSet);
    }


    addToSceneGraph(ctx, rCfg, rType, dataSet, false);
    return true;
}

void SceneGraphController::addToSceneGraph(OGL::Context* ctx, const SEAScope::RenderingCfg& rCfg, RendererType rendererType, Renderable2* renderable, bool ownRenderer)
{
    // Create a new scene graph node
    GranuleNode* node = new GranuleNode();

    node->setRenderable(renderable);

    // Save the granule in the dictionnary
    _dictionnary.add(renderable);

    // Check the rendering target
    if (true == rCfg.target.granuleLevel) // single granule
    {
        // The granule will be independant, create a rendering config
        node->setRendererType(rendererType);
        node->setRenderingParam(createRenderingParam(ctx, rCfg));
        node->setFillType(fillTypeFromRenderingCfg(node->dataRenderingParam()));

        // Add the node to the scene graph
        _sceneGraph->addTo(nullptr, node, rCfg.target);
    }
    else // granule in collection
    {
        const SEAScope::RenderableId& collectionId = rCfg.target;

        Node* collectionNode = nullptr;

        // Check the granule's collection node.
        collectionNode = _sceneGraph->getNode(collectionId);

        // Node exist ?
        if (nullptr == collectionNode)
        {
            // The collection node doesn't exist.
            collectionNode = new Node();

            // Set the rendering configuration
            collectionNode->setRendererType(rendererType);
            collectionNode->setRenderingParam(createRenderingParam(ctx, rCfg));
            collectionNode->setFillType(fillTypeFromRenderingCfg(collectionNode->dataRenderingParam()));

            // Add the collection node to the scene graph
            _sceneGraph->addTo(nullptr, collectionNode, collectionId);
        }

//        if (ownRenderer)
//        {
//            node->setRendererType(rendererType);
////            node->setRenderingParam(createRenderingParam(ctx, rCfg));
//            node->setFillType(fillTypeFromRenderingCfg(collectionNode->dataRenderingParam()));
//        }

        // Add the granule node to the scene graph
        _sceneGraph->addTo(collectionNode, node, renderable->renderableId(), ownRenderer);
    }

    SEAScope::RenderableAddedToSceneGraphEvent event(renderable);
    this->_appEventsDispatcher.dispatch<SEAScope::RenderableAddedToSceneGraphEvent>(event);
}

bool SceneGraphController::addTrajectory(OGL::Context* ctx,
                                         std::vector<Geodetic2D>& points, 
                                         std::vector<float>& fData, 
                                         std::vector<bool>& mask, 
                                         const SEAScope::RenderingCfg& rCfg, 
                                         const SEAScope::RenderableId& rId)
{
    Renderable2* renderable = nullptr;
    
    // Check if the granule already exist
    if (_dictionnary.exist(rId, renderable))
    {
        return false;
    }
    
    PolylineMesh* polylineMesh = ShapesAllocator::getMesh<PolylineMesh>();
    
    // Send the granule data to the GPU
	if
		(false == fData.empty())
	{
		polylineMesh->setAttribute(0, std::move(fData));
		polylineMesh->setMask(std::move(mask));
		polylineMesh->initialize(ctx, rId, std::move(points),
                                 ShapesAllocator::getShape<Polyline>(),
                                 nullptr);
	}
	else
	{
		polylineMesh->initialize(ctx, rId, std::move(points),
                                 ShapesAllocator::getShape<Polyline>());
	}

    addToSceneGraph(ctx, rCfg, renderer_trajectory, polylineMesh, false);
    return true;
}

bool SceneGraphController::addShape(OGL::Context* ctx, Pickable* shape, const SEAScope::RenderingCfg& rCfg)
{
    bool ok(false);
    RendererType rType = toRendererType(rCfg._renderMethod, shape, ok);
    if
        (false == ok)
    {
        SEAScope::Log::log(SEAScope::LogLevel::Error, "SceneGraphController",
                           "Cannot render shape with the selected render method");
        return false;
    }

	addToSceneGraph(ctx, rCfg, rType, shape, false);
    return true;
}

void SceneGraphController::remove(const SEAScope::RenderableId& rId)
{
    Node* node = nullptr;

    if (_sceneGraph->remove(rId, node))
    {
        delete node;
		node = nullptr;
    }

    Renderable2* renderable = nullptr;

    if (false == _dictionnary.remove(rId, renderable))
    {
        return;
    }

    // Notify other components (the memory pointed by renderable has not been freed yet)
    SEAScope::RenderableRemovedFromSceneGraphEvent event(renderable);
    this->_appEventsDispatcher.dispatch<SEAScope::RenderableRemovedFromSceneGraphEvent>(event);

    DataSet* dataset = dynamic_cast<DataSet*>(renderable);
    ShapeMesh* mesh = dynamic_cast<ShapeMesh*>(renderable);

    if
        (nullptr != dataset)
    {
        DatasetsAllocator::release(dataset);
    }
    else if
        (nullptr != mesh)
    {
        ShapesAllocator::release(mesh);
    }
    else
    {
        delete renderable;
    }

    dataset = nullptr;
    mesh = nullptr;
    renderable = nullptr;
}

void SceneGraphController::setVisible(const SEAScope::RenderableId& rId, bool status)
{
    _sceneGraph->setVisible(rId, status);
}

void SceneGraphController::setSceneGraph(OGL::Context* /*ctx*/, SceneGraph* sg)
{
    _sceneGraph = sg;
}

void SceneGraphController::setColormapManager(OGL::ColormapManagerGPU* cm)
{
    _colormapMngr = cm;
}


DataRenderingParam* SceneGraphController::createRenderingParam(OGL::Context* ctx, const SEAScope::RenderingCfg& renderingCfg)
{
    // Set rendering parameters
    DataRenderingParam* param = new DataRenderingParam;

    updateRenderingCfg(ctx, param, renderingCfg);
    
    return param;
}

void SceneGraphController::updateRenderingCfg(OGL::Context* ctx, DataRenderingParam* param, const SEAScope::RenderingCfg& renderingCfg)
{
    param->alpha = renderingCfg.opacity();

    if (renderingCfg.colormap().empty())
    {
        // Empty string, use the color variable
        param->color = glm::vec3( static_cast<float>(renderingCfg.color()[0]) / 255.f
                                , static_cast<float>(renderingCfg.color()[1]) / 255.f
                                , static_cast<float>(renderingCfg.color()[2]) / 255.f);

        // Set the colormap pointer to null
        param->colormap = nullptr;
    }
    else
    {
        // Colormap string is not empty, get the texture object.
        param->colormap = _colormapMngr->getColormap(ctx, renderingCfg.colormap());
        
        // No colormap found, use color renderer.
        if (nullptr == param->colormap)
        {
            param->color = glm::vec3(0.f, 1.f, 1.f);
        }
    }

    param->log = renderingCfg.logscale();
    param->range = glm::vec2(renderingCfg.min(), renderingCfg.max());
    param->visible = true;
    param->zIndex = renderingCfg.zindex();
    param->filterMode = seascopeFilterModeEnumToGLWrapperEnum(renderingCfg.filterMode());
    param->particlesCount = renderingCfg.particlesCount();
    param->streamlineLength = renderingCfg.streamlineLength();
    param->streamlineSpeed = renderingCfg.streamlineSpeed();
}

void SceneGraphController::updateDisplayParameters(OGL::Context* ctx, const SEAScope::RenderingCfg& renderingCfg)
{
    if (false == renderingCfg.target.granuleLevel)
    {
        Node* node = _sceneGraph->getNode(renderingCfg.target);
        
        if(nullptr != node)
        {
            DataRenderingParam* param = node->dataRenderingParam();
            
            // Compare the z order
            bool updateZIndex = (false == SEAScope::floating_point_equality(param->zIndex, renderingCfg.zindex()));
            
            // Check filter mode change
            OGL::FilterMode filterMode = seascopeFilterModeEnumToGLWrapperEnum(renderingCfg._filterMode);

            if (param->filterMode != filterMode)
            {
                auto array = node->children();

                // Loop on collection childrens
                for (Node* child : *array)
                {
                    // Normally it's GranuleNode objects
                    GranuleNode* childGranuleNode = dynamic_cast<GranuleNode*>(child);

                    if (childGranuleNode)
                    {
                        // If we change the filter mode it's for DataSet, but check for more safety
                        DataSet* dataSet = dynamic_cast<DataSet*>(childGranuleNode->renderable());

                        if (dataSet)
                        {
                            // It's ok, change the filter mode
                            dataSet->setFilterMode(filterMode);
                        }
                    }
                }
            }

            if
                ( (SEAScope::RenderMethod::STREAMLINES == renderingCfg._renderMethod)
                && ((param->particlesCount != renderingCfg.particlesCount())
                 || (param->streamlineLength != renderingCfg.streamlineLength())
                 || (false == SEAScope::floating_point_equality(param->streamlineSpeed, renderingCfg.streamlineSpeed()))))
            {
                auto array = node->children();

                // Loop on collection childrens
                for (Node* child : *array)
                {
                    // Normally it's GranuleNode objects
                    GranuleNode* childGranuleNode = dynamic_cast<GranuleNode*>(child);

                    if (childGranuleNode)
                    {
                        // If we change the filter mode it's for DataSet, but check for more safety
                        DataSet* dataSet = dynamic_cast<DataSet*>(childGranuleNode->renderable());

                        if (dataSet)
                        {
                            //dataSet->setStreamlinesData(new StreamlinesData);
                            dataSet->getStreamlinesData()->create(ctx, renderingCfg.particlesCount(), renderingCfg.streamlineLength(), renderingCfg.streamlineSpeed(), dataSet);
                        }
                    }
                }
            }

            updateRenderingCfg(ctx, param, renderingCfg);
            
            node->setFillType(fillTypeFromRenderingCfg(param));
            _sceneGraph->updateRenderer(node);
            
            // Z changed, update the node list (== draw order)
            if (updateZIndex)
            {
                _sceneGraph->zChanged(node);
            }
        }
    }
}

FillType SceneGraphController::fillTypeFromRenderingCfg(DataRenderingParam* param)
{
    FillType ft;
    
    if (nullptr == param->colormap)
    {
        ft = fill_color;
    }
    else
    {
        ft = fill_false_colors;
    }
    
    return ft;
}

bool SceneGraphController::exist(const SEAScope::RenderableId& rId, float resolution)
{
    Renderable2* obj = nullptr;
    bool result = true;
    
    // Search in dictionnary if the id exist
    if (_dictionnary.exist(rId, obj))
    {
        // Try to cast the renderable object in DataSet
        DataSet* dataSet = dynamic_cast<DataSet*>(obj);
        
        // Test if it's a dataset 
        if (   nullptr != dataSet 
            && false == SEAScope::floating_point_equality(dataSet->getResolution(), resolution))
        {
            // It's a dataset but the resolution is different
            result = false;
        }
    }
    else
    {
        result = false;
    }
    
    return result;
}

bool SceneGraphController::exist(const SEAScope::RenderableId& rId)
{
    return _dictionnary.exist(rId);
}

void SceneGraphController::setVisibleData(const std::set<SEAScope::RenderableId>& rIds)
{
    std::list<std::pair<SEAScope::RenderableId, Renderable2*>> existing;
    std::list<std::pair<SEAScope::RenderableId, Renderable2*>> notFound;

    _dictionnary.fillListToRemovePair(rIds, existing, notFound);
    
    for (auto it : existing)
    {
        setVisible(it.first, true);
    }
    
    for (auto it : notFound)
    {
        if (true == it.second->flag(RenderableFlags::erasable))
        {
            remove(it.first);
        }
        else
        {
            setVisible(it.first, false);
        }
    }
}

void SceneGraphController::addGlobe(Globe* globe)
{
    DataRenderingParam* _worldRenderingParam = new DataRenderingParam();
    _worldRenderingParam->zIndex = 0.f;

    _world = new GranuleNode();  // scene graph root takes ownership and will handle deletion
    _world->setRenderingParam(_worldRenderingParam);
    _world->setRenderable(globe);
    _world->setRendererType(renderer_globe);
    _world->setFillType(fill_color);
    _worldRenderingParam = nullptr;

    _sceneGraph->addTo(nullptr, _world, globe->renderableId());
}

void SceneGraphController::drawWorld(OGL::Context* ctx, const glm::mat4& mvp, FillType type)
{
    switch(type)
    {
        case fill_raw:
            _world->drawRaw(ctx, mvp);
        break;
        case fill_false_colors:
        case fill_color:
        case fill_count:
            break;
    }
}

const SceneDictionnary* SceneGraphController::getSceneDictionnary() const
{
    return &this->_dictionnary;
}

SceneGraph* SceneGraphController::getSceneGraph() const
{
    return _sceneGraph;
}

bool SceneGraphController::handleAppEvent(std::type_index eventType, const SEAScope::AppEvent& event)
{
    if
        (std::type_index(typeid(SEAScope::AnnotationRemovedEvent)) == eventType)
    {
        const SEAScope::RenderableId& rId = static_cast<const SEAScope::AnnotationRemovedEvent&>(event).getRenderableId();
        this->remove(rId);
        return true;
    }

    return false;
}
