// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/streamlines_data.h"

#include "3d/data_set.h"
#include "3d/core_objects.h"
#include "3d/program_bind_helper.h"
#include "3d/unit_sphere.h"
#include "3d/geodetic.h"
#include "3d/random.h"
#include "3d/spherical_math.h"
#include "3d/default_value.h"
#include "3d/gcp_utils.h"

StreamlinesData::StreamlinesData()
	: _streamlineCountTmp(0),
	_update(false),
	_streamlineLengthTmp(0),
	//_particleTTL(0),
	_sourceId(0),
	_targetId(1),
	_gcp(nullptr),
	_gcpNorthEast(nullptr),
	//_minX(0.f),
	_saveConfig(false)
{
	memset(_vbo, 0, sizeof(void*) * 2);
	_vertices_local_id_buffer = nullptr;
	_indices_buffer = nullptr;

	for (int i = 0; i < 2; ++i)
	{
		_vaoPhysic[i] = nullptr;
		_vaoRendering[i] = nullptr;
	}

	setSpeed(1.f);
	setArea(1.f);
}

StreamlinesData::~StreamlinesData()
{
	for (int i = 0; i < 2; ++i)
	{
		delete _vaoPhysic[i];
		_vaoPhysic[i] = nullptr;
		delete _vaoRendering[i];
		_vaoRendering[i] = nullptr;
	}

	delete _vbo[0];
	_vbo[0] = nullptr;
	delete _vbo[1];
	_vbo[1] = nullptr;
	delete _indices_buffer;
	_indices_buffer = nullptr;
	delete _vertices_local_id_buffer;
	_vertices_local_id_buffer = nullptr;
	delete _gcp;
	delete _gcpNorthEast;
}


void StreamlinesData::create(OGL::Context* ctx, unsigned int streamlineCount, unsigned int streamlineLength, float speed, DataSet* dataSet)
{
	_streamlineCount = streamlineCount;
	_streamlineLength = streamlineLength;
	setSpeed(speed);
	setCurrentStreamline(0);

	float ttl = 5.f;

	_streamlinePerSec = static_cast<float>(getStreamlineCount()) / ttl;

	computeGcps(ctx, dataSet, _gcp, _gcpNorthEast, _uvMin, _uvRange);

	setParticleCount(getStreamlineCount() * getStreamlineLength());

	initBuffers(ctx);

	const std::vector<glm::vec3>& gcps = dataSet->getGcps();
	const unsigned int lonSize = dataSet->getLon_size();

	const glm::vec3& p0 = gcps[lonSize + 1];
	const glm::vec3& p1 = gcps[lonSize + 2];
	const glm::vec3& p2 = gcps[lonSize * 2 + 1];

	double doubleTmp = static_cast<double>(glm::distance(p0, p1)) * GLOBE_RADIUS_EQUATOR_KM;
	float lonDist = static_cast<float>(doubleTmp);
	doubleTmp = static_cast<double>(glm::distance(p0, p2)) * GLOBE_RADIUS_EQUATOR_KM;
	float latDist = static_cast<float>(doubleTmp);

	setArea(lonDist * latDist);

	// Start all chronos
	_origin.start();
	getLastUpdateTime().start();
	getLastTime().start();
}

void StreamlinesData::setStreamlineLength(unsigned int val)
{
	this->_streamlineLengthTmp = val;
	this->_update = true;
}

void StreamlinesData::setStreamlineCount(unsigned int val)
{
	this->_streamlineCountTmp = val;
	this->_update = true;
}

void StreamlinesData::swap()
{
	std::swap(this->_sourceId, this->_targetId);
}

void StreamlinesData::update(OGL::Context* ctx)
{
	if
		(true == this->_update)
	{
		if
			(0 != this->_streamlineCountTmp &&
				this->_streamlineCount != this->_streamlineCountTmp)
		{
			this->_streamlineCount = this->_streamlineCountTmp;
			this->_streamlineCountTmp = 0;
		}

		if
			(0 != this->_streamlineLengthTmp &&
				this->_streamlineLength != this->_streamlineLengthTmp)
		{
			this->_streamlineLength = this->_streamlineLengthTmp;
			this->_streamlineLengthTmp = 0;
		}

		setParticleCount(this->_streamlineCount * this->_streamlineLength);
		setCurrentStreamline(0);

		// Destroy all buffers
		clearBuffers();

		// Create new buffers according to the new size
		initBuffers(ctx);

		this->_update = false;
	}
}

void StreamlinesData::clearBuffers()
{
	for (int i = 0; i < 2; ++i)
	{
		delete _vbo[i];
	}

	delete _vertices_local_id_buffer;
	delete _indices_buffer;

	for (int i = 0; i < 2; ++i)
	{
		delete _vaoRendering[i];
	}

	for (int i = 0; i < 2; ++i)
	{
		delete _vaoPhysic[i];
	}
}

void StreamlinesData::initBuffers(OGL::Context* ctx)
{
	for (int i = 0; i < 2; ++i)
	{
		_vbo[i] = OGL::Helper::createBuffer(ctx, OGL::array_buffer, Array(BufferDesc(getParticleCount(), 2, OGL::float32), nullptr));
	}

	glm::vec2* particulesPosition = new glm::vec2[getParticleCount()];

	glm::vec2 uv;

	for (size_t i = 0; i < getStreamlineCount(); ++i)
	{
		uv = glm::vec2(randomOne(), randomOne());

		for (size_t j = 0; j < getStreamlineLength(); ++j)
		{
			particulesPosition[i * getStreamlineLength() + j] = uv;
		}
	}

	// Fill the source vbo
	_vbo[_sourceId]->copy(particulesPosition);

	delete[] particulesPosition;

	std::vector<unsigned int> indices;
	indices.reserve(getParticleCount());
	std::vector<float> vertex_local_id(getParticleCount() + 1);

	unsigned int offset;

	for (unsigned int i = 0; i < getStreamlineCount(); ++i)
	{
		offset = i * getStreamlineLength();

		for (unsigned int j = 0; j < getStreamlineLength() - 1; ++j)
		{
			indices.push_back(offset + j);
			indices.push_back(offset + j + 1);
		}

		for (unsigned int k = 0; k < getStreamlineLength(); ++k)
		{
			vertex_local_id[offset + k] = static_cast<float>(k);
		}
	}

	setIndiceCount(static_cast<unsigned int>(indices.size()));

	_vertices_local_id_buffer = OGL::Helper::createBuffer(ctx, OGL::array_buffer, makeArray(vertex_local_id));
	_indices_buffer = OGL::Helper::createBuffer(ctx, OGL::element_array_buffer, makeArray(indices));

	unsigned int vertices_location = 0; //renderingProgram->getVerticesLocation();

	// Rendering
	for (int i = 0; i < 2; ++i)
	{
		// 		OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 2, _vbo[i], OGL::float32, sizeof(float) * 3);
		OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 2, _vbo[i]);
		unsigned int localID = 1; //renderingProgram->getProgram()->getAttribLocation("localID");

		// localId to compute alpha
		OGL::VertexAttribArray* vertices_attrib2 = ctx->createVertexAttribArray(localID, 1, _vertices_local_id_buffer);

		_vaoRendering[i] = ctx->createVertexArrayObject();
		_vaoRendering[i]->create(ctx);

		_vaoRendering[i]->addObject(vertices_attrib);
		_vaoRendering[i]->addObject(vertices_attrib2);
		_vaoRendering[i]->addObject(_indices_buffer);

		_vaoRendering[i]->fill();
	}

	// Physic
	for (int i = 0; i < 2; ++i)
	{
		// Longitude / latitude array
// 		OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 3, _vbo[i]);
// 		OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 2, _vbo[i], OGL::float32, sizeof(float) * 3);
		OGL::VertexAttribArray* vertices_attrib = ctx->createVertexAttribArray(vertices_location, 2, _vbo[i]);

		// "Position in the streamline" array
		unsigned int localID = 2;
		OGL::VertexAttribArray* vertices_attrib2 = ctx->createVertexAttribArray(localID, 1, _vertices_local_id_buffer);

		// Longitude / latitude array with an offset. The current vertex knows the position of his neighbour.
		unsigned int neighbourId = 1;
		OGL::VertexAttribArray* vertices_attrib_next = ctx->createVertexAttribArray(neighbourId, 2, _vbo[i], OGL::float32, 0, sizeof(float) * 2);
		// 		OGL::VertexAttribArray* vertices_attrib_next = ctx->createVertexAttribArray(neighbourId, 2, _vbo[i], OGL::float32, sizeof(float) * 3, sizeof(float) * 3);
		// 		OGL::VertexAttribArray* vertices_attrib_next = ctx->createVertexAttribArray(neighbourId, 3, _vbo[i], OGL::float32, 0, sizeof(float)*3);

		// 		OGL::VertexAttribArray* vertices_time = ctx->createVertexAttribArray(3, 1, _vbo[i], OGL::float32, sizeof(float) * 3, sizeof(float) * 2);

				// Vao creation
		_vaoPhysic[i] = ctx->createVertexArrayObject();
		_vaoPhysic[i]->create(ctx);

		// Fill the vao, cpu side
		_vaoPhysic[i]->addObject(vertices_attrib);
		_vaoPhysic[i]->addObject(vertices_attrib_next);
		// 		_vaoPhysic[i]->addObject(vertices_time);
		_vaoPhysic[i]->addObject(vertices_attrib2);
		_vaoPhysic[i]->addObject(_indices_buffer);

		// bind all objects to fill vao OpenGL side.
		_vaoPhysic[i]->fill();
	}

	// Start all chronos
	_origin.reset();
	getLastUpdateTime().reset();
	getLastTime().reset();
}

void StreamlinesData::setSaveConfig(bool b)
{
	_saveConfig = b;
}

bool StreamlinesData::getSaveConfig() const
{
	return _saveConfig;
}
