// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "3d/style_manager.h"

#include "vendors/_ini_parser.h"
#include "core/workspace_folders.h"
#include "core/log.h"

using namespace SEAScope;

namespace
{
	bool splitList(const std::string& input, const std::string& delim, std::vector<std::string>& output)
	{
		size_t start = 0U;
		auto end = input.find(delim);

		while
			(end != std::string::npos)
		{
			output.push_back(input.substr(start, end - start));
			start = end + delim.length();
			end = input.find(delim, start);
		}
		output.push_back(input.substr(start, end));
		return 0 < output.size();
	};

	template<typename T> T convert(const std::string& /*str*/) { return T(); }

	template<> glm::vec3 convert(const std::string& str)
	{
		glm::vec3 result;
		std::vector<std::string> output;

		if
			(true == splitList(str, " ", output)
				&& output.size() == 3)
		{
			result.x = std::stof(output[0]);
			result.y = std::stof(output[1]);
			result.z = std::stof(output[2]);
		}

		return result;
	}

	template<> float convert(const std::string& str)
	{
		return std::stof(str);
	}

	template<typename T> bool readValue(INI::Level& level, const std::string& name, T& val)
	{
		auto it = level.values.find(name);

		if (it != level.values.end())
		{
			val = convert<T>(level[name]);
			return true;
		}

		return false;
	}
}

bool StyleManager::loadIniFile()
{
	bool ok = false;

	loadDefault();

	std::string stylePath = WorkspaceFolders::getInstance()->getStyleFilePath();

	if (false == stylePath.empty())
	{
		INI::Level root;

		try
		{
			INI::Parser p(stylePath.c_str());
			root = p.top();

			/* paths */
			INI::Level& main = root("main");
			readValue(main, "background_color", this->_params.backgroundColor);

			INI::Level& anchor = root("anchor");
			
			glm::vec3 tmp3;

			// Initialize all anchor color variables with the same value
			if (true == readValue(anchor, "default_color", tmp3))
			{
				this->_params.moveable_normal_color		   = tmp3;
				this->_params.moveable_highlight_color	   = tmp3;
				this->_params.moveable_selected_color	   = tmp3;
				this->_params.add_normal_color			   = tmp3;
				this->_params.add_highlight_color		   = tmp3;
			}
			
			// Initialize all "moveable anchor" color variables with the same value
			if (true == readValue(anchor, "moveable_default_color", tmp3))
			{
				this->_params.moveable_normal_color = tmp3;
				this->_params.moveable_highlight_color = tmp3;
				this->_params.moveable_selected_color = tmp3;
			}

			// Initialize all "add anchor" color variables with the same value
			if (true == readValue(anchor, "add_default_color", tmp3))
			{
				this->_params.add_normal_color = tmp3;
				this->_params.add_highlight_color = tmp3;
			}

			readValue(anchor, "moveable_normal_color", this->_params.moveable_normal_color);
			readValue(anchor, "moveable_highlight_color", this->_params.moveable_highlight_color);
			readValue(anchor, "moveable_selected_color", this->_params.moveable_selected_color);

			readValue(anchor, "add_normal_color", this->_params.add_normal_color);
			readValue(anchor, "add_highlight_color", this->_params.add_highlight_color);

			float val;

			if (true == readValue(anchor, "default_size", val))
			{
				this->_params.moveable_normal_size = val;
				this->_params.moveable_highlight_size = val;
				this->_params.moveable_selected_size = val;
				this->_params.add_normal_size = val;
				this->_params.add_highlight_size = val;
				this->_params.add_selected_size = val;
			}

			if (true == readValue(anchor, "add_default_size", val))
			{
				this->_params.add_normal_size = val;
				this->_params.add_highlight_size = val;
				this->_params.add_selected_size = val;
			}

			if (true == readValue(anchor, "moveable_default_size", val))
			{
				this->_params.moveable_normal_size = val;
				this->_params.moveable_highlight_size = val;
				this->_params.moveable_selected_size = val;
			}


			readValue(anchor, "moveable_normal_size", this->_params.moveable_normal_size);
			readValue(anchor, "moveable_highlight_size", this->_params.moveable_highlight_size);
			readValue(anchor, "moveable_selected_size", this->_params.moveable_selected_size);

			readValue(anchor, "add_normal_size", this->_params.add_normal_size);
			readValue(anchor, "add_highlight_size", this->_params.add_highlight_size);
			readValue(anchor, "add_selected_size", this->_params.add_selected_size);

			INI::Level& polyline = root("polyline");
			readValue(polyline, "normal_color", this->_params.polyline_normal_color);
			readValue(polyline, "highlight_color", this->_params.polyline_highlight_color);
			readValue(polyline, "selected_color", this->_params.polyline_selected_color);

			INI::Level& polygon = root("polygon");
			readValue(polygon, "normal_color", this->_params.polygon_normal_color);
			readValue(polygon, "highlight_color", this->_params.polygon_highlight_color);
			readValue(polygon, "selected_color", this->_params.polygon_selected_color);

			ok = true;
		}
		catch
			(std::runtime_error& e)
		{
			Log::log(LogLevel::Debug, "Style ini parser"
				, "{}"
				, e.what());
		}
	}

	return ok;
}

const StyleParameters& StyleManager::getParams() const
{
	return _params;
}

StyleManager::StyleManager()
{

}

StyleManager::~StyleManager()
{

}

void StyleManager::loadDefault()
{
	this->_params.backgroundColor = glm::vec3(0.1f, 0.1f, 0.1f);
	this->_params.moveable_normal_color = green3f;
	this->_params.moveable_highlight_color = green3f;
	this->_params.moveable_selected_color = green3f;
	this->_params.add_normal_color = green3f;
	this->_params.add_highlight_color = green3f;

	this->_params.moveable_normal_size = 32.f;
	this->_params.moveable_highlight_size = 32.f;
	this->_params.moveable_selected_size = 32.f;
	this->_params.add_normal_size = 32.f;
	this->_params.add_highlight_size = 32.f;
	this->_params.add_selected_size = 32.f;

	this->_params.polyline_normal_color = green3f;
	this->_params.polyline_highlight_color = cyan3f;
	this->_params.polyline_selected_color = orange3f;

	this->_params.polygon_normal_color = yellow3f;
	this->_params.polygon_highlight_color = cyan3f;
	this->_params.polygon_selected_color = orange3f;
}
