// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/api_in.h"

using namespace SEAScope;

namespace
{
	// Similar to std::assign but without warnings
	// Warning with std::assign : 
	// warning C4244: 'argument': conversion from '_Diff' to 'const flatbuffers::uoffset_t', possible loss of data
	template<typename T>
	void flatbuffersVectorToStdVector(const flatbuffers::Vector<T>& fbVector, std::vector<T>& stdVector)
	{
		stdVector.resize(fbVector.size());

		auto itStdVector = stdVector.begin();

		for (auto val : fbVector)
		{
			*itStdVector = val;
			itStdVector++;
		}
	}
}

/**
 *
 */
bool APIIn::renderableId( const API::RenderableId* input
								, RenderableId& output)
{
	output.granuleLevel = input->granuleLevel();
	output.collection.first = input->sourceId();
	output.collection.second = input->collectionId();
	output.granule.first = input->sourceId();
	output.granule.second = input->granuleId();
	output.variable = input->variableId();

	return true;
};

/**
 *
 */
bool APIIn::renderingCfg( const API::RenderingCfg* input
								, RenderingCfg& output)
{
	output._rendered = input->rendered();
	output._logScale = input->logscale();
	output._min = input->min();
	output._max = input->max();
	output._opacity = input->opacity();
	output._zindex = input->zindex();
	output._color[0] = static_cast<uint8_t>(input->color()->r());
	output._color[1] = static_cast<uint8_t>(input->color()->g());
	output._color[2] = static_cast<uint8_t>(input->color()->b());
	output._colormap = input->colormap()->str();
	output._particlesCount = input->particlesCount();
	output._particleTTL = static_cast<uint8_t>(input->particleTTL());
	output._streamlineLength = input->streamlineLength();
	output._streamlineSpeed = input->streamlineSpeed();
	APIIn::renderableId(input->target(), output.target);

	auto renderingMethod = input->renderMethod();
	if
		(API::RenderingMethod_RASTER == renderingMethod)
	{
		output._renderMethod = RenderMethod::RASTER;
	}
	else if
		(API::RenderingMethod_ARROWS == renderingMethod)
	{
		output._renderMethod = RenderMethod::ARROWS;
	}
	else if
		(API::RenderingMethod_BARBS == renderingMethod)
	{
		output._renderMethod = RenderMethod::BARBS;
	}
	else if
		(API::RenderingMethod_TRAJECTORIES == renderingMethod)
	{
		output._renderMethod = RenderMethod::TRAJECTORIES;
	}
	else if
		(API::RenderingMethod_DOTSCLOUD == renderingMethod)
	{
		output._renderMethod = RenderMethod::DOTSCLOUD;
	}
	else if
		(API::RenderingMethod_STREAMLINES == renderingMethod)
	{
		output._renderMethod = RenderMethod::STREAMLINES;
	}
	else if
		(API::RenderingMethod_RAWRGB == renderingMethod)
	{
		output._renderMethod = RenderMethod::RAWRGB;
	}

	auto filterMode = input->filterMode();
	if
		(API::FilterMode_NEAREST == filterMode)
	{
		output._filterMode = FilterMode::NEAREST;
	}
	else if
		(API::FilterMode_BILINEAR == filterMode)
	{
		output._filterMode = FilterMode::BILINEAR;
	}

	return true;
};

/**
 *
 */
bool APIIn::variable(const API::Variable* input, VariableCfg& output)
{
	output.label = input->label()->str();
	output.units = input->units()->str();

	auto _fields = input->fields();
	auto _fieldsLength = _fields->Length();
	for
		(unsigned int i = 0; i < _fieldsLength; ++i)
	{
		output.fields.insert(_fields->Get(i)->str());
	}

	auto _tags = input->tags();
	auto _tagsLength = _tags->Length();
	for
		(unsigned int i = 0; i < _tagsLength; ++i)
	{
		auto _tag = _tags->Get(i);
		output.tags[_tag->name()->str()] = _tag->value()->str();
	}

	return true;
};

/**
 *
 */
bool APIIn::collection( const API::Collection* input
							  , CollectionCfg& output)
{
	output.label = input->label()->str();
	output.defaultVariable = input->defaultVariable();
	output.mustBeCurrent = input->mustBeCurrent();
	output.xSeamless = input->xSeamless();
	output.ySeamless = input->ySeamless();
	output.NEWSAligned = input->newsAligned();

	auto tags = input->tags();
	unsigned int tagsLength = static_cast<unsigned int>(tags->Length());
	for
		(unsigned int i = 0; i < tagsLength; ++i)
	{
		auto t = tags->Get(i);
		output.tags[t->name()->str()] = t->value()->str();
	}

	auto variables = input->variables();
    
    if (nullptr != variables)
    {
        unsigned int variablesLength = static_cast<unsigned int>(variables->Length());
        for
            (unsigned int i = 0; i < variablesLength; ++i)
        {
            auto v = variables->Get(i);

            output.variables[v->id()];
            VariableCfg& vCfg =  output.variables[v->id()];
            APIIn::variable(v, vCfg);
        }
    }

	return true;
};

/**
 *
 */
bool APIIn::shape(const API::Shape* input, Shape& output)
{
	output.xArity = static_cast<std::uint16_t>(input->xArity());
	output.yArity = static_cast<std::uint16_t>(input->yArity());

	auto _gcps = input->gcps();
	auto gcpsLength = _gcps->Length();
	for
		(unsigned int i = 0; i < gcpsLength; ++i)
	{
		auto _gcp = _gcps->Get(i);
		output.gcps.emplace_back( _gcp->lat(), _gcp->lon()
								, static_cast<uint16_t>(_gcp->i()), static_cast<uint16_t>(_gcp->j()));
	}

	return true;
};

/**
 *
 */
bool APIIn::idfDescriptor( const API::IDFDescriptor* input
								 , IDFDescriptor& output)
{
	output.uri = input->uri()->str();
	output.resolution = input->resolution();
    output.subsamplingFactor = static_cast<std::int32_t>(input->subsamplingFactor());
	output.xArity = input->xArity();
	output.yArity = input->yArity();

    Log::log( LogLevel::Debug, "APIIn"
            , "IDF descriptor: {} {} {} {} {}"
            , output.uri, output.resolution, output.subsamplingFactor
            , output.xArity, output.yArity);
	auto _shape = input->shape();
	APIIn::shape(_shape, output.shape);

	return true;
};

/**
 *
 */
bool APIIn::granuleMetadata( const API::GranuleMetadata* input
								   , GranuleMetadata& output)
{
	output.id.first = input->sourceId();
	output.id.second = input->granuleId();
	output.dataId = input->dataId()->str();
	output.collectionId = input->collectionId();
	output.start = input->start();
	output.stop = input->stop();

	auto _dataModel = input->dataModel();

	if
		(SEAScope::API::DataModel_LAT_LON == _dataModel)
	{
		output.dataModel = DataModel::LAT_LON;
	}
	else if
		(SEAScope::API::DataModel_Y_X == _dataModel)
	{
		output.dataModel = DataModel::Y_X;
	}
	else if
		(SEAScope::API::DataModel_NJ_NI == _dataModel)
	{
		output.dataModel = DataModel::NJ_NI;
	}
	else if
		(SEAScope::API::DataModel_ROW_CELL == _dataModel)
	{
		output.dataModel = DataModel::ROW_CELL;
	}
	else if
		(SEAScope::API::DataModel_TIME_STATION == _dataModel)
	{
		output.dataModel = DataModel::TIME_STATION;
	}
	else if
		(SEAScope::API::DataModel_TIME == _dataModel)
	{
		output.dataModel = DataModel::TIME;
	}

	auto uris = input->uris();
	auto urisLength = uris->Length();
	for
		(unsigned int i = 0; i < urisLength; ++i)
	{
		IDFDescriptor idf;	//

		auto _idf = uris->Get(i);
		APIIn::idfDescriptor(_idf, idf);
		output.uris.push_back(idf);
	}

    /* Handle optional attributes */
    output.availableAttributes.reset();

    if
        (true == input->hasTitle())
    {
        output.title = input->title()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::TITLE), 1);
    }

    if
        (true == input->hasInstitution())
    {
        output.institution = input->institution()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::INSTITUTION), 1);
    }

    if
        (true == input->hasComment())
    {
        output.comment = input->comment()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::COMMENT), 1);
    }

    if
        (true == input->hasFileId())
    {
        output.fileId = input->fileId()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::FILE_ID), 1);
    }

    if
        (true == input->hasProductVersion())
    {
        output.productVersion = input->productVersion()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::PRODUCT_VERSION), 1);
    }

    if
        (true == input->hasLatMin())
    {
        output.latMin = input->latMin();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::LAT_MIN), 1);
    }

    if
        (true == input->hasLatMax())
    {
        output.latMax = input->latMax();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::LAT_MAX), 1);
    }

    if
        (true == input->hasLonMin())
    {
        output.lonMin = input->lonMin();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::LON_MIN), 1);
    }

    if
        (true == input->hasLonMax())
    {
        output.lonMax = input->lonMax();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::LON_MAX), 1);
    }

    if
        (true == input->hasCreatorEmail())
    {
        output.creatorEmail = input->creatorEmail()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::CREATOR_EMAIL), 1);
    }

    if
        (true == input->hasStationId())
    {
        output.stationId = input->stationId()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::STATION_ID), 1);
    }

    if
        (true == input->hasPlatform())
    {
        output.platform = input->platform()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::PLATFORM), 1);
    }

    if
        (true == input->hasSensor())
    {
        output.sensor = input->sensor()->str();
        output.availableAttributes.set(
            static_cast<std::size_t>(IDFAttributes::SENSOR), 1);
    }

    Log::log(LogLevel::Debug, "APIIn"
            , "{}/{}/{} {} [{}, {}] {} | {} | {} | {} | {} | {} {} {} {} | {} > {} {} {}"
            , output.id.first, output.collectionId, output.id.second
            , output.dataId, output.start, output.stop
            , output.title, output.institution, output.comment
            , output.fileId, output.productVersion
            , output.latMin, output.latMax, output.lonMin, output.lonMax
            , output.creatorEmail, output.stationId, output.platform, output
            .sensor);

	return true;
};

/**
 *
 */
bool APIIn::dataInfo( const API::DataInfo* input
							, DataInfo& output)
{
	output.channels = static_cast<std::uint8_t>(input->channels());
	output.xArity = static_cast<std::uint16_t>(input->xArity());
	output.yArity = static_cast<std::uint16_t>(input->yArity());

	auto _dataType = input->dataType();
	if
		(API::DataType_NONE == _dataType)
	{
		output.dataType = OGL::DataType::none;
	}
	else if
		(API::DataType_UBYTE == _dataType)
	{
		output.dataType = OGL::unsigned_byte;
	}
	else if
		(API::DataType_USHORT == _dataType)
	{
		output.dataType = OGL::unsigned_short;
	}
	else if
		(API::DataType_UINT == _dataType)
	{
		output.dataType = OGL::unsigned_int;
	}
	else if
		(API::DataType_FLOAT32 == _dataType)
	{
		output.dataType = OGL::float32;
	}

	flatbuffersVectorToStdVector(*input->offsets(), output.offsets);
	flatbuffersVectorToStdVector(*input->scaleFactors(), output.scaleFactors);
	flatbuffersVectorToStdVector(*input->fillValues(), output.fillValues);

	return true;
};

/**
 *
 */
bool APIIn::dataBucket( const API::DataBucket* input
							  , DataBucket& output)
{
	APIIn::dataInfo(input->info(), output.info);
	output.buffer.allocate( output.info.xArity
						  , output.info.yArity
						  , output.info.channels
						  , output.info.dataType
						  , reinterpret_cast<void*>(const_cast<unsigned int*>(input->buffer()->data())));

	return true;
};

/**
 *
 */
bool APIIn::granuleData( const API::GranuleData* input
							   , GranuleData& output)
{
	auto _fields = input->fields();
	auto _fieldsLength = _fields->Length();
	auto _buckets = input->buckets();
	for
		(unsigned int i = 0; i < _fieldsLength; ++i)
	{
		auto _field = _fields->Get(i);
		auto _bucket = _buckets->Get(i);

		output[_field->str()];
		DataBucket& bucket = output[_field->str()];
		APIIn::dataBucket(_bucket, bucket);
	}

	return true;
};
