// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <array>
#include <sstream>

#include "core/application.h"
#include "core/path.h"
#include "core/log.h"

#include "core/dynamic.h"
#include "core/interfaces/icatalogue_service_connector.h"
#include "core/interfaces/iindex_service_connector.h"
#include "core/interfaces/idatastream_service_connector.h"
#include "core/interfaces/iprotocol_handler.h"
#include "core/control_protocol_handler.h"
#include "core/workspace_folders.h"

#include "3d/data_set.h"
#include "3d/datasets_allocator.h"
#include "3d/scene_view.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"

#include "core/app_events_dispatcher.h"

using namespace SEAScope;

/**
 * Constructor.
 */
Application::Application(AppEventsDispatcher& appEventsDispatcher)
: settingsStore(nullptr),
  stateStore(nullptr),
  window(nullptr),
  engine(appEventsDispatcher)
{

};

/**
 * Destructor.
 */
Application::~Application()
{
};

/**
 *
 */
bool Application::init( const std::string& installDir
					  , const std::string& configPath
					  , ISettingsStore& settingsStoreIn
					  , IStateStore& stateStoreIn
					  , IWindow& windowIn
					  , RendererResources& resources
					  , const std::string& logFilter)
{
	bool ok;	//

    this->settingsStore = &settingsStoreIn;
	this->stateStore = &stateStoreIn;
	this->window = &windowIn;

	ok = this->loadSettings(configPath, logFilter);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not load settings from {}", this->settingsPath);
		return false;
	}

	ok = this->initWorkspace(installDir);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Could not initialize workspace");
		return false;
	}

	ok = this->initWorkers();
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not initialize workers");
		return false;
	}

	ok = this->initRendererResources(installDir, resources);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not initialize renderer resources");
		return false;
	}

	ok = this->initWindow();
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not initialize window");
		return false;
	}

	return ok;
};

/**
 *
 */
bool Application::initRendererResources(const std::string& installDir,
										RendererResources& resources) const
{
	std::vector<std::string> shadersDirElems;	//
	std::vector<std::string> windBarbsDirElems;	//
	std::vector<std::string> arrowsDirElems;	//
	std::vector<std::string> worldmapPathElems;	//

	shadersDirElems.push_back(installDir);
	shadersDirElems.push_back(std::string("GLSL"));
	shadersDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(shadersDirElems, resources.shadersDir);

	windBarbsDirElems.push_back(installDir);
	windBarbsDirElems.push_back(std::string("images"));
	windBarbsDirElems.push_back(std::string("windBarbs"));
	windBarbsDirElems.push_back(std::string("pgm"));
	windBarbsDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(windBarbsDirElems, resources.windBarbsDir);

	arrowsDirElems.push_back(installDir);
	arrowsDirElems.push_back(std::string("images"));
	arrowsDirElems.push_back(std::string("arrows"));
	arrowsDirElems.push_back(std::string("pgm"));
	arrowsDirElems.push_back(std::string("")); // required to keep the trailing slash
	Path::joinPaths(arrowsDirElems, resources.arrowsDir);

	if
		("highres" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("earth_highres.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else if
		("lowres" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("earth_lowres.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else if
		("black" == this->engine.settings.worldMap)
	{
		worldmapPathElems.push_back(installDir);
		worldmapPathElems.push_back(std::string("images"));
		worldmapPathElems.push_back(std::string("black.png"));
		Path::joinPaths(worldmapPathElems, resources.worldmapPath);
	}
	else
	{
		resources.worldmapPath = this->engine.settings.worldMap;
	}

	Log::log(LogLevel::Debug, "Application", "SHADERS DIR: {}~", resources.shadersDir);
	Log::log(LogLevel::Debug, "Application", "ARROWS DIR: {}#", resources.arrowsDir);
	Log::log(LogLevel::Debug, "Application", "WIND BARBS DIR: {}#", resources.windBarbsDir);
	return true;
};

/**
 *
 */
bool Application::loadSettings(const std::string& configPath, const std::string& logFilter)
{
	bool ok;

	this->settingsPath = configPath;
	ok = this->settingsStore->load(this->engine.settings, this->settingsPath);

	this->engine.settings.logFilter = logFilter;

	return ok;
};

/**
 *
 */
bool Application::initWorkspace(const std::string& installDir)
{
	bool ok;
	std::vector<std::string> pathElems;
	std::array<std::string, 4> appDirs;

    if
        (false == WorkspaceFolders::getInstance()->setWorkspacePath(installDir))
    {
        return false;
    }

	/* Set default values for paths */
	if
		(0 >= this->engine.settings.pluginsPath.length())
	{
		Path::joinPaths({installDir, "plugins"},
						this->engine.settings.pluginsPath);
	}

	if
		(0 >= this->engine.settings.colormapsPath.length())
	{
		Path::joinPaths({installDir, "colormaps"},
						this->engine.settings.colormapsPath);
	}

	if
		(0 >= this->engine.settings.operatorsPath.length())
	{
		Path::joinPaths({installDir, "operators"},
						this->engine.settings.operatorsPath);
	}

	if
		(0 >= this->engine.settings.dataPath.length())
	{
		Path::joinPaths({installDir, "data"},
						this->engine.settings.dataPath);
	}

	/* Normalize paths to avoid issues on some platforms */
	Path::setPathValid(this->engine.settings.pluginsPath);
	Path::setPathValid(this->engine.settings.colormapsPath);
	Path::setPathValid(this->engine.settings.operatorsPath);
	Path::setPathValid(this->engine.settings.dataPath);
	Path::setPathValid(this->engine.settings.indexPath);
	Path::setPathValid(this->engine.settings.statePath);
    Path::setPathValid(this->engine.settings.annotationsPath);

	/* Create application directories if need be */
	appDirs[0] = this->engine.settings.pluginsPath;
	appDirs[1] = this->engine.settings.colormapsPath;
	appDirs[2] = this->engine.settings.operatorsPath;
	appDirs[3] = this->engine.settings.dataPath;

	for
		(const auto& path: appDirs)
	{
		ok = Path::dirExists(path);
		if
			(false == ok)
		{
			ok = Path::createDirs(path, 0755);
			if
				(false == ok)
			{
				Log::log( LogLevel::Error, "Application"
                        , "Could not create application directory '{}'"
						, path);
				return false;
			}
		}
	}

	return true;
}

/**
 *
 */
bool Application::saveSettings()
{
	bool ok;	//

	ok = this->settingsStore->save(this->engine.settings, this->settingsPath);
	return ok;
};

/**
 *
 */
bool Application::loadState()
{
	bool ok;	//

	ok = false;

	ok = this->stateStore->setStatePath(this->engine.settings.statePath);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Could not set state path");
		return false;
	}
	ok = this->stateStore->load(this->engine.state);
	return ok;
};

/**
 *
 */
bool Application::saveState()
{
	bool ok;	//

	ok = this->stateStore->save(this->engine.state);
	return ok;
};

/**
 *
 */
bool Application::initWindow()
{
	bool ok	//
	;

	ok = this->window->init(this->engine.settings);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not initialize window");
		return false;
	}

    ok = this->window->create( this->engine.settings.windowWidth
							 , this->engine.settings.windowHeight);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Application"
				, "Could not create window");
		return false;
	}

	return true;
}

/**
 *
 */
bool Application::run(SceneView& sceneView)
{
    bool ok(false); //
	bool pursue(true);	//
    std::set<RenderableId> rendered;    //
    ISceneRenderer* renderer;   //

    renderer = sceneView.getRenderer();

	while
		(true == pursue)
	{
        if
            (true == this->engine.updateRequested)
        {
            Log::log(LogLevel::Debug, "Application", "Engine update requested");
            this->engine.update();
        }

		while
			(false == this->engine.displayQueue.queue.empty())
		{
			this->engine.displayQueue.stageFirst();
            DisplayJob& j = this->engine.displayQueue.jobOnStage;

			switch
				(j.jobType)
			{
				case DisplayJobType::UPDATE_STARTED:
                    Log::log(LogLevel::Debug, "Application", "Update started");
					renderer->updateStarted(sceneView.getScene());
					renderer->setToRender(j.toRender);
                    rendered = j.toRender;
					break;
				case DisplayJobType::LOAD_DATA:
                    if
                        (rendered.end() != std::find(rendered.begin(),
                                                     rendered.end(),
                                                     j.rId))
                    {
                        if
                            (nullptr != j.dataset)
                        {
                            ok = sceneView.getScene()->addGranule(renderer->getContext(), j.dataset, j.renderingCfg, sceneView.getDataSetVisibility());
                        }
                        else
                        {
                            ok = sceneView.getScene()->addTrajectory(renderer->getContext(), j.points, j.fData, j.mask, j.renderingCfg, j.rId);
                        }

                        if
                            (false == ok)
                        {
                            // Dataset was already in the renderer and this duplicate will not be used, delete it to free memory
                            Log::log(LogLevel::Debug, "Application", "Dataset already in the renderer");
                            DatasetsAllocator::release(j.dataset);
                        }
                    }
                    else
                    {
                        // Dataset will not be used, delete it to free memory
                        Log::log(LogLevel::Debug, "Application", "Dataset not used because it does not need to be rendered");
                        DatasetsAllocator::release(j.dataset);
                    }

                    // Avoid keeping a ref on the dataset that has either been destroyed or passed down to the renderer
                    j.dataset = nullptr;

					break;
				case DisplayJobType::UPDATE_RENDERING:
					renderer->setDisplayCfg(sceneView.getScene(), j.renderingCfg);
					break;
				case DisplayJobType::UPDATE_COMPLETE:
                    Log::log(LogLevel::Debug, "Application", "Update complete");
					renderer->updateFinished(sceneView.getScene());
					break;
                case DisplayJobType::NOOP:
                    break;
			}

			this->engine.displayQueue.unstage();
		}
		pursue = this->window->renderFrame();
	}
	return true;
}

/**
 *
 */
bool Application::initWorkers()
{
	bool ok;	//

	ok = this->dataWorkers.init(1, &this->dataScheduler);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Failed to initialize data workers");
		Log::log(LogLevel::Debug, "Application", "Init parameters: {}", 3);
		return false;
	}

	ok = this->processWorkers.init(5, &this->processScheduler);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Failed to initialize processing workers");
		Log::log(LogLevel::Debug, "Application", "Init parameters: {}", 5);
		return false;
	}

	ok = this->internalWorkers.init(2, &this->internalScheduler);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Failed to initialize internal workers");
		Log::log(LogLevel::Debug, "Application", "Init parameters: {}", 2);
		return false;
	}

	return true;
}

/**
 *
 */
bool Application::shutdown()
{
	bool ok;	//

	ok = this->dataWorkers.shutdown();
	ok = this->processWorkers.shutdown();
	ok = this->internalWorkers.shutdown();

	ok = this->engine.shutdown();

    if
        (nullptr != this->window)
    {
        this->window->clean();
        this->window = nullptr;
    }
	return ok;
}

/**
 *
 */
bool Application::openControlSocket(const std::string& interface, std::uint64_t port)
{
	bool ok;	//

	ok = this->engine.sockets.registerProtocol( "SEASCOPE_CONTROL",
		[this] () -> IProtocolHandler* {
			ControlProtocolHandler* handler;	//

			handler = new ControlProtocolHandler();
			handler->workers = &(this->internalWorkers);
			handler->grunt.engine = &(this->engine);
			return handler;
		}
	);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Failed to register the SEASCOPE_CONTROL protocol");
		return false;
	}

	ok = this->engine.sockets.createServer(interface, static_cast<std::int64_t>(port), "SEASCOPE_CONTROL");
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Application", "Failed to create sockets server on {}:{}", interface, port);
		return false;
	}

	return true;
}
