// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_flatbuffers.h"
#include "core/command_grunt.h"
#include "core/rendering_cfg.h"
#include "core/api_in.h"
#include "core/api_out.h"
#include "core/log.h"

#include "tools/scene_view_controller_base.h"
#include "3d/scene_view.h"
#include "3d/iscene_renderer.h"
#include "core/engine.h"

using namespace SEAScope;

/**
 *
 */
CommandGrunt::CommandGrunt()
: engine(nullptr)
{
};

/**
 *
 */
CommandGrunt::~CommandGrunt()
{
};


/**
 *
 */
std::uint8_t* CommandGrunt::execute( flatbuffers::FlatBufferBuilder& builder
								   , std::uint8_t* buffer
								   , std::uint64_t& resultSize
								   , bool& ok)
{
	std::uint8_t* resultBuffer;	//

    Log::log(LogLevel::Debug, "Command grunt", "Execute:getting command");
    auto cmd = SEAScope::API::GetCommand(buffer);
    Log::log(LogLevel::Debug, "Command grunt", "Execute:command ok");
    auto opcode = cmd->opcode();
    Log::log(LogLevel::Debug, "Command grunt", "Execute:opcode ok");

	resultBuffer = nullptr;
	resultSize = 0;
	ok = false;

    if
        (SEAScope::API::OpCode_setCurrentDatetime == opcode)
    {
        auto args = static_cast<const API::SetCurrentDatetime*>(cmd->args());
		resultBuffer = this->setCurrentDatetime(args, resultSize, ok);
    }
	else if
		(SEAScope::API::OpCode_getCurrentDatetime == opcode)
	{
		resultBuffer = this->getCurrentDatetime(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getRendering == opcode)
	{
		auto args = static_cast<const API::GetRendering*>(cmd->args());
		resultBuffer = this->getRendering(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_setRendering == opcode)
	{
		auto args = static_cast<const API::SetRendering*>(cmd->args());
		resultBuffer = this->setRendering(args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_searchGranules == opcode)
	{
		auto args = static_cast<const API::SearchGranules*>(cmd->args());
		resultBuffer = this->searchGranules(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getExtraction == opcode)
	{
		resultBuffer = this->getExtraction(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_addCollection == opcode)
	{
		auto args = static_cast<const SEAScope::API::AddCollection*>(cmd->args());
		resultBuffer = this->addCollection(args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_addGranule == opcode)
	{
		auto args = static_cast<const SEAScope::API::AddGranule*>(cmd->args());
		resultBuffer = this->addGranule(args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_addVariable == opcode)
	{
		auto args = static_cast<const SEAScope::API::AddVariable*>(cmd->args());
		resultBuffer = this->addVariable(args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getCatalogues == opcode)
	{
		resultBuffer = this->getCatalogues(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_listVisibleData == opcode)
	{
		resultBuffer = this->listVisibleData(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_resetCamera == opcode)
	{
		this->resetCamera(ok);
	}
	else if
		(SEAScope::API::OpCode_zoomIn == opcode)
	{
		this->zoomIn(ok);
	}
	else if
		(SEAScope::API::OpCode_zoomOut == opcode)
	{
		this->zoomOut(ok);
	}
	else if
		(SEAScope::API::OpCode_lookAt == opcode)
	{
		auto args = static_cast<const SEAScope::API::LookAt*>(cmd->args());
		this->lookAt(args, ok);
	}
	else if
		(SEAScope::API::OpCode_setAltitude == opcode)
	{
		auto args = static_cast<const SEAScope::API::SetAltitude*>(cmd->args());
		this->setAltitude(args, ok);
	}
	else if
		(SEAScope::API::OpCode_selectVariable == opcode)
	{
		auto args = static_cast<const SEAScope::API::SelectVariable*>(cmd->args());
		this->selectVariable(args, ok);
	}
	else if
		(SEAScope::API::OpCode_selectVariableByLabel == opcode)
	{
		auto args = static_cast<const SEAScope::API::SelectVariableByLabel*>(cmd->args());
		this->selectVariableByLabel(args, ok);
	}
	else if
		(SEAScope::API::OpCode_getVariableIdentifier == opcode)
	{
		auto args = static_cast<const SEAScope::API::GetVariableIdentifier*>(cmd->args());
		resultBuffer = this->getVariableIdentifier(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getTimestep == opcode)
	{
		resultBuffer = this->getTimestep(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_setTimestep == opcode)
	{
		auto args = static_cast<const SEAScope::API::SetTimestep*>(cmd->args());
		resultBuffer = this->setTimestep(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_listTimesteps == opcode)
	{
		resultBuffer = this->listTimesteps(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getTimespan == opcode)
	{
		resultBuffer = this->getTimespan(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_setTimespan == opcode)
	{
		auto args = static_cast<const SEAScope::API::SetTimespan*>(cmd->args());
		resultBuffer = this->setTimespan(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_listTimespans == opcode)
	{
		resultBuffer = this->listTimespans(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getNearestMode == opcode)
	{
		resultBuffer = this->getNearestMode(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_setNearestMode == opcode)
	{
		auto args = static_cast<const SEAScope::API::SetNearestMode*>(cmd->args());
		resultBuffer = this->setNearestMode(builder, args, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_getDataAlwaysVisibleMode == opcode)
	{
		resultBuffer = this->getDataAlwaysVisibleMode(builder, resultSize, ok);
	}
	else if
		(SEAScope::API::OpCode_setDataAlwaysVisibleMode == opcode)
	{
		auto args = static_cast<const SEAScope::API::SetDataAlwaysVisibleMode*>(cmd->args());
		resultBuffer = this->setDataAlwaysVisibleMode(builder, args, resultSize, ok);
	}

	if
		(0 < resultSize)
	{
		Log::log( LogLevel::Debug, "Command grunt"
				, "Result built: {} bytes to send", resultSize);
	}

    return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::setCurrentDatetime( const API::SetCurrentDatetime* args
											  , std::uint64_t& resultSize
											  , bool& ok)
{
	auto timestamp = args->timestamp();
	this->engine->timeline.model.setCurrent(timestamp);

	ok = true;
	resultSize = 0;
	return nullptr;
};

/**
 *
 */
std::uint8_t* CommandGrunt::setRendering( const API::SetRendering* args
										, std::uint64_t& resultSize, bool& ok)
{
	RenderingCfg* renderingCfg;	//
	RenderingCfg* trueCfg;	//

	auto rCfgAPI = args->cfg();

	renderingCfg = new RenderingCfg();
	APIIn::renderingCfg(rCfgAPI, *renderingCfg);

	ok = false;
	RenderingControlModel& rModel = this->engine->renderingControl.model;
	if
		(renderingCfg->target.granuleLevel)
	{
		trueCfg = rModel.getGranuleCfg( renderingCfg->target.granule
									  , renderingCfg->target.variable
									  , ok);
	}
	else
	{
		trueCfg = rModel.getCollectionCfg( renderingCfg->target.collection
										 , renderingCfg->target.variable
										 , ok);
	}

	if
		(true == ok)
	{
		renderingCfg->applyTo(trueCfg);
		delete renderingCfg; /* will not be used anymore */
		renderingCfg = nullptr;
	}
	else
	{
		Log::log( LogLevel::Error, "Command grunt"
				, "Could not find the true cfg for the renderable, registering the one passed with the command");
		rModel.registerCfg(renderingCfg);
		rModel.cfgChanged( renderingCfg, RenderingParam::ANY);
	}

	ok = true;
	resultSize = 0;
	return nullptr;
};

/**
 *
 */
std::uint8_t* CommandGrunt::getRendering( flatbuffers::FlatBufferBuilder& builder
										, const API::GetRendering* args
										, std::uint64_t& resultSize
										, bool& ok)
{
	RenderingCfg* rCfg;	//
	RenderableId rId;	//
	std::uint8_t* resultBuffer;	//

	/* RenderableId */
	auto _rId = args->target();
	APIIn::renderableId(_rId, rId);

	ok = false;
	rCfg = nullptr;
	rCfg = this->engine->renderingControl.model.getCfg(rId.collection,
													   rId.granule,
													   rId.variable,
													   ok);
	Log::log(LogLevel::Debug, "Command grunt", "rCfg retrieved from engine: {}", ok);

	auto _rCfg = APIOut::renderingCfg(builder, *rCfg, ok);
	Log::log(LogLevel::Debug, "Command grunt", "rCfg converted to flatbuffers: {}", ok);

	auto result = SEAScope::API::CreateGetRenderingResponse(builder, _rCfg);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::searchGranules( flatbuffers::FlatBufferBuilder& builder
										  , const API::SearchGranules* args
										  , std::uint64_t& resultSize
										  , bool& ok)
{
	std::map<CollectionId, std::vector<GranuleMetadata> > searchResults;	//
	std::vector<std::pair<CollectionId, bool>> collectionMustBeCurrent;	//
	std::uint8_t* resultBuffer;	//
	CollectionCfg colCfg;	//
	unsigned int i;	//

	/* Parse command */
	Log::log(LogLevel::Debug, "Command grunt", "Parsing searchGranules command");
	auto current = args->current();
	auto start = args->start();
	auto stop = args->stop();
	auto zoom = args->zoom();

	Log::log( LogLevel::Debug, "Command grunt"
			, "Search with time constraints: [{} < {} < {}]"
			, start, current, stop);
	Log::log( LogLevel::Debug, "Command grunt"
			, "Search with spatial constraints: @{}"
			, zoom);

	auto collections = args->collections();
	auto collectionsLength = collections->Length();
	i = 0;
	for
		(; i < collectionsLength; ++i)
	{
		auto collection = collections->Get(i);
		auto sId = collection->sourceId();
		auto cId = collection->collectionId();

		CollectionId colid;
		colid.first = sId;
		colid.second = cId;
		ok = this->engine->metadataMediator.getCollectionCfg(colid, colCfg);
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "Command grunt"
					, "Failed to retrieve collection configuration for {}/{}"
					, sId, cId);
			ok = false;
			// TODO: send an error message to the client.
			return nullptr;
		}
		collectionMustBeCurrent.push_back(std::make_pair(colid, colCfg.mustBeCurrent));
	}

	/* Execute command */
	ok = this->engine->metadataMediator.searchGranules( current
													  , start
													  , stop
													  , zoom
													  , collectionMustBeCurrent
													  , searchResults);

	/* Build response */
	std::vector<flatbuffers::Offset<SEAScope::API::GranuleMetadata> > granulesVector;	//

	for
		(const auto& it : searchResults)
	{
		for
			(const auto& itGranule : it.second)
		{
			auto granule = APIOut::granuleMetadata(builder, itGranule, ok);
			if
				(false == ok)
			{
				/* TODO: send error message */
				resultSize = 0;
				return nullptr;
			}
			granulesVector.push_back(granule);
		}
	}

	auto granules = builder.CreateVector(granulesVector);
	auto result = SEAScope::API::CreateSearchGranulesResponse(builder, granules);
	builder.Finish(result);

	Log::log(LogLevel::Debug, "Command grunt", "Result built");

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::getExtraction( flatbuffers::FlatBufferBuilder& builder
										 , std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	std::set<std::string> fields;	//
	std::vector<flatbuffers::Offset<SEAScope::API::Extraction> > extractionsVector;	//

    Log::log(LogLevel::Debug, "Command grunt", "getExtraction...");
    Log::log(LogLevel::Debug, "Command grunt", "has engine: {}", (nullptr != this->engine));
	for
		(const auto& itExtraction: this->engine->getExtractedData())
	{
        Log::log(LogLevel::Debug, "Command grunt", "extracting layer {}/{}/{}:{}...",
                itExtraction.first.collection.first, itExtraction.first.collection.second,
                itExtraction.first.granule.second, itExtraction.first.variable);

		auto extraction = APIOut::extraction( builder, itExtraction.first
											, itExtraction.second, ok);
		if
			(false == ok)
		{
			/* TODO: send an error message */
            Log::log(LogLevel::Error, "Command grunt", "API ou failed for extraction");
			resultSize = 0;
			return nullptr;
		}
		extractionsVector.push_back(extraction);
	}

	auto extractions = builder.CreateVector(extractionsVector);
	auto result = SEAScope::API::CreateGetExtractionResponse(builder, extractions);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::addCollection( const API::AddCollection* args
										 , std::uint64_t& resultSize
										 , bool& ok)
{
	CollectionCfg colCfg;	//
	std::vector<RenderingCfg*> renderings;	//
	std::uint32_t collectionId;	//

	auto collection = args->collection();

	collectionId = collection->id()->collectionId();
	APIIn::collection(collection, colCfg);

	/* Collections can only be added to the user-generated source, hence 
	   the "1" source identifier */
	CollectionId cId(1u, collectionId);
	this->engine->metadataMediator.collectionCfgLoaded( cId
													  , colCfg
													  , renderings);

	ok = true;
	resultSize = 0;
	return nullptr;
};

/**
 *
 */
std::uint8_t* CommandGrunt::addGranule( const API::AddGranule* args
									  , std::uint64_t& resultSize, bool& ok)
{
	GranuleMetadata metadata;	//
	RenderableId rId;	//

	/* RenderableId */
	auto _rId = args->id();
	APIIn::renderableId(_rId, rId);

	/* GranuleMetadata */
	auto _metadata = args->metadata();
	APIIn::granuleMetadata(_metadata, metadata);

	if
		(0 >= metadata.uris.size())
	{
		/* At least one uri must be provided!! */
		Log::log( LogLevel::Error, "CommandGrunt"
				, "No URI provided in the metadata of the granule to add");
		ok = false;
		resultSize = 0;
		return nullptr;
	}

	/* Granule data */
	GranuleData& gData = this->engine->dataMediator.store->getDataBucket(metadata.uris[0].uri);
	auto _data = args->data();
	APIIn::granuleData(_data, gData);

	/* Granules can only be added to user-generated source */
	this->engine->getUserGenerated()->index.add(metadata, false);

	ok = true;
	resultSize = 0;
	return nullptr;
};

/**
 *
 */
std::uint8_t* CommandGrunt::addVariable( const API::AddVariable* args
									   , std::uint64_t& resultSize, bool& ok)
{
	RenderingCfg* rCfg;	//
	CollectionId cId;	//
	std::set<CollectionId> collections;	//
	RenderableId vId;	//
	VariableCfg vCfg;	//

	auto _var = args->variable();

	APIIn::renderableId(args->collectionId(), vId);
	APIIn::variable(_var, vCfg);

	/* Variables can only be added to the user-generated source, hence 
	   the "1" source identifier */
	cId.first = 1;
	cId.second = vId.collection.second;
	this->engine->metadataMediator.store->addVariable( cId
													 , _var->id()
													 , vCfg);
	collections.insert(cId);
	this->engine->collectionsControl.model.addCollections(collections);

	/* Create a default rendering configuration for the variable at the 
	   collection level. */
	rCfg = new RenderingCfg();
	rCfg->target.granuleLevel = false;
	rCfg->target.collection.first = 1;
	rCfg->target.collection.second = vId.collection.second;
	rCfg->target.variable = vId.variable;
	this->engine->renderingControl.model.registerCfg(rCfg);

	ok = true;
	resultSize = 0;
	return nullptr;
};

/**
 *
 */
std::uint8_t* CommandGrunt::getCatalogues( flatbuffers::FlatBufferBuilder& builder
										 , std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	SourceInfo srcInfo;	//
	CollectionCfg colCfg;	//
	std::vector<flatbuffers::Offset<SEAScope::API::Source> > srcsVector;	//
	std::vector<flatbuffers::Offset<SEAScope::API::Collection> > colsVector;	//

	/* Add source for user-generated ata */
	auto _label = builder.CreateString("User data");
	auto _descr = builder.CreateString("Data generated by the user");
	auto src = SEAScope::API::CreateSource( builder
										  , 1
										  , _label
										  , _descr
										  , {});
	srcsVector.push_back(src);

	/* List sources */
	SourcesControlModel& srcModel = this->engine->sourcesControl.model;
	for
		(const auto& it : srcModel.uris)
	{
		srcModel.getSource(it.second, srcInfo);
		auto srcOut = APIOut::source(builder, it.first, srcInfo, ok);
		if
			(false == ok)
		{
			/* TODO: send an error message */
			resultSize = 0;
			return nullptr;
		}
		srcsVector.push_back(srcOut);
	}

	/* List collections */
	std::set<CollectionId> cols;	//
	this->engine->metadataMediator.store->listCollections(cols);
	for
		(const auto& it: cols)
	{
		this->engine->metadataMediator.store->getCollectionCfg( it
															  , colCfg);
		auto collection = APIOut::collection(builder, it, colCfg, ok);
		if
			(false == ok)
		{
			/* TODO: send an error message */
			resultSize = 0;
			return nullptr;
		}
		colsVector.push_back(collection);
	}

	auto sources = builder.CreateVector(srcsVector);
	auto collections = builder.CreateVector(colsVector);
	auto result = SEAScope::API::CreateGetCataloguesResponse( builder
															, sources
															, collections);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::listVisibleData( flatbuffers::FlatBufferBuilder& builder
										   , std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	std::vector<flatbuffers::Offset<flatbuffers::String> > urisVector;	//
	std::vector<API::RenderableId> idsVector;	//

    SceneView* sceneView = this->engine->getTools()->getSceneViewController()->getSceneView();
	for
		(const auto& it: sceneView->getRenderer()->getToRender())
	{
		idsVector.emplace_back(it.granuleLevel
							 , (it.granuleLevel) ? it.granule.first : it.collection.first
							 , it.collection.second
							 , it.granule.second
							 , it.variable);

		auto found = this->engine->uris.find(it);
		if
			(this->engine->uris.end() == found)
		{
			urisVector.push_back(builder.CreateString(""));
		}
		else
		{
			urisVector.push_back(builder.CreateString(found->second));
		}
	}

	auto uris = builder.CreateVector(urisVector);
	API::RenderableId* _idsVector = idsVector.data();
	auto ids = builder.CreateVectorOfStructs(_idsVector, idsVector.size());
	auto result = SEAScope::API::CreateListVisibleDataResponse( builder
															  , uris
															  , ids);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
void CommandGrunt::resetCamera(bool& ok)
{
    SceneViewControllerBase* controller(nullptr);   //

    controller = this->engine->getTools()->getSceneViewController();
    controller->resetCamera();
	ok = true;
};

/**
 *
 */
void CommandGrunt::zoomIn(bool& ok)
{
	SceneViewControllerBase* controller(nullptr);   //

    controller = this->engine->getTools()->getSceneViewController();
    controller->zoomIn();
	ok = true;
};

/**
 *
 */
void CommandGrunt::zoomOut(bool& ok)
{
	SceneViewControllerBase* controller(nullptr);   //

    controller = this->engine->getTools()->getSceneViewController();
    controller->zoomOut();
	ok = true;
};

/**
 *
 */
void CommandGrunt::lookAt( const API::LookAt* args, bool& ok)
{
	SceneViewControllerBase* controller(nullptr);   //
	float lon;	//
	float lat;	//

	lon = args->lon();
	lat = args->lat();

    controller = this->engine->getTools()->getSceneViewController();
    controller->lookAt(lon, lat);
	ok = true;
};

/**
 *
 */
void CommandGrunt::setAltitude( const API::SetAltitude* args, bool& ok)
{
	SceneViewControllerBase* controller(nullptr);   //
	float altitude;	//

	altitude = args->altitude();

    controller = this->engine->getTools()->getSceneViewController();
    controller->setZoom(altitude);
	ok = true;
};

/**
 *
 */
void CommandGrunt::selectVariable(const API::SelectVariable* args, bool& ok)
{
	RenderableId rId;	//
	CollectionVariable colvar;	//

	/* RenderableId */
	auto _rId = args->id();
	APIIn::renderableId(_rId, rId);
	colvar.first = rId.collection;
	colvar.second = rId.variable;

	bool selected = args->selected();
	bool exclusive = args->exclusive();

	Log::log(LogLevel::Debug, "Command grunt",
             "Job received: select variable {} [{}]", selected, exclusive);
	this->engine->collectionsControl.model.select(colvar, selected, exclusive);

	ok = true;
};

/**
 *
 */
void CommandGrunt::selectVariableByLabel(const API::SelectVariableByLabel* args, bool& ok)
{
	std::string collectionLabel;	//
	std::string variableLabel;	//
	CollectionVariable colvar;	//

	ok = false;
	collectionLabel = args->collectionLabel()->str();
	variableLabel = args->variableLabel()->str();
	ok = this->engine->collectionsControl.model.getCollectionVariableByLabel(collectionLabel,
																			 variableLabel,
																			 colvar);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Command grunt",
				 "Could not find a variable labelled \"{}\" for collection \"{}\"",
				 variableLabel, collectionLabel);
		return;
	}

	bool selected = args->selected();
	bool exclusive = args->exclusive();

	Log::log(LogLevel::Debug, "Command grunt",
             "Job received: select variable {} [{}]", selected, exclusive);
	this->engine->collectionsControl.model.select(colvar, selected, exclusive);
	ok = true;
};

/**
 *
 */
std::uint8_t* CommandGrunt::getVariableIdentifier( flatbuffers::FlatBufferBuilder& builder
												 , const API::GetVariableIdentifier* args
												 , std::uint64_t& resultSize
												 , bool& ok)
{
	std::string collectionLabel;	//
	std::string variableLabel;	//
	CollectionVariable colvar;	//
	std::uint8_t* resultBuffer;	//

	collectionLabel = args->collectionLabel()->str();
	variableLabel = args->variableLabel()->str();
	ok = false;


	/* Parse command */
	collectionLabel = args->collectionLabel()->str();
	variableLabel = args->variableLabel()->str();

	/* Execute command */
	ok = this->engine->collectionsControl.model.getCollectionVariableByLabel(collectionLabel,
																			 variableLabel,
																			 colvar);
	SEAScope::API::RenderableId rId(false, /* granuleLevel */
                                    colvar.first.first, /* source */
                                    colvar.first.second, /* collection */
                                    0, /* granule */
                                    colvar.second); /* variable */

	/* Build response */
	auto result = SEAScope::API::CreateGetVariableIdentifierResponse(builder, &rId);
	builder.Finish(result);

	Log::log(LogLevel::Debug, "Command grunt", "Result built");

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
};

/**
 *
 */
std::uint8_t* CommandGrunt::getTimestep( flatbuffers::FlatBufferBuilder& builder
									   , std::uint64_t& resultSize, bool& ok)
{
	std::string timestepLabel;	//
	timestamp_t timestepValue;	//
	std::uint8_t* resultBuffer;	//

	ok = false;
	timestepValue = this->engine->timeline.model.timestep;
	ok = this->engine->timeline.getTimestepLabel(timestepValue, timestepLabel);
	if
		(false == ok)
	{
		/* TODO: send error message */
		resultSize = 0;
		return nullptr;
	}

	Log::log(LogLevel::Debug, "Command grunt", "Timestep retrieved from engine: {}", ok);

	auto _timestep = APIOut::timestep(builder, timestepLabel, timestepValue, ok);
	if
		(false == ok)
	{
		/* TODO: send error message */
		resultSize = 0;
		return nullptr;
	}

	/* Build response */
	auto result = SEAScope::API::CreateGetTimestepResponse(builder, _timestep);
	builder.Finish(result);

	Log::log(LogLevel::Debug, "Command grunt", "Result built");

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::setTimestep( flatbuffers::FlatBufferBuilder& /*builder*/,
                                         const SEAScope::API::SetTimestep* args,
										 std::uint64_t& resultSize,
										 bool& ok)
{
	std::string timestepLabel;	//

	timestepLabel = args->label()->str();

	auto found = this->engine->timeline.timesteps.find(timestepLabel);
	if
		(this->engine->timeline.timesteps.end() == found)
	{
		// TODO: send error message
		resultSize = 0;
		return nullptr;
	}

	ok = this->engine->timeline.model.setTimestep(found->second);
	if
		(false == ok)
	{
		// TODO: send error message
		resultSize = 0;
		return nullptr;
	}

	resultSize = 0;
	return nullptr;
}

/**
 *
 */
std::uint8_t* CommandGrunt::listTimesteps( flatbuffers::FlatBufferBuilder& builder,
										   std::uint64_t& resultSize,
										   bool& ok)
{
	std::uint8_t* resultBuffer;	//
	std::vector<flatbuffers::Offset<SEAScope::API::Timestep> > timestepsVector;	//

	for
		(const auto& itTimestep : this->engine->timeline.timesteps)
	{
		auto _timestep = APIOut::timestep(builder, itTimestep.first,
                                          itTimestep.second, ok);
		if
			(false == ok)
		{
			/* TODO: send an error message */
            Log::log(LogLevel::Error, "Command grunt", "API ou failed for extraction");
			resultSize = 0;
			return nullptr;
		}
		timestepsVector.push_back(_timestep);
	}

	auto timesteps = builder.CreateVector(timestepsVector);
	auto result = SEAScope::API::CreateListTimestepsResponse(builder, timesteps);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::getCurrentDatetime( flatbuffers::FlatBufferBuilder& builder
									          , std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	timestamp_t currentDatetime;

	currentDatetime = this->engine->timeline.model.current;
	/* Build response */
	auto result = SEAScope::API::CreateGetCurrentDatetimeResponse(builder, currentDatetime);
	builder.Finish(result);

	Log::log(LogLevel::Debug, "Command grunt", "Result built");

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::getTimespan( flatbuffers::FlatBufferBuilder& builder
									   , std::uint64_t& resultSize, bool& ok)
{
	std::string timespanLabel;	//
	timestamp_t pastOffset;	//
	timestamp_t futureOffset;	//
	std::uint8_t* resultBuffer;	//

	ok = false;
	pastOffset = this->engine->timeline.model.timeframePastOffset;
	futureOffset = this->engine->timeline.model.timeframeFutureOffset;
	ok = this->engine->timeline.getTimespanLabel(pastOffset, futureOffset,
												 timespanLabel);
	if
		(false == ok)
	{
		/* TODO: send error message */
		resultSize = 0;
		return nullptr;
	}

	Log::log(LogLevel::Debug, "Command grunt", "Timespan retrieved from engine: {}", ok);

	auto _timespan = APIOut::timespan(builder, timespanLabel, pastOffset, futureOffset, ok);
	if
		(false == ok)
	{
		/* TODO: send error message */
		resultSize = 0;
		return nullptr;
	}

	/* Build response */
	auto result = SEAScope::API::CreateGetTimespanResponse(builder, _timespan);
	builder.Finish(result);

	Log::log(LogLevel::Debug, "Command grunt", "Result built");

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::setTimespan( flatbuffers::FlatBufferBuilder& /*builder*/,
                                         const SEAScope::API::SetTimespan* args,
										 std::uint64_t& resultSize,
										 bool& ok)
{
	std::string timespanLabel;	//

	timespanLabel = args->label()->str();

	auto found = this->engine->timeline.timeframes.find(timespanLabel);
	if
		(this->engine->timeline.timeframes.end() == found)
	{
		// TODO: send error message
		resultSize = 0;
		return nullptr;
	}

	ok = this->engine->timeline.model.setTimeframeOffsets(found->second[0],
														  found->second[1]);
	if
		(false == ok)
	{
		// TODO: send error message
		resultSize = 0;
		return nullptr;
	}

	resultSize = 0;
	return nullptr;
}

/**
 *
 */
std::uint8_t* CommandGrunt::listTimespans( flatbuffers::FlatBufferBuilder& builder,
										   std::uint64_t& resultSize,
										   bool& ok)
{
	std::uint8_t* resultBuffer;	//
	std::vector<flatbuffers::Offset<SEAScope::API::Timespan> > timespansVector;	//

	for
		(const auto& itTimespan : this->engine->timeline.timeframes)
	{
		auto _timespan = APIOut::timespan(builder, itTimespan.first,
                                          itTimespan.second[0],
										  itTimespan.second[1], ok);
		if
			(false == ok)
		{
			/* TODO: send an error message */
            Log::log(LogLevel::Error, "Command grunt", "API out failed for timespan");
			resultSize = 0;
			return nullptr;
		}
		timespansVector.push_back(_timespan);
	}

	auto timespans = builder.CreateVector(timespansVector);
	auto result = SEAScope::API::CreateListTimespansResponse(builder, timespans);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::getNearestMode(flatbuffers::FlatBufferBuilder& builder,
                                           std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	bool active;	//
	active = this->engine->timeline.model.nearestMode;

	auto result = SEAScope::API::CreateGetNearestModeResponse(builder, active);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::setNearestMode(flatbuffers::FlatBufferBuilder& /*builder*/,
                                           const SEAScope::API::SetNearestMode* args,
                                           std::uint64_t& resultSize, bool& ok)
{
	ok = this->engine->timeline.view->setNearestMode(args->active());
	ok = this->engine->timeline.updateOptions();
	resultSize = 0;
	return nullptr;
}

/**
 *
 */
std::uint8_t* CommandGrunt::getDataAlwaysVisibleMode(flatbuffers::FlatBufferBuilder& builder,
                                                     std::uint64_t& resultSize, bool& ok)
{
	std::uint8_t* resultBuffer;	//
	bool active;	//
	active = this->engine->timeline.model.nearestInDayMode;

	auto result = SEAScope::API::CreateGetDataAlwaysVisibleModeResponse(builder, active);
	builder.Finish(result);

	resultBuffer = builder.GetBufferPointer();
	resultSize = builder.GetSize();
	ok = true;
	return resultBuffer;
}

/**
 *
 */
std::uint8_t* CommandGrunt::setDataAlwaysVisibleMode(flatbuffers::FlatBufferBuilder& /*builder*/,
                                                     const SEAScope::API::SetDataAlwaysVisibleMode* args,
                                                     std::uint64_t& resultSize, bool& ok)
{
	ok = this->engine->timeline.view->setNearestInDayMode(args->active());
	ok = this->engine->timeline.updateOptions();
	resultSize = 0;
	return nullptr;
}
