// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <memory>
#include <cstring>

#include "core/log.h"
#include "core/job.h"
#include "core/control_protocol_handler.h"
#include "core/engine.h"

#include "flatbuffers/flatbuffers.h"

using namespace SEAScope;

/**
 *
 */
ControlProtocolHandler::ControlProtocolHandler()
: msgSize(0), cumulatedRead(0), buffer(nullptr), workers(nullptr)
{
};

/**
 *
 */
ControlProtocolHandler::~ControlProtocolHandler()
{
};

/**
 *
 */
bool ControlProtocolHandler::parseStream(socket_t sock, std::uint8_t* stream, std::uint64_t offset, std::uint64_t available, std::uint64_t& consumed)
{
	std::allocator<std::uint8_t> allocator;	//
	std::string jobId;	//
	Job job;	//
	bool ok;	//

	consumed = 0;
	if
		(nullptr == this->buffer)
	{
		Log::log(LogLevel::Debug, "Control protocol handler", "Create new message: +{} /{}", offset, available);

		/* Read message size: needs at least 8 bytes to read a uint64_t */
		if
			(8 > available)
		{
			consumed = 0;
			return false;
		}

		this->msgSize = static_cast<std::uint64_t>(stream[offset+7]) | \
						static_cast<std::uint64_t>(stream[offset+6]) << 8 | \
						static_cast<std::uint64_t>(stream[offset+5]) << 16 | \
						static_cast<std::uint64_t>(stream[offset+4]) << 24 | \
						static_cast<std::uint64_t>(stream[offset+3]) << 32 | \
						static_cast<std::uint64_t>(stream[offset+2]) << 40 | \
						static_cast<std::uint64_t>(stream[offset+1]) << 48 | \
						static_cast<std::uint64_t>(stream[offset+0]) << 56;
		this->cumulatedRead = 0;
		this->buffer = allocator.allocate(this->msgSize);
		consumed += 8;
		Log::log(LogLevel::Debug, "Control protocol handler", "Message size: {}", this->msgSize);
	}

	if
		(available - consumed < this->msgSize - this->cumulatedRead)
	{
		/* Read all remaining bytes in the buffer */
		memcpy(&(this->buffer[this->cumulatedRead]), &(stream[offset + consumed]), available - consumed);
		this->cumulatedRead += available - consumed;
		consumed = available;
	}
	else
	{
		memcpy(&(this->buffer[this->cumulatedRead]), &(stream[offset + consumed]), this->msgSize - this->cumulatedRead);
		consumed += this->msgSize - this->cumulatedRead;

		Log::log(LogLevel::Debug, "Control protocol handler", "Message complete: {}", this->buffer);
		Log::log(LogLevel::Debug, "Control protocol handler", "Workers? {}", nullptr!=this->workers);

		/* TODO: send message to job queue (let workers perform heavy tasks) */
		if
			(nullptr != this->workers)
		{
			std::uint8_t* bfr;	//
			std::uint64_t bfrSize;	//

			bfr = this->buffer;
			bfrSize = this->msgSize;

			Log::log(LogLevel::Debug, "Control protocol handler", "Creating job...");
			job.name = "parseControlCommand";
			job.task = [bfr, bfrSize, sock, this] (std::string& /*s*/) -> bool {
				std::allocator<std::uint8_t> localAllocator;	//
				std::uint8_t* result;	//
				std::uint64_t resultSize;	//
				bool success;	//

				Log::log(LogLevel::Debug, "Control protocol handler: worker", "Message complete from {}: [{}] {}", sock, bfrSize, bfr);

				/* Execute command */
				flatbuffers::FlatBufferBuilder builder;
				result = this->grunt.execute(builder, bfr, resultSize, success);
				if
					(false == success)
				{
					Log::log(LogLevel::Error, "Control protocol handler: worker", "Execution failed");
					// TODO: cleanup
					return false;
				}

				/* Send data back [TEST] */
				if
					(0 < resultSize)
				{
					Log::log(LogLevel::Debug, "Control protocol handler", "Result ready to be sent: {} bytes [{}]", resultSize, nullptr == result);

					success = this->sendData(sock, result, resultSize);
					if
						(false == success)
					{
						Log::log(LogLevel::Error, "Control protocol handler", "Failed to send data");
						// TODO: cleanup memory
						return false;
					}
					// TODO: cleanup memory
				}

				/* Free the buffer */
				localAllocator.deallocate(bfr, bfrSize);
				return true;
			};

			Log::log(LogLevel::Debug, "Control protocol handler", "Submitting job...");
			ok = this->workers->submitJob(job, jobId);
			if
				(false == ok)
			{
				Log::log( LogLevel::Error, "Control protocol handler"
						, "Failed to submit message parsing job");
				/*TODO: find a way to handle this error: discarding the message
				is not an option since it could have side-effects */
				return false;
			}
			job.id = jobId;
			Log::log(LogLevel::Debug, "Control protocol handler", "Job submitted: {}", jobId);
		}
		else
		{
			Log::log(LogLevel::Debug, "Control protocol handler", "!!!! WORKERS IS NULL !!!!");
		}


		this->buffer = nullptr; // Stream content will feed another message buffer

		return true;
	}

	return false;
};

/**
 *
 */
bool ControlProtocolHandler::sendData(socket_t sock, std::uint8_t* inBuffer, std::uint64_t dataSize)
{
	std::uint8_t dataSizeBytes[8];	//
    bool ok;    //

	dataSizeBytes[0] = static_cast<std::uint8_t>(dataSize & 0x00000000000000ff);
	dataSizeBytes[1] = static_cast<std::uint8_t>((dataSize & 0x000000000000ff00) >> 8);
	dataSizeBytes[2] = static_cast<std::uint8_t>((dataSize & 0x0000000000ff0000) >> 16);
	dataSizeBytes[3] = static_cast<std::uint8_t>((dataSize & 0x00000000ff000000) >> 24);
	dataSizeBytes[4] = static_cast<std::uint8_t>((dataSize & 0x000000ff00000000) >> 32);
	dataSizeBytes[5] = static_cast<std::uint8_t>((dataSize & 0x0000ff0000000000) >> 40);
	dataSizeBytes[6] = static_cast<std::uint8_t>((dataSize & 0x00ff000000000000) >> 48);
	dataSizeBytes[7] = static_cast<std::uint8_t>((dataSize & 0xff00000000000000) >> 56);

    ok = this->grunt.engine->sockets.monitor->sendData(sock, &(dataSizeBytes[0]), 8);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Control protocol handler", "Failed to send data size through the socket");
        return false;
    }

    ok = this->grunt.engine->sockets.monitor->sendData(sock, inBuffer, dataSize);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Control protocol handler", "Failed to send data through the socket");
        return false;
    }

    return true;
};

/**
 *
 */
bool ControlProtocolHandler::reset()
{
	return true;
};
