// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/data_mediator.h"
#include "core/dynamic.h"
#include "core/job.h"
#include "core/utils.h"
#include "core/interfaces/idatastream_service_connector.h"

using namespace SEAScope;

/**
 *
 */
DataMediator::DataMediator()
{
};

/**
 *
 */
DataMediator::~DataMediator()
{
};

/**
 *
 */
bool DataMediator::init(IDataStore* inStore, SourcesControlModel* inSources)
{
	this->store = inStore;
	this->sources = inSources;
	return true;
};

/**
 *
 */
bool DataMediator::loadData( const GranuleId& granuleId
						   , const std::string& uri
						   , const std::set<std::string>& fields
						   , DataModel dataModel
						   , const std::function<void()>& callback)
{
	bool ok;	//
	bool alreadyLoaded;	//
	std::set<std::string> missingFields;	//
	std::string jobId;	//
	DataLoadingJob job;	//

	alreadyLoaded = true;
	for
		(auto& field : fields)
	{
		ok = this->store->hasData(uri, field);
		alreadyLoaded = alreadyLoaded && ok;
		if
			(false == ok)
		{
			missingFields.insert(field);
		}
	}

	/* Create a job and delegate the download task to a worker */
	std::string jobURI(uri);
	GranuleId jobGranule;
	jobGranule.first= granuleId.first;
	jobGranule.second = granuleId.second;
	std::set<std::string> jobMissingFields;
	for
		(auto& it: missingFields)
	{
		jobMissingFields.insert(it);
	}
	DataModel jobDataModel;
	jobDataModel = dataModel;

	if
		(true == alreadyLoaded)
	{
		job.name = "loadData";
	}
	else
	{
		job.name = "downloadData";
		job.granule = jobGranule;
		job.uri = jobURI;
		job.fields = jobMissingFields;
	}

	job.task = [this, jobGranule, jobURI, jobMissingFields, jobDataModel, callback, alreadyLoaded] (std::string& /*s*/) -> bool {
		dynamic<IDatastreamServiceConnector>* datastream;	//
		bool success;	//

		Log::log(LogLevel::Debug, "DataMediator", "IN A THREAD");
        Log::log(LogLevel::Debug, "DataMediator", "Already loaded? {}, {}", alreadyLoaded, jobURI);

		success = false;
		if
			(false == alreadyLoaded)
		{
			auto& data = this->store->getDataBucket(jobURI);
			datastream = this->sources->getDatastream(jobGranule.first, success);
			if
				(false == success)
			{
				Log::log( LogLevel::Error, "DataMediator"
						, "Could not find a suitable datastream plugin for source {}"
						, jobGranule.first);
				return false;
			}

			Log::log(LogLevel::Debug, "DataMediator", "Load data from plugin: granule = {}/{}", jobGranule.first, jobGranule.second);
			Log::log(LogLevel::Debug, "DataMediator", "Load data from plugin: uri = {}", jobURI);
			Log::log(LogLevel::Debug, "DataMediator", "Load data from plugin: missing fields = {}", jobMissingFields.size());
			Log::log(LogLevel::Debug, "DataMediator", "Load data from plugin: model = {}", static_cast<int>(jobDataModel));
			success = (*datastream)->loadData(jobGranule, jobURI, jobMissingFields, jobDataModel, data);
			Log::log(LogLevel::Debug, "DataMediator", "Data loading complete? {}", success);
		}
		//this->dataLoaded(jobGranule, jobURI, jobMissingFields);
		if
			(alreadyLoaded || success)
		{
			callback();
		}
		else
		{
			Log::log(LogLevel::Error, "DataMediator", "No extracted data");
		}
		return success;
	};

	ok = this->workers->submitJob(job, jobId);
    if
        (false == ok)
    {
        job.name = "loadData";
            job.task = [this, fields, uri, callback] (std::string& /*s*/) -> bool {
            bool hasAllFields = true;
            for
                (auto& field : fields)
            {
                hasAllFields = hasAllFields && this->store->hasData(uri, field);
                if
                    (false == hasAllFields)
                {
                    return false;
                }
            }

            callback();
            return true;
        };
            ok = this->workers->submitJob(job, jobId);
    }

	job.id = jobId;

	return ok;
};

/**
 *
 */
void DataMediator::dataLoaded(const GranuleId& granuleId, const std::string& uri, const std::set<std::string>& fields)
{
	Log::log(LogLevel::Debug, "DataMediator", "dataLoaded: {}/{}: {} fields @ {}", granuleId.first, granuleId.second, fields.size(), uri);
	for
		(auto& field : fields)
	{
		Log::log(LogLevel::Debug, "DataMediator", " - {}", field);
	}
};
