// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <sstream>
#include "core/log.h"
#include "core/types.h"
#include "core/guid.h"
#include "core/data_scheduler.h"

using namespace SEAScope;

/**
 * Trick to silence warnings about unused variables in a portable way.
 *
 * https://stackoverflow.com/a/39746834
 */
template <typename T>
void trickForPotentiallyUnusedVariable(T&& /*No name*/) { /*Empty*/ }

/**
 *
 */
DataScheduler::DataScheduler()
{
};

/**
 *
 */
DataScheduler::~DataScheduler()
{
};

/**
 *
 */
bool DataScheduler::addJob(Job& job, std::string& jobId)
{
	bool ok;	//
	DataLoadingJob* j;	//
	std::ostringstream buffer;	//

	ok = false;
	j = nullptr;

	j = dynamic_cast<DataLoadingJob*>(&job);
	if
		(nullptr == j)
	{
		Log::log(LogLevel::Error, "DataScheduler", "The job is not a DataLoadingJob");
		return false;
	}

	if
		("downloadData" == job.name)
	{
		buffer << j->granule.first << "/" << j->granule.second << "_";
		buffer << j->uri << "_" ;
		for
			(const auto& field: j->fields)
		{
			buffer << field << "+";
		}
		jobId = buffer.str();

		auto found = std::find( this->downloaded.begin()
							  , this->downloaded.end()
							  , jobId);
		if
			(found != this->downloaded.end())
		{
			return false;
		}

		this->downloaded.push_back(jobId);
	}
	else
	{
		ok = GUID::generate(jobId);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "DataScheduler"
					, "Could not generate GUID for new job");
			return false;
		}
	}

	Log::log( LogLevel::Debug, "Data scheduler"
			, "Add data job: {} => {}", job.name, jobId);

	job.status = JobStatus::PENDING;
	this->jobs.push_back(job);
	return true;
};

/**
 *
 */
bool DataScheduler::removeJob(const std::string& jobId)
{
	auto it = std::remove_if( this->jobs.begin()
				  , this->jobs.end()
				  , [&jobId] (Job j) -> bool { return j.id == jobId; });

    trickForPotentiallyUnusedVariable(it);

	return true;
};

/**
 *
 */
bool DataScheduler::selectJob(Job& job)
{
	if
		(this->jobs.empty())
	{
		return false;
	}

	job = this->jobs[0];
	this->jobs.erase(this->jobs.begin());
	return true;
};

/**
 *
 */
bool DataScheduler::updateJobStatus(const std::string& jobId, const JobStatus& status)
{
	auto found = std::find_if( this->jobs.begin()
							 , this->jobs.end()
							 , [jobId] (Job j) { return j.id == jobId; });

	if
		(  this->jobs.end() == found
		|| found->status == status)
	{
		return false;
	}

	found->status = status;
	return true;
};

/**
 *
 */
bool DataScheduler::listJobs(std::vector<Job>& result)
{
	result = this->jobs;
	return true;
};

/**
 *
 */
bool DataScheduler::listJobs(std::vector<std::string>& result)
{
	for
		(const auto& job : this->jobs)
	{
		result.push_back(job.id);
	}

	return true;
};

/**
 *
 */
std::size_t DataScheduler::countPendingJobs() const
{
	std::int64_t result;	//

	result = std::count_if( this->jobs.begin()
						  , this->jobs.end()
						  , [](Job j) { return JobStatus::PENDING == j.status; });
	return static_cast<std::size_t>(result);
};
