// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>

#include "core/log.h"
#include "core/display_queue.h"
#include "num_utils.h"

using namespace SEAScope;

/**
 *
 */
DisplayQueue::DisplayQueue()
{
};

/**
 *
 */
DisplayQueue::~DisplayQueue()
{
};

/**
 *
 */
bool DisplayQueue::contains(const RenderableId& rId, float resolution)
{
	Log::log(LogLevel::Debug, "DisplayQueue", "contains");
	std::unique_lock<std::mutex> mlock(this->mutex);

    if
        (  DisplayJobType::LOAD_DATA == this->jobOnStage.jobType
        && rId == this->jobOnStage.rId
        && floating_point_equality(resolution, this->jobOnStage.resolution))
    {
        return true;
    }

	for
		(const auto& it: this->queue)
	{
		if
			(  DisplayJobType::LOAD_DATA == it.jobType
			&& it.rId == rId
			&& floating_point_equality(it.resolution, resolution))
		{
			return true;
		}
	}
	return false;
};

/**
 *
 */
bool DisplayQueue::isLoadDataDeprecated(const DisplayJob& j) const
{
    return (DisplayJobType::LOAD_DATA == j.jobType) && (this->toRender.end() == std::find(this->toRender.begin(), this->toRender.end(), j.rId));
}


/**
 *
 */
bool DisplayQueue::addJob(DisplayJob& j)
{
	Log::log(LogLevel::Debug, "DisplayQueue", "addJob");
    if
        (true == this->isLoadDataDeprecated(j))
    {
        Log::log(LogLevel::Debug, "DisplayQueue", "OOO Discard deprecated load data job");
        return false;
    }

    if
        (  DisplayJobType::LOAD_DATA == j.jobType
        && j.rId == this->jobOnStage.rId
        && floating_point_equality(j.resolution, this->jobOnStage.resolution))
    {
        return false;
    }

	std::unique_lock<std::mutex> mlock(this->mutex);

    // Remove jobs that perform the same type of action and target the same
    // renderable
	if
		(  DisplayJobType::LOAD_DATA == j.jobType
		|| DisplayJobType::UPDATE_RENDERING == j.jobType)
	{
        this->queue.erase(std::remove_if(this->queue.begin(), this->queue.end(), [&j](const DisplayJob& job) -> bool {return j.jobType == job.jobType && j.rId == job.rId;}),
                          this->queue.end());
	}

    // Remove deprecated data loading jobs if a new batch has started
    if
        (DisplayJobType::UPDATE_STARTED == j.jobType)
    {
        this->latestBatch = j.batchId;
        this->toRender = j.toRender;
        this->queue.erase(std::remove_if(this->queue.begin(), this->queue.end(), [this](const DisplayJob& tmpJob) -> bool {return this->isLoadDataDeprecated(tmpJob);}), this->queue.end());
    }

	this->queue.push_back(j);
	mlock.unlock();
	this->cond.notify_one();

    return true;
};

/**
 *
 */
DisplayJob& DisplayQueue::stageFirst()
{
    this->pop(this->jobOnStage);
    return this->jobOnStage;
}

/**
 *
 */
void DisplayQueue::unstage()
{
    this->jobOnStage.jobType = DisplayJobType::NOOP;
}
