// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <chrono>

#include "core/log.h"
#include "core/path.h"
#include "core/utils.h"
#include "core/engine.h"
#include "core/app_events_dispatcher.h"
#include "core/measurements.h"
#include "3d/data_set.h"

#include "3d/datasets_allocator.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"
#include "3d/scene_view.h"
#include "3d/scene_graph.h"
#include "3d/scene_graph_controller.h"
#include "3d/polyline_mesh.h"

#include "events/selected_pickables_changed_event.h"
#include "events/selected_pickable_with_info_changed_event.h"
#include "events/center_time_on_granule_request.h"
#include "events/display_granule_info_request.h"
#include "events/annotation_removed_event.h"
#include "events/granule_info_ready_for_display_event.h"
#include "events/remote_pickle_export_request.h"
#include "events/remote_png_export_request.h"
#include "events/extract_from_renderable_id_request.h"
#include "events/extraction_request.h"
#include "events/transect_request.h"
#include "events/collections_updated_event.h"
#include "events/drawing_complete_event.h"
#include "events/add_prepared_annotation_request.h"
#include "events/collection_registration_request.h"
#include "events/granule_added_event.h"
#include "events/rendering_cfg_update_event.h"
#include "events/renderables_changed_event.h"
#include "events/collection_removal_request.h"
#include "events/collection_removed_event.h"
#include "events/variables_added_event.h"
#include "events/variable_selection_request.h"
#include "events/drawing_started_event.h"
#include "events/drawing_complete_event.h"
#include "events/pixel_size_changed_event.h"
#include "events/annotation_time_coverage_updated_event.h"

#include "vendors/_mini.h"

using namespace SEAScope;

/**
 *
 */
Engine::Engine(AppEventsDispatcher& appEventsDispatcher)
: _appEventsDispatcher(appEventsDispatcher),
  maxTextureSize(0), updateRequested(false),
  _tools(appEventsDispatcher),
  _userGenerated(1u, appEventsDispatcher)
{
    this->lastUpdate = std::chrono::high_resolution_clock::now();

    this->_appEventsDispatcher.subscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<CenterTimeOnGranuleRequest>(this);
    this->_appEventsDispatcher.subscribe<DisplayGranuleInfoRequest>(this);
    this->_appEventsDispatcher.subscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<ExtractFromRenderableIdRequest>(this);
    this->_appEventsDispatcher.subscribe<RemotePickleExportRequest>(this);
    this->_appEventsDispatcher.subscribe<RemotePNGExportRequest>(this);
    this->_appEventsDispatcher.subscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.subscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<CollectionRegistrationRequest>(this);
    this->_appEventsDispatcher.subscribe<GranuleAddedEvent>(this);
    this->_appEventsDispatcher.subscribe<CollectionRemovalRequest>(this);
    this->_appEventsDispatcher.subscribe<CollectionsUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<VariableSelectionRequest>(this);
    this->_appEventsDispatcher.subscribe<PixelSizeChangedEvent>(this);
    this->_appEventsDispatcher.subscribe<AnnotationTimeCoverageUpdatedEvent>(this);
};

/**
 *
 */
Engine::~Engine()
{
    this->_appEventsDispatcher.unsubscribe<SelectedPickablesChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CenterTimeOnGranuleRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayGranuleInfoRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ExtractFromRenderableIdRequest>(this);
    this->_appEventsDispatcher.unsubscribe<RemotePickleExportRequest>(this);
    this->_appEventsDispatcher.unsubscribe<RemotePNGExportRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingStartedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DrawingCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRegistrationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleAddedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRemovalRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionsUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<VariableSelectionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<PixelSizeChangedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationTimeCoverageUpdatedEvent>(this);

    this->clearExtractedData();
};


/**
 *
 */
bool Engine::clearExtractedData()
{
	for
		(auto& it: this->_extractedData)
	{
		it.second.buffer.deallocate();
	}
	this->_extractedData.clear();
    this->_extractedLabels.clear();

	return true;
};

/**
 *
 */
bool Engine::initExtractedData()
{
	RenderableId trajId;	//
	VariableCfg varCfg;	//
	std::vector<float> fillValues;	//
	std::vector<std::string> fieldNames;	//
	timestamp_t tStart;	//
	timestamp_t tStop;	//
	bool ok;	//

    SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
	for
		(auto& rId: sceneView->getRenderer()->getToRender())
	{
		Log::log( LogLevel::Debug, "Engine"
				, "Create extraction data for {}/{}/{}:{}"
				, rId.collection.first, rId.collection.second
				, rId.granule.second, rId.variable);

		fillValues.clear();
		fieldNames.clear();

		ok = this->metadataMediator.getVariableCfg( rId.collection
												  , rId.variable
												  , varCfg);


		if
			(false == ok)
		{
			Log::log( LogLevel::Warn, "Engine"
					, "Failed to retrieve rendering configuration for extractable granule");
			continue;
		}

		auto cFound = this->renderables.find({rId.collection});
		if
			(this->renderables.end() == cFound)
		{
			Log::log(LogLevel::Error, "Engine", "Collection not found: {}/{}", rId.collection.first, rId.collection.second);
			continue;
		}

		tStart = 0;
		tStop = 0;
		ok = false;
		for
			(const auto& gMeta : cFound->second)
		{
			if
				(gMeta.id.first == rId.granule.first && gMeta.id.second == rId.granule.second)
			{
				tStart = static_cast<timestamp_t>(0.001 * static_cast<double>(gMeta.start));
				tStop = static_cast<timestamp_t>(0.001 * static_cast<double>(gMeta.stop));
				ok = true;
				Log::log(LogLevel::Debug, "Engine",
						 "Found time range for granule {}/{}: [{}, {}]",
						 rId.granule.first, rId.granule.second, gMeta.start, gMeta.stop);
				break;
			}
		}
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "Engine", "Could not find metadata for granule {}/{}", rId.granule.first, rId.granule.second);
			continue;
		}

		auto uriFound = this->uris.find(rId);
		if
			(this->uris.end() == uriFound)
		{
			continue;
		}
		GranuleData& gData = this->dataMediator.store->getDataBucket(uriFound->second);
		for
			(auto& field: varCfg.fields)
		{
			auto fieldFound = gData.find(field);
			if
				(gData.end() == fieldFound)
			{
				// TODO: check if we should not skip the granule instead
				// of just skipping the field
				continue;
			}

			float unpackedFill = gData[field].info.offsets[0] + gData[field].info.scaleFactors[0] * gData[field].info.fillValues[0];
			fillValues.push_back(unpackedFill);
			fieldNames.push_back(field);
		}

        RenderableId tmpId(rId);
        std::string uri("");
		if
			(RenderMethod::TRAJECTORIES == varCfg.defaultRendering)
		{
            /* Extraction is performed for all granules at once for collections
            rendered as trajectories, only display one entry in the results */
            tmpId.granuleLevel = false;
			tmpId.granule.first = 0;
			tmpId.granule.second = 0;

			auto found = this->_extractedData.find(tmpId);
			if
				(this->_extractedData.end() != found)
			{
				// PolylinePool already registered
                /* There is one pool of polylines per collection and all the
                trajectories from a collection are rendered at the same time.
                The extraction mechanism is similar, so the buffer holding the
                result of the extraction for a trajectory actually contains the
                data extracted from all the trajectories that belong to this
                collection.

                Having the renderable ids of all individual trajectories
                registered in the extraction results would only slow down the
                and consume more resources to store duplicate information, that
                is why only one renderable identifier is registered in the
                extraction results for this collection.
                */
				continue;
			}

            uri = "[all trajectories]";
        }
        else
        {
            uri = uriFound->second;
        }

        this->_extractedData[tmpId];
        this->_extractedData[tmpId].fillValues = fillValues;
        this->_extractedData[tmpId].uri = uri;
        this->_extractedData[tmpId].fields = fieldNames;
        this->_extractedData[tmpId].start = tStart;
        this->_extractedData[tmpId].stop = tStop;


        // Build label for this extraction
        std::forward_list<std::string> pathElements;	//
        std::string pathElement;	//
        std::string colvarLabel("");
        ok = this->getRenderableLabel(tmpId, colvarLabel, "\n");
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "Engine",
                     "Could not build label for collection:variable {}/{}:{}",
                     tmpId.collection.first, tmpId.collection.second, tmpId.variable);
            return false;
        }

		pathElements.clear();
		pathElement = "";
		Path::splitPaths(uri, "/", pathElements);
		for
			(const auto& itPath : pathElements)
		{
			pathElement = itPath;
		}

        this->_extractedLabels[tmpId] = colvarLabel + "\n" + pathElement;
	}

	return true;
};

/**
 *
 */
bool Engine::init(SceneView* sceneView, IMetadataStore* metaStore, IDataStore* dataStore, IPluginsLoader* pluginsLoader, IBlueprintFactory* bpFactory)
{
    std::set<std::string> individualPaths;  //
    bool ok;    //

	if
		(nullptr == sceneView)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Rendering is not possible with a null scene view.");
		return false;
	}

    this->_tools.getSceneViewController()->setSceneView(sceneView);
	this->maxTextureSize = sceneView->getRenderer()->getMaxTextureSize();
    this->state.blueprintFactory = bpFactory;

    /* Initialize metadata mediator */
    ok = this->metadataMediator.init(metaStore, &(this->sourcesControl.model));
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Failed to initialize metadata mediator");
        return false;
    }

    /* Initialize data mediator */
    ok = this->dataMediator.init(dataStore, &(this->sourcesControl.model));
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Failed to initialize data mediator");
        return false;
    }

    /* Initialize plugins */
    this->plugins.init(pluginsLoader);
    ok = this->plugins.discoverPlugins( this->settings.pluginsPath
                                      , individualPaths);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Plugins discovery failed");
        return false;
    }
    dynamic<ICatalogueServiceConnector>::manager = &this->plugins;
    dynamic<IIndexServiceConnector>::manager = &this->plugins;
    dynamic<IDatastreamServiceConnector>::manager = &this->plugins;

    /* Initialize colormaps */
    ok = this->initColormaps();
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Colormaps initialization failed");
        return false;
    }

    /* Initialize sources control */
    ok = this->sourcesControl.model.init(this->plugins, this->settings);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Sources control initialization failed");
        return false;
    }

	/* Metadata callbacks */
	this->metadataMediator.variableCfgAdded = [this] (RenderingCfg* rCfg) -> bool{
        bool success;    //

		Log::log( LogLevel::Debug, "Engine"
				, "Variable configuration added: {}/{}:{}"
				, rCfg->target.collection.first
				, rCfg->target.collection.second
				, rCfg->target.variable);
		success = this->renderingControl.model.registerCfg(rCfg);
        return success;
	};


	/* Set rendering callbacks */
	this->renderingControl.callback = [this] ( const RenderingCfg* rCfg
                                             , const RenderingParam& param) {
		this->updateRendering(rCfg, param);
	};


	this->renderingControl.toLabel = [this] ( const CollectionVariable& colvar
                                            , std::string& label) -> bool {
        bool success;    //
        RenderableId rId;   //

        rId.collection = colvar.first;
        rId.variable = colvar.second;
        success = this->getRenderableLabel(rId, label);
        return success;
	};

    this->renderingControl.getUnits = [this] ( const CollectionVariable& colvar
                                             , std::string& units) -> bool {
        bool success;    //
        RenderableId rId;   //

        rId.collection = colvar.first;
        rId.variable = colvar.second;
        success = this->getRenderableUnits(rId, units);
        Log::log(LogLevel::Debug, "Engine", "-------> get renderable units {} {}",
                 success, units);
        return success;
    };

    // Update variables displayed on the map when the checkboxes status changes
    // in the rendering menu
    this->renderingControl.variableVisibilityChangedCallback = [this] () {
        this->renderablesChanged();
    };

	this->renderingControl.advanced.saveConfigCallback = [this](SEAScope::RenderingCfg* cfg) {

		// Get the config.ini path
		std::string path = (*this->sourcesControl.model.cataloguePlugins.begin())->getPath(cfg->target.collection);

		// First check, the path is not empty
		if (false == path.empty())
		{
			SEAScope::CollectionCfg cCfg;
			std::vector<RenderingCfg*> renderings;

			// Get the CollectionCfg for the variable list
			bool res = (*this->sourcesControl.model.cataloguePlugins.begin())->getCollectionCfg(cfg->target.collection.second, cCfg, renderings);

			// CollectionCfg found
			if (true == res)
			{
				// Write to config.ini
				saveConfig(path, cfg, cCfg);

				// Update the default RenderingCfg
				renderingControl.model.updateDefaultCfg(cfg);
			}
		}
	};

	return true;
};

/**
 *
 */
bool Engine::initColormaps()
{
    bool ok;    //
    std::forward_list<std::string> colormaps;   //

    ok = Path::listFiles(this->settings.colormapsPath, colormaps);
    if
        (false == ok)
    {
        return false;
    }

    for
        (const auto& it : colormaps)
    {
        if
            (  4 >= it.size()
            || ".rgb" != it.substr(it.size() -4, 4))
        {
            /* Not an RGB file */
            continue;
        }
        Log::log(LogLevel::Debug, "Engine", "Load colormap: {}", it);
        ok = this->cmaps.loadColormap(it);
        if
            (false == ok)
        {
            return false;
        }
    }

    return true;
};

bool Engine::getRenderableUnits(const RenderableId& rId, std::string& units)
{
    std::ostringstream buffer; //
    VariableCfg varCfg; //
    bool ok;    //

    ok = this->metadataMediator.getVariableCfg( rId.collection
                                              , rId.variable
                                              , varCfg);
    if
        (false == ok)
    {
        Log::log( LogLevel::Error, "Engine"
                , "Could not retrieve variable cfg for {}/{}:{}"
                , rId.collection.first, rId.collection.second, rId.variable);
        return false;
    }

    units = varCfg.units;

    Log::log(LogLevel::Debug, "Engine", "getUnits for {}/{}:{}  => {}",
            rId.collection.first, rId.collection.second, rId.variable, units);

    return true;
};

/**
 *
 */
bool Engine::getRenderableLabel(const RenderableId& rId, std::string& label, const std::string& separator)
{
    std::ostringstream buffer;	//
    CollectionCfg colCfg;	//
    VariableCfg varCfg;	//
    bool ok;	//

    ok = this->metadataMediator.getCollectionCfg(rId.collection, colCfg);
    if
        (false == ok)
    {
        Log::log( LogLevel::Error, "Main"
                , "Could not retrieve collection cfg for {}/{}"
                , rId.collection.first, rId.collection.second);
        return false;
    }

    ok = this->metadataMediator.getVariableCfg( rId.collection
                                              , rId.variable
                                              , varCfg);
    if
        (false == ok)
    {
        Log::log( LogLevel::Error, "Main"
                , "Could not retrieve variable cfg for {}/{}:{}"
                , rId.collection.first, rId.collection.second, rId.variable);
        return false;
    }
    buffer << colCfg.label << separator << varCfg.label;
    label = buffer.str();

    return true;
}

/**
 *
 */
bool Engine::getRenderableLabel(const RenderableId& rId, std::string& label)
{
    return this->getRenderableLabel(rId, label, ": ");
};

/**
 *
 */
bool Engine::setRenderables( const std::map<CollectionId
                           , std::vector<GranuleMetadata> >& renderablesIn)
{
	std::map<CollectionId, uint32_t> gCounters;	//

	this->renderables = renderablesIn;

	/* Update granules count in rendering control */
	for
		(auto& renderable : this->renderables)
	{
		gCounters[renderable.first] = static_cast<unsigned int>(renderable.second.size());
	}
	this->renderingControl.model.setCounters(gCounters);

    this->renderablesChanged();
	return true;
};

/**
 *
 */
bool Engine::loadOneGranule( const RenderableId& rId
						   , const VariableCfg& varCfg
						   , const GranuleMetadata& granule
						   , float viewportResolution, int viewportWidth
						   , int viewportHeight
						   , const DisplayBatchId& batchId)
{
    RenderingCfg* gCfg; //
	unsigned int resolutionIndex;	//
	std::string granuleURI;	//
	float granuleResolution;	//
	bool ok;	//

	gCfg = nullptr;
	resolutionIndex = 0;
	granuleResolution = 0.f;
	ok = false;

	/* Request rendering cfg since there might be a configuration
	customized for this granule */
	gCfg = this->renderingControl.model.getCfg( rId.collection
											  , rId.granule
											  , rId.variable
											  , ok);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Could not retrieve rendering cfg for {}/{}:{} @ {}.{}"
				, rId.collection.first, rId.collection.second
				, rId.variable, rId.granule.first, rId.granule.second);
		return false;
	}

	/* Select the most suitable resolution */
	ok = this->selectResolution( granule, viewportResolution
							   , viewportWidth, viewportHeight
							   , granuleURI, granuleResolution
							   , resolutionIndex);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Failed to select a resolution for granule {}/{}"
				, granule.id.first, granule.id.second);
		for
			(const auto& descriptor: granule.uris)
		{
			Log::log( LogLevel::Debug, "Engine"
					, "\tGranule IDF descriptor: {} {}"
					, descriptor.resolution, descriptor.uri);
		}

		/* Granule should not be displayed since there is no valid
		resolution */
		return false;
	}

	this->uris[rId] = granuleURI;

    SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
	ok = sceneView->getScene()->getNodeFactory()->exist(rId, granuleResolution);
    sceneView = nullptr;

	Log::log( LogLevel::Debug, "Engine"
			, "Check if granule is loaded: {}/{} @ {} => {}"
			, rId.granule.first, rId.granule.second, granuleResolution, ok);
	if
		(true == ok)
	{
		/* Granule data is already present in the renderer */
		return true;
	}

	ok = this->loadGranuleData( granule.uris[resolutionIndex]
							   , varCfg.fields
							   , granule.id
							   , granule.dataModel
							   , gCfg
							   , rId
							   , batchId);

	return ok;
};

/**
 *
 */
bool Engine::loadVariableData( const RenderableId& rId
                             , const DisplayBatchId& batchId)
{
    VariableCfg varCfg; //
	float viewportResolution;	//
	int viewportWidth;	//
	int viewportHeight;	//
    bool ok;    //

    SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
	viewportResolution = sceneView->getPixelSize();
	viewportWidth = sceneView->getViewport().x;
	viewportHeight = sceneView->getViewport().y;
    sceneView = nullptr;

    /* Retrieve information for the variable */
    ok = this->metadataMediator.getVariableCfg( rId.collection
                                              , rId.variable
                                              , varCfg);
    if
        (false == ok)
    {
        Log::log( LogLevel::Error, "Engine"
                , "Could not retrieve configuration for {}/{}:{}"
                , rId.collection.first
                , rId.collection.second
                , rId.variable);
        return false;
    }

	if
		(true == rId.granuleLevel)
	{
		for
			(auto& granule: this->renderables[rId.collection])
		{
			if
				(  granule.id.first == rId.granule.first
				&& granule.id.second == rId.granule.second)
			{
				this->loadOneGranule( rId
									, varCfg
									, granule
									, viewportResolution
									, viewportWidth
									, viewportHeight
                                    , batchId);
				break;
			}
		}
	}
	else
	{
		for
			(auto& granule : this->renderables[rId.collection])
		{
			RenderableId gId;
			gId.granuleLevel = true;
			gId.collection = rId.collection;
			gId.granule = granule.id;
			gId.variable = rId.variable;

			this->loadOneGranule( gId
								, varCfg
								, granule
								, viewportResolution
								, viewportWidth
								, viewportHeight
								, batchId);
		}
	}

    return true;
};

/**
 *
 */
bool Engine::selectResolution( const GranuleMetadata& granule
							 , float viewportResolution
							 , std::int32_t viewportWidth
							 , std::int32_t viewportHeight
							 , std::string& dataURI
							 , float& dataRes
							 , unsigned int& index) const
{
	std::int64_t j;	//
	float deltaRes;	//
	float viewportSize;	//
	float granuleSize;	//

	j = 0;
	deltaRes = 0.f;
	viewportSize = 0.f;
	granuleSize = 0.f;

	/* IMPORTANT: Resolution should be called spacing */

	if
		(0 >= granule.uris.size())
	{
		return false;
	}

	viewportSize = static_cast<float>(static_cast<double>(viewportResolution) * std::sqrt(viewportWidth * viewportWidth + viewportHeight * viewportHeight));
	if
		(0.f >= viewportSize)
	{
		return false;
	}

	Log::log( LogLevel::Debug, "Engine", "Select resolution for {}/{} @ {}"
		    , granule.id.first, granule.id.second, viewportResolution);
	unsigned int defaultIndex = 0;
	j = static_cast<std::int64_t>(granule.uris.size()) - 1;
	for
		(; j > -1; --j)
	{
        std::size_t k = static_cast<std::size_t>(j);

		/* Temporary workaround for high resolution images */
		if
			(  granule.uris[k].xArity > static_cast<const std::uint32_t>(this->maxTextureSize)
			|| granule.uris[k].yArity > static_cast<const std::uint32_t>(this->maxTextureSize))
		{
			Log::log(LogLevel::Error, "Engine", "A resolution was skipped because the GPU would not have been able to handle it");
			defaultIndex += 1;
			continue;
		}

		deltaRes = viewportResolution - granule.uris[k].resolution;
		Log::log( LogLevel::Debug, "Engine", "Check res {}: 0 < {} ? {}"
				, granule.uris[k].resolution, deltaRes, 0.f < deltaRes);
		if
			(0.f < deltaRes)
		{
			granuleSize = static_cast<float>(static_cast<double>(granule.uris[k].resolution) * std::sqrt(granule.uris[k].xArity * granule.uris[k].xArity + granule.uris[k].yArity * granule.uris[k].yArity));
			Log::log(LogLevel::Debug, "Engine", "Screen estate: {}m / {}m => {}"
					, granuleSize, viewportSize, this->settings.minGranuleScreenEstate);
			if
				(this->settings.minGranuleScreenEstate > granuleSize / viewportSize)
			{
				// Less than 1% of the viewport is occupied by the granule
				if
					(k < granule.uris.size() - 1)
				{
					// Select lower resolution since the higher resolution
					// is too far away from the viewport res
					index = static_cast<unsigned int>(k + 1);
					dataRes = granule.uris[index].resolution;
					dataURI = granule.uris[index].uri;
					return true;
				}

				// no lower resolution, higher resolution too far
				return false;
			}

			// found a suitable higher res
			index = static_cast<unsigned int>(k);
			dataRes = granule.uris[index].resolution;
			dataURI = granule.uris[index].uri;
			return true;
		}

		// Ignore lower res data
	}

	// No suitable higher res data found, use closest low res
	Log::log(LogLevel::Debug, "Engine", "Use first resolution by default");
	if
		(defaultIndex >= granule.uris.size())
	{
		Log::log(LogLevel::Error, "Engine",
                 "Even first resolution is not suitable (probably a dimension exceeding GPU memory capacity)");
		return false;
	}
	index = defaultIndex;
	dataRes = granule.uris[index].resolution;
	dataURI = granule.uris[index].uri;
	return true;
};

/**
 *
 */
bool Engine::loadGranuleData( const IDFDescriptor& idf
                              , const std::set<std::string>& fields
							  , const GranuleId& granuleId
							  , const DataModel& dataModel
							  , const RenderingCfg* rCfg
							  , const RenderableId& rId
							  , const DisplayBatchId& batchId)
{
    bool ok;    //

	auto callback = [this, idf, rCfg, rId, fields, batchId] () {
		DisplayJob j;	//
		Image* img(nullptr);	//
		CollectionCfg colCfg;	//
		DataInfo dataInfo;	//
		bool success(false);	//

		/* Check if data loading is still required */
        SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
        success = sceneView->getScene()->getNodeFactory()->exist(rId, idf.resolution);
        sceneView = nullptr;
		if
			(true == success)
		{
			Log::log( LogLevel::Debug, "Engine"
					, "Job cancelled because data is already there");
			return;
		}
        SEAScope::Log::log(SEAScope::LogLevel::Debug, "Engine", "isDataResLoaded? {} {}/{}/{}:{} @ {}", success, rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, idf.resolution);

        success = this->displayQueue.contains(rId, idf.resolution);
		if
			(true == success)
		{
			Log::log( LogLevel::Debug, "Engine"
					, "Job cancelled because data is already being sent to the renderer");
			return;
		}
        Log::log(LogLevel::Debug, "Engine", "In display queue? {} {}/{}/{}:{} @ {}", success, rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, idf.resolution);

		success = this->metadataMediator.getCollectionCfg(rId.collection, colCfg);
		if (false == success)
		{
			Log::log(LogLevel::Error, "Engine",
					 "Job cancelled due to an error while load the collection configuration");
			return;
		}

		Log::log(LogLevel::Debug, "Engine", "LOAD GRANULE DATA FOR {}", idf.uri);

		img = this->dataMediator.store->getData( idf.uri
											   , fields
											   , dataInfo
											   , success);
		if
			(false == success)
		{
			Log::log(LogLevel::Error, "Engine", "Getting data from the store failed for {}", idf.uri);
			return;
		}

		if
			(RenderMethod::TRAJECTORIES != rCfg->_renderMethod)
		{
            j.dataset = DatasetsAllocator::get(rId);
            success = this->gpuData.buildDataSet(j.dataset,
                                            img->data(),
										    dataInfo,
										    colCfg,
											idf.shape,
											idf.resolution);
		}
		else
		{
            j.dataset = nullptr;
			this->gpuData.buildPolyline( img->data()
									   , dataInfo
									   , colCfg
									   , idf.shape
									   , j.points
									   , j.fData
									   , j.mask);
		}

		/* Send data to the renderer */
		j.batchId = batchId;
		j.jobType = DisplayJobType::LOAD_DATA;
		j.renderingCfg = *rCfg;
		j.rId = rId;
		j.resolution = idf.resolution;
		success = this->displayQueue.addJob(j);
        if
            (false == success)
        {
            // Data loading job has been rejected, free dataset memory
            Log::log(LogLevel::Debug, "Engine", "Data load job rejected by display queue");
            DatasetsAllocator::release(j.dataset);
            j.dataset = nullptr;
        }

        Log::log(LogLevel::Debug, "Engine", "Added to display queue? {}/{}/{}:{} @ {}", rId.collection.first, rId.collection.second, rId.granule.second, rId.variable, idf.resolution);
	};

	Log::log(LogLevel::Debug, "Engine", "Will try to load data from {}", idf.uri);

    ok = this->dataMediator.loadData( granuleId
									, idf.uri
									, fields
									, dataModel
									, callback);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Could not load data");
        return false;
    }

    return true;
};

/**
 *
 */
bool Engine::updateRenderSet( const CollectionId& collection
							, const std::vector<GranuleMetadata>& granules
							, std::set<RenderableId>& toRender)
{
	std::set<VariableId> vars;	//
	bool ok;	//

	ok = false;

	ok = this->renderingControl.renderedVariables(collection, vars);
	if
		(false == ok)
	{
		// Should not happen
		Log::log( LogLevel::Error, "Engine"
				, "No selected variable found for collection {}/{}"
				, collection.first, collection.second);
		return false;
	}

	for
		(auto& varId : vars)
	{
		for
			(const auto& granule: granules)
		{
			toRender.emplace(true, collection, granule.id, varId);
		}
	}

	return true;
};

/**
 *
 */
void Engine::renderablesChanged()
{
    bool ok;    //
	DisplayJob j;	//
	DisplayBatchId batchId;	//

	ok = false;
	batchId = std::chrono::high_resolution_clock::now();

	/* Update the set of unique renderables */
	j.batchId = batchId;
	j.jobType = DisplayJobType::UPDATE_STARTED;
	for
		(auto& renderable : this->renderables)
	{
		ok = this->updateRenderSet( renderable.first
								  , renderable.second
								  , j.toRender);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "Engine"
					, "Failed to update render set with collection {}/{}"
					, renderable.first.first, renderable.first.second);
			return;
		}
	}
	this->displayQueue.addJob(j);

	/* Load data when needed */
	for
		(const auto& rId: j.toRender)
	{
        if
            (2u == rId.collection.first)
        {
            /* Annotations */
            continue;
        }
		this->loadVariableData(rId, batchId);
	}

	/* Tell the renderer that everything is ready */
	j.batchId = batchId;
	j.jobType = DisplayJobType::UPDATE_COMPLETE;
	this->displayQueue.addJob(j);

	/* deactivate unused collections */
	/* TODO: Add a marker to say that there is no visible granule for some collections */

    RenderablesChangedEvent event(this->renderables);
    this->_appEventsDispatcher.dispatch<RenderablesChangedEvent>(event);
};

/**
 *
 */
void Engine::updateRendering(const RenderingCfg* rCfg, const RenderingParam& /*param*/)
{
	DisplayJob j;

	j.jobType = DisplayJobType::UPDATE_RENDERING;
	j.renderingCfg = *rCfg;
	j.rId = rCfg->target;

	this->displayQueue.addJob(j);

    RenderingCfgUpdateEvent event(rCfg);
    this->_appEventsDispatcher.dispatch<RenderingCfgUpdateEvent>(event);
};


/**
 *
 */
bool Engine::setCollectionsUI(ICollectionsControlView* view)
{
    this->collectionsControl.init(&this->metadataMediator, view);
    this->collectionsControl.setVariableStateCallback([this] (const CollectionVariable& colvar, bool selected) {
		if
			(selected)
		{
			this->renderingControl.addRendering(colvar);
		}
		else
		{
			this->renderingControl.removeRendering(colvar);
		}

        // Update renderables, otherwise the variable will be displayed/hidden
        // on the map only if it is/was the only one selected in the collection
        this->renderablesChanged();
	});

	this->collectionsControl.exclusiveCallback = [this] (const CollectionVariable& colvar) {
		for
			(const auto& it: this->collectionsControl.model.store)
		{
			auto found = this->renderingControl.widgets.find(it.first);
			if
				(it.first == colvar)
			{
				if
					(this->renderingControl.widgets.end() == found)
				{
					this->renderingControl.addRendering(it.first);
				}
			}
			else if
				(this->renderingControl.widgets.end() != found)
			{
				this->renderingControl.removeRendering(it.first);
			}
		}

        // Update renderables in this case too, otherwise the variables from
        // the same collection that were selected previously will remain on
        // screen and the new one will not
        this->renderablesChanged();
	};

    return true;
};

/**
 *
 */
bool Engine::setRenderingUI(IRenderingControlView* view)
{
    this->renderingControl.init(view, &this->cmaps);
    return true;
};

/**
 *
 */
bool Engine::setAdvancedRenderingUI(IAdvancedRenderingControlView* view)
{
    view->init(&this->cmaps);
    this->renderingControl.advanced.init(view);
    return true;
};

/**
 *
 */
bool Engine::setTimelineUI(ITimelineView* view)
{
    this->timeline.init(view, this->settings.timelineCfg);
    return true;
};

/**
 *
 */
bool Engine::setSourcesUI(ISourcesControlView* view)
{
    this->sourcesControl.init(view);
    return true;
};

/**
 *
 */
bool Engine::shutdown()
{
	DisplayJob j;	//
    bool ok;    //

    this->sourcesControl.model.reset();

	ok = this->plugins.releaseAll();
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Engine", "Could not release all plugins");
	}

    dynamic<ICatalogueServiceConnector>::manager = nullptr;
    dynamic<IIndexServiceConnector>::manager = nullptr;
    dynamic<IDatastreamServiceConnector>::manager = nullptr;

    ok =  ok && this->sockets.shutdown();
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Engine", "Could not stop the sockets manager properly");
	}

	while
		(false == this->displayQueue.queue.empty())
	{
		this->displayQueue.pop(j);
		// TODO: check that there is no memory leak here
	}

    return ok;
};

/**
 *
 */
bool Engine::getRenderableInfo(const RenderableId& rId, GranuleInfo& info)
{
	bool ok;	//

	ok = false;
	info.metadata = nullptr;

	auto found = this->uris.find(rId);
	if
		(this->uris.end() == found)
	{
		Log::log( LogLevel::Error, "Engine"
				, "No URI associated with renderable object");
		return false;
	}
	info.uri = found->second;

	ok = this->metadataMediator.getCollectionCfg( rId.collection
												, info.collection);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Could not retrieve collection configuration for renderable object");
		return false;
	}

	ok = this->metadataMediator.getVariableCfg( rId.collection, rId.variable
											  , info.variable);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Could not retrieve variable configuration for renderable object");
		return false;
	}

	ok = this->metadataMediator.getDataCoverage(rId.collection, info.coverage);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Could not retrieve data coverage for the collection of renderable object");
		return false;
	}

	auto foundMeta = this->renderables.find(rId.collection);
	if
		(this->renderables.end() == foundMeta)
	{
		Log::log( LogLevel::Error, "Engine"
				, "Could not find granules list for the collection of renderable object");
		return false;
	}

	for
		(const auto& it : foundMeta->second)
	{
		if
			(  it.id.first == rId.granule.first
			&& it.id.second == rId.granule.second)
		{
			info.metadata = &it;
			break;
		}
	}

	if
		(nullptr == info.metadata)
	{
		Log::log( LogLevel::Error, "Main"
				, "Could not retrieve granule metadata for picked object");
		return false;
	}

	if
		(info.collection.xSeamless && info.collection.ySeamless)
	{
		info.area = 510072000000000.f; /* Earth surface */
		info.perimeter = std::numeric_limits<float>::quiet_NaN();
	}
	else
	{
		for
			(const auto& it : info.metadata->uris)
		{
			if
				(it.uri == info.uri)
			{
				if
					(info.collection.xSeamless || info.collection.ySeamless)
				{
					/* ring band around the Earth => no perimeter */
					info.perimeter = std::numeric_limits<float>::quiet_NaN();
				}
				else
				{
					info.perimeter = static_cast<float>(SEAScope::perimeterFromShape(it.shape));
				}

				info.area = static_cast<float>(SEAScope::areaFromShape(it.shape));
				break;
			}
		}
	}
	return true;
};

/**
 *
 */
bool Engine::sendProcessingOpcode(std::uint64_t opcode)
{
	std::uint8_t opcodeBytes[8];	//
    socket_t sock;  //
    bool ok;    //

    ok = false;

	sock = this->sockets.factory->connect(this->settings.processorAddress,
										  this->settings.processorPort, ok);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Engine", "Could not connect to server");
		return false;
	}

	opcodeBytes[0] = static_cast<std::uint8_t>(opcode & 0x00000000000000ff);
    opcodeBytes[1] = static_cast<std::uint8_t>((opcode & 0x000000000000ff00) >> 8);
	opcodeBytes[2] = static_cast<std::uint8_t>((opcode & 0x0000000000ff0000) >> 16);
	opcodeBytes[3] = static_cast<std::uint8_t>((opcode & 0x00000000ff000000) >> 24);
	opcodeBytes[4] = static_cast<std::uint8_t>((opcode & 0x000000ff00000000) >> 32);
	opcodeBytes[5] = static_cast<std::uint8_t>((opcode & 0x0000ff0000000000) >> 40);
	opcodeBytes[6] = static_cast<std::uint8_t>((opcode & 0x00ff000000000000) >> 48);
	opcodeBytes[7] = static_cast<std::uint8_t>((opcode & 0xff00000000000000) >> 56);

    ok = this->sockets.monitor->sendData(sock, &(opcodeBytes[0]), 8);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Engine", "Failed to send opcode through the socket");
        return false;
    }
    return true;
};

/**
 *
 */
void Engine::update()
{
	timestamp_t current;	//
	timestamp_t begin;	//
	timestamp_t end;	//
	bool ok;	//
	float zoom;	//
	std::vector<GranuleId> granules;	//
	std::map<CollectionId, std::vector<GranuleMetadata> > searchResults;	//
	std::vector<std::pair<CollectionId, bool> > collectionMustBeCurrent;	//
	CollectionCfg colCfg;	//

	std::chrono::high_resolution_clock::time_point now = std::chrono::high_resolution_clock::now();
    std::chrono::duration<double> timeDiff = now - this->lastUpdate;

    Log::log(LogLevel::Debug, "Engine", "Search request delay: {}", this->settings.searchRequestMinDelay);
	if
		(timeDiff.count() < this->settings.searchRequestMinDelay)
	{
        Log::log(LogLevel::Debug, "Engine", "Update delayed due to throttling");
		return;
	}

	this->updateRequested = false;

    SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
	zoom = sceneView->getPixelSize();
    sceneView = nullptr;
	this->timeline.getBounds(current, begin, end);

	/* Retrieve match criteria for each rendered collection */
	for
		(const auto& collection : this->renderingControl.model.renderedCollections)
	{
		ok = this->metadataMediator.getCollectionCfg(collection, colCfg);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "Engine"
					, "Could not retrieve configuration for collection {}/{}"
					, collection.first, collection.second);
			return;
		}
		collectionMustBeCurrent.push_back(std::make_pair(collection, colCfg.mustBeCurrent));
	}

	ok = this->metadataMediator.searchGranules( current
											   , begin
											   , end
											   , zoom
											   , collectionMustBeCurrent
											   , searchResults);

    /*
	Log::log(LogLevel::Debug, "Engine", "{} granules found", searchResults.size());
	for
		(const auto& it : searchResults)
	{
		for
			(const auto& itGranule : it.second)
		{
			Log::log( LogLevel::Debug, "Main"
            , "  ==|> {}/{}/{} {}"
					, itGranule.id.first, itGranule.collectionId, itGranule.id.second, itGranule.uris[0].uri);
		}
	}
    */

	this->setRenderables(searchResults);
	this->lastUpdate = now;
};

/**
 *
 */
bool Engine::updateCollections(const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel)
{
	bool ok;	//
	timestamp_t newPrev;	//
	timestamp_t newNext;	//
	EventsCoverage coverage;	//
	timestamp_t minDelta;	//

	minDelta = 1;

	Log::log(LogLevel::Debug, "Engine", "Rendered collections changed: -{} +{}", toDel.size(), toAdd.size());

	// Remove unselected collections from data coverage
	auto itDel = toDel.begin();
	for
		(;itDel != toDel.end(); itDel++)
	{
        Log::log(LogLevel::Debug, "Engine", "Remove coverage from collection {}/{}", (*itDel).first, (*itDel).second);
		this->timeline.model.removeCoverage(*itDel, true); // silent
	}

	// Add newly selected collections to data coverage
	auto itAdd = toAdd.begin();
	for
		(;itAdd != toAdd.end(); itAdd++)
	{
        Log::log(LogLevel::Debug, "Engine", "Add coverage from collection {}/{}", (*itAdd).first, (*itAdd).second);
		ok = this->metadataMediator.getDataCoverage(*itAdd, coverage);
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "Engine", "Could not retrieve data coverage");
			return false;
		}
		this->timeline.model.setCoverage(*itAdd, coverage, true); // silent
	}

    if
        (0 < toAdd.size() + toDel.size())
    {
        ok = this->timeline.model.updateCoverages();
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "Engine", "Failed to update timeline coverage");
            return false;
        }
    }

	// Update previous/next in the timeline
	this->metadataMediator.searchPrevNext( this->timeline.model.current
										  , this->renderingControl.model.renderedCollections
										  , newPrev
										  , newNext
										  , minDelta);
	this->timeline.model.setPrevNext(newPrev, newNext);

	// Apply timeline modes "nearest day" and "closest to 12:00" only if a single
	//   collection is selected.
	bool applyModes = (1 == this->renderingControl.model.renderedCollections.size());
	this->timeline.model.update( applyModes && this->timeline.model.nearestMode
							 , applyModes && this->timeline.model.nearestInDayMode);

	// Search for granules in the new configuration
	Log::log(LogLevel::Debug, "Engine", "UPDATE REQUESTED");
	this->updateRequested = true;
	return true;
};

/**
 *
 */
Tools* Engine::getTools()
{
    return &(this->_tools);
}

/**
 *
 */
bool Engine::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SelectedPickablesChangedEvent)) == eventType)
    {
        GranuleInfo info;

        const std::vector<Pickable*>& selected = static_cast<const SelectedPickablesChangedEvent&>(event).getSelected();
        if
            (1 == selected.size())
        {
            const RenderableId& rId= selected[0]->renderableId();

            // TODO: find a way to replace hardcoded identifier for the source
            if
                (2u == rId.collection.first)
            {
                return true;  // event has been processed
            }


            bool ok(false);
            ok = this->getRenderableInfo(rId, info);
            if
                (false == ok)
            {
                Log::log(LogLevel::Error, "Engine",
                         "Could not retrieve picked object info");
                return true;  // event has been processed, even if it failed
            }

            SelectedPickableWithInfoChangedEvent outEvent(selected[0], info, RenderableType::GRANULE);
            this->_appEventsDispatcher.dispatch<SelectedPickableWithInfoChangedEvent>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(CenterTimeOnGranuleRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const CenterTimeOnGranuleRequest&>(event).getRenderableId();
		bool success(false);    //
		GranuleInfo info;   //
		timestamp_t center; //
        double centerTmp(0.0);  //
        double timeExtent(0.0); //

		Log::log(LogLevel::Debug, "Engine", "center time callback");

		success = this->getRenderableInfo(rId, info);
		if
			(false == success)
		{
			Log::log( LogLevel::Error, "Engine"
					, "Could not retrieve picked object info");
			return true;  // event has been processed, even if it failed
		}

        timeExtent = static_cast<double>(info.metadata->start + info.metadata->stop);
		centerTmp = std::floor(0.0005 * timeExtent);
        center = static_cast<timestamp_t>(centerTmp);
		Log::log(LogLevel::Debug, "Main", "Center on {}", center);
		this->timeline.model.setCurrent(center);

        return true;
    }
    else if
        (std::type_index(typeid(DisplayGranuleInfoRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const DisplayGranuleInfoRequest&>(event).getRenderableId();
		bool success(false);	//
		GranuleInfo info;	//

		success = this->getRenderableInfo(rId, info);
		if
			(false == success)
		{
			Log::log( LogLevel::Error, "Engine"
					, "Could not retrieve picked object info");
			return true;    // event has been processed, even if it failed
		}

        GranuleInfoReadyForDisplayEvent outEvent(info);
        this->_appEventsDispatcher.dispatch<GranuleInfoReadyForDisplayEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(ExtractFromRenderableIdRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const ExtractFromRenderableIdRequest&>(event).getRenderableId();

        // Get footprint for the renderable acting as extraction mask
        SceneView* sceneView = this->_tools.getSceneViewController()->getSceneView();
        Renderable2* footprint(nullptr);
        bool ok(false);
        ok = sceneView->getScene()->getNodeFactory()->getSceneDictionnary()->exist(rId, footprint);
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "Engine",
                     "Could not find renderable {}/{}/{}:{} in the scene, so extraction is not possible",
                     rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
            return true;    // event has been processed, even if it failed
        }

        VariableCfg varCfg; //
		ok = this->metadataMediator.getVariableCfg(rId.collection,
                                                   rId.variable,
												   varCfg);
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "Engine", "Cannot get rendering configuration for selected granule/variable");
            return true;    // event has been processed, even if it failed
        }

        this->clearExtractedData();
        this->initExtractedData();

        Log::log(LogLevel::Info, "Engine",
                 "Extracting using {}/{}/{}:{} as mask (rendering method for mask granule: {}, trajectories: {}, polyline: {}, point: {})",
                 rId.collection.first, rId.collection.second, rId.granule.second, rId.variable,
                 int(varCfg.defaultRendering), int(RenderMethod::TRAJECTORIES), int(RenderMethod::POLYLINE), int(RenderMethod::POINT));

		if
			(  (RenderMethod::TRAJECTORIES == varCfg.defaultRendering)
            || (RenderMethod::POLYLINE == varCfg.defaultRendering)
            || (RenderMethod::POINT == varCfg.defaultRendering))
		{
            ShapeMesh* shapeMesh = dynamic_cast<ShapeMesh*>(footprint);
            TransectRequest outEvent(sceneView, shapeMesh,
                                     &(this->_extractedData),
                                     &(this->_extractedLabels));
            this->_appEventsDispatcher.dispatch<TransectRequest>(outEvent);
        }
        else
        {
            ExtractionRequest outEvent(sceneView, footprint,
                                       &(this->_extractedData),
                                       &(this->_extractedLabels));
            this->_appEventsDispatcher.dispatch<ExtractionRequest>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(RemotePickleExportRequest)) == eventType)
    {
		bool success(false);    //

		success = this->sendProcessingOpcode(0 /* noop */);
		if
			(false == success)
		{
			Log::log(LogLevel::Info, "Engine", "Remote processing server is not available");
            return true;    // event has been processed, even if it failed
		}

		success = this->sendProcessingOpcode(3 /* serialize extraction */);
        return true;    // event has been processed, regardless of success
    }
    else if
        (std::type_index(typeid(RemotePNGExportRequest)) == eventType)
    {
		bool success(false);    //

		success = this->sendProcessingOpcode(0 /* noop */);
		if
			(false == success)
		{
			Log::log(LogLevel::Info, "Engine", "Remote processing server is not available");
            return true;    // event has been processed, even if it failed
		}

		success = this->sendProcessingOpcode(1 /* plot extraction */);
        return true;    // event has been processed, regardless of success
    }
    else if
        (std::type_index(typeid(CollectionsUpdatedEvent)) == eventType)
    {
        const std::set<CollectionId>& toAdd = static_cast<const CollectionsUpdatedEvent&>(event).getToAdd();
        const std::set<CollectionId>& toRemove = static_cast<const CollectionsUpdatedEvent&>(event).getToRemove();

        this->updateCollections(toAdd, toRemove);

        return true;
    }
    else if
        (std::type_index(typeid(DrawingStartedEvent)) == eventType)
    {
        const CollectionVariable& collectionVariable = static_cast<const DrawingStartedEvent&>(event).getCollectionVariable();

        this->collectionsControl.view->allowDeletion(collectionVariable, false);

        return true;
    }
    else if
        (std::type_index(typeid(DrawingCompleteEvent)) == eventType)
    {
        OGL::Shape* shape = static_cast<const DrawingCompleteEvent&>(event).getShape();
        const CollectionVariable& collectionVariable = static_cast<const DrawingCompleteEvent&>(event).getCollectionVariable();

        this->collectionsControl.view->allowDeletion(collectionVariable, true);

        if
            (nullptr == shape)
        {
            // User did not draw anything and switched the drawing tool off
            return true;
        }
        // Default time coverage: 24h centred on current time
        double currentTimeInSeconds = static_cast<double>(this->timeline.model.current);
        double defaultStart = 1000.0 * (currentTimeInSeconds - 43200.0);
        double defaultStop = 1000.0 * (currentTimeInSeconds + 43200.0);

        AddPreparedAnnotationRequest outEvent(collectionVariable, shape, defaultStart, defaultStop);
        this->_appEventsDispatcher.dispatch<AddPreparedAnnotationRequest>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(CollectionRegistrationRequest)) == eventType)
    {
        const CollectionId& collectionId = static_cast<const CollectionRegistrationRequest&>(event).getCollectionId();
        const CollectionCfg& collectionCfg = static_cast<const CollectionRegistrationRequest&>(event).getCollectionCfg();
        const std::vector<RenderingCfg*>& collectionRenderingCfgs = static_cast<const CollectionRegistrationRequest&>(event).getCollectionRenderingCfgs();

        this->metadataMediator.collectionCfgLoaded(collectionId,
                                                   collectionCfg,
                                                   collectionRenderingCfgs);

        this->collectionsControl.model.addCollections({collectionId});

        std::map<RenderingCfg*, std::string> addedVariables;
        for
            (const auto rCfg : collectionRenderingCfgs)
        {
            std::string colvarStr("");
            SEAScope::getColvarLabel(collectionCfg.label,
                                     collectionCfg.variables.at(rCfg->target.variable).label,
                                     colvarStr);
            addedVariables[rCfg] = colvarStr;
        }
        VariablesAddedEvent outEvent(addedVariables);
        this->_appEventsDispatcher.dispatch<VariablesAddedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(CollectionRemovalRequest)) == eventType)
    {
        const CollectionId& colId = static_cast<const CollectionRemovalRequest&>(event).getCollectionId();

        // We will need the collection configuration to notify other components
        // of the variables that have been deleted
        CollectionCfg colCfg;
        this->metadataMediator.getCollectionCfg(colId, colCfg);
        std::vector<VariableId> variables;
        for
            (const auto& variable : colCfg.variables)
        {
            variables.push_back(variable.first);
        }

        // Remove RenderingCfg objects for variables related to the removed collection
        this->renderingControl.model.removeCollection(colId);

        this->collectionsControl.model.removeCollections({colId});

        // Remove VariableCfg and CollectionCfg objects
        this->metadataMediator.store->deleteCollectionCfg(colId);

        CollectionRemovedEvent outEvent(colId, variables);
        this->_appEventsDispatcher.dispatch<CollectionRemovedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(GranuleAddedEvent)) == eventType)
    {
        std::uint32_t sourceId = static_cast<const GranuleAddedEvent&>(event).getSourceId();
        std::uint32_t collectionId = static_cast<const GranuleAddedEvent&>(event).getCollectionId();
        //std::uint64_t granuleId = static_cast<const GranuleAddedEvent&>(event).getGranuleId();

		std::set<CollectionId> selected;	//
		std::set<CollectionId> toAdd;	//
		std::set<CollectionId> toDel;	//
		CollectionId userCollection;	//
		bool success(false);	//

		success = this->collectionsControl.model.getSelectedCollections(selected);
		if
			(false == success)
		{
			return true;  // event has been processed, regardless of success
		}

		userCollection.first = sourceId;
		userCollection.second = collectionId;
		auto found = selected.find(userCollection);
		if
			(selected.end() == found)
		{
			return true;  // event has been processed, regardless of success
		}

		// Update data availability
		toAdd.insert(userCollection);
		this->updateCollections(toAdd, toDel);

        return true;
    }
    else if
        (std::type_index(typeid(VariableSelectionRequest)) == eventType)
    {
        const CollectionVariable& collectionVariable = static_cast<const VariableSelectionRequest&>(event).getCollectionVariable();
        bool selected = static_cast<const VariableSelectionRequest&>(event).getSelected();

        this->collectionsControl.model.select(collectionVariable, selected, false);
        return true;
    }
    else if
        (std::type_index(typeid(PixelSizeChangedEvent)) == eventType)
    {
        // Trigger an update so that the engine will adapt the granules to the
        // resolution. Required when zooming in and out
        this->renderablesChanged();
        return true;
    }
    else if
        (std::type_index(typeid(AnnotationTimeCoverageUpdatedEvent)) == eventType)
    {
        const RenderableId& rId = static_cast<const AnnotationTimeCoverageUpdatedEvent&>(event).getRenderableId();
        this->updateCollections({rId.collection}, {});
        return true;
    }

    return false;
}

const std::map<RenderableId, Extraction>& Engine::getExtractedData() const
{
    return this->_extractedData;
}

const std::map<RenderableId, std::string>& Engine::getExtractedLabels() const
{
    return this->_extractedLabels;
}

UserGenerated* Engine::getUserGenerated()
{
    return &(this->_userGenerated);
}

//////////////////////////////////////////////////////////////////////////////////////////////
// TMP
//////////////////////////////////////////////////////////////////////////////////////////////

void Engine::saveConfig(const std::string& path, SEAScope::RenderingCfg* cfg, const SEAScope::CollectionCfg& cCfg)
{
	// first, create a file instance
	mINI::INIFile file(path);

	// next, create a structure that will hold data
	mINI::INIStructure ini;

	// now we can read the file
	bool res = file.read(ini);

	bool saveFile = false;

	if 
		(true == res)
	{
		// Get the variable targeted by the RenderingCfg
		auto it = cCfg.variables.find(cfg->target.variable);

		if
			(it != cCfg.variables.end())
		{
			// Loop on the ini map to find the corresponding ini section
			auto itMap = ini.begin();
			for 
				(; itMap != ini.end(); ++itMap)
			{
				// Security check, we need the label parameter
				bool bRes = itMap->second.has("label");

				if 
					(true == bRes)
				{
					// Get the string label
					std::string str = itMap->second.get("label");

					// Compare labels to match the VariableCfg with the right ini section
					if
						(str == it->second.label)
					{
						// Section found, copy new values
						std::string varKey = itMap->first;

						ini[varKey]["logscale"] = (true == cfg->logscale()) ? "true" : "false";
						ini[varKey]["min"] = std::to_string(cfg->min());
						ini[varKey]["max"] = std::to_string(cfg->max());
						ini[varKey]["opacity"] = std::to_string(cfg->opacity());
						ini[varKey]["zindex"] = std::to_string(cfg->zindex());
						ini[varKey]["particlesCount"] = std::to_string(cfg->particlesCount());
						ini[varKey]["particleTTL"] = std::to_string(cfg->particleTTL());
						ini[varKey]["streamlineLength"] = std::to_string(cfg->streamlineLength());
						ini[varKey]["streamlineSpeed"] = std::to_string(cfg->streamlineSpeed());
						ini[varKey]["colormap"] = cfg->colormap();

						std::stringstream ss;
						ss << std::to_string(cfg->color()[0]) << "," << std::to_string(cfg->color()[1]) << "," << std::to_string(cfg->color()[2]);
						ini[varKey]["color"] = ss.str();
						ini[varKey]["filterMode"] = (SEAScope::FilterMode::BILINEAR == cfg->filterMode()) ? "BILINEAR" : "NEAREST";

						saveFile = true;

						// Variable found, break the loop
						break;
					}
				}
			}
		}
		else
		{
			Log::log(LogLevel::Error, "Engine",
				"Variable {} not found, impossible to save in config.ini",
				cfg->target.variable);
		}
	}
	else
	{
		Log::log(LogLevel::Error, "Engine",
			"Could not read {} file",
			path);
	}

	if 
		(true == saveFile)
	{
		// write updates to file
		res = file.write(ini);

		if
			(false == res)
		{
			Log::log(LogLevel::Error, "Engine",
				"Could not write to {} file",
				path);
		}
	}
}
