// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/rendering_cfg.h"
#include "num_utils.h"

using namespace SEAScope;

/**
 *
 */
RenderingCfg::RenderingCfg()
: _rendered(false), _logScale(false), _min(0.f), _max(1.f), _opacity(0.f),
  _zindex(0.f), _color({0, 0, 0}), _colormap(""),
  _renderMethod(RenderMethod::RASTER), _particlesCount(0), _particleTTL(0),
  _streamlineLength(0), _streamlineSpeed(0.f),
  _filterMode(FilterMode::NEAREST), callback(nullptr)
{
};

/**
 *
 */
bool RenderingCfg::applyTo(RenderingCfg* targetCfg) const
{
	if
		(nullptr == targetCfg)
	{
		return false;
	}
	targetCfg->_logScale = this->_logScale;
	targetCfg->_min = this->_min;
	targetCfg->_max = this->_max;
	targetCfg->_opacity = this->_opacity;
	targetCfg->_zindex = this->_zindex;
	targetCfg->_colormap = this->_colormap;
	targetCfg->_color[0] = this->_color[0];
	targetCfg->_color[1] = this->_color[1];
	targetCfg->_color[2] = this->_color[2];
	targetCfg->_renderMethod = this->_renderMethod;
	targetCfg->_particlesCount = this->_particlesCount;
	targetCfg->_particleTTL = this->_particleTTL;
	targetCfg->_streamlineLength = this->_streamlineLength;
	targetCfg->_streamlineSpeed = this->_streamlineSpeed;
	targetCfg->_filterMode = this->_filterMode;
	if
		(nullptr != targetCfg->callback)
	{
		targetCfg->callback(targetCfg, RenderingParam::ANY);
	}
	return true;
};

/**
 *
 */
RenderingCfg::~RenderingCfg()
{
};

/**
 *
 */
bool RenderingCfg::setRendered(bool value)
{
	if
		(value == this->_rendered)
	{
		return false;
	}

	this->_rendered = value;
	if
		(nullptr != this->callback)
	{
		this->callback(this, RenderingParam::RENDERED);
	}
	return true;
};

/**
 *
 */
bool RenderingCfg::rendered() const
{
	return this->_rendered;
};

/**
 *
 */
bool RenderingCfg::setLogscale(bool value)
{
    if
        (value == this->_logScale)
    {
        return false;
    }

    this->_logScale = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::LOGSCALE);
    }
    return true;
};

/**
 *
 */
bool RenderingCfg::logscale() const
{
    return this->_logScale;
};

/**
 *
 */
bool RenderingCfg::setMin(float value)
{
    if
        (floating_point_equality(value, this->_min))
    {
        return false;
    }

    this->_min = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::MIN);
    }
    return true;
};

/**
 *
 */
float RenderingCfg::min() const
{
    return this->_min;
};

/**
 *
 */
bool RenderingCfg::setMax(float value)
{
    if
        (floating_point_equality(value, this->_max))
    {
        return false;
    }

    this->_max = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::MAX);
    }
    return true;
};

/**
 *
 */
float RenderingCfg::max() const
{
    return this->_max;
};

/**
 *
 */
bool RenderingCfg::setOpacity(float value)
{
    if
        (floating_point_equality(value, this->_opacity))
    {
        return false;
    }

    this->_opacity = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::OPACITY);
    }
    return true;
};

/**
 *
 */
float RenderingCfg::opacity() const
{
    return this->_opacity;
};

/**
 *
 */
bool RenderingCfg::setZindex(float value)
{
    if
        (floating_point_equality(value, this->_zindex))
    {
        return false;
    }

    this->_zindex = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::ZINDEX);
    }
    return true;
};

/**
 *
 */
float RenderingCfg::zindex() const
{
    return this->_zindex;
};

/**
 *
 */
bool RenderingCfg::setColor(const std::array<std::uint8_t, 3>& value)
{
	if
		(  this->_color[0] == value[0]
		&& this->_color[1] == value[1]
		&& this->_color[2] == value[2])
	{
		return false;
	}

	this->_colormap = "";
	this->_color[0] = value[0];
	this->_color[1] = value[1];
	this->_color[2] = value[2];
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::COLOR);
    }
	return true;
};

/**
 *
 */
const std::array<std::uint8_t, 3>& RenderingCfg::color() const
{
	return this->_color;
};

/**
 *
 */
bool RenderingCfg::setColormap(const std::string& value)
{
    if
        (value == this->_colormap)
    {
        return false;
    }

    this->_colormap = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::COLORMAP);
    }
    return true;
};

/**
 *
 */
const std::string& RenderingCfg::colormap() const
{
    return this->_colormap;
};

/**
 *
 */
bool RenderingCfg::setRenderMethod(const RenderMethod& value)
{
    if
        (value == this->_renderMethod)
    {
        return false;
    }

    this->_renderMethod = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::RENDER_METHOD);
    }
    return true;
};

/**
 *
 */
const RenderMethod& RenderingCfg::renderMethod() const
{
    return this->_renderMethod;
};

/**
 *
 */
bool RenderingCfg::setParticlesCount(uint32_t value)
{
    if
        (value == this->_particlesCount)
    {
        return false;
    }

    this->_particlesCount = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::PARTICLES_COUNT);
    }
    return true;
};

/**
 *
 */
uint32_t RenderingCfg::particlesCount() const
{
    return this->_particlesCount;
};

/**
 *
 */
bool RenderingCfg::setParticleTTL(uint8_t value)
{
    if
        (value == this->_particleTTL)
    {
        return false;
    }

    this->_particleTTL = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::PARTICLE_TTL);
    }
    return true;
};

/**
 *
 */
uint8_t RenderingCfg::particleTTL() const
{
    return this->_particleTTL;
};

/**
 *
 */
bool RenderingCfg::setStreamlineLength(uint32_t value)
{
    if
        (value == this->_streamlineLength)
    {
        return false;
    }

    this->_streamlineLength = value;
    if
        (nullptr != this->callback)
    {
        this->callback(this, RenderingParam::STREAMLINE_LENGTH);
    }
    return true;
};

/**
 *
 */
uint32_t RenderingCfg::streamlineLength() const
{
    return this->_streamlineLength;
};

/**
 *
 */
bool RenderingCfg::setStreamlineSpeed(float value)
{
	if
		(floating_point_equality(value, this->_streamlineSpeed))
	{
		return false;
	}

	this->_streamlineSpeed = value;
	if
		(nullptr != this->callback)
	{
		this->callback(this, RenderingParam::STREAMLINE_SPEED);
	}
	return true;
};

/**
 *
 */
float RenderingCfg::streamlineSpeed() const
{
	return this->_streamlineSpeed;
}

/**
 *
 */
bool RenderingCfg::setFilterMode(FilterMode value)
{
	if
		(value == this->_filterMode)
	{
		return false;
	}

	this->_filterMode = value;
	if
		(nullptr != this->callback)
	{
		this->callback(this, RenderingParam::FILTER_MODE);
	}
	return true;
}

/**
 *
 */
FilterMode RenderingCfg::filterMode() const
{
	return this->_filterMode;
};

/**
 *
 */
bool RenderingCfg::setScale(float min, float max, bool log)
{
	if
		(  floating_point_equality(min, this->_min)
		&& floating_point_equality(max, this->_max)
		&& log == this->_logScale)
	{
		return false;
	}

	this->_min = min;
	this->_max = max;
	this->_logScale = log;

	Log::log(LogLevel::Debug, "RenderingCfg", "Scale changed");
	if
		(nullptr != this->callback)
	{
		this->callback(this, RenderingParam::SCALE);
	}
	return true;
};

/**
 *
 */
bool RenderingCfg::scrollExtrema(float value, bool targetIsMin)
{
	float extent(0.f);	//
	float delta(0.f);	//
	float newMin(0.f);	//
	float newMax(0.f);	//
	float sgn(0.f);	//
	bool logScale(false);	//

	if
		(false == this->_logScale)
	{
		extent = this->_max - this->_min;
		delta = 0.02f * extent;
	}
	else if
		(0 < this->_min && this->_min < this-> _max)
	{
		extent = static_cast<float>(log(static_cast<double>(this->_max)) - log(static_cast<double>(this->_min)));
		delta = static_cast<float>(exp(0.02 * static_cast<double>(extent)));
	}

	sgn = (0.f < value) ? 1.f:-1.f;
	if
		(targetIsMin)
	{
		newMin = this->_min + sgn * delta;
		newMax = this->_max;
	}
	else
	{
		newMin = this->_min;
		newMax = this->_max + sgn * delta;
	}
	logScale = this->_logScale && (0.f < newMin) && (newMin < newMax);
	this->setScale(newMin, newMax, logScale);
	return true;
};

/**
 *
 */
bool RenderingCfg::setCallback(const std::function<void(const RenderingCfg*, const RenderingParam&)>& value)
{
    this->callback = value;
    return true;
};
