// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "core/guid.h"
#include "core/threads_pool.h"

using namespace SEAScope;

/**
 *
 */
ThreadsPool::ThreadsPool(): terminationRequested(false), terminated(false), scheduler(nullptr)
{
};

/**
 *
 */
ThreadsPool::~ThreadsPool()
{
	if
		(false == this->terminated)
	{
		this->shutdown();
	}
};

/**
 *
 */
bool ThreadsPool::init(std::uint16_t threadsCount, IJobScheduler* schedulerIn)
{
#ifdef _WIN32
	threadsCount = 1;
#endif

	this->scheduler = schedulerIn;
	while
		(0 < threadsCount--)
	{
		this->threads.emplace_back(std::thread(&ThreadsPool::processJobs, this));
	}
	return nullptr != this->scheduler;
};

/**
 *
 */
bool ThreadsPool::submitJob(Job& job, std::string& jobId)
{
	bool ok;	//

	// Scope based locking.
	{
		// Put unique lock on task mutex.
		std::unique_lock<std::mutex> lock(this->tasksMutex);

		// Push task into queue.
		ok = this->scheduler->addJob(job, jobId);
	}

	Log::log(LogLevel::Debug, "ThreadsPool", "job submitted: [{}] {}", ok, jobId);

	// Wake up one thread.
	if
		(true == ok)
	{
		this->condition.notify_one();
	}
	else
	{
		Log::log(LogLevel::Error, "Threads pool", "Scheduler refused to add job");
	}

	return ok;
};

/**
 *
 */
void ThreadsPool::processJobs()
{
	Job job;	//
	//bool ok;	//

	while
		(true)
	{
		// Scope based locking.
		{
			// Put unique lock on task mutex.
			std::unique_lock<std::mutex> lock(this->tasksMutex);

			// Wait until queue is not empty or termination signal is sent.
			this->condition.wait(lock, [this]{ return (0 < this->scheduler->countPendingJobs()) || this->terminationRequested; });

			// If termination signal received and queue is empty then exit else continue clearing the queue.
			if
				(  this->terminationRequested
				&& 0 >= this->scheduler->countPendingJobs())
			{
				return;
			}

			// Get next task in the queue.
			this->scheduler->selectJob(job);
		}

		// Execute the task.
		job.start();
		/*ok = */job.task(job.errorMsg);
		job.done();

		// Required if the task is a lambda coming from a plugin
		job.task = nullptr;
	}
};

/**
 *
 */
bool ThreadsPool::cancelJob(const std::string& jobId)
{
	bool ok;

	ok = this->scheduler->removeJob(jobId);

	return ok;
};

/**
 *
 */
bool ThreadsPool::shutdown()
{
	// Scope based locking.
	{
		// Put unique lock on task mutex.
		std::unique_lock<std::mutex> lock(this->tasksMutex);

		// Set termination flag to true.
		this->terminationRequested = true;
	}

	// Wake up all threads.
	this->condition.notify_all();

	// Join all threads.
	for(std::thread& thread : this->threads)
	{
		thread.join();
	}

	// Empty workers vector.
	this->threads.clear();

	// Indicate that the pool has been shut down.
	this->terminated = true;

	return true;
};
