// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/timeline_cfg.h"

using namespace SEAScope;

/**
 *
 */
Timeframe::Timeframe(const std::string& labelIn, const std::string& descriptionIn, timestamp_t pastOffsetIn, timestamp_t futureOffsetIn)
: label(labelIn), description(descriptionIn), pastOffset(pastOffsetIn), futureOffset(futureOffsetIn)
{
};

/**
 *
 */
Timeframe::Timeframe()
{
};

/**
 *
 */
Timeframe::~Timeframe()
{
};

/**
 *
 */
Timestep::Timestep(const std::string& labelIn, const std::string& descriptionIn, timestamp_t valueIn)
:label(labelIn), description(descriptionIn), value(valueIn)
{
};

/**
 *
 */
Timestep::Timestep()
{

};

/**
 *
 */
Timestep::~Timestep()
{
};

/**
 *
 */
TimelineCfg::TimelineCfg()
: minYear(1970), maxYear(9999), defaultTimespan(""), defaultTimestep(""),
  useNearestMode(false), useNearestInDayMode(false), _initialDatetime(0)
{
};

/**
 *
 */
TimelineCfg::~TimelineCfg()
{
};

/**
 *
 */
bool TimelineCfg::addTimeframe(const std::string& label, const std::string& description, timestamp_t pastOffset, timestamp_t futureOffset)
{
	/* TODO: check that label is available */
	this->timeframes.emplace_back(label, description, pastOffset, futureOffset);
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimeframe(const std::string& label, Timeframe& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); it++)
	{
		if
			(label == it->label)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimeframe(timestamp_t pastOffset, timestamp_t futureOffset, Timeframe& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); it++)
	{
		if
			(pastOffset == it->pastOffset && futureOffset == it->futureOffset)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimeframes(std::vector<std::string>& labels, std::vector<std::string>& descriptions) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); it++)
	{
		labels.push_back(it->label);
		descriptions.push_back(it->description);
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimeframes(std::map<std::string, std::array<timestamp_t, 2>>& result) const
{
	std::vector<Timeframe>::const_iterator it;	//

	it = this->timeframes.begin();
	for
		(; it != this->timeframes.end(); it++)
	{
		result.insert(std::pair<std::string, std::array<timestamp_t, 2>>(it->label, {{it->pastOffset, it->futureOffset}}));
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::addTimestep(const std::string& label, const std::string& description, timestamp_t value)
{
	/* TODO: check that label is available */
	this->timesteps.emplace_back(label, description, value);
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimestep(const std::string& label, Timestep& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); it++)
	{
		if
			(label == it->label)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimestep(timestamp_t value, Timestep& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); it++)
	{
		if
			(value == it->value)
		{
			result = *it;
			return true;
		}
	}

	return false;
};

/**
 *
 */
bool TimelineCfg::getTimesteps(std::vector<std::string>& labels, std::vector<std::string>& descriptions) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); it++)
	{
		labels.push_back(it->label);
		descriptions.push_back(it->description);
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::getTimesteps(std::map<std::string, timestamp_t>& result) const
{
	std::vector<Timestep>::const_iterator it;	//

	it = this->timesteps.begin();
	for
		(; it != this->timesteps.end(); it++)
	{
		result.insert(std::pair<std::string, timestamp_t>(it->label, it->value));
	}
	return true;
};

/**
 *
 */
bool TimelineCfg::setInitialDatetime(timestamp_t initialDatetime)
{
    this->_initialDatetime = initialDatetime;
    return true;
}

/**
 *
 */
timestamp_t TimelineCfg::getInitialDatetime() const
{
    return this->_initialDatetime;
}
