// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <vector>
#include <algorithm>

#include "core/log.h"
#include "core/timeline_model.h"

using namespace SEAScope;

/**
 *
 */
TimelineModel::TimelineModel(): currentCallback(nullptr), boundsCallback(nullptr), timeframeCallback(nullptr), timestepCallback(nullptr), coverageCallback(nullptr), nearestDataCallback(nullptr), nearestDataBoundCallback(nullptr), current(0), lowerBound(0), upperBound(0), timeframePastOffset(0), timeframeFutureOffset(0), timestep(std::numeric_limits<uint64_t>::max()), nearestInDayMode(false), nearestMode(false), forceNearestMode(false), forceNearestInDayMode(false)
{
};

/**
 *
 */
TimelineModel::~TimelineModel()
{
};

/**
 *
 */
bool TimelineModel::getCoverage(const CollectionId& /*collectionId*/, EventsCoverage& /*result*/) const
{
	return false;
};

/**
 *
 */
bool TimelineModel::getCoverage(const std::string& formula, EventsCoverage& result) const
{
	auto itCoverages = this->coverages.begin();
	if
		(this->coverages.end() == itCoverages)
	{
		return false;
	}

	if
		(0 == formula.compare("union"))
	{
        Log::log(LogLevel::Debug, "TimelineModel", "Starting Merge");
		for
			(; itCoverages != this->coverages.end(); itCoverages++)
		{
            Log::log(LogLevel::Debug, "TimelineModel", "\t Merging collection {}/{}", itCoverages->first.first, itCoverages->first.second);
			result.merge(itCoverages->second);
		}
        Log::log(LogLevel::Debug, "TimelineModel", "Merge complete");
		return true;
	}

	if
		(0 == formula.compare("intersection"))
	{
		result.clone(itCoverages->second);
		itCoverages++;
		for
			(; itCoverages != this->coverages.end(); itCoverages++)
		{
			result.intersect(itCoverages->second);
		}
		return true;
	}

	return false;
};

/**
 *
 */
bool TimelineModel::setCoverage(const CollectionId& collectionId, const EventsCoverage& coverage, bool silent)
{
	this->coverages[collectionId] = coverage;
	if
		(false == silent && nullptr != this->coverageCallback)
	{
		this->coverageCallback();
	}
	return true;
};

/**
 *
 */
bool TimelineModel::removeCoverage(const CollectionId& collectionId, bool silent)
{
	auto found = this->coverages.find(collectionId);
	if
		(this->coverages.end() == found)
	{
		return false;
	}
	this->coverages.erase(found);
	if
		(false == silent && nullptr != this->coverageCallback)
	{
		this->coverageCallback();
	}
	return true;
};

/**
 *
 */
bool TimelineModel::updateCoverages()
{
	if
		(nullptr != this->coverageCallback)
	{
		this->coverageCallback();
        return true;
	}
	return false;
}

/**
 *
 */
bool TimelineModel::setCurrent(timestamp_t timestamp, bool applyModes)
{
	Log::log(LogLevel::Debug, "Timeline model", "set datetime to {} vs {}", timestamp, this->current);
	bool yearChanged;	//
	bool monthChanged;	//
	bool dayChanged;	//
	std::tm* oldTime;	//
	std::tm* newTime;	//
	std::time_t oldT;	//
	std::time_t newT;	//
	std::int16_t oldYear;	//
	std::uint8_t oldMonth;	//
	std::uint8_t oldDay;	//

	if
		(true == applyModes)
	{
		oldT = static_cast<std::time_t>(this->current);
		oldTime = std::localtime(&oldT);
		oldYear = static_cast<std::int16_t>(oldTime->tm_year);
		oldMonth = static_cast<std::uint8_t>(oldTime->tm_mon);
		oldDay = static_cast<std::uint8_t>(oldTime->tm_mday);
		newT = static_cast<std::time_t>(timestamp);
		newTime = std::localtime(&newT);

		Log::log( LogLevel::Debug, "Timeline model", "old: {} {} {} {} --- new: {} {} {} {}"
				, oldT, 1900 + oldYear, 1 + oldMonth, oldDay
				, newT, 1900 + newTime->tm_year, 1 + newTime->tm_mon, newTime->tm_mday);

		yearChanged = (oldYear != newTime->tm_year);
		monthChanged = (oldMonth != newTime->tm_mon);
		dayChanged = (oldDay != newTime->tm_mday);

		if
			(true == this->nearestMode)
		{
			this->applyNearestMode(timestamp, yearChanged, monthChanged, dayChanged);
		}
		if
			(true == this->nearestInDayMode)
		{
			this->applyNearestInDayMode(timestamp, yearChanged, monthChanged, dayChanged);
		}
	}

	if
		(  timestamp == this->current
		|| timestamp < this->lowerBound
		|| timestamp >= this->upperBound)
	{
		Log::log(LogLevel::Debug, "TimelineModel", "Current not changed {} -> [{}, {}, {}]", this->current, this->lowerBound, timestamp, this->upperBound);
		return false;
	}
	this->current = timestamp;

	if
		(nullptr != this->currentCallback)
	{
		this->currentCallback(timestamp);
	}

	return true;
};

/**
 *
 */
bool TimelineModel::setBounds(timestamp_t lowerBoundIn, timestamp_t upperBoundIn)
{
	Log::log(LogLevel::Debug, "Timeline model", "set bounds to [{}, {}]", lowerBoundIn, upperBoundIn);
	if
		(  lowerBoundIn == this->lowerBound
		&& upperBoundIn == this->upperBound)
	{

		return false;
	}
	this->lowerBound = lowerBoundIn;
	this->upperBound = upperBoundIn;

	if
		(nullptr != this->boundsCallback)
	{
		this->boundsCallback(lowerBoundIn, upperBoundIn);
	}

	return true;
};

/**
 *
 */
bool TimelineModel::setTimeframeOffsets(timestamp_t pastOffset, timestamp_t futureOffset)
{
	Log::log(LogLevel::Debug, "Timeline model", "set timeframe to [{}, {}]", pastOffset, futureOffset);
	if
		(  pastOffset == this->timeframePastOffset
		&& futureOffset == this->timeframeFutureOffset)
	{
		return false;
	}
	this->timeframePastOffset = pastOffset;
	this->timeframeFutureOffset = futureOffset;

	if
		(nullptr != this->timeframeCallback)
	{
		this->timeframeCallback(pastOffset, futureOffset);
	}
	return true;
};

/**
 *
 */
bool TimelineModel::setTimestep(timestamp_t timestepIn)
{
	Log::log(LogLevel::Debug, "Timeline model", "set timestep to {}", timestepIn);
	if
		(timestepIn == this->timestep)
	{
		return false;
	}
	this->timestep = timestepIn;

	if
		(nullptr != this->timestepCallback)
	{
		this->timestepCallback(timestepIn);
	}
	return true;
};

/**
 *
 */
bool TimelineModel::setCurrentCallback(const std::function<void(timestamp_t)>& callback)
{
	this->currentCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setBoundsCallback(const std::function<void(timestamp_t, timestamp_t)>& callback)
{
	this->boundsCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setTimeframeCallback(const std::function<void(timestamp_t, timestamp_t)>& callback)
{
	this->timeframeCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setTimestepCallback(const std::function<void(timestamp_t)>& callback)
{
	this->timestepCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::setCoverageCallback(const std::function<void()>& callback)
{
	this->coverageCallback = callback;
	return true;
};

/**
 *
 */
bool TimelineModel::goToPrev()
{
	timestamp_t modified;	//
	bool ok;	//

	if
		(0 == this->timestep)
	{
		modified = this->prev;
	}
	else
	{
		modified = this->current - this->timestep;
	}

	modified = std::max(this->lowerBound, modified);
	ok = this->setCurrent(modified, false);

	Log::log( LogLevel::Debug, "Timeline model"
			, "go to prev: {} {} {} => {}"
			, this->current, this->prev, this->lowerBound, modified);

	return ok;
};

/**
 *
 */
bool TimelineModel::goToNext()
{
	timestamp_t modified;	//
	bool ok;	//

	if
		(0 == this->timestep)
	{
		modified = this->next;
	}
	else
	{
		modified = this->current + this->timestep;
	}
	modified = std::min(this->upperBound, modified);
	ok = this->setCurrent(modified, false);

	Log::log( LogLevel::Debug, "Timeline model"
			, "go to next: {} {} {} => {}"
			, this->current, this->next, this->upperBound, modified);
	return ok;
};

/**
 *
 */
bool TimelineModel::setPrevNext(timestamp_t prevIn, timestamp_t nextIn)
{
	this->prev = prevIn;
	this->next = nextIn;
	return true;
};

/**
 *
 */
bool TimelineModel::setModes(bool nearestInDayModeIn, bool nearestModeIn)
{
	bool changed = false;	//

	if
		(nearestInDayModeIn != this->nearestInDayMode)
	{
		this->nearestInDayMode = nearestInDayModeIn;
		if
			(true == nearestInDayModeIn)
		{
			this->forceNearestInDayMode = true;
		}
		changed = true;
	}

	if
		(nearestModeIn != this->nearestMode)
	{
		this->nearestMode = nearestModeIn;
		if
			(true == nearestModeIn)
		{
			this->forceNearestMode = true;
		}
		changed = true;
	}

	if
		(true == changed)
	{
		this->setCurrent(this->current, true);
		if
			(nullptr != this->currentCallback)
		{
			this->currentCallback(this->current);
		}
	}
	return changed;
};

/**
 *
 */
bool TimelineModel::applyNearestInDayMode(timestamp_t& timestamp, bool /*yearChanged*/, bool /*monthChanged*/, bool dayChanged)
{
	timestamp_t newTimestamp;	//
	std::time_t t;	//
	std::tm* ts;	//
	bool changed;	//
	bool ok;	//

	if
		(  false == this->forceNearestInDayMode
		&& false == dayChanged)
	{
		return false;
	}

	this->forceNearestInDayMode = false;
	changed = false;
	t = static_cast<std::time_t>(timestamp);
	ts = std::localtime(&t);
	ok = this->isDataAvailable(static_cast<std::int16_t>(1900 + ts->tm_year), static_cast<std::uint8_t>(1 + ts->tm_mon), static_cast<std::uint8_t>(ts->tm_mday));
	if
		(ok)
	{
		ts->tm_hour = 12;
		ts->tm_min = 0;
		ts->tm_sec = 0;
		t = std::mktime(ts);
		newTimestamp = static_cast<timestamp_t>(t);
		if
			(nullptr != this->nearestDataBoundCallback)
		{
			this->nearestDataBoundCallback(newTimestamp);
			changed = (newTimestamp != timestamp);
			timestamp = newTimestamp;
		}
	}

	return changed;
};

/**
 *
 */
bool TimelineModel::update(bool forceNearestModeIn, bool forceNearestInDayModeIn)
{
	bool ok;	//

	Log::log( LogLevel::Debug, "Timeline model"
			, "Update timeline model");
	this->forceNearestMode = forceNearestModeIn;
	this->forceNearestInDayMode = forceNearestInDayModeIn;
	ok = this->setCurrent(this->current, true);
	return ok;
};

/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint16_t> years;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getYears(years);
	if
		(false == ok)
	{
		Log::log(LogLevel::Debug, "Timeline model", "Could not get years from coverage");
		return false;
	}
	auto yearFound = std::find( years.begin()
							  , years.end()
							  , year);
	return years.end() != yearFound;
}

/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year, std::uint8_t month) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint8_t> months;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getMonths(static_cast<std::uint16_t>(year), months);
	if
		(false == ok)
	{
		Log::log(LogLevel::Debug, "Timeline model", "Could not get months from coverage");
		return false;
	}
	auto monthFound = std::find( months.begin()
							   , months.end()
							   , month);
	return months.end() != monthFound;
};

/**
 *
 */
bool TimelineModel::isDataAvailable(std::int16_t year, std::uint8_t month, std::uint8_t day) const
{
	bool ok;	//
	EventsCoverage currentCoverage;	//
	std::vector<std::uint8_t> days;	//

	ok = this->getCoverage("union", currentCoverage);
	if
		(false == ok)
	{
		return false;
	}

	ok = currentCoverage.getDays(static_cast<std::uint16_t>(year), month, days);
	if
		(false == ok)
	{
		Log::log(LogLevel::Debug, "Timeline model", "Could not get days from coverage");
		return false;
	}
	auto dayFound = std::find( days.begin()
							 , days.end()
							 , day);
	return days.end() != dayFound;
};

/**
 *
 */
bool TimelineModel::applyNearestMode(timestamp_t& timestamp, bool yearChanged, bool monthChanged, bool dayChanged)
{
	std::tm* ts;	//
	std::time_t t;	//
	bool goToNearest;	//

	Log::log(LogLevel::Debug, "Timeline model", "applyNearestMode: {} {} {} {}", timestamp, yearChanged, monthChanged, dayChanged);

	t = static_cast<std::time_t>(timestamp);
	ts = std::localtime(&t);

	goToNearest = this->forceNearestMode && ! this->isDataAvailable(static_cast<std::int16_t>(1900 + ts->tm_year), static_cast<std::uint8_t>(1 + ts->tm_mon), static_cast<std::uint8_t>(ts->tm_mday));
	goToNearest |= yearChanged && this->isDataAvailable(static_cast<std::int16_t>(1900 + ts->tm_year));
	goToNearest |= monthChanged && this->isDataAvailable(static_cast<std::int16_t>(1900 + ts->tm_year), static_cast<std::uint8_t>(1 + ts->tm_mon));
	this->forceNearestMode = false;

	if
		(  goToNearest
		&& nullptr != this->nearestDataCallback)
	{
		this->nearestDataCallback(timestamp);
	}
	return goToNearest;
};

/**
 *
 */
bool TimelineModel::shiftByYears(std::int16_t shift)
{
	std::tm* oldTime;	//
	std::time_t oldT;	//
    std::time_t t;  //
    timestamp_t newTimestamp;   //

    oldT = static_cast<std::time_t>(this->current);
    oldTime = std::localtime(&oldT);

    oldTime->tm_year = oldTime->tm_year + shift;
    t = std::mktime(oldTime);
    newTimestamp = static_cast<timestamp_t>(t);
    return this->setCurrent(newTimestamp);
};

/**
 *
 */
bool TimelineModel::shiftByMonths(std::int32_t shift)
{
	std::tm* oldTime;	//
	std::time_t oldT;	//
    std::time_t t;  //
    timestamp_t newTimestamp;   //
    std::int32_t shiftedMonth; //
    std::int16_t yearsShift;    //

    oldT = static_cast<std::time_t>(this->current);
    oldTime = std::localtime(&oldT);

    shiftedMonth = shift + oldTime->tm_mon - 1;
    yearsShift = static_cast<std::int16_t>(std::floor(static_cast<float>(shiftedMonth) / 12.f));
    shiftedMonth = 1 + (shiftedMonth - 12 * yearsShift) % 12;

    oldTime->tm_year = oldTime->tm_year + yearsShift;
    oldTime->tm_mon = static_cast<std::uint8_t>(shiftedMonth);

    t = std::mktime(oldTime);
    newTimestamp = static_cast<timestamp_t>(t);
    return this->setCurrent(newTimestamp);
};

/**
 *
 */
bool TimelineModel::shiftByDays(std::int32_t shift)
{
    return this->setCurrent(this->current + static_cast<std::uint64_t>(shift * 86400));
};

/**
 *
 */
bool TimelineModel::shiftByTimeframeLength(std::int32_t shift, float portion = 1.f)
{
    float secondsShift; //
    timestamp_t seconds;    //
    timestamp_t timeframeLength;    //

    timeframeLength = this->timeframeFutureOffset + this->timeframePastOffset;
    secondsShift = portion * static_cast<float>(timeframeLength);
    seconds = static_cast<timestamp_t>(static_cast<float>(shift) * std::round(secondsShift));
    Log::log(LogLevel::Debug, "TimelineModel", "SHIFT {} {} {} {} {}", this->timeframePastOffset, this->timeframeFutureOffset, timeframeLength, secondsShift, seconds);
    return this->setCurrent(this->current + seconds);
};
