// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "data/events_coverage.h"

using namespace SEAScope;

/**
 *
 */
void EventsCoverage::reset()
{
    this->store.clear();
}

/**
 *
 */
bool EventsCoverage::clone(const EventsCoverage& other)
{
    this->reset();

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		this->store.insert(std::make_pair(itYears->first, itYears->second));
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::merge(const EventsCoverage& other)
{
	size_t i;	//

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		auto found = this->store.find(itYears->first);
		if
			(this->store.end() == found)
		{
            Log::log(LogLevel::Debug, "TimelineModel", "\t\t\tMerge inserted year {}", itYears->first);
			this->store.insert(std::make_pair(itYears->first, itYears->second));
		}
		else
		{
            Log::log(LogLevel::Debug, "TimelineModel", "\t\t\tMerge months for year {}", itYears->first);
			i = 0;
			for
				(; i<12; ++i)
			{
				this->store[itYears->first][i] |= itYears->second[i];
			}
		}
	}
	return false;
};

/**
 *
 */
bool EventsCoverage::intersect(const EventsCoverage& other)
{
	coverage_t result;	//
	uint32_t myMonth;	//
	uint32_t otherMonth;	//
	size_t i;	//
	bool atLeastOneMonth;	//

	auto itYears = other.store.begin();
	for
		(; itYears != other.store.end(); itYears++)
	{
		auto yearFound = this->store.find(itYears->first);
		if
			(this->store.end() == yearFound)
		{
			/* No coverage in this year */
			continue;
		}

		std::array<uint32_t, 12> yearData;
		atLeastOneMonth = false;
		i = 0;
		for
			(; i<12; ++i)
		{
			myMonth = yearFound->second[i];
			otherMonth = itYears->second[i];
			yearData[i] = myMonth & otherMonth;
			atLeastOneMonth = atLeastOneMonth || (0 < yearData[i]);
		}

		if
			(true == atLeastOneMonth)
		{
			result.insert(std::make_pair(itYears->first, yearData));
		}
	}

	this->store = result;
	return false;
};

/**
 *
 */
bool EventsCoverage::getDays(uint16_t year, uint8_t month, std::vector<uint8_t>& values) const
{
	uint8_t i;	//
	auto found = this->store.find(year);
	if
		(this->store.end() == found)
	{
		return false;
	}

	i = 0;
	for
		(; i < 31; ++i)
	{
		if
			(1 == ((found->second[month-1u] >> i) & 1) )
		{
			values.push_back(i+1u);
		}
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::getMonths(uint16_t year, std::vector<uint8_t>& values) const
{
	uint8_t i;	//
	auto found = this->store.find(year);
	if
		(this->store.end() == found)
	{
		return false;
	}

	i = 0;
	for
		(; i<12; ++i)
	{
		if
			(0 < found->second[i])
		{
			values.push_back(i+1u);
		}
	}
	return true;
};

/**
 *
 */
bool EventsCoverage::getYears(std::vector<uint16_t>& values) const
{
	auto it = this->store.begin();
	for
		(; it != this->store.end(); it++)
	{
		values.push_back(it->first);
	}
	return 0 < values.size();
};

/**
 *
 */
void EventsCoverage::display() const
{
	coverage_t::const_iterator it;
	std::array<uint8_t, 12> days_per_months = {{ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 }};
	uint8_t month_adjust;
	uint8_t days_count;
	size_t i;
	size_t j;

	it = this->store.begin();
	for
		(;it != this->store.end(); it++)
	{
		Log::log(LogLevel::Debug, "EventsCoverage", "{}",  static_cast<int>(it->first));
		i = 1u;
		for
			(;i<13; ++i)
		{
			month_adjust = (0 == it->first % 4 && 2 == i) ? 1u : 0u;	
			days_count = static_cast<uint8_t>(days_per_months[i-1u] + month_adjust);
			Log::log(LogLevel::Debug, "EventsCoverage", "\t{} [{}]: {}", static_cast<int>(i), static_cast<int>(days_count), static_cast<int>(it->second[i-1]));
			for
				(j=0; j<days_count; ++j)
			{
				Log::log(LogLevel::Debug, "EventsCoverage", "\t\t{}", (((it->second[i-1] >> j) & 1) == 1));
			}
		}		
	}
}
