// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <bitset>

#include "core/log.h"
#include "core/opengl_enum.h"
#include "data/data_handler.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/bbox.h"
#include "data/gpu_data_builder.h"
#include "3d/data_set.h"
#include "num_utils.h"

using namespace SEAScope;


/**
 *
 */
GPUDataBuilder::GPUDataBuilder()
{
};

/**
 *
 */
GPUDataBuilder::~GPUDataBuilder()
{
};

/**
 *
 */
bool GPUDataBuilder::buildDataSet(DataSet* data_set,
                                  void* raw_data,
                                  const DataInfo& dataInfo,
								  const CollectionCfg& colCfg,
                                  const Shape& shape,
                                  float resolution)
{
    Log::log(LogLevel::Debug, "GPUDataBuilder", "Create new dataset");

    // Initialize variables
    data_set->setWidth(dataInfo.xArity);              // number of elements in the X axis
    data_set->setHeight(dataInfo.yArity);             // number of elements in the Y axis
    data_set->setDimensionCount(dataInfo.channels);   // Number of variables interleaved in the data stream
    data_set->setResolution(resolution);              //

    float* fValuePtr = new float[dataInfo.channels];

    data_set->setOffset(fValuePtr);

    memcpy( fValuePtr
          , dataInfo.offsets.data()
          , sizeof(float) * dataInfo.offsets.size());

    fValuePtr = new float[dataInfo.channels];

    data_set->setScaleFactor(fValuePtr);

    memcpy( fValuePtr
          , dataInfo.scaleFactors.data()
          , sizeof(float) * dataInfo.scaleFactors.size());


    if (dataInfo.dataType == OGL::unsigned_byte)
    {
        data_set->setScaleFactorGL(255.f);
    }
    else
    {
        data_set->setScaleFactorGL(1.f);
    }

    // GCP
    unsigned short* gcpx = new unsigned short[static_cast<size_t>(shape.xArity * shape.yArity)];
    unsigned short* gcpy = new unsigned short[static_cast<size_t>(shape.xArity * shape.yArity)];
    float* lons = new float[static_cast<size_t>(shape.xArity * shape.yArity)];
    float* lats = new float[static_cast<size_t>(shape.xArity * shape.yArity)];
    int lon_size = shape.xArity;
    int lat_size = shape.yArity;

    bool sameVariationForLonAndX;
	bool sameVariationForLatAndY;
    bool lonsGrowing;
    bool latsGrowing;

	Log::log(LogLevel::Debug, "GPUDataBuilder", "Number of GCPs: {} {}x{}", shape.gcps.size(), shape.xArity, shape.yArity);

    for (std::size_t i = 0; i < shape.gcps.size(); ++i)
    {
        const SEAScope::GCP& gcp = shape.gcps[i];

        gcpx[i] = gcp.i;
        gcpy[i] = gcp.j;
        lons[i] = gcp.lon;
        lats[i] = gcp.lat;

		Log::log(LogLevel::Debug, "GPUDataBuilder", "@{} {} {} {} {}", i, gcpy[i], gcpx[i], lats[i], lons[i]);
    }

    data_set->setGcpx(gcpx);
    data_set->setGcpy(gcpy);
    data_set->setLon_size(shape.xArity);
    data_set->setLat_size(shape.yArity);
    data_set->setLons(lons);
    data_set->setLats(lats);

	data_set->setSeamLess(colCfg.xSeamless, colCfg.ySeamless);
	data_set->setNEWSAligned(colCfg.NEWSAligned);

	/* Compute the axis aligned bounding box
	    0 --------- 1
        |           |
        |           |
        3 --------- 2
	*/
    glm::vec2 corners[4];

    corners[0] = glm::vec2(lons[0], lats[0]);
    corners[1] = glm::vec2(lons[lon_size-1], lats[lon_size-1]);
    corners[2] = glm::vec2(lons[lon_size * lat_size - 1], lats[lon_size * lat_size - 1]);
    corners[3] = glm::vec2(lons[(lat_size-1) * lon_size], lats[(lat_size-1) * lon_size]);

    for (int i = 0; i < 4; ++i)
    {
        data_set->setCorner(i, glm::vec4(UnitSphere::ToVector3D(Geodetic2D(glm::radians(corners[i]))), 1.f));
    }

    glm::vec4 x0 = data_set->getCorner(1) - data_set->getCorner(0);
    glm::vec4 y0 = data_set->getCorner(3) - data_set->getCorner(0);

    glm::vec4 x = glm::normalize(x0);
    glm::vec4 y = glm::normalize(y0);

    glm::vec3 nDataSet = glm::normalize(glm::cross(glm::vec3(x.x, x.y, x.z), glm::vec3(y.x, y.y, y.z)));

    glm::vec2 middle2(data_set->getLons()[((data_set->getLat_size() / 2 - 1) * data_set->getLon_size() + data_set->getLon_size() / 2)], data_set->getLats()[((data_set->getLat_size() / 2 - 1) * data_set->getLon_size() + data_set->getLon_size() / 2)]);
    glm::vec4 middle3 = glm::vec4(UnitSphere::ToVector3D(Geodetic2D(glm::radians(middle2))), 1.f);

    glm::vec4 fmr = glm::normalize(middle3 - data_set->getCorner(0));

    glm::vec3 middleV = glm::vec3(fmr.x, fmr.y, fmr.z);

    float dotValue = glm::dot(nDataSet, middleV);

    if (dotValue < 0.f)
    {
        nDataSet = -nDataSet;
    }

    data_set->setNormal(nDataSet);
    data_set->setCorner(4, middle3);
    data_set->setCenterPlane(data_set->getCorner(0) + ((data_set->getCorner(2) - data_set->getCorner(0)) * 0.5f));

    glm::vec4 bboxDegrees = SEAScope::DataHandler::computeBbox(lons, lats, gcpx, gcpy, shape.xArity, shape.yArity, lonsGrowing, latsGrowing, sameVariationForLonAndX, sameVariationForLatAndY);
	if (colCfg.NEWSAligned && !colCfg.xSeamless)
	{
		if ((corners[0].x > corners[1].x) && lonsGrowing)
		{
			Log::log(LogLevel::Debug, "GPUDataBuilder", "Longitudes are supposed to increase between these two corners but that is not the case");
			bboxDegrees.x = corners[0].x;
			bboxDegrees.z = std::fmod(corners[1].x + 360.f - corners[0].x, 360.f) + corners[0].x;
			Log::log(LogLevel::Debug, "GPUDataBuilder", "corner0 = {} {}, corner1 = {} {}", corners[0].x, corners[0].y, corners[1].x, corners[1].y);
			Log::log(LogLevel::Debug, "GPUDataBuilder", "BBOX: {} {} {} {}", bboxDegrees.w, bboxDegrees.x, bboxDegrees.y, bboxDegrees.z);
		}
		else if ((corners[0].x < corners[1].x) && !lonsGrowing)
		{
			Log::log(LogLevel::Debug, "GPUDataBuilder", "Longitudes are supposed to decrease between these two corners but that is not the case");

			bboxDegrees.z = corners[0].x;
			bboxDegrees.x = std::fmod(corners[1].x - 360.f - corners[0].x, 360.f) + corners[0].x;
			Log::log(LogLevel::Debug, "GPUDataBuilder", "corner0 = {} {}, corner1 = {} {}", corners[0].x, corners[0].y, corners[1].x, corners[1].y);
			Log::log(LogLevel::Debug, "GPUDataBuilder", "BBOX: {} {} {} {}", bboxDegrees.w, bboxDegrees.x, bboxDegrees.y, bboxDegrees.z);
		}
	}

    data_set->setBbox(glm::vec4(bboxDegrees.x, bboxDegrees.w, bboxDegrees.z, bboxDegrees.y), !sameVariationForLonAndX, !sameVariationForLatAndY);
    data_set->addToBboxArray(glm::radians(bboxDegrees));

    // Buffer for the raw data
    Log::log(LogLevel::Debug, "GPUDataBuilder", "Copy data from void*");
    data_set->setData(dataInfo.xArity, dataInfo.yArity, dataInfo.channels, dataInfo.dataType, raw_data);

    // compute a mask to display only valid data
    Log::log(LogLevel::Debug, "GPUDataBuilder", "Compute mask");
    data_set->setDataMask(dataInfo.xArity, dataInfo.yArity, 1, OGL::unsigned_byte, nullptr);
    SEAScope::DataHandler::computeMask(*data_set->getData(), dataInfo.fillValues[0], *data_set->getDataMask());

    // Dilation for bilinear filtering
    SEAScope::DataHandler::dilate(data_set);

    return true;
};

/**
 *
 */
bool GPUDataBuilder::buildPolyline( void* raw_data
                                  , const DataInfo& dataInfo
								  , const CollectionCfg& /*colCfg*/
                                  , const Shape& shape
                                  , std::vector<Geodetic2D>& points
                                  , std::vector<float>& fData
								  , std::vector<bool>& mask)
{
	std::uint8_t* uintBuffer;	//
	float* floatBuffer;	//
	float scaleFactor;	//
	float offset;	//
	float physicalValue;	//
	float packedValue;	//
	bool unmasked;	//

	floatBuffer = nullptr;
	uintBuffer = nullptr;
    points.reserve(shape.xArity);

    for
        (unsigned int i = 0; i < shape.gcps.size(); ++i)
    {
        const SEAScope::GCP& gcp = shape.gcps[i];
        points.push_back(Geodetic2D( static_cast<double>(glm::radians(gcp.lon))
                                   , static_cast<double>(glm::radians(gcp.lat))));
    }

    if
        (1 == dataInfo.channels)
    {
		fData.reserve(shape.xArity);
		mask.reserve(shape.xArity);
		scaleFactor = dataInfo.scaleFactors[0];
		offset = dataInfo.offsets[0];
		if
			(dataInfo.dataType == OGL::float32)
		{
			floatBuffer = static_cast<float*>(raw_data);
		}
		else
		{
			uintBuffer = static_cast<std::uint8_t*>(raw_data);
		}
        for (int i = 0; i < shape.xArity; ++i)
        {
			if
				(dataInfo.dataType == OGL::float32)
			{
				packedValue = static_cast<float>(*(floatBuffer + i));
			}
			else
			{
				packedValue = static_cast<float>(*(uintBuffer + i));
			}
			unmasked = (false == floating_point_equality(dataInfo.fillValues[0], packedValue));
			physicalValue = offset + scaleFactor * packedValue;
			Log::log(LogLevel::Debug, "GPUDataBuilder", "value: {} + {} * {} = {}", offset, scaleFactor, packedValue, physicalValue);
			mask.push_back(unmasked);
            fData.push_back(physicalValue);
        }
    }

	floatBuffer = nullptr;
	uintBuffer = nullptr;

    return true;
};
