// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <fstream>
#include <ctime>

#include "vendors/_ini_parser.h"

#include "core/log.h"
#include "core/path.h"
#include "core/utils.h"
#include "ini_settings.h"

using namespace SEAScope;

/**
 * Constructor.
 */
INISettings::INISettings()
{
};

/**
 * Destructor.
 */
INISettings::~INISettings()
{

};

/**
 * Load settings from INI configuration file.
 */
bool INISettings::load(Settings& result, std::string& configPath)
{
	bool ok;	//

	ok = this->loadINI(result, configPath);
	return ok;
};

/**
 *
 */
bool INISettings::splitList(const std::string& input, const std::string& delim, std::vector<std::string>& output) const
{
	size_t start = 0U;
	auto end = input.find(delim);

	while
		(end != std::string::npos)
	{
		output.push_back(input.substr(start, end - start));
		start = end + delim.length();
		end = input.find(delim, start);
	}
	output.push_back(input.substr(start, end));
	return 0 < output.size();
};

/**
 *
 */
bool INISettings::loadTimespans(Settings& result, const std::string& timespansStr)
{
	std::vector<std::string> timespans;

	this->splitList(timespansStr, ",", timespans);
	for
		(const auto& timespan : timespans)
	{
		Log::log(LogLevel::Debug, "Settings INI", "Extracted timespan: {}" , timespan);
		std::vector<std::string> timespanComponents;
		this->splitList(timespan, "|", timespanComponents);
		if
			(4 > timespanComponents.size())
		{
			for
				(const auto& cmp : timespanComponents)
			{
				Log::log(LogLevel::Warn, "Settings INI", "Incomplete timespan: {}" , cmp);
			}
			continue;
		}
		result.timelineCfg.addTimeframe( timespanComponents[0]
									   , timespanComponents[1]
									   , static_cast<timestamp_t>(std::atol(timespanComponents[2].c_str()))
									   , static_cast<timestamp_t>(std::atol(timespanComponents[3].c_str())));
	}

	return true;
};

/**
 *
 */
bool INISettings::loadTimesteps(Settings& result, const std::string& timestepsStr)
{
	std::vector<std::string> timesteps;

	this->splitList(timestepsStr, ",", timesteps);
	for
		(const auto& timestep : timesteps)
	{
		Log::log(LogLevel::Debug, "Settings INI", "Extracted timestep: {}" , timestep);
		std::vector<std::string> timestepComponents;
		this->splitList(timestep, "|", timestepComponents);
		if
			(3 > timestepComponents.size())
		{
			for
				(const auto& cmp : timestepComponents)
			{
				Log::log(LogLevel::Warn, "Settings INI", "Incomplete timestep: {}" , cmp);
			}
			continue;
		}
		result.timelineCfg.addTimestep( timestepComponents[0]
									  , timestepComponents[1]
									  , static_cast<timestamp_t>(std::atol(timestepComponents[2].c_str())));
	}

	return true;
};

static bool validate(const std::string& uri, INI::Level& root)
{
    std::vector<std::string> missing;   //
    const std::array<std::string, 7> requiredPaths = {"plugins",
                                                      "colormaps",
                                                      "operators",
                                                      "data",
                                                      "state",
                                                      "index",
                                                      "annotations"};
    const std::array<std::string, 11> requiredRendering = {"glRenderer",
                                                           "glContextMajor",
                                                           "glContextMinor",
                                                           "windowWidth",
                                                           "windowHeight",
                                                           "antialiasing",
                                                           "antialiasingType",
                                                           "antialiasingFactor",
                                                           "fullscreen",
                                                           "vsync",
                                                           "minGranuleScreenEstate"};
    const std::array<std::string, 10> requiredTimeline = {"minYear",
                                                          "maxYear",
                                                          "timespans",
                                                          "timesteps",
                                                          "defaultTimespan",
                                                          "defaultTimestep",
                                                          "useNearestMode",
                                                          "useNearestInDayMode",
                                                          "advancedScrolling",
                                                          "initialDatetime"};
    const std::array<std::string, 3> requiredMisc = {"worldMap",
                                                     "skipIndexCheck",
                                                     "searchRequestMinDelay"};
    const std::array<std::string, 3> requiredProcessor = {"enabled",
                                                          "address",
                                                          "port"};

	try
	{
		INI::Parser p(uri.c_str());
		root = p.top();
	}
	catch
		(std::runtime_error& e)
	{
		Log::log( LogLevel::Error, "Settings INI"
				, "Error during parsing of {}"
				, uri);
		Log::log( LogLevel::Debug, "CollectionINIParser"
				, "{}"
				, e.what());
		return false;
	}

	// Check [paths] section
	if
		(root.sections.end() == root.sections.find("paths"))
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "[paths] section missing in {}", uri);
		return false;
	}

    INI::Level& paths = root("paths");
	for
		(const auto& field : requiredPaths)
	{
		if
			(paths.values.end() == paths.values.find(field))
		{
			missing.push_back(fmt::format("Field: [paths] {}", field));
		}
	}


	// Check [rendering] section
	if
		(root.sections.end() == root.sections.find("rendering"))
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "[rendering] section missing in {}", uri);
		return false;
	}

    INI::Level& rendering = root("rendering");
	for
		(const auto& field : requiredRendering)
	{
		if
			(rendering.values.end() == rendering.values.find(field))
		{
			missing.push_back(fmt::format("Field: [rendering] {}", field));
		}
	}

	// Check [timeline] section
	if
		(root.sections.end() == root.sections.find("timeline"))
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "[timeline] section missing in {}", uri);
		return false;
	}

    INI::Level& timeline = root("timeline");
	for
		(const auto& field : requiredTimeline)
	{
		if
			(timeline.values.end() == timeline.values.find(field))
		{
			missing.push_back(fmt::format("Field: [timeline] {}", field));
		}
	}

	// Check [misc] section
	if
		(root.sections.end() == root.sections.find("misc"))
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "[misc] section missing in {}", uri);
		return false;
	}

    INI::Level& misc = root("misc");
	for
		(const auto& field : requiredMisc)
	{
		if
			(misc.values.end() == misc.values.find(field))
		{
			missing.push_back(fmt::format("Field: [misc] {}", field));
		}
	}

	// Check [processor] section
	if
		(root.sections.end() == root.sections.find("processor"))
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "[processor] section missing in {}", uri);
		return false;
	}

    INI::Level& processor = root("processor");
	for
		(const auto& field : requiredProcessor)
	{
		if
			(processor.values.end() == processor.values.find(field))
		{
			missing.push_back(fmt::format("Field: [processor] {}", field));
		}
	}

	// Return false if some fields are missing
	if
		(0 < missing.size())
	{
		Log::log(LogLevel::Error, "Settings iNI",
				 "Configuration file {} is incomplete, the following elements have not been found:", uri);
		for
			(const auto& msg : missing)
		{
			Log::log(LogLevel::Error, "Settings INI", msg);
		}
		return false;
	}

    return true;
}

/**
 *
 */
bool INISettings::loadINI(Settings& result, const std::string& uri)
{
    bool ok(false); //
	INI::Level root;	//

    ok = validate(uri, root);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Settings INI",
                 "Validation of the {} configuration file failed", uri);
		return false;
	}

    /* paths */
    INI::Level& paths = root("paths");
    result.pluginsPath = paths["plugins"];
    result.colormapsPath = paths["colormaps"];
    result.operatorsPath = paths["operators"];
    result.dataPath = paths["data"];
    result.statePath = paths["state"];
    result.indexPath = paths["index"];
    result.annotationsPath = paths["annotations"];

    /* rendering */
    INI::Level& rendering = root("rendering");
    result.glRenderer = rendering["glRenderer"];
    result.glContextMajor = static_cast<uint8_t>(std::atoi(rendering["glContextMajor"].c_str()));
    result.glContextMinor = static_cast<uint8_t>(std::atoi(rendering["glContextMinor"].c_str()));
    result.windowWidth = static_cast<uint16_t>(std::atoi(rendering["windowWidth"].c_str()));
    result.windowHeight = static_cast<uint16_t>(std::atoi(rendering["windowHeight"].c_str()));
    result.vsync =  "true" == rendering["vsync"];
    result.antialiasing = "true" == rendering["antialiasing"];
    result.antialiasingType = rendering["antialiasingType"];
    result.antialiasingFactor = static_cast<uint8_t>(std::atoi(rendering["antialiasingFactor"].c_str()));
    result.fullscreen = "true" == rendering["fullscreen"];
    result.minGranuleScreenEstate = static_cast<float>(std::atof(rendering["minGranuleScreenEstate"].c_str()));

    /* timeline */
    INI::Level& timeline = root("timeline");
    result.timelineCfg.minYear = static_cast<std::uint16_t>(std::atoi(timeline["minYear"].c_str()));
    const std::string& maxYear = timeline["maxYear"];
    if
        (maxYear.empty())
    {
        std::time_t timestamp = time(nullptr);
        std::tm* now = std::gmtime(&timestamp);
        result.timelineCfg.maxYear = 1900u + static_cast<std::uint16_t>(now->tm_year);
    }
    else
    {
        result.timelineCfg.maxYear = static_cast<std::uint16_t>(std::atoi(maxYear.c_str()));
    }
    this->loadTimespans(result, timeline["timespans"]);
    result.timelineCfg.defaultTimespan = timeline["defaultTimespan"];
    this->loadTimesteps(result, timeline["timesteps"]);
    result.timelineCfg.defaultTimestep = timeline["defaultTimestep"];
    result.timelineCfg.useNearestMode = "true" == timeline["useNearestMode"];
    result.timelineCfg.useNearestInDayMode = "true" == timeline["useNearestInDayMode"];
    result.timelineCfg.advancedScrolling = "true" == timeline["advancedScrolling"];
    const std::string& initialDatetime = timeline["initialDatetime"];
    timestamp_t timestamp(0);
    if
        (initialDatetime.empty())
    {
	timestamp = static_cast<timestamp_t>(std::time(nullptr));  // now
    }
    else
    {
        ok = SEAScope::getTimestampFromDatetime(initialDatetime, timestamp);
        if
            (false == ok)
        {
            // failed to parse initialDatetime
            return false;
        }
    }
    result.timelineCfg.setInitialDatetime(timestamp);

    /* misc */
    INI::Level& misc = root("misc");
    result.defaultColormap = misc["defaultColormap"];
    result.baseLayer = misc["baseLayer"];
    result.worldMap = misc["worldMap"];
    result.skipIndexCheck = "true" == misc["skipIndexCheck"];

    result.searchRequestMinDelay = 0.1; // 100 ms
    if
        ("" != misc["searchRequestMinDelay"])
    {
        result.searchRequestMinDelay = std::atof(misc["searchRequestMinDelay"].c_str());
    }

    /* processor */
    INI::Level& processor = root("processor");
    result.processorEnabled = "true" == processor["enabled"];
    result.processorAddress = processor["address"];
    result.processorPort = std::atoll(processor["port"].c_str()); // assumes int64_t <=> long long

	return true;

};


/**
 * Save settings to file.
 */
bool INISettings::save(const Settings& settings, const std::string& configPath) const
{
	bool ok;	//
	std::string envConfigPath;	//

	ok = this->saveINI(settings, configPath);
	return ok;
};

/**
 *
 */
bool INISettings::saveINI(const Settings& settings, const std::string& uri) const
{
	std::stringstream buffer;	//
	std::ostringstream tmp;	//
	std::ofstream output;		//

	buffer << "[paths]\n";
	buffer << "plugins = " << settings.pluginsPath << "\n";
	buffer << "colormaps = " << settings.colormapsPath << "\n";
	buffer << "operators = " << settings.operatorsPath << "\n";
	buffer << "data = " << settings.dataPath << "\n";
	buffer << "state = " << settings.statePath << "\n";
	buffer << "index = " << settings.indexPath << "\n";
    buffer << "annotations = " << settings.annotationsPath << "\n";
	buffer << "[rendering]\n";
	buffer << "glRenderer = " << settings.glRenderer << "\n";
	buffer << "glContextMajor = " << static_cast<uint16_t>(settings.glContextMajor) << "\n";
	buffer << "glContextMinor = " << static_cast<uint16_t>(settings.glContextMinor) << "\n";
	buffer << "windowWidth = " << settings.windowWidth << "\n";
	buffer << "windowHeight = " << settings.windowHeight << "\n";
	buffer << "antialiasing = " << (settings.antialiasing ? "true":"false") << "\n";
	buffer << "antialiasingType = " << settings.antialiasingType << "\n";
	buffer << "antialiasingFactor = " << static_cast<uint16_t>(settings.antialiasingFactor) << "\n";
	buffer << "fullscreen = " << (settings.fullscreen ? "true":"false") << "\n";
	buffer << "vsync = " << (settings.vsync ? "true":"false") << "\n";
	buffer << "minGranuleScreenEstate = " << settings.minGranuleScreenEstate << "\n";
	buffer << "[timeline]\n";
	buffer << "minYear = " << settings.timelineCfg.minYear << "\n";
	buffer << "maxYear = " << settings.timelineCfg.maxYear << "\n";

	/* Timespans */
	tmp << "timespans = ";
	for
		(const auto& timespan : settings.timelineCfg.timeframes)
	{
		tmp << timespan.label << "|" << timespan.description << "|" << timespan.pastOffset << "|" << timespan.futureOffset << ",";
	}
	if
		(14 < tmp.str().size())
	{
		buffer << tmp.str().substr(0, tmp.str().size() - 1) << "\n";
	}
	else
	{
		buffer << tmp.str() << "\n";
	}
	tmp.str("");

	/* Timesteps */
	tmp << "timesteps = ";
	for
		(const auto& timestep : settings.timelineCfg.timesteps)
	{
		tmp << timestep.label << "|" << timestep.description << "|" << timestep.value << ",";
	}
	if
		(12 < tmp.str().size())
	{
		buffer << tmp.str().substr(0, tmp.str().size() - 1) << "\n";
	}
	else
	{
		buffer << tmp.str() << "\n";
	}
	tmp.str("");

	buffer << "defaultTimespan = " << settings.timelineCfg.defaultTimespan << "\n";
	buffer << "defaultTimestep = " << settings.timelineCfg.defaultTimestep << "\n";
    buffer << "useNearestMode = " << settings.timelineCfg.useNearestMode << "\n";
    buffer << "useNearestInDayMode = " << settings.timelineCfg.useNearestInDayMode << "\n";
    buffer << "initialDatetime = " << "" << "\n";

	buffer << "[misc]\n";
	buffer << "defaultColormap = " << settings.defaultColormap << "\n";
	buffer << "baseLayer = " << settings.baseLayer << "\n";
	buffer << "worldMap = " << settings.worldMap << "\n";
	buffer << "skipIndexCheck = " << (settings.skipIndexCheck ? "true":"false") << "\n";

	buffer << "[processor]\n";
	buffer << "enabled = " << (settings.processorEnabled ? "true":"false") << "\n";
	buffer << "address = " << settings.processorAddress << "\n";
	buffer << "port = " << settings.processorPort << "\n";

	output.open(uri, std::ofstream::out);
	if
		(false == output.is_open())
	{
		Log::log( LogLevel::Error, "Settings INI"
				, "Could not open '{}' with write permissions"
				, uri);
		return false;
	}

	try
	{
		INI::Parser p(buffer);	//
		p.dump(output);
	}
	catch
		(std::runtime_error& e)
	{
		Log::log( LogLevel::Error, "Settings INI"
				, "Could not save settings in '{}'"
			    , uri);
		Log::log( LogLevel::Debug, "Settings INI"
				, e.what());
		output.close();
		return false;
	}

	return true;
};
