// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <iomanip>
#include "core/log.h"
#include "local/granules_index.h"
#include <mutex>

using namespace SEAScope;

/**
 *
 */
GranulesIndex::GranulesIndex(): idGenerator(1), granuleAddedCallback(nullptr)
{

};

/**
 *
 */
GranulesIndex::~GranulesIndex()
{

};

/**
 *
 */
std::uint64_t GranulesIndex::generateGranuleId()
{
	return this->idGenerator++; // return current value, then increment
}

/**
 *
 */
bool GranulesIndex::hasGranule(const std::string& filePath) const
{
	auto found = this->ids.find(filePath);
	return this->ids.end() != found;
};

/**
 *
 */
bool GranulesIndex::hasGranule(std::uint64_t granuleId) const
{
	auto found = this->data.find(granuleId);
	return this->data.end() != found;
};

/**
 *
 */
bool GranulesIndex::add(GranuleMetadata& info, bool silent)
{
    std::uint64_t localId;  //

    auto found = this->ids.find(info.dataId);
    if
        (this->ids.end() != found)
    {
        localId = this->ids[info.dataId];

        // Just add the new uri to the list
        this->data[localId].uris.push_back(info.uris[0]);
		std::sort( this->data[localId].uris.begin()
				 , this->data[localId].uris.end()
				 , [](const IDFDescriptor& i, const IDFDescriptor& j) -> bool {
			return i.resolution < j.resolution;
		});
        return true;
    }

	if
		(0 < info.id.second)
	{
		localId = info.id.second;
	}
	else
	{
		localId = this->generateGranuleId();
		info.id.second = localId;
	}

    info.id.first = 0; // default source id
    {  // Lock scope
        std::lock_guard<std::mutex> lock(this->mapsMutex);
	    this->ids[info.dataId] = localId;
	    this->data[localId] = info;
    }

	if
		(false == silent && nullptr != this->granuleAddedCallback)
	{
		this->granuleAddedCallback(localId, info.collectionId);
	}
    Log::log( LogLevel::Debug, "GranulesIndex"
            , "Add granule {} {}/{}"
            , info.dataId, info.id.first, info.id.second);
	return true;
};

/**
 *
 */
bool GranulesIndex::updateTimeCoverage(std::uint64_t granuleId, const timestamp_t start, const timestamp_t stop)
{
	auto found = this->data.find(granuleId);
    if
        (this->data.end() == found)
    {
        return false;
    }

    {  // Lock scope
        std::lock_guard<std::mutex> lock(this->mapsMutex);
        found->second.start = start;
        found->second.stop = stop;
    }

    return true;
}

/**
 *
 */
bool GranulesIndex::remove(const std::string& dataId, bool /*silent*/)
{
    std::uint64_t localId;  //

    auto found = this->ids.find(dataId);
    if
        (this->ids.end() == found)
    {
        return false;
    }

    localId = found->second;
    this->data.erase(localId);
    this->ids.erase(found);
    return true;
};

/**
 *
 */
std::array<bool, 2> GranulesIndex::adjacentGranules( const timestamp_t current
									, const timestamp_t minDelta
									, const std::set<std::uint32_t>& collections
									, GranuleMetadata& prev
									, GranuleMetadata& next
									) const
{
	timestamp_t before(0);	//
	timestamp_t after(0);	//
	timestamp_t central(0);	//
	bool prevFound(false);	//
	bool nextFound(false);	//
	timestamp_t pastDelta(0);	//
	timestamp_t futureDelta(0);	//
	timestamp_t delta(0);	//

	pastDelta = std::numeric_limits<uint64_t>::max();
	futureDelta = std::numeric_limits<uint64_t>::max();

	Log::log(LogLevel::Debug, "GranulesIndex", "current: {}, minDelta: {}", current, minDelta);
	for
		(auto& it : this->data)
	{
		Log::log(LogLevel::Debug, "GranulesIndex", "Check granule {}", it.first);
		/* Check that granule belongs to one of the requested collections */
		auto found = std::find( collections.begin()
							  , collections.end()
							  , it.second.collectionId);
		if
			(collections.end() == found)
		{
			Log::log( LogLevel::Debug, "GranulesIndex"
					, "Granule skipped: wrong collection {}"
					, it.second.uris[0].uri);
			continue;
		}

		central = static_cast<timestamp_t>(0.5 * static_cast<double>(it.second.start) + 0.5 * static_cast<double>(it.second.stop));
		Log::log(LogLevel::Debug, "GranulesIndex", "{}: central is {}", it.first, central);
		if
			(central < current)
		{
			/* Granule in the past */
			delta = current - central;

			Log::log(LogLevel::Debug, "GranulesIndex", "granule is a past candidate: {} {} {}", delta, minDelta, central - current);
			if
				(minDelta > delta)
			{
				Log::log(LogLevel::Debug, "GranulesIndex", "Skipped 'coz minDelta not ok: {} {} {}", minDelta, central, current);
				continue;
			}
			if
				(  false == prevFound
				|| delta < pastDelta
				|| central > before)
			{
				pastDelta = delta;
				before = central;
				prev = it.second;
				prevFound = true;
			}
		}
		else if
			(central > current)
		{
			/* Granule in the future */
			delta = central - current;

			Log::log(LogLevel::Debug, "GranulesIndex", "granule is a future candidate: {} {} {}", delta, minDelta, central - current);
			if
				(minDelta > delta)
			{
				Log::log(LogLevel::Debug, "GranulesIndex", "Skipped 'coz minDelta not ok: {} {} {}", minDelta, central, current);
				continue;
			}

			if
				(  false == nextFound
				|| delta < futureDelta
				|| central < after)
			{
				futureDelta = delta;
				after = central;
				next = it.second;
				nextFound = true;
			}
		}
	}

	return std::array<bool, 2>({{ prevFound, nextFound }});
};

/**
 *
 */
bool GranulesIndex::closestGranuleBound( const timestamp_t current
									, const std::set<std::uint32_t>& collections
									, GranuleMetadata& closest
									) const
{
	timestamp_t lowerBound;	//
	timestamp_t upperBound;	//
	bool closestFound;	//
	timestamp_t absDiff;	//
	timestamp_t delta;	//

	closestFound = false;
	lowerBound = current - 43200000; // -12h
	upperBound = current + 43200000; // +12h
	delta = std::numeric_limits<uint64_t>::max();

	for
		(auto& it : this->data)
	{
		/* Check that granule belongs to one of the requested collections */
		auto found = std::find( collections.begin()
							  , collections.end()
							  , it.second.collectionId);
		if
			(collections.end() == found)
		{
			Log::log( LogLevel::Debug, "GranulesIndex"
					, "Granule skipped: wrong collection {}"
					, it.second.uris[0].uri);
			continue;
		}

		if
			(  it.second.start <= current
			&& it.second.stop >= current)
		{
			closest = it.second;
			if
				(it.second.stop == current)
			{
				delta = 1000;
			}
			else
			{
				return true;
			}
		}

		if
			(  it.second.start >= lowerBound
			&& it.second.start < upperBound)
		{
			if
				(it.second.start > current)
			{
				absDiff = it.second.start - current;
			}
			else
			{
				absDiff = current - it.second.start;
			}

			if
				(absDiff < delta)
			{
				delta = absDiff;
				closest = it.second;
				closestFound = true;
			}
		}

		if
			(  it.second.stop >= lowerBound
			&& it.second.stop < upperBound)
		{
			if
				(it.second.stop > current)
			{
				absDiff = it.second.stop - current;
			}
			else
			{
				absDiff = current - it.second.stop;
			}

			if
				(absDiff < delta)
			{
				delta = absDiff;
				closest = it.second;
				closestFound = true;
			}
		}
	}

	return closestFound;
};

/**
 *
 */
bool GranulesIndex::getCoverage(std::uint32_t collectionId, EventsCoverage& result) const
{
	timestamp_t d;	// Milliseconds
	timestamp_t end;	// Milliseconds
	std::uint16_t year; //
	std::uint8_t month;	//
	std::uint8_t day;	//
	time_t timestamp;	// Seconds
	std::tm* t;	//

    result.reset();

	for
		(auto& it : this->data)
	{
        if
            (it.second.collectionId != collectionId)
        {
			/* granule does not belong to a requested collection */
            continue;
        }

		d = it.second.start;
		timestamp = static_cast<time_t>(0.001 * static_cast<double>(d));
		Log::log(LogLevel::Debug, "GranulesIndex", "Original timestamp: {}", timestamp);
		t = std::localtime(&timestamp);
		t->tm_hour = 0;
		t->tm_min = 0;
		/* Up to debate: do we consider granules to be available one day 
		if their end date is at 00:00? The following implementation excludes granules
		that do not spend at least one second the day.
		*/
		t->tm_sec = 1;
		timestamp = std::mktime(t);
		Log::log(LogLevel::Debug, " GranulesIndex", "Modified timestamp: {}", 1000 * timestamp);

		d = static_cast<timestamp_t>(timestamp * static_cast<time_t>(1000));
		end = it.second.stop;
		while
			(d <= end)
		{
			timestamp = static_cast<long>(0.001 * static_cast<double>(d));
			t = std::localtime(&timestamp);
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//Log::log(LogLevel::Debug, "GranulesIndex", "Coverage {} {} {}", year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);
			d = d + 86400000;
		}
    }
	return true;
};

/**
 *
 */
bool GranulesIndex::getCoverage(const std::set<std::uint32_t>& collections, EventsCoverage& result) const
{
	timestamp_t d;	// Milliseconds
	timestamp_t end;	// Milliseconds
	std::uint16_t year;
	std::uint8_t month;	//
	std::uint8_t day;	//
	time_t timestamp;	// Seconds
	std::tm* t;	//

    result.reset();

	for
		(auto& it : this->data)
	{
        auto found = std::find( collections.begin()
                              , collections.end()
                              , it.second.collectionId);
        if
            (collections.end() == found)
        {
			/* granule does not belong to a requested collection */
            continue;
        }

		d = it.second.start;
		timestamp = static_cast<time_t>(0.001 * static_cast<double>(d));
		Log::log(LogLevel::Debug, "GranulesIndex", "Original timestamp: {}", timestamp);
		t = std::localtime(&timestamp);
		t->tm_hour = 0;
		t->tm_min = 0;
		/* Up to debate: do we consider granules to be available one day 
		if their end date is at 00:00? The following implementation excludes granules
		that do not spend at least one second the day.
		*/
		t->tm_sec = 1;
		timestamp = std::mktime(t);

		Log::log(LogLevel::Debug, " GranulesIndex", "Modified timestamp: {}", 1000 * timestamp);

		d = static_cast<timestamp_t>(timestamp * static_cast<time_t>(1000));
		end = it.second.stop;
		while
			(d <= end)
		{
			timestamp = static_cast<long>(0.001 * static_cast<double>(d));
			t = std::localtime(&timestamp);
			year = static_cast<std::uint16_t>(1900 + t->tm_year);
			month = static_cast<std::uint8_t>(t->tm_mon);
			day = static_cast<std::uint8_t>(t->tm_mday);

			//Log::log(LogLevel::Debug, "GranulesIndex", "Collection coverage {}: {} {} {}", it.second.collectionId, year, month+1, day);
			if
				(result.store.end() == result.store.find(year))
			{
				result.store[year] = std::array<std::uint32_t, 12>();
			}
			result.store[year][month] |= 1 << (day - 1);
			d = d + 86400000;
		}
	}
	return true;
};

/**
 *
 */
bool GranulesIndex::findGranules(const SearchCriteria& criteria, std::vector<GranuleMetadata>& result) const
{
    bool collectionFound(false);   //
    bool mustBeCurrent(false); //

    for
        (auto& it: this->data)
    {
        collectionFound = false;
        for
            (const auto& collection : criteria.collections)
        {
            if
                (collection.first == it.second.collectionId)
            {
                mustBeCurrent = collection.second;
                collectionFound = true;
                break;
            }
        }

        if
            (false == collectionFound)
        {
            Log::log( LogLevel::Debug, "GranulesIndex"
                    , "Granule skipped because it does not belong to a requested collection: {} {}"
                    , it.first, it.second.collectionId);
            continue;
        }

        if
            (  it.second.start < 1000 * criteria.stop
            && it.second.stop > 1000 * criteria.start)
        {
            if
                (  true == mustBeCurrent
                && ( 1000 * criteria.current < it.second.start
                   || 1000 * criteria.current >= it.second.stop))
            {
                // Skip
                Log::log( LogLevel::Debug, "GranulesIndex"
                        , "Skipped 'coz not current");
                continue;
            }

            result.push_back(it.second);
        }
    }
	return 0 < result.size();
};
