// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <mutex>

#include "core/log.h"
#include "core/utils.h"
#include "local/in_memory_data_store.h"

using namespace SEAScope;

/**
 *
 */
InMemoryDataStore::InMemoryDataStore()
{
};

/**
 *
 */
InMemoryDataStore::~InMemoryDataStore()
{

};

/**
 *
 */
GranuleData& InMemoryDataStore::getDataBucket(const std::string& uri)
{
    const std::lock_guard<std::mutex> lock(this->granuleDataMutex);

	auto found = this->data.find(uri);
	if
		(this->data.end() == found)
	{
        // Create a new data bucket
        this->data.emplace(std::make_pair(uri, GranuleData()));
	}

	return this->data[uri];
};

/**
 *
 */
bool InMemoryDataStore::hasData(const std::string& uri, const std::string& field) const
{
	auto found = this->data.find(uri);
	if
		(this->data.end() == found)
	{
		return false;
	}

	auto fieldFound = found->second.find(field);
	if
		(found->second.end() == fieldFound)
	{
		return false;
	}
	return true;
};

/**
 *
 */
Image* InMemoryDataStore::getData(const std::string& uri, const std::set<std::string>& fields, DataInfo& info, bool& ok)
{
	Image* result(nullptr);	//
	uint8_t* buffer(nullptr);	//
	uint8_t* resultBuffer(nullptr);	//
	unsigned int dataWidth(0);	//
	unsigned int dataHeight(0);	//
	std::map<std::string, DataBucket>::const_iterator fieldFound;	//
	uint8_t pixelsOffset(0);	//
	size_t i(0);	//
	size_t j(0);	//

	if
		(0 >= fields.size())
	{
		Log::log(LogLevel::Error, "Data store", "No field requested");
		ok = false;
		return nullptr;
	}

	auto found = this->data.find(uri);
	if
		(this->data.end() == found)
	{
		Log::log( LogLevel::Error, "Data store"
				, "Data for granule {}  not found"
				, uri);
		ok = false;
		return nullptr;
	}

	if
		(1 == fields.size())
	{
		auto it = fields.begin();
		fieldFound = found->second.find(*it);
		if
			(found->second.end() == fieldFound)
		{
			Log::log( LogLevel::Error, "Data store"
					, "Data for granule {} not found for field {}"
					, uri, *it);
			auto itDebug = found->second.begin();
			for
				(; itDebug != found->second.end(); itDebug++)
			{
				Log::log(LogLevel::Debug, "Data store", " - field {}", itDebug->first);
			}
			ok = false;
			return nullptr;
		}

		info = fieldFound->second.info;
		result = const_cast<Image*>(&fieldFound->second.buffer);
		ok = true;
		return result;
	}
	else
	{
		std::string key;	//
		key = uri + ":" + SEAScope::implodeStringSet(fields, ",", key);
		auto compositeFound = this->compositeData.find(key);
		if
			(this->compositeData.end() != compositeFound)
		{
			info = compositeFound->second.info;
			result = &compositeFound->second.buffer;
			ok = true;
			return result;
		}

		this->compositeData[key];
		DataInfo& inf = compositeData[key].info;
		Image& img = compositeData[key].buffer;

		// Check that all requested fields are in store
		auto itFields = fields.begin();
		for
			(; itFields != fields.end(); itFields++)
		{
			fieldFound = found->second.find(*itFields);
			if
				(found->second.end() == fieldFound)
			{
				Log::log( LogLevel::Error, "Data store"
						, "Data for granule {} not found for field {} **"
						, uri, *itFields);
				ok = false;
				return nullptr;
			}

			dataWidth = fieldFound->second.buffer.width();
			dataHeight = fieldFound->second.buffer.height();
		}

		img.allocate( dataWidth, dataHeight, static_cast<unsigned int>(fields.size())
					, OGL::unsigned_byte, nullptr);
		resultBuffer = static_cast<uint8_t*>(img.data());
		pixelsOffset = static_cast<uint8_t>(img.channels());

		inf.xArity = static_cast<std::uint16_t>(dataWidth);
		inf.yArity = static_cast<std::uint16_t>(dataHeight);
		inf.channels = static_cast<std::uint8_t>(fields.size());
		inf.dataType = OGL::unsigned_byte;

		itFields = fields.begin();
		for
			(j=0; itFields != fields.end(); itFields++, ++j)
		{
			fieldFound = found->second.find(*itFields);
			buffer = static_cast<uint8_t*>(fieldFound->second.buffer.data());

			// Interleave data values
			for
				(i=0; i<dataWidth*dataHeight;++i)
			{
				resultBuffer[i * pixelsOffset + j] = buffer[i];
			}

			size_t fieldChannels = fieldFound->second.info.offsets.size();
			for
				(size_t channel=0; channel < fieldChannels; ++channel)
			{
				inf.offsets.push_back(fieldFound->second.info.offsets[channel]);
				inf.scaleFactors.push_back(fieldFound->second.info.scaleFactors[channel]);
				inf.fillValues.push_back(fieldFound->second.info.fillValues[channel]);
			}
		}

		info = this->compositeData[key].info;
		result = &this->compositeData[key].buffer;

		ok = true;
		return result;
	}

    /*
	Log::log(LogLevel::Error, "Data store", "getData should never reach here");
	ok = false;
	return nullptr;
    */
};

/**
 *
 */
bool InMemoryDataStore::getFieldData(const std::string& uri, const std::string& field, DataInfo& info, Image& result) const
{
	std::map<std::string, DataBucket>::const_iterator fieldFound;	//

	auto found = this->data.find(uri);
	if
		(this->data.end() == found)
	{
		Log::log( LogLevel::Error, "Data store"
				, "Data for granule {}  not found"
				, uri);
		return false;
	}

	fieldFound = found->second.find(field);
	if
		(found->second.end() == fieldFound)
	{
		Log::log( LogLevel::Error, "Data store"
				, "Data for granule {} not found for field {}"
				, uri, field);
		return false;
	}

	info = fieldFound->second.info;
	result.allocate( fieldFound->second.buffer.width()
				   , fieldFound->second.buffer.height()
				   , fieldFound->second.buffer.channels()
				   , fieldFound->second.buffer.dataType()
				   , fieldFound->second.buffer.data());

	return true;
};
