// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "local/user_generated.h"
#include "core/app_events_dispatcher.h"
#include "events/granule_added_event.h"

using namespace SEAScope;

/**
 *
 */
UserGenerated::UserGenerated(SourceId sourceId, AppEventsDispatcher& appEventsDispatcher)
: _sourceId(sourceId),
  _appEventsDispatcher(appEventsDispatcher)
{
	this->index.granuleAddedCallback = [this] (std::uint64_t granuleId, std::uint32_t collectionId) {
        GranuleAddedEvent event(this->_sourceId, collectionId, granuleId);
        this->_appEventsDispatcher.dispatch<GranuleAddedEvent>(event);
    };
};

/**
 *
 */
UserGenerated::~UserGenerated()
{
};

/**
 *
 */
bool UserGenerated::init(const Settings&)
{
    return true;
};

/**
 *
 */
bool UserGenerated::reset()
{
    return true;
}

/**
 *
 */
bool UserGenerated::searchGranules( const SearchCriteria& criteria
                                  , std::vector<GranuleMetadata>& result) const
{
	bool ok;	//
	ok = this->index.findGranules(criteria, result);
	return ok;
};

/**
 *
 */
std::array<bool, 2> UserGenerated::searchPrevNext( timestamp_t current
                                                 , const std::set<CollectionId>& collections
                                                 , timestamp_t& prev
                                                 , timestamp_t& next
                                                 , timestamp_t minDelta) const
{
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata prevGranule;	//
	GranuleMetadata nextGranule;	//

	Log::log( LogLevel::Debug, "UserIndex"
			, "searchPrevNext: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	Log::log( LogLevel::Debug, "UserIndex"
			, "Local collections list built");

	auto oks = this->index.adjacentGranules( 1000 * current
										   , 1000 * minDelta
										   , localCollections
										   , prevGranule, nextGranule);

	Log::log( LogLevel::Debug, "UserIndex"
			, "Adjacent granules ok");
	Log::log( LogLevel::Debug, "UserIndex"
			, "{} {}"
			, oks[0], oks[1]);
	if
		(true == oks[0])
	{
		prev = static_cast<timestamp_t>(0.0005 * static_cast<double>(prevGranule.start)  + 0.0005 * static_cast<double>(prevGranule.stop));
	}
	if
		(true == oks[1])
	{
		next = static_cast<timestamp_t>(0.0005 * static_cast<double>(nextGranule.start) + 0.0005 * static_cast<double>(nextGranule.stop));
	}

	Log::log( LogLevel::Debug, "UserIndex"
			, "Returning from searchPrevNext");

    return oks;
};

/**
 *
 */
bool UserGenerated::searchClosestBound( timestamp_t current
                                      , const std::set<CollectionId>& collections
                                      , timestamp_t& closest) const
{
	timestamp_t delta;	//
	timestamp_t closestTmp;	//
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata closestGranule;	//
	bool ok;	//

	Log::log( LogLevel::Debug, "UserIndex"
			, "searchClosestBoundt: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	Log::log( LogLevel::Debug, "UserIndex"
			, "Local collections list built");
    Log::log( LogLevel::Debug, "UserIndex"
            , "number of local collections: {}", localCollections.size());

	ok = this->index.closestGranuleBound(1000 * current, localCollections, closestGranule);
	if
		(true == ok)
	{
        Log::log( LogLevel::Debug, "UserIndex", "closest granule found: {} {} {}"
                , closestGranule.uris[0].uri
                , closestGranule.start
                , closestGranule.stop);

		if
			(  closestGranule.start <= 1000 * current
			&& closestGranule.stop >= 1000 * current)
		{
			/* A granule contains the current datetime, no bounds required */
			closest = current;
			if
				(closestGranule.stop ==  1000 * current)
			{
				/* if granule end, remove one second or the granule will not be displayed */
				closest -= 1;
			}
			return true;
		}
		closestTmp = current;
		delta = std::numeric_limits<uint64_t>::max();
		if
			(  closestGranule.start < current
			&& delta > current - closestGranule.start)
		{
			delta = current - closestGranule.start;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.start));
		}
		else if
			(  closestGranule.start > current
			&& delta > closestGranule.start - current)
		{
			delta = closestGranule.start - current;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.start));
		}

		/* Remove one second when using the end of the granule as a result
		   or the granule will no be displayed */
		if
			(  closestGranule.stop < current
			&& delta > current - closestGranule.stop)
		{
			delta = current - closestGranule.stop;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.stop)) - 1;
		}
		else if
			(  closestGranule.stop > current
			&& delta > closestGranule.stop - current)
		{
			delta = closestGranule.stop - current;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.stop)) - 1;
		}

		closest = closestTmp;
	}

	Log::log(LogLevel::Debug, "UserIndex", "closest bound: {}", closest);
    return ok;
};

/**
 *
 */
bool UserGenerated::getDataCoverage( const std::set<CollectionId>& collections
                                   , EventsCoverage& result) const
{
	std::set<std::uint32_t> localCollections;	//
	bool ok;	//

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	ok = this->index.getCoverage(localCollections, result);
	return ok;
};

/**
 *
 */
bool UserGenerated::getDataCoverage( const CollectionId& collectionId
                                   , EventsCoverage& result) const
{
	bool ok;	//

	ok = this->index.getCoverage(collectionId.second, result);
	return ok;
};

/**
 *
 */
bool UserGenerated::canHandle(const std::string&) const
{
    return false;
}

const SourceId& UserGenerated::getSourceId() const
{
    return this->_sourceId;
}

/**
 *
 */
bool UserGenerated::generateRenderableId(const CollectionId& colId, RenderableId& result)
{
    if
        (this->_sourceId != colId.first)
    {
        Log::log(LogLevel::Error, "User generated",
                 "Generation of granule identifier for an external collection is not allowed");
        return false;
    }

    result.granuleLevel = true;
    result.collection = colId;
    result.granule.first = colId.first; /* same source as collection */
    result.granule.second = this->index.generateGranuleId();
    result.variable = 0; /* default value */

    return true;
}
