// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstdio>
#include <cstdint>
#include <cstdlib>
#include <map>
#include <set>
#include <array>
#include <vector>
#include <string>
#include <sstream>
#include <iomanip>

#include "vendors/_cxxopts.h"
#include "vendors/_nanogui.h"

#include "num_utils.h"

#include "core/log.h"
#include "core/dialog.h"
#include "core/path.h"
#include "core/guid.h"
#include "core/application.h"
#include "simplelogger.h"
#include "filter_logger.h"
#include "ini_settings.h"
#include "sqlite_state.h"
#include "glfw_window.h"
#include "ui/nanogui/nanogui_ui.h"
#include "ui/nanogui/nanogui_blueprint_factory.h"
#include "sink_eventhandler.h"

#include "local/in_memory_data_store.h"
#include "dummy_metadata_store.h"

#include "data/events_coverage.h"

#include "core/types.h"
#include "core/structs.h"
#include "core/timeline_cfg.h"
#include "3d/resources.h"
#include "3d/shape.h"
#include "3d/texture_loader.h"

#include "ui/nanogui/nanogui_advanced_rendering_control.h"
#include "ui/nanogui/nanogui_timeline.h"
#include "ui/nanogui/nanogui_map_caption.h"
#include "ui/nanogui/nanogui_main_menu.h"
#include "ui/nanogui/nanogui_rendering_menu.h"

#include "platform/plugins_loader_impl.h"
#include "platform/dialog_impl.h"
#include "platform/path_impl.h"
#include "platform/guid_impl.h"
#include "platform/sockets_monitor_impl.h"
#include "platform/socket_factory_impl.h"

#include "seascope_version.h"

#include "ui/nanogui/nanogui_camera_control.h"

//#include "core/measurements.h"

#include <fstream>
#include "sqlite_annotations.h"

#include "3d/memory_manager.h"
#include "3d/shapes_allocator.h"
#include "3d/datasets_allocator.h"

#include "3d/globe_scene.h"
#include "3d/scene_view.h"
#include "3d/scene_renderer.h"

#include "glfw_events_dispatcher.h"
#include "tools/window_controller.h"
#include "tools/debug_controller.h"
#include "tools/extraction_controller.h"
#include "tools/contextual_menu.h"

#include "core/app_events_dispatcher.h"
#include "events/camera_moved_event.h"

#include "tools/side_menu.h"
#include "events/side_menu_resize_request.h"
#include "events/display_catalogue_request.h"

#include "tools/annotations_controller.h"
#include "ui/nanogui/nanogui_toolbar.h"
#include "tools/drawing_controller.h"

#include "3d/icons.h"
#include "events/scene_view_updated_event.h"
#include "events/scene_view_controller_ready_event.h"
#include "events/drawing_controller_ready_event.h"
#include "core/workspace_folders.h"

using namespace SEAScope;


/**
 *
 */
int main(int argc, char** argv)
{
	/* Set specific implementations */
	Dialog::implementation = new DialogImpl();
	Path::implementation = new PathImpl();
	GUID::implementation = new GUIDImpl();

    { // Scope for stack-allocated objects

	INISettings settingsStore;	//
	SQLiteState stateStore;	//
	GLFWWindow window;			//
	PluginsLoaderImpl pLoader;	//
	SocketsMonitorImpl usm;	//
	SocketFactoryImpl sFactory;	//
	InMemoryDataStore imds;	//
	DummyMetadataStore dmds;	//

    AppEventsDispatcher appEventsDispatcher;
    SQLiteAnnotations annotations(appEventsDispatcher);  //

	NanoguiBlueprintFactory blueprintFactory;	//
	NanoguiUI ui(appEventsDispatcher);				//
	NanoguiTimeline* timelineUI;	//
	NanoguiAdvancedRenderingControl* advancedRenderingUI;	//
	NanoguiRenderingControl* renderingUI;	//
    NanoguiCameraControl* cameraUI;
    NanoguiToolbar* toolbarUI;

    GLFWEventsDispatcher inputEventsDispatcher;
	Application app(appEventsDispatcher);			//
    Icons icons;    //

	std::string configPath;	//
	std::string installDir;	//
	RendererResources resources;	//
	timestamp_t initTime;	//
	bool ok;					//
	std::vector<std::string> txt;
    std::string logFilter("");  //

#if defined __linux__ || defined __APPLE__
	/* Force UTC timezone */
	setenv("TZ", "UTC", 1);
#elif defined _WIN32
	_putenv_s("TZ", "UTC");
#endif

	/* Parse command line options */
	cxxopts::Options cliParser(argv[0], "Geolocated data visualization and manipulation tool.");
	cliParser.add_options()
		("v,version", "Print SEAScope version and build information")
        ("f,filter", "Level max of verbosity when debug is enabled (fatal, error, warn, info, debug)", cxxopts::value<std::string>(), "debug")
		("l,listen", "Open a control socket", cxxopts::value<std::string>(), "IFACE:PORT")
		("c,config", "Path to the configuration file", cxxopts::value<std::string>(), "")
        ("w,workspace", "Path of the directory where SEAScope is installed", cxxopts::value<std::string>(), "")
		("h,help", "Show this help message")
	;

	std::string listenURL = "";
    try
    {
        cxxopts::ParseResult options = cliParser.parse(argc, argv);

        /* Handle -h/--help */
        if
            (options.count("help"))
        {
            std::cout << cliParser.help({""}) << std::endl;
            return 0;
        }

        std::cout << "SEAScope is a 3D visualisation and analysis application for satellite, in-situ and\n";
        std::cout << "numerical model data.\n\n";
        std::cout << "Copyright (C) 2014-2022 OceanDataLab\n";
        std::cout << "License LGPLv3+\n\n";
        if
            (0 < options.count("version"))
        {
            std::cout << "Build information:\n";
            for
                (const auto& it : SEAScope::getVersion())
            {
                std::cout << "\t " << it.first << ": " << it.second << "\n";
            }
            return 0;
        }

        /* Handle -w/--workspace */
        if
            (0 == options.count("workspace"))
        {
            std::cerr << "You must specify a value for the -w argument" << std::endl;
            return 1;
        }
        installDir = options["workspace"].as<std::string>();
        Path::setPathValid(installDir);

        /* Handle -c/--config */
        if
            (0 == options.count("config"))
        {
            std::cerr << "You must specify a value for the -c argument" << std::endl;
            return 1;
        }
        configPath = options["config"].as<std::string>();
        Path::setPathValid(configPath);

		/* Handle -l/--listen */
		if
			(options.count("listen"))
		{
			listenURL = options["listen"].as<std::string>();
		}

		/* Handle -f/--filter */
        if (0 < options.count("filter"))
        {
            logFilter = options["filter"].as<std::string>();
        }
        FilterLogger* logger = new FilterLogger();
        ok = logger->init(logFilter);
        if
            (false == ok)
        {
            std::cerr << "Failed to initialize logger, please check the filter you provided for the -f argument" << std::endl;
            return 1;
        }
        Log::implementation = logger;
        logger = nullptr;

	}
    catch(cxxopts::OptionParseException& e)
    {
        std::cerr << e.what() << std::endl;
        return 1;
    }


	/* Initialize application */
	ok = app.init(installDir, configPath, settingsStore, stateStore, window, resources, logFilter);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Main", "Failed to initialize application");
		return 1;
	}

	/* Initialize UI */
	ok = ui.init(window, app.engine.settings, &app.engine.state);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Main", "Could not initialize user interface");
		return 1;
	}

    SideMenu sideMenu(appEventsDispatcher, ui.getRenderer(), app.engine.settings.processorEnabled);

    NanoguiRenderingMenu* rMenu = ui.getRenderingMenu();
	rMenu->window->display(Eigen::Vector2i(20, 124));
	renderingUI = rMenu->control;
	advancedRenderingUI = rMenu->advanced;

	timelineUI = new NanoguiTimeline(ui.getRenderer());

	cameraUI = new NanoguiCameraControl(ui.getRenderer(), appEventsDispatcher, inputEventsDispatcher);

	int width, height, fbWidth, fbHeight;

	// Nanogui computes a pixel ratio and may changes the window size.
	// Size in ini settings can be different than size used by nanogui.
	window.getFrameBufferSize(fbWidth, fbHeight);
    window.getSize(width, height);

    //float scaleFactor = static_cast<float>(fbWidth) / static_cast<float>(width);

    // Assemble the Scene MVC
    SceneRenderer sceneRenderer;    //
    sceneRenderer.initialize(app.engine.settings.glRenderer, &(app.engine.cmaps), &resources/*, scaleFactor*/);

	// Construct the path for the worldmap picture file. Data path + worldmap file name.
	std::string worldMapPath = app.engine.settings.dataPath + '/' + app.engine.settings.worldMap;

    // Load worldmap picture file (ownership will be given to the Globe object)
    OGL::Texture* worldMapTexture = TextureLoader::loadFromFile(sceneRenderer.getContext(), resources.worldmapPath);

    // Load pictograms used for drawing tools
    // TODO: use style manager for this
    std::string iconsDir = installDir + Path::separator() + "images" + Path::separator() + "icons";
    icons.setTexture(IconType::CIRCLE_OUTLINE_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "circle_outline_32.png");
    icons.setTexture(IconType::CIRCLE_DOT_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "circle_dot_32.png");
    icons.setTexture(IconType::CIRCLE_FILLED_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "circle_filled_32.png");
    icons.setTexture(IconType::LOSANGE_OUTLINE_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "losange_outline_32.png");
    icons.setTexture(IconType::CIRCLE_PLUS_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "circle_plus_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_FILLED_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "small_circle_filled_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_MOVE_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "small_circle_move_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_PLUS_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "small_circle_plus_32.png");
    icons.setTexture(IconType::SMALL_CIRCLE_OUTLINE_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "small_circle_outline_32.png");
    icons.setTexture(IconType::SMALL_CROSSHAIR_32, sceneRenderer.getContext(),
                     iconsDir + Path::separator() + "small_crosshair_32.png");

    GlobeScene scene(appEventsDispatcher);   //
    SceneView sceneView(appEventsDispatcher);    //
    sceneView.init(&scene, &sceneRenderer, worldMapTexture);
    sceneView.setViewport(glm::ivec2(fbWidth, fbHeight));

	/* Initialize engine */
	blueprintFactory.init(ui.getRenderer());
	app.engine.init(&sceneView, &dmds, &imds, &pLoader, &blueprintFactory);
	app.engine.setCollectionsUI(sideMenu.getMainMenu()->collectionsControl);
	app.engine.setRenderingUI(renderingUI);
	app.engine.setAdvancedRenderingUI(advancedRenderingUI);
	app.engine.setTimelineUI(timelineUI);
	app.engine.setSourcesUI(sideMenu.getMainMenu()->sourcesControl);
	app.engine.sockets.init(&usm, &sFactory);
	app.engine.dataMediator.workers = &app.dataWorkers;

	// Initialize UI
	ui.addComponent(timelineUI, "Timeline", "100%", "136px", "0%", "-0%", false);
	ui.addComponent(cameraUI, "Camera", "300px", "24px", "30%", "10px", false);

    toolbarUI = new NanoguiToolbar(appEventsDispatcher, inputEventsDispatcher);
    toolbarUI->init(ui.getRenderer());
    toolbarUI->showAt(Eigen::Vector2i(20, 48));

	ok = app.loadState();
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "Main", "Could not load state");
		return 1;
	}

	/* -------------------------------------------------------------------- */
	/*						C A L L B A C K S								*/
	/* -------------------------------------------------------------------- */


    /*
    mapRenderer->setDeleteVisibleUserShapesCallback([&app, &mapRenderer]() {
        bool success(false); //
        std::vector<RenderableId> toRemove;    //
        Log::log(LogLevel::Debug, "Main", "Delete all visible user shapes");

        // export all rendered shapes
        for
            (const auto& rId : mapRenderer->toRender)
        {
            if
                (  1 != rId.collection.first
                || (  1 != rId.collection.second
                   && 2 != rId.collection.second))
            {
                // Only user generated polylines and polygons can be deleted
                continue;
            }

            toRemove.push_back(rId);
        }

        if
            (0 >= toRemove.size())
        {
            Log::log(LogLevel::Debug, "Main", "No annotation to delete");
            return;
        }

        for
            (const auto& rId : toRemove)
        {
            mapRenderer->removeUserShape(rId);

            success = app.annotations->remove(rId.granule.second);
            if
                (false == success)
            {
                Log::log(LogLevel::Error, "Main", "Failed to remove annoation {}", rId.granule.second);
            }

            std::ostringstream buffer("");
		    buffer << "usershape_" << rId.collection.first << "/" << rId.collection.second << "/" << rId.granule.second << ":" <<rId.variable;
		    app.engine.userGenerated.index.remove(buffer.str(), false);
        }
        toRemove.clear();

        // Trigger update
        app.engine.updateCollections({{1u, 1u}, {1u, 2u}}, {});
    });
    */


	ui.setSaveLayoutCallback([&app]() {
		bool success(false);	//

		Log::log(LogLevel::Debug, "Main", "save layout...");
		success = app.saveState();
		if
			(false == success)
		{
			Log::log( LogLevel::Error, "Main", "Could not save layout");
		}
	});

	ui.setComponentResizeCallback([&appEventsDispatcher, &ui] (const std::string& cId, int w, int h) {
		Log::log(LogLevel::Debug, "Main", "Component resized: {} {}x{}", cId, w, h);
		if
			("Timeline" == cId)
		{
            SideMenuResizeRequest event(320, ui.getRenderer()->height() - h);
            appEventsDispatcher.dispatch<SideMenuResizeRequest>(event);
		}
	});

	ui.setResizeCallback([&appEventsDispatcher, 
#if defined _WIN32
		&ui,
#endif
		& timelineUI] (int /*width*/, int heightIn) {
        SideMenuResizeRequest event(320, heightIn - timelineUI->fixedHeight());
        appEventsDispatcher.dispatch<SideMenuResizeRequest>(event);

		/*Windows does not seem to receive the resize event for the timeline in
		  this case: force call to the resize callback for the timeline component*/
#if defined _WIN32
		std::function<void(const std::string&, int, int)> callback = ui.getComponentResizeCallback();
        callback("Timeline", timelineUI->fixedWidth(), timelineUI->fixedHeight());
#endif
	});

	/* Set timeline callbacks */
	app.engine.timeline.model.nearestDataCallback = [&app](timestamp_t& value) {
		bool success(false);	//
		bool validPast;	//
		bool validFuture;	//
		timestamp_t nearest;	//
		timestamp_t newPrev;	//
		timestamp_t newNext;	//
		timestamp_t pastDelta;	//
		timestamp_t futureDelta;	//
		timestamp_t minDelta;	//

		minDelta = 0;
		success = app.engine.metadataMediator.searchPrevNext( value
											  , app.engine.renderingControl.model.renderedCollections
											  , newPrev
											  , newNext
											  , minDelta);

		Log::log( LogLevel::Debug, "Main", "nearestDataCallback: {} {} -> {} or {}?"
				, success, value, newPrev, newNext);
		if
			(false == success)
		{
			// Do nothing if there is no closest granule.
			return;
		}

		pastDelta = value - newPrev;
		futureDelta = newNext - value;
		validPast = 0 < pastDelta;
		validFuture = 0 < futureDelta;
		if
			(  validPast
			&& validFuture)
		{
			nearest = (pastDelta < futureDelta) ? newPrev : newNext;
		}
		else if
			(validPast)
		{
			nearest = newPrev;
		}
		else if
			(validFuture)
		{
			nearest = newNext;
		}
		else
		{
			nearest = value;
		}

		Log::log(LogLevel::Debug, "Main", "******* {} {} {} {}", newPrev, newNext, value, nearest);
		value = nearest;
	};

	app.engine.timeline.model.nearestDataBoundCallback = [&app](timestamp_t& value) {
		bool success(false);	//
		timestamp_t closestBound;	//

		closestBound = 0;

		success = app.engine.metadataMediator.searchClosestBound( value
											  , app.engine.renderingControl.model.renderedCollections
											  , closestBound);
		if
			(false == success)
		{
			return;
		}

		value = closestBound;
	};

	app.engine.timeline.setCurrentChangedCallback([&app](timestamp_t value) {
		timestamp_t newPrev;	//
		timestamp_t newNext;	//
		timestamp_t minDelta;	//

		minDelta = 1;
		app.engine.metadataMediator.searchPrevNext( value
											  , app.engine.renderingControl.model.renderedCollections
											  , newPrev
											  , newNext
											  , minDelta);
		app.engine.timeline.model.setPrevNext(newPrev, newNext);
		Log::log(LogLevel::Debug, "Main", "{} <= {} <= {}", value, newPrev, newNext);

		Log::log(LogLevel::Debug, "Main", "UPDATE REQUESTED");
		app.engine.updateRequested = true;
	});
	app.engine.timeline.setTimeframeChangedCallback([&app](timestamp_t /*pastOffset*/, timestamp_t /*futureOffset*/) {
		Log::log(LogLevel::Debug, "Main", "UPDATE REQUESTED");
		app.engine.updateRequested = true;
	});

	app.engine.renderingControl.renderedCollectionsCallback = [&app](const std::set<CollectionId>& toAdd, const std::set<CollectionId>& toDel) -> bool{
		bool success(false);	//
		success = app.engine.updateCollections(toAdd, toDel);
		return success;
	};

	app.engine.sourcesControl.sourceInitialized = [&app] (const SourceInfo& /*info*/, const std::set<CollectionId>& collections) {
		bool success(false);	//
		Log::log(LogLevel::Debug, "Main", "Source initialized!!!!");
		for
			(const auto& it: collections)
		{
			Log::log(LogLevel::Debug, "Main", " - collection {}/{}", it.first, it.second);
		}
		success = app.engine.collectionsControl.model.addCollections(collections);
		if
			(false == success)
		{
			Log::log(LogLevel::Error, "Main", "Failed to add collections");
		}
	};

	rMenu->moreDataCallback = [&appEventsDispatcher]() {
        DisplayCatalogueRequest event;
        appEventsDispatcher.dispatch<DisplayCatalogueRequest>(event);
	};

	/* -------------------------------------------------------------------- */
	/*						D A T A    I N I T								*/
	/* -------------------------------------------------------------------- */

	/* init timeline data */
	//initTime = static_cast<timestamp_t>(std::time(nullptr));  // now
    initTime = app.engine.settings.timelineCfg.getInitialDatetime();
	if
		(initTime < app.engine.timeline.model.lowerBound)
	{
		initTime = app.engine.timeline.model.lowerBound;
	}
	if
		(initTime > app.engine.timeline.model.upperBound)
	{
		initTime = app.engine.timeline.model.upperBound - 1;
	}
	app.engine.timeline.model.setCurrent(initTime);

	/* Bootstrap demo */
	app.engine.sourcesControl.addSourceFromURI("embedded");

	// Try to locate a control plugin
	if
		(0 < listenURL.length())
	{
		std::forward_list<std::string> tokens;	//
		std::string iface;	//
		std::uint64_t port;	//

		Path::implementation->splitPaths(listenURL, ":", tokens);
		iface = tokens.front();
		tokens.pop_front();
		port = static_cast<std::uint64_t>(std::atol(tokens.front().c_str()));
		ok = app.openControlSocket(iface, port);
	}

    app.engine.metadataMediator.registerUserGeneratedIndex(1u, app.engine.getUserGenerated());

    /* --------------------------------------------------------------------- */
    /*                   I N I T  U S E R  S H A P E S                       */
    /* --------------------------------------------------------------------- */
    AnnotationsController annotationsController(appEventsDispatcher,
                                                annotations,
                                                app.engine.collectionsControl,
                                                app.engine.renderingControl,
                                                sceneView,
                                                icons);

    app.engine.metadataMediator.registerUserGeneratedIndex(2u, annotationsController.getUserGenerated());

    annotationsController.init(app.engine.settings.annotationsPath);

	/* -------------------------------------------------------------------- */
	/*						R E N D E R I N G								*/
	/* -------------------------------------------------------------------- */

	/* Init renderers */
	window.addRenderer(sceneView);
	window.addRenderer(*(ui.getRenderer()));

	/* Init events handlers */
	window.addHandler(*(ui.getRenderer()));

    // Make sure mouse move events will be expressed in framebuffer coords
    inputEventsDispatcher.setFrameBufferRatio(static_cast<double>(fbWidth / width),
                                              static_cast<double>(fbHeight / height));
    inputEventsDispatcher.initBindings(); // Must be done AFTER GLFW initialization
    window.addHandler(inputEventsDispatcher);

    WindowController* windowController = app.engine.getTools()->getWindowController();
    windowController->setWindow(&window);
    inputEventsDispatcher.addSubscriber(windowController);
    windowController->setActive(true);

    SceneViewControllerBase* sceneViewController = app.engine.getTools()->getSceneViewController();
    inputEventsDispatcher.addSubscriber(sceneViewController);
    sceneViewController->setActive(true);

    // notify toolbar that scene view controller is ready (to read keybindings)
    SceneViewControllerReadyEvent event(sceneViewController->getId());
    appEventsDispatcher.dispatch<SceneViewControllerReadyEvent>(event);

    ExtractionController* extractionController = app.engine.getTools()->getExtractionController();
    inputEventsDispatcher.addSubscriber(extractionController);
    extractionController->setActive(true);

    TransectController* transectController = app.engine.getTools()->getTransectController();
    inputEventsDispatcher.addSubscriber(transectController);
    transectController->setActive(true);

    DebugController* debugController = app.engine.getTools()->getDebugController();
    inputEventsDispatcher.addSubscriber(debugController);
    debugController->setActive(true);

    SelectionController* selectionController = app.engine.getTools()->getSelectionController();
    inputEventsDispatcher.addSubscriber(selectionController);
    sceneRenderer.registerPreRenderer(selectionController);
    sceneRenderer.registerPostRenderer(selectionController);
    selectionController->setActive(true);

    ContextualMenu ctxMenu(appEventsDispatcher, ui.getRenderer(), &window);

    DrawingController drawingController(appEventsDispatcher, sceneView, icons);
    drawingController.setId(6);
    drawingController.possibleActionsUpdated();
    inputEventsDispatcher.addSubscriber(&drawingController);
    sceneRenderer.registerPostRenderer(&drawingController);
    drawingController.setActive(true);

    DrawingControllerReadyEvent drawingControllerReadyEvent(drawingController.getId());
    appEventsDispatcher.dispatch<DrawingControllerReadyEvent>(drawingControllerReadyEvent);

    windowController = nullptr;
    sceneViewController = nullptr;
    extractionController = nullptr;
    transectController = nullptr;
    selectionController = nullptr;
    debugController = nullptr;

	/* Display catalogue in the sidemenu */
	rMenu->moreDataCallback();

    // Required to initialize altitude in the camera controller
    SEAScope::SceneViewUpdatedEvent eventForAltitudeInit(&sceneView);
    appEventsDispatcher.dispatch<SEAScope::SceneViewUpdatedEvent>(eventForAltitudeInit);

	SinkEventHandler seh;
	window.addHandler(seh);

    /* This is required on some systems (tested on Ubuntu 18.04 and 20.04)
    otherwise the Nanogui UI is not correctly initialized until the user
    manually resizes the window */
    int windowWidth(0);
    int windowHeight(0);
    window.getSize(windowWidth, windowHeight);
    ui.resize(windowWidth, windowHeight);

	ok = app.run(sceneView);

	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "main"
				, "Could not run application");
		return 1;
	}

	app.shutdown();

    delete toolbarUI;
    toolbarUI = nullptr;

    } // Scope for stack-allocated objects

	delete GUID::implementation;
	GUID::implementation = nullptr;

	delete Log::implementation;
	Log::implementation = nullptr;

	delete Path::implementation;
	Path::implementation = nullptr;

    delete Dialog::implementation;
    Dialog::implementation = nullptr;

    // Memory leaks
    MemoryManager::killInstance();

    ShapesAllocator::killInstance();

    DatasetsAllocator::killInstance();

    WorkspaceFolders::killInstance();

	return 0;
};
