// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <vector>

#include "vendors/_sqlite3.h"
#include "core/log.h"
#include "core/path.h"
#include "sqlite_state.h"

using namespace SEAScope;

/**
 *
 */
SQLiteState::SQLiteState()
{
};

/**
 *
 */
SQLiteState::~SQLiteState()
{
};

/**
 *
 */
bool SQLiteState::setStatePath(const std::string& path)
{
	if
		(false == Path::writeFile(path))
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "The state file {} is not writable"
				, path);
		return false;
	}
	this->statePath = path;
	return true;
};

/**
 *
 */
bool SQLiteState::load(State& result)
{
	std::int64_t size(0);	//
	bool creationNeeded(true);	//
	bool ok(false);	//

	ok = Path::fileExists(this->statePath);
	if
		(true == ok)
	{
		ok = Path::fileSize(this->statePath, size);
		creationNeeded = ok && 0 >= size;
	}

	if
		(true == creationNeeded)
	{
		/* Create an empty state */
		ok = this->createSQLite(this->statePath);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "State SQLite"
					, "Could not initialize state file");
			Path::deleteFile(statePath);
			return false;
		}
		Log::log(LogLevel::Debug, "State SQLite", "SQLite state created? {}", ok);
	}

	ok = this->loadSQLite(result, this->statePath);
	return ok;
};

/**
 *
 */
bool SQLiteState::save(const State& state) const
{
	bool ok(false);	//

	ok = this->saveSQLite(state, this->statePath);
	return ok;
};

/**
 *
 */
bool SQLiteState::createSQLite(const std::string& uri) const
{
	sqlite3* db(nullptr);	// SQLite3 database handle
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::string sql("");	// SQL query
	std::vector<std::string> queries;	// SQL queries

	/* Open database */
	rc = sqlite3_open(uri.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not open database {}: {}", uri, sqlite3_errmsg(db));
		return false;
	}

	/* Create metadata table */
	sql =	"CREATE TABLE seascope(" \
			"VERSION_MAJOR INT NOT NULL," \
			"VERSION_MINOR INT NOT NULL," \
			"VERSION_BUILD TEXT);";
	queries.push_back(sql);

	/* Populate metadata table */
	sql =	"INSERT INTO seascope(VERSION_MAJOR, VERSION_MINOR, VERSION_BUILD) " \
			"VALUES(0, 3, 'dev');";
	queries.push_back(sql);

	/* Create layout table */
	sql =	"CREATE TABLE layout(" \
			"NAME TEXT NOT NULL," \
			"ID TEXT NOT NULL," \
			"X TEXT NOT NULL," \
			"Y TEXT NOT NULL," \
			"WIDTH TEXT NOT NULL," \
			"HEIGHT TEXT NOT NULL," \
			"MOVABLE BOOLEAN NOT NULL," \
			"RESIZABLE BOOLEAN NOT NULL," \
			"PRIMARY KEY (NAME, ID));";
	queries.push_back(sql);

	/* Create rendering params table */
	sql =	"CREATE TABLE rendering(" \
			"SRC TEXT NOT NULL," \
			"ID INT NOT NULL," \
			"LOGSCALE BOOL," \
			"MIN REAL," \
			"MAX REAL," \
			"OPACITY REAL," \
			"ZINDEX REAL," \
			"LABEL TEXT NOT NULL," \
			"COLORMAP TEXT," \
			"METHOD INT NOT NULL," \
			"PARTICLES_COUNT INT," \
			"PARTICLES_TTL INT," \
			"STREAMLINE_LENGTH INT," \
			"STREAMLINE_SPRING REAL," \
			"X_SEAMLESS BOOL," \
			"Y_SEAMLESS BOOL," \
			"VARIABLES TEXT NOT NULL," \
			"MUST_BE_CURRENT BOOL," \
			"PRIMARY KEY(SRC,ID));";
	queries.push_back(sql);

	/* Create spacetime table */
	sql =	"CREATE TABLE spacetime(" \
			"CAM0 REAL," \
			"CAM1 REAL," \
			"CAM2 REAL," \
			"CAM3 REAL," \
			"CAM4 REAL," \
			"CAM5 REAL," \
			"CAM6 REAL," \
			"CAM7 REAL," \
			"CAM8 REAL," \
			"CAM9 REAL," \
			"CAM10 REAL," \
			"CAM11 REAL," \
			"CAM12 REAL," \
			"CAM13 REAL," \
			"CAM14 REAL," \
			"CAM15 REAL," \
			"WORLD0 REAL," \
			"WORLD1 REAL," \
			"WORLD2 REAL," \
			"WORLD3 REAL," \
			"WORLD4 REAL," \
			"WORLD5 REAL," \
			"WORLD6 REAL," \
			"WORLD7 REAL," \
			"WORLD8 REAL," \
			"WORLD9 REAL," \
			"WORLD10 REAL," \
			"WORLD11 REAL," \
			"WORLD12 REAL," \
			"WORLD13 REAL," \
			"WORLD14 REAL," \
			"WORLD15 REAL," \
			"TIMEFRAME_PAST INT," \
			"TIMEFRAME_FUTURE INT," \
			"TIMESTEP INT," \
			"CURRENT_TIMESTAMP);";
	queries.push_back(sql);

	auto it = queries.begin();
	for
		(; it != queries.end(); it++)
	{
		rc = sqlite3_exec( db
						 , it->c_str()
						 , nullptr
						 , nullptr
						 , &errMsg);
		if
			(SQLITE_OK != rc)
		{
			Log::log( LogLevel::Error, "State SQLite"
					, "Could not initialize database: {}", errMsg);
			Log::log( LogLevel::Debug, "State SQLite"
					, "Failed SQL query: {}", *it);
			sqlite3_free(errMsg);
			sqlite3_close(db);
			return false;
		}
	}

	sqlite3_close(db);
	return true;
};

/**
 *
 */
int SQLiteState::layoutRetrievedCallback(void* result, int argc, char** argv, char** colName)
{
	State* state(nullptr);	//
	std::string id("");	//
	std::string x("");	//
	std::string y("");	//
	std::string w("");	//
	std::string h("");	//
	bool movable(false);	//
	bool resizable(false);	//
	IBlueprint* blueprint(nullptr);	//
	int i(0);	//

	state = reinterpret_cast<State*>(result);
	if
		(nullptr == state)	
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not apply results to state");
		return 1;
	}

	Log::log( LogLevel::Debug, "State SQLite"
			, "layout retrieved");
	i = 0;
	for
		(; i<argc; ++i)
	{
		Log::log( LogLevel::Debug, "State SQLite"
				, "Key: {}, Value: {}", colName[i], argv[i]);
		if
			(0 == std::strncmp(colName[i], "ID", 2))
		{
			id = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "X", 1))
		{
			x = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "Y", 1))
		{
			y = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "WIDTH", 5))
		{
			w = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "HEIGHT", 6))
		{
			h = argv[i];
		}
		else if
			(0 == std::strncmp(colName[i], "MOVABLE", 7))
		{
			movable = (0 < std::atoi(argv[i]));
		}
		else if
			(0 == std::strncmp(colName[i], "RESIZABLE", 9))
		{
			resizable = (0 < std::atoi(argv[i]));
		}
	}

	auto found = state->layout.find(id);
	if
		(state->layout.end() == found)
	{
		blueprint = state->createBlueprint(resizable, movable);
		blueprint->setLayout(x,y,w,h);
		state->layout.insert(std::make_pair(id, blueprint));
	}
	else
	{
		found->second->setLayout(x,y,w,h);
	}
	return 0;
};

/**
 *
 */
int SQLiteState::renderParamsRetrievedCallback(void* /*result*/, int /*argc*/, char** /*argv*/, char** /*colName*/)
{
	return 1;
};

/**
 *
 */
int SQLiteState::timelineRetrievedCallback(void* /*result*/, int /*argc*/, char** /*argv*/, char** /*colName*/)
{
	return 1;
};

/**
 *
 */
int SQLiteState::cameraRetrievedCallback(void* /*result*/, int /*argc*/, char** /*argv*/, char** /*colName*/)
{
	return 1;
};

/**
 *
 */
bool SQLiteState::loadSQLite(State& result, const std::string& uri)
{
	sqlite3* db(nullptr);	// SQLite3 database handle
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::string sql("");	// SQL query

	/* Open database */
	rc = sqlite3_open(uri.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not open database {}: {}", uri, sqlite3_errmsg(db));
		return false;
	}

	/* Read layout */
	sql = "SELECT * FROM layout WHERE 1;";
	rc = sqlite3_exec( db
					 , sql.c_str()
					 , SQLiteState::layoutRetrievedCallback
					 , &result
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not retrieve layout: {}", errMsg);
		sqlite3_free(errMsg);
		sqlite3_close(db);
		return false;
	}

	sqlite3_free(errMsg);
	sqlite3_close(db);
	Log::log( LogLevel::Debug, "State SQLite"
			, "Load from SQLite complete");
	return true;
};

/**
 *
 */
bool SQLiteState::saveSQLite(const State& state, const std::string& uri) const
{
	sqlite3* db(nullptr);	// SQLite3 database handle
	int rc(-1);	// Return code for sqlite3 functions
	char* errMsg(nullptr);	// Buffer for SQLite error messages
	std::string sql("");	// SQL query
	std::ostringstream buffer("");	//
	std::string x("");	//
	std::string y("");	//
	std::string w("");	//
	std::string h("");	//
	IComponent* component(nullptr);	//

	Log::log(LogLevel::Debug, "State SQLite", "Save state");

	/* Open database */
	rc = sqlite3_open(uri.c_str(), &db);
	if
		(SQLITE_OK != rc)
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not open database {}: {}", uri, sqlite3_errmsg(db));
		return false;
	}

	/* Save layout */
	auto it = state.layout.begin();
	for
		(; it != state.layout.end(); it++)
	{
		component = it->second->getTarget();
		if
			(nullptr == component)
		{
			Log::log( LogLevel::Warn, "State SQLite"
					, "Component {} is null", it->first);
			continue;
		}
		it->second->getLayout(x,y,w,h);
		buffer << "INSERT OR REPLACE INTO layout ";
		buffer << "(NAME,ID,X,Y,WIDTH,HEIGHT,MOVABLE,RESIZABLE) ";
		buffer << "VALUES('USER','" << it->first << "'";
		buffer << ",'" << x << "','" << y << "'";
		buffer << ",'" << w << "','" << h << "'";
		buffer << "," << (component->isMovable() ? 1:0);
		buffer << "," << (component->isResizable() ? 1:0);
		buffer << ");";
	}

	Log::log( LogLevel::Debug, "State SQLite"
			, "Save layout: \n {}", buffer.str());
	rc = sqlite3_exec( db
					 , buffer.str().c_str()
					 , nullptr
					 , nullptr
					 , &errMsg);
	if
		(SQLITE_OK != rc)
	{
		Log::log( LogLevel::Error, "State SQLite"
				, "Could not save layout: {}", errMsg);
		sqlite3_free(errMsg);
		sqlite3_close(db);
		return false;
	}

	sqlite3_free(errMsg);
	sqlite3_close(db);
	return true;
};
