// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <fstream>

#include "tools/annotations_controller.h"


#include "core/log.h"

#include "core/interfaces/iannotations_store.h"

#include "core/app_events_dispatcher.h"
#include "local/user_generated.h"
#include "json_annotation.h"
#include "core/collections_control.h"
#include "core/rendering_control.h"
#include "3d/scene_view.h"
#include "3d/iscene.h"
#include "3d/iscene_renderer.h"
#include "3d/icons.h"

#include "events/collections_updated_event.h"
#include "events/add_prepared_annotation_request.h"
#include "events/collection_registration_request.h"
#include "events/save_one_annotation_to_file_request.h"
#include "events/load_annotations_from_file_request.h"
#include "events/save_annotations_to_file_request.h"
#include "events/create_annotations_collection_request.h"
#include "events/annotation_removed_event.h"
#include "events/rendering_cfg_update_event.h"
#include "events/granule_removal_request.h"
#include "events/collection_removed_event.h"
#include "events/shape_edition_complete_event.h"
#include "events/enable_shape_creation_request.h"
#include "events/enable_shape_creation_ready_event.h"
#include "events/annotation_time_coverage_update_request.h"
#include "events/annotation_time_coverage_updated_event.h"

using namespace SEAScope;

/**
 *
 */
AnnotationsController::AnnotationsController(AppEventsDispatcher& appEventsDispatcher, IAnnotationsStore& store, CollectionsControl& collectionsControl, RenderingControl& renderingControl, SceneView& sceneView, const Icons& icons)
: _appEventsDispatcher(appEventsDispatcher),
  _userGenerated(2u, appEventsDispatcher),
  _store(store),
  _collectionsControl(collectionsControl),
  _renderingControl(renderingControl),
  _sceneView(sceneView),
  _icons(icons),
  _collectionIdGenerator(1)
{
    this->_defaultColvars[ShapeType::POINT] = {{this->_userGenerated.getSourceId(), 1u}, 0u};
    this->_defaultColvars[ShapeType::POLYLINE] = {{this->_userGenerated.getSourceId(), 2u}, 0u};
    this->_defaultColvars[ShapeType::POLYGON] = {{this->_userGenerated.getSourceId(), 3u}, 0u};
    this->_collectionIdGenerator = 4u;

    this->_appEventsDispatcher.subscribe<AddPreparedAnnotationRequest>(this);
    this->_appEventsDispatcher.subscribe<SaveOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<LoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.subscribe<SaveAnnotationsToFileRequest>(this);
    this->_appEventsDispatcher.subscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.subscribe<GranuleRemovalRequest>(this);
    this->_appEventsDispatcher.subscribe<CreateAnnotationsCollectionRequest>(this);
    this->_appEventsDispatcher.subscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.subscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.subscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.subscribe<AnnotationTimeCoverageUpdateRequest>(this);
}

/**
 *
 */
AnnotationsController::~AnnotationsController()
{
    this->_appEventsDispatcher.unsubscribe<AddPreparedAnnotationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SaveOneAnnotationToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<LoadAnnotationsFromFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<SaveAnnotationsToFileRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<RenderingCfgUpdateEvent>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleRemovalRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CreateAnnotationsCollectionRequest>(this);
    this->_appEventsDispatcher.unsubscribe<CollectionRemovedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<ShapeEditionCompleteEvent>(this);
    this->_appEventsDispatcher.unsubscribe<EnableShapeCreationRequest>(this);
    this->_appEventsDispatcher.unsubscribe<AnnotationTimeCoverageUpdateRequest>(this);
}

UserGenerated* AnnotationsController::getUserGenerated()
{
    return &(this->_userGenerated);
}

/**
 *
 */
bool AnnotationsController::handleDefaultCollectionVariable(const CollectionVariable& inColvar, CollectionVariable& outColvar, ShapeType shapeType, bool createCollectionIfNeeded)
{
    if
        ((0u == inColvar.second) && (0u == inColvar.first.first) && (0u == inColvar.first.second))
    {
        auto found = this->_defaultColvars.find(shapeType);
        if
            (this->_defaultColvars.end() == found)
        {
            Log::log(LogLevel::Error, "AnnotationsController",
                     "Requesting default CollectionVariable for unsupported shape type");
            return false;
        }
        outColvar = found->second;

        // Create collection and variable if they do not exist yet
        Log::log(LogLevel::Debug, "AnnotationsController", "Using default colvar: {} {} {}",
                 outColvar.first.first, outColvar.first.second, outColvar.second);

        if
            (true == createCollectionIfNeeded)
        {
            CollectionCfg colCfg;
            bool ok(false);
            ok = this->_store.getCollection(outColvar.first.second, colCfg);
            if
                (false == ok)
            {
                Log::log(LogLevel::Error, "AnnotationsController",
                         "Error while fetching coonfiguration for collection {}.{}",
                         outColvar.first.first, outColvar.first.second);
                return false;
            }

            if
                ("" == colCfg.label) // means that no matching collection has been found
            {
                std::string shapeLabel("");
                if
                    (ShapeType::POINT == shapeType)
                {
                    shapeLabel = "points";
                }
                else if
                    (ShapeType::POLYLINE == shapeType)
                {
                    shapeLabel = "polylines";
                }
                else if
                    (ShapeType::POLYGON == shapeType)
                {
                    shapeLabel = "polygons";
                }
                else
                {
                    Log::log(LogLevel::Error, "AnnotationsController",
                             "Cannot find label for default collection with shape type: {}", int(shapeType));
                    return false;
                }
                Log::log(LogLevel::Debug, "AnnotationsController", "Creating missing default collection");
                this->createAnnotationsCollection(shapeType, "Annotations", shapeLabel, true, outColvar.first);
            }
        }
    }
    else
    {
        outColvar = inColvar;
    }
    return true;
}

/**
 *
 */
bool AnnotationsController::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(AddPreparedAnnotationRequest)) == eventType)
    {
        const CollectionVariable& inColvar = static_cast<const AddPreparedAnnotationRequest&>(event).getCollectionVariable();
        OGL::Shape* shape = static_cast<const AddPreparedAnnotationRequest&>(event).getShape();
        double defaultStart = static_cast<const AddPreparedAnnotationRequest&>(event).getDefaultStart();
        double defaultStop = static_cast<const AddPreparedAnnotationRequest&>(event).getDefaultStop();

        CollectionVariable colvar;
        bool ok(false);
        ok = this->handleDefaultCollectionVariable(inColvar, colvar, shape->getType(), true);
        if
            (false == ok)
        {
            return ok;
        }

        RenderableId rId;
        this->_userGenerated.generateRenderableId(colvar.first, rId);
        rId.variable = colvar.second;

        this->addAnnotation(rId, shape, defaultStart, defaultStop);

        // Trigger update
        CollectionsUpdatedEvent outEvent({colvar.first}, {});
        this->_appEventsDispatcher.dispatch<CollectionsUpdatedEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(SaveOneAnnotationToFileRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const SaveOneAnnotationToFileRequest&>(event).getRenderableId();
        const std::string& path = static_cast<const SaveOneAnnotationToFileRequest&>(event).getPath();
        this->saveAnnotation(rId, path);
        return true;
    }
    else if
        (std::type_index(typeid(LoadAnnotationsFromFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const LoadAnnotationsFromFileRequest&>(event).getPath();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const LoadAnnotationsFromFileRequest&>(event).getCollectionsMapping();

        std::map<ShapeType, CollectionId> fixedCollectionsMapping;
        for
            (const auto& it: collectionsMapping)
        {
            CollectionVariable colvar;
            CollectionVariable inColvar({it.second, 0u});
            bool ok(false);
            ok = this->handleDefaultCollectionVariable(inColvar, colvar, it.first, true);
            if
                (false == ok)
            {
                return ok;
            }
            fixedCollectionsMapping[it.first] = colvar.first;
        }

        this->loadFromFile(path, fixedCollectionsMapping);
        return true;
    }
    else if
        (std::type_index(typeid(SaveAnnotationsToFileRequest)) == eventType)
    {
        const std::string& path = static_cast<const SaveAnnotationsToFileRequest&>(event).getPath();
        const std::map<ShapeType, CollectionId>& collectionsMapping = static_cast<const SaveAnnotationsToFileRequest&>(event).getCollectionsMapping();

        std::map<ShapeType, CollectionId> fixedCollectionsMapping;
        for
            (const auto& it: collectionsMapping)
        {
            CollectionVariable colvar;
            CollectionVariable inColvar({it.second, 0u});
            bool ok(false);
            ok = this->handleDefaultCollectionVariable(inColvar, colvar, it.first, false);
            if
                (false == ok)
            {
                return ok;
            }
            fixedCollectionsMapping[it.first] = colvar.first;
        }

        this->saveToFile(path, fixedCollectionsMapping);
        return true;
    }
    else if
        (std::type_index(typeid(CreateAnnotationsCollectionRequest)) == eventType)
    {
        const ShapeType& shapeType = static_cast<const CreateAnnotationsCollectionRequest&>(event).getAnnotationsType();
        const std::string& colLabel = static_cast<const CreateAnnotationsCollectionRequest&>(event).getCollectionLabel();
        const std::string& varLabel = static_cast<const CreateAnnotationsCollectionRequest&>(event).getVariableLabel();
        bool persist = static_cast<const CreateAnnotationsCollectionRequest&>(event).getPersist();

        this->createAnnotationsCollection(shapeType, colLabel, varLabel, persist);
    }
    else if
        (std::type_index(typeid(AnnotationRemovedEvent)) == eventType)
    {
		std::ostringstream buffer("");	//

        const RenderableId& rId = static_cast<const AnnotationRemovedEvent&>(event).getRenderableId();
		buffer << "usershape_" << rId.collection.first << "/" << rId.collection.second << "/" << rId.granule.second << ":" <<rId.variable;

        if
            (this->_userGenerated.index.hasGranule(buffer.str()))
        {
            this->_userGenerated.index.remove(buffer.str(), false);
            CollectionsUpdatedEvent outEvent({rId.collection}, {});
            this->_appEventsDispatcher.dispatch<CollectionsUpdatedEvent>(outEvent);
        }
        return true;
    }
    else if
        (std::type_index(typeid(RenderingCfgUpdateEvent)) == eventType)
    {
        const RenderingCfg* rCfg = static_cast<const RenderingCfgUpdateEvent&>(event).getRenderingCfg();

        // Update rendering in the annotations store only if the collection belongs to the annotations source
        if
            (this->_userGenerated.getSourceId() == rCfg->target.collection.first)
        {
            this->_store.setRendering(*rCfg);
        }
        return true;
    }
    else if
        (std::type_index(typeid(GranuleRemovalRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const GranuleRemovalRequest&>(event).getRenderableId();
        if
            (this->_userGenerated.getSourceId() != rId.collection.first)
        {
            // Cannot remove since granule does not belong to user-generated content
            Log::log(LogLevel::Debug, "Annotations controller",
                     "will not remove annotation: wrong source ({}.{}/{})",
                     rId.collection.first, rId.collection.second, rId.granule.second);
            return true;
        }

        this->_store.remove(rId);
        return true;
    }
    else if
        (std::type_index(typeid(CollectionRemovedEvent)) == eventType)
    {
        const CollectionId& collectionId = static_cast<const CollectionRemovedEvent&>(event).getCollectionId();
        if
            (this->_userGenerated.getSourceId() != collectionId.first)
        {
            // Cannot remove since the collection does not belong to user-generated content
            Log::log(LogLevel::Debug, "Annotations controller",
                     "will not remove collection: wrong source ({}.{})",
                     collectionId.first, collectionId.second);
            return true;
        }

        std::vector<std::uint64_t> granuleIds;
        this->_store.listCollectionGranules(collectionId.second, granuleIds);

        this->_store.removeCollection(collectionId.second);

        for
            (const auto& id : granuleIds)
        {
            RenderableId rId(true, collectionId, {collectionId.first, id}, 0);
            AnnotationRemovedEvent outEvent(rId);
            this->_appEventsDispatcher.dispatch<AnnotationRemovedEvent>(outEvent);
        }

        return true;
    }
    else if
        (std::type_index(typeid(ShapeEditionCompleteEvent)) == eventType)
    {
        OGL::Shape* shape = static_cast<const ShapeEditionCompleteEvent&>(event).getShape();
        const RenderableId& rId = static_cast<const ShapeEditionCompleteEvent&>(event).getRenderableId();

        Annotation annotation;
        std::uint32_t collection;
        bool success(false);
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
            return true;
        }

        annotation.setShape(shape);

        // update database content
        success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Could not save annotation");
            return true;
        }
        return true;
    }
    else if
        (std::type_index(typeid(EnableShapeCreationRequest)) == eventType)
    {
        const ShapeType& shapeType =  static_cast<const EnableShapeCreationRequest&>(event).getShapeType();
        const CollectionVariable& collectionVariable = static_cast<const EnableShapeCreationRequest&>(event).getCollectionVariable();

        CollectionVariable fixedCollectionVariable;
        bool ok(false);
        ok = this->handleDefaultCollectionVariable(collectionVariable, fixedCollectionVariable, shapeType, true);
        if
            (false == ok)
        {
            return false;
        }

        EnableShapeCreationReadyEvent outEvent(shapeType, fixedCollectionVariable);
        this->_appEventsDispatcher.dispatch<EnableShapeCreationReadyEvent>(outEvent);

        return true;
    }
    else if
        (std::type_index(typeid(AnnotationTimeCoverageUpdateRequest)) == eventType)
    {
        const RenderableId& rId = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getRenderableId();
        double start = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getStart();
        double stop = static_cast<const AnnotationTimeCoverageUpdateRequest&>(event).getStop();

        if
            (this->_userGenerated.getSourceId() != rId.collection.first)
        {
            return true;
        }

        Annotation annotation;
        std::uint32_t collection;
        bool success(false);
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
            return true;
        }

        annotation.setStart(start);
        annotation.setStop(stop);

        // update database content
        success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Could not save annotation");
            return true;
        }

        bool ok(false);
        ok = this->_userGenerated.index.updateTimeCoverage(rId.granule.second,
                                                           static_cast<timestamp_t>(start),
                                                           static_cast<timestamp_t>(stop));
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Could not update annotation time coverage in the index");
            return true;
        }

        AnnotationTimeCoverageUpdatedEvent outEvent(rId, start, stop);
        this->_appEventsDispatcher.dispatch<AnnotationTimeCoverageUpdatedEvent>(outEvent);

        return true;
    }

    return false;
}


/**
 *
 */
bool AnnotationsController::loadFromFile(const std::string& jsonPath, const std::map<ShapeType, CollectionId>& collectionsMapping)
{
    bool success(false);    //

    //> JSON annotations
    std::ifstream json_file(jsonPath);
    if
        (false == json_file.is_open())
    {
        Log::log(LogLevel::Error, "AnnotationsController", "JSON input file is closed");
        return false;
    }

    JSONAnnotation jsonAnnotation;
    std::vector<Annotation> annotationsList;
    success = jsonAnnotation.deserialize(json_file, annotationsList);
    json_file.close();

    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "Failed to deserialize annotations from JSON file");
        return false;
    }
    //< JSON annotations

    //
    std::vector<Annotation> annotations;
    std::vector<std::uint64_t> granuleIds;
    std::vector<std::uint32_t> collectionIds;

    annotations.reserve(annotationsList.size());
    granuleIds.reserve(annotationsList.size());
    collectionIds.reserve(annotationsList.size());
    for
        (auto& annotation : annotationsList)
    {
        OGL::Shape* shape = annotation.getShape();
        auto collectionFound = collectionsMapping.find(shape->getType());
        shape = nullptr;

        if
            (collectionsMapping.end() == collectionFound)
        {
            Log::log(LogLevel::Warn, "Annotations controller",
                     "Annotation skipped due to unsupported shape type");
            continue;
        }

        // Register user shape in the index first otherwise it will not be
        // possible to set the temporal coverage
        RenderableId rId;
        success = this->_userGenerated.generateRenderableId(collectionFound->second, rId);

        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Failed to acquire an identifier for the user shape");
            Log::log(LogLevel::Error, "AnnotationsController", "Data import aborted");
            return false;
        }

        annotations.emplace_back(std::move(annotation));
        granuleIds.push_back(rId.granule.second);
        collectionIds.push_back(rId.collection.second);
    }
    annotationsList.clear();

    success = this->_store.set(granuleIds, collectionIds, annotations);
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "Could not save annotations in the local database");
        Log::log(LogLevel::Error, "AnnotationsController", "Data import aborted");
        return false;
    }

    auto granuleIt = granuleIds.cbegin();
    auto collectionIt = collectionIds.cbegin();
    auto annotationIt = annotations.cbegin();
    for
        (; granuleIt != granuleIds.end() && collectionIt != collectionIds.end() && annotationIt != annotations.end(); ++granuleIt, ++collectionIt, ++annotationIt)
    {
        this->initAnnotation(*granuleIt, *collectionIt, *annotationIt);
    }

    std::set<CollectionId> userCols;
    std::set<CollectionId> selectedCollections;
    this->_collectionsControl.model.getSelectedCollections(selectedCollections);
    for
        (const auto& it : collectionsMapping)
    {
        auto collectionFound = std::find(selectedCollections.begin(),
                                  selectedCollections.end(),
                                  it.second);
        if
            (selectedCollections.end() != collectionFound)
        {
            userCols.emplace(*collectionFound);
        }
    }

    // Trigger update
    CollectionsUpdatedEvent event(userCols, {});
    this->_appEventsDispatcher.dispatch<CollectionsUpdatedEvent>(event);

    return true;
}

/**
 *
 */
bool AnnotationsController::saveToFile(const std::string& jsonPath, const std::map<ShapeType, CollectionId>& collectionsMapping)
{
    bool success(false);    //
    std::uint32_t collection(0);    //
    std::vector<Annotation> annotationsList;    //
    Log::log(LogLevel::Debug, "AnnotationsController", "Save user shapes for Syntool");

    // export all rendered shapes
    for
        (const auto& rId : this->_sceneView.getRenderer()->getToRender())
    {
        bool isRequested(false);
        for
            (const auto& it : collectionsMapping)
        {
            if
                (  (rId.collection.first == it.second.first)
                && (rId.collection.second == it.second.second))
            {
                isRequested = true;
                break;
            }
        }
        if
            (false == isRequested)
        {
            continue;
        }

        Annotation annotation;

        // Export granule
        success = this->_store.get(rId.granule.second, collection, annotation);
        if
            (false == success)
        {
            Log::log(LogLevel::Error, "AnnotationsController", "Failed to get annotation {} from local database", rId.granule.second);
            return false;
        }

        annotationsList.push_back(std::move(annotation));
    }

    if
        (0 >= annotationsList.size())
    {
        Log::log(LogLevel::Debug, "AnnotationsController", "No annotation to export");
        return false;
    }

    std::ofstream jsonFile(jsonPath);
    if
        (false == jsonFile.is_open())
    {
        Log::log(LogLevel::Error, "AnnotationsController", "JSON output file is closed");
    }

    JSONAnnotation jsonAnnotation;
    jsonAnnotation.serialize(annotationsList, jsonFile);

    return true;
}


/**
 *
 */
bool AnnotationsController::addAnnotation(const RenderableId& rId, OGL::Shape* shape, double defaultStart, double defaultStop)
{
    if
        (nullptr == shape)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "A null shape has been passed to the user shape added callback");
        return false;
    }

    // Try to load info from the database
    Annotation annotation;
    std::uint32_t collection;
    bool success(false);
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        // Default time coverage
        annotation.setStart(defaultStart);
        annotation.setStop(defaultStop);
    }

    annotation.setShape(shape);

    // update database content
    success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "Could not save annotation");
        return false;
    }

    this->initAnnotation(rId.granule.second, rId.collection.second, annotation);

    return true;
}

/**
 *
 */
bool AnnotationsController::updateAnnotation(const RenderableId& rId, OGL::Shape* shape)
{
    Log::log(LogLevel::Debug, "AnnotationsController", "User shape {}/{}/{}:{} updated", rId.collection.first, rId.collection.second, rId.granule.second, rId.variable);
    if
        (nullptr == shape)
    {
        Log::log(LogLevel::Error, "AnnotaitonsController", "A null shape has been passed to the user shape updated callback");
        return false;
    }

    Annotation annotation;
    std::uint32_t collection;
    bool success(false);
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "Edited user shape not found in the annotations database");
        return false;
    }

    annotation.setShape(shape);

    // update database content
    success = this->_store.set(rId.granule.second, rId.collection.second, annotation);
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController", "Could not save annotation");
        return false;
    }

    return true;
}

/**
 *
 */
bool AnnotationsController::saveAnnotation(const RenderableId& rId, const std::string& jsonPath)
{
    bool success(false); //
    std::uint32_t collection(0);    //
    Annotation annotation;  //
    std::vector<Annotation> annotationsList;    //
    JSONAnnotation jsonAnnotation;  //

    // Export granule
    success = this->_store.get(rId.granule.second, collection, annotation);
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "AnnotationsController",
                 "Failed to get annotation {} from local database", rId.granule.second);
        return false;
    }
    annotationsList.push_back(std::move(annotation));

    std::ofstream jsonFile(jsonPath);
    if
        (false == jsonFile.is_open())
    {
        Log::log(LogLevel::Error, "AnnotationsController", "JSON output file is closed");
        return false;
    }

    jsonAnnotation.serialize(annotationsList, jsonFile);
    Log::log(LogLevel::Info, "AnnotationsController", "User shape saved in {}", jsonPath);

    return true;
}

/**
 *
 */
bool AnnotationsController::init(const std::string& path)
{
    bool ok(false); //

    ok = this->_store.init(path);
    if
        (false == ok)
    {
        return false;
    }

    // Make sure new collections will not overwrite existing ones
    ok = this->_store.getMaxCollectionId(this->_collectionIdGenerator);
    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Annotations controller", "Could not initialize identifiers for annotations collections");
        return false;
    }
    Log::log(LogLevel::Debug, "Annotations controller",
             "Collections max identifier: {}",
              this->_collectionIdGenerator);
    this->_collectionIdGenerator = this->_collectionIdGenerator + 1;

    this->_store.applyAllCollections([this] (std::uint32_t collectionId, CollectionCfg& collectionCfg, std::vector<RenderingCfg>& rCfgs) {
        CollectionId cId({this->_userGenerated.getSourceId(), collectionId});

        this->_renderingCfgs[cId];
        this->_renderingCfgs[cId].target.granuleLevel = rCfgs[0].target.granuleLevel;
        this->_renderingCfgs[cId].target.collection = cId;
        this->_renderingCfgs[cId].target.variable = rCfgs[0].target.variable;
        this->_renderingCfgs[cId].setRenderMethod(rCfgs[0].renderMethod());
        this->_renderingCfgs[cId].setLogscale(rCfgs[0].logscale());
        this->_renderingCfgs[cId].setMin(rCfgs[0].min());
        this->_renderingCfgs[cId].setMax(rCfgs[0].max());
        this->_renderingCfgs[cId].setOpacity(rCfgs[0].opacity());
        this->_renderingCfgs[cId].setZindex(rCfgs[0].zindex());
        this->_renderingCfgs[cId].setColor(rCfgs[0].color());
        this->_renderingCfgs[cId].setColormap(rCfgs[0].colormap()); // must be done AFTER setColor because setColor resets colormap
        this->_renderingCfgs[cId].setParticlesCount(rCfgs[0].particlesCount());
        this->_renderingCfgs[cId].setParticleTTL(rCfgs[0].particleTTL());
        this->_renderingCfgs[cId].setStreamlineLength(rCfgs[0].streamlineLength());
        this->_renderingCfgs[cId].setStreamlineSpeed(rCfgs[0].streamlineSpeed());
        this->_renderingCfgs[cId].setFilterMode(rCfgs[0].filterMode());

        std::vector<RenderingCfg*> renderings;	//
        renderings.push_back(&(this->_renderingCfgs[cId]));

        CollectionRegistrationRequest event(cId, collectionCfg, renderings);
        this->_appEventsDispatcher.dispatch<CollectionRegistrationRequest>(event);
    });

    std::uint64_t maxGranuleId = 1;
    this->_store.applyall([this, &maxGranuleId](std::uint64_t granule, std::uint32_t collection, Annotation& annotation) {
        this->initAnnotation(granule, collection, annotation);

        if
            (maxGranuleId <= granule)
        {
            maxGranuleId = granule + 1;
        }
    });

    // Some user shapes have been loaded, update the ID generator
    this->_userGenerated.index.idGenerator = maxGranuleId;

	return true;
}


/**
 *
 */
void AnnotationsController::createAnnotationsCollection(const ShapeType& shapeType, const std::string& collectionLabel, const std::string& variableLabel, bool persist, const CollectionId& cId)
{
	VariableCfg varCfg;	//
	varCfg.label = variableLabel;

	CollectionCfg colCfg;   //
	colCfg.label = collectionLabel;
	colCfg.xSeamless = false;
	colCfg.ySeamless = false;
	colCfg.mustBeCurrent = false;
	colCfg.defaultVariable = 0;

    this->_renderingCfgs[cId];
	this->_renderingCfgs[cId].target.granuleLevel = false;
	this->_renderingCfgs[cId].target.collection = cId;
	this->_renderingCfgs[cId].target.variable = 0;

    if
        (ShapeType::POLYGON == shapeType)
    {
        varCfg.defaultRendering = RenderMethod::POLYGON;
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POLYGON);
        this->_renderingCfgs[cId].setOpacity(0.7f);
        this->_renderingCfgs[cId].setColor({{200,200,75}});
        this->_renderingCfgs[cId].setZindex(5.0);
    }
    else if
        (ShapeType::POLYLINE == shapeType)
    {
        varCfg.defaultRendering = RenderMethod::POLYLINE;
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POLYLINE);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{30,250,75}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }
    else if
        (ShapeType::ELLIPSE == shapeType)
    {
        varCfg.defaultRendering = RenderMethod::ELLIPSE;
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::ELLIPSE);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{255, 95, 200}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }
    else if
        (ShapeType::POINT == shapeType)
    {
        varCfg.defaultRendering = RenderMethod::POINT;
        this->_renderingCfgs[cId].setRenderMethod(RenderMethod::POINT);
        this->_renderingCfgs[cId].setOpacity(1.0);
        this->_renderingCfgs[cId].setColor({{255, 0, 0}});
        this->_renderingCfgs[cId].setZindex(6.0);
    }

	colCfg.variables[0] = varCfg;

	std::vector<RenderingCfg*> renderings;	//
	renderings.push_back(&(this->_renderingCfgs[cId]));

    if
        (true == persist)
    {
        this->_store.setCollection(cId.second, colCfg, renderings);
    }

    CollectionRegistrationRequest event(cId, colCfg, renderings);
    this->_appEventsDispatcher.dispatch<CollectionRegistrationRequest>(event);
}

/**
 *
 */
void AnnotationsController::createAnnotationsCollection(const ShapeType& shapeType, const std::string& collectionLabel, const std::string& variableLabel, bool persist)
{
    // Get a new collection ID
    CollectionId cId({this->_userGenerated.getSourceId(),
                      this->_collectionIdGenerator++});

    this->createAnnotationsCollection(shapeType, collectionLabel, variableLabel, persist, cId);
}


/**
 *
 */
void AnnotationsController::initAnnotation(std::uint64_t granule, std::uint32_t collection, const Annotation& annotation)
{
    GranuleMetadata metadata;   //
    std::ostringstream buffer;  //
    IDFDescriptor idf;  //
    std::uint32_t sourceId; //

    sourceId = this->_userGenerated.getSourceId();
    RenderableId rId(true, CollectionId(sourceId, collection), GranuleId(sourceId, granule), 0);

    bool success(false); //
    RenderingCfg* rCfg(nullptr);
    bool ok(false);
	rCfg = this->_renderingControl.model.getCfg(rId.collection,
                                                rId.granule,
                                                rId.variable, ok);

    if
        (false == ok)
    {
        Log::log(LogLevel::Error, "Annotations controller", "Could not get rendering configuration for user generated content");
        return;
    }

    success = this->_sceneView.getScene()->createShapeGpu(this->_sceneView.getRenderer()->getContext(), annotation.getShape(), rId, *rCfg, this->_icons);
    rCfg = nullptr;
    if
        (false == success)
    {
        Log::log(LogLevel::Error, "Annotations controller", "Skipping annotation {} because its shape could not be created in the renderer, it is probably invalid", granule);
        return;
    }

    buffer << "usershape_" << sourceId << "/" << collection << "/" << granule << ":" << 0;
    metadata.dataId = buffer.str();
    metadata.id = rId.granule;
    metadata.collectionId = collection;
    metadata.start = static_cast<timestamp_t>(annotation.getStart());
    metadata.stop = static_cast<timestamp_t>(annotation.getStop());
    idf.uri = buffer.str();
    idf.resolution = 10000000;
    metadata.uris.push_back(idf);

    this->_userGenerated.index.add(metadata, true);
}
