// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "tools/extraction_controller.h"

#include "core/log.h"
#include "core/input_event.h"
#include "core/input_devices_state.h"
#include "3d/camera.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/default_value.h"
#include "3d/scene_view.h"
#include "3d/iscene_renderer.h"
#include "core/metadata_mediator.h"
#include "core/data_mediator.h"

#include "core/app_events_dispatcher.h"
#include "events/extraction_request.h"
#include "events/display_extraction_info_request.h"

#include <iostream>

using namespace SEAScope;

/**
 *
 */
ExtractionController::ExtractionController(AppEventsDispatcher& appEventsDispatcher)
: BaseInputEventsSubscriber("Extraction controller"),
  _appEventsDispatcher(appEventsDispatcher)
{
    // Configure actions

    // - Resize framebuffer
    this->defineAction(static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER),
                       "Resize extractor framebuffer", "#?");

    this->_possibleActions = {static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER),};

    this->_appEventsDispatcher.subscribe<ExtractionRequest>(this);
}

/**
 *
 */
ExtractionController::~ExtractionController()
{
    this->_appEventsDispatcher.unsubscribe<ExtractionRequest>(this);
}

/**
 *
 */
bool ExtractionController::isActive() const
{
    return this->_active && (nullptr != this->_extractor.getFBO());
}

/**
 *
 */
bool ExtractionController::performAction(const ActionType& actionType,
                                         const InputEvent& event,
                                         const InputDevicesState& /*devicesState*/)
{
    if
        (static_cast<ActionType>(ExtractionControllerActions::RESIZE_FRAMEBUFFER) == actionType)
    {
        Log::log(LogLevel::Debug, "ExtractionController", "Resize framebuffer");
        this->_extractor.resize(event.getWidth(), event.getHeight());
        return true;
    }

    return false;
}

/**
 *
 */
bool ExtractionController::extractData(SceneView& sceneView, Renderable2* footprint, std::map<RenderableId, Extraction>& extractionData)
{
    bool ok(false); //

    if
        (nullptr == this->_extractor.getFBO())
    {
        this->_extractor.create(sceneView.getRenderer()->getContext(),
                                sceneView.getViewport().x,
                                sceneView.getViewport().y);
    }
    ok = this->_extractor.extract(sceneView.getRenderer()->getContext(),
                                  footprint,
                                  sceneView,
                                  extractionData);
    return ok;
}

/**
 *
 */
bool ExtractionController::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(ExtractionRequest)) == eventType)
    {
        bool ok(false);

        SceneView* sceneView = static_cast<const ExtractionRequest&>(event).getSceneView();
        Renderable2* footprint = static_cast<const ExtractionRequest&>(event).getFootprint();
        std::map<RenderableId, Extraction>* extractedData = static_cast<const ExtractionRequest&>(event).getExtractedData();
        std::map<RenderableId, std::string>* extractedLabels = static_cast<const ExtractionRequest&>(event).getExtractedLabels();

        ok = this->extractData((*sceneView), footprint, (*extractedData));
        if
            (false == ok)
        {
            Log::log(LogLevel::Error, "ExtractionController",
                     "Extraction failed");
            return true;    // event has been processed, even if it failed
        }

        DisplayExtractionInfoRequest outEvent((*extractedData), (*extractedLabels));
        this->_appEventsDispatcher.dispatch<DisplayExtractionInfoRequest>(outEvent);

        return true;
    }
    return false;
}
