// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "3d/stdafx.h"
#include "tools/scene_view_controller_original.h"

#include "3d/scene_view.h"
#include "3d/default_value.h"
#include "core/types.h"
#include "core/log.h"
#include "core/input_devices_state.h"
#include "3d/geodetic.h"
#include "3d/unit_sphere.h"
#include "3d/math_func.h"
#include "3d/iscene_renderer.h"

#include "core/input_event.h"
#include "events/change_rasterization_mode_request.h"
#include "events/zoom_in_request.h"
#include "events/zoom_out_request.h"
#include "events/camera_reset_request.h"
#include "events/camera_aligned_with_poles_request.h"
#include "events/camera_poles_lock_request.h"
#include "events/cursor_location_changed_event.h"
#include "events/lock_view_request.h"
#include "events/change_rasterization_mode_request.h"
#include "events/poles_axis_lock_event.h"


namespace SEAScope
{

SceneViewControllerOriginal::SceneViewControllerOriginal(AppEventsDispatcher& appEventsDispatcher)
:	SceneViewControllerBase("Camera controller", appEventsDispatcher),
	_zoomSpeed(1.f),
	_angles(0, 0),
	_zoom(GLOBE_ZOOM_DEFAULT),
	_lockPolesAxis(true),
	_viewLocked(false),
    _wasOnGlobe(false)
{
	initEvents();
}

SceneViewControllerOriginal::~SceneViewControllerOriginal()
{
	cleanEvents();
}

bool SceneViewControllerOriginal::isActive() const
{
	return this->_active && (nullptr != this->_sceneView);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////
// Camera actions
/////////////////////////////////////////////////////////////////////////////////////////////////////

bool SceneViewControllerOriginal::startCameraMovement(const InputEvent& /*event*/,
													  const InputDevicesState& devicesState)
{
    bool onGlobe(false);    //

    onGlobe = getPickingPosition(this->_sceneView->getViewport(),
                                 static_cast<int>(devicesState.getMouseX()),
                                 static_cast<int>(devicesState.getMouseY()),
                                 this->_arcBallHitPoint,
                                 this->_q,
                                 this->_sceneView->getCamera()->getViewMatrix(),
                                 this->_sceneView->getCamera()->getProjectionMatrix());
	if
		(true == onGlobe)
	{
        this->_arcBallQ = this->_q;

		// Save quaternion as it will be used as origin for future movements
		this->updatePossibleActions({ SceneViewControllerOriginalActions::MOVE,
									  SceneViewControllerOriginalActions::MOVE_END },
									{ SceneViewControllerOriginalActions::MOVE_START });
		this->_preemptEvents = true;
	}

    return onGlobe;
}

bool SceneViewControllerOriginal::moveCamera(const InputEvent& event,
											 const InputDevicesState& devicesState,
											 bool, float)
{
	if
		(true == this->_viewLocked)
	{
		return false;
	}

    glm::vec3 hitPoint; //
    glm::quat rotationQuat = this->_arcBallQ;
    glm::vec3 refPoint = this->_arcBallHitPoint;
    if
        (true == this->_lockPolesAxis)
    {
        rotationQuat = this->_q;
        refPoint = glm::vec3(1.f, 0.f, 0.f)  * this->_q;
    }

    bool onGlobe(false);
    onGlobe = getPickingPosition(this->_sceneView->getViewport(),
                                 static_cast<int>(devicesState.getMouseX()),
                                 static_cast<int>(devicesState.getMouseY()),
                                 hitPoint,
                                 rotationQuat,
                                 this->_sceneView->getCamera()->getViewMatrix(),
                                 this->_sceneView->getCamera()->getProjectionMatrix());

	if
        (true == this->_lockPolesAxis)
	{
		double pixelSizeTmp(0.0);   //

		double perimeterTmp(0.0);   //
		double tmpValue(0.0);   //
		bool ok(false); //

		pixelSizeTmp = static_cast<double>(this->_sceneView->getPixelSize());

		// Default perimeter used when the cursor is not on the globe
		perimeterTmp = 0.5 * GLOBE_EQUATOR_LENGTH;

		ok = this->cursorPosition(glm::vec2(devicesState.getMouseX(), devicesState.getMouseY()), hitPoint);

		if
			(ok)
		{
			Geodetic2D geo = UnitSphere::ToGeodetic2D(hitPoint);
			tmpValue = GLOBE_EQUATOR_LENGTH * std::max(0.001, std::cos(geo.Latitude()));
		}

		tmpValue = -1.0 * event.getMouseDx() * pixelSizeTmp * 360.0 / perimeterTmp;
		this->_angles.x += static_cast<float>(tmpValue);

		tmpValue = -1.0 * event.getMouseDy() * pixelSizeTmp * 360.0 / GLOBE_EQUATOR_LENGTH;
		this->_angles.y += static_cast<float>(tmpValue);

		this->checkAngles();

		this->updateCameraRotation();
	}
	else if
        (onGlobe)
    {
        if
            (false == this->_wasOnGlobe)
        {
            this->_arcBallHitPoint = hitPoint;
            refPoint = this->_arcBallHitPoint;
        }

        glm::quat q = rotationBetweenVectors(refPoint, hitPoint);
        this->_q = this->_arcBallQ * q;
        this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
    }
    else
    {
        // Use theta-phi

        double perimeter(0.5 * GLOBE_EQUATOR_LENGTH);    // default value when the cursor is not on the globe

        double theta(event.getMouseDx() * static_cast<double>(this->_sceneView->getPixelSize()) * 2 * M_PI / perimeter);
        double phi(event.getMouseDy() * static_cast<double>(this->_sceneView->getPixelSize()) * 2 * M_PI / GLOBE_EQUATOR_LENGTH);
        glm::quat zRot(static_cast<float>(std::cos(0.5 * theta)), 0.f, 0.f, static_cast<float>(std::sin(0.5 * theta)));
        glm::quat yRot(static_cast<float>(std::cos(0.5 * phi)), 0.f, static_cast<float>(std::sin(0.5 * phi)), 0.f);
        this->_q = zRot * yRot * this->_q;

        this->_arcBallQ = this->_q;

        this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
    }

    this->_lastMousePosition = glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY());
    if
        (onGlobe)
    {
        this->_lastGlobePos = hitPoint;
    }


    this->_wasOnGlobe = onGlobe;

    return true;
}

bool SceneViewControllerOriginal::stopCameraMovement()
{
	ActionType actionType;  //

	// Remove MOVE to possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE);
	this->_possibleActions.erase(this->_possibleActions.find(actionType));

	// Remove MOVE_END to possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END);
	this->_possibleActions.erase(this->_possibleActions.find(actionType));

	// Add MOVE_START from possible actions
	actionType = static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START);
	this->_possibleActions.insert(actionType);

	this->_preemptEvents = false;

	return true;
}

void SceneViewControllerOriginal::setZoom(float value)
{
	if
		(	(value >= GLOBE_MAX_ZOOM_DISTANCE)
		 || (value <= GLOBE_MIN_ZOOM_DISTANCE))
	{
		// out of range
		return;
	}

	_zoom = value;

	this->_sceneView->getCamera()->lookAt(glm::vec3(value, 0, 0),
										  glm::vec3(0.f), 
										  glm::vec3(0, 0, 1));
}

void SceneViewControllerOriginal::resetCamera()
{
	this->setZoom(GLOBE_ZOOM_DEFAULT);
	this->_q = glm::quat(1.f, 0.f, 0.f, 0.f);
	this->_angles = glm::vec2(0.f, 0.f);
	this->_zoom = GLOBE_ZOOM_DEFAULT;
	this->updateCameraRotation();
}

void SceneViewControllerOriginal::zoomIn()
{
	float zoom = _zoom;
	zoom = zoom - (zoom - GLOBE_RADIUS) * this->_zoomSpeed * GLOBE_ZOOM_IN_SPEED;
	this->setZoom(zoom);
}

void SceneViewControllerOriginal::zoomOut()
{
	float zoom = _zoom;
	zoom = zoom + (zoom - GLOBE_RADIUS) * this->_zoomSpeed * GLOBE_ZOOM_OUT_SPEED;
	this->setZoom(zoom);
}

void SceneViewControllerOriginal::checkAngles()
{
	if
        (this->_angles.x > 360.f)
    {
		this->_angles.x -= 360.f;
    }

	if
        (this->_angles.x < 0.f)
    {
		this->_angles.x += 360.f;
    }

	this->_angles.y = glm::clamp(this->_angles.y, -89.9f, 89.9f);
}

void SceneViewControllerOriginal::updateCameraRotation()
{
	glm::mat4 m(1.f);

	m = glm::rotate(m, glm::radians(-1 * this->_angles.y), glm::vec3(0.f, 1.f, 0.f));
	m = glm::rotate(m, glm::radians(-1 * this->_angles.x), glm::vec3(0.f, 0.f, 1.f));

	this->_q = glm::toQuat(m);

	this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
}

void SceneViewControllerOriginal::zoomTo(float distFromCenter)
{
	if
		(SEAScope::floating_point_equality(distFromCenter, this->_zoom))
	{
		return;
	}

	if (distFromCenter >= GLOBE_MAX_ZOOM_DISTANCE
		&& distFromCenter <= GLOBE_MIN_ZOOM_DISTANCE)
	{
		return;
	}

	this->_zoom = distFromCenter;

	updateCameraZoom();
}

void SceneViewControllerOriginal::updateCameraZoom()
{
	// compute the new camera matrix.
	this->_sceneView->getCamera()->lookAt(glm::vec3(this->_zoom, 0, 0), glm::vec3(0.f), glm::vec3(0, 0, 1));
	this->_sceneView->updatePixelSize();
}

bool SceneViewControllerOriginal::cursorPosition(const glm::vec2& mousePos, glm::vec3& hitPoint)
{
	bool ok = getPickingPosition(this->_sceneView->getViewport(),
								 static_cast<int>(mousePos.x),
								 static_cast<int>(mousePos.y),
								 hitPoint,
								 this->_q,
								 this->_sceneView->getCamera()->getViewMatrix(),
								 this->_sceneView->getCamera()->getProjectionMatrix());
	return ok;
}

void SceneViewControllerOriginal::setLockPolesAxis(bool lockPolesAxis)
{
    if
        (lockPolesAxis == this->_lockPolesAxis)
    {
        return;
    }

	this->_lockPolesAxis = lockPolesAxis;

	if (true == this->_lockPolesAxis)
	{
		glm::vec3 pCam = glm::vec3(1.f, 0.f, 0.f) * this->_q;
		Geodetic2D geo = UnitSphere::ToGeodetic2D(pCam);
		this->lookAt(static_cast<float>(glm::degrees(geo.Longitude())), static_cast<float>(glm::degrees(geo.Latitude())));
	}

    PolesAxisLockEvent event(this->_lockPolesAxis);
    this->_appEventsDispatcher.dispatch<PolesAxisLockEvent>(event);
}

void SceneViewControllerOriginal::cursorMoveEvent(const InputDevicesState& devicesState)
{
	glm::vec3 hitPoint;

	bool ok = this->cursorPosition(glm::vec2(devicesState.getMouseX(), devicesState.getMouseY()), hitPoint);

	if
		(true == ok)
	{
		this->_lastGlobePos = hitPoint;
	}

	this->_lastMousePosition = glm::ivec2(devicesState.getMouseX(), devicesState.getMouseY());
}

void SceneViewControllerOriginal::zoomInTo()
{
	this->zoomIn();

	glm::vec3 hitPoint;
	bool result = this->cursorPosition(this->_lastMousePosition, hitPoint);

	if
        (result)
	{
		if (true == this->_lockPolesAxis)
		{
			glm::vec2 lastPos = UnitSphere::ToGeodetic2D(this->_lastGlobePos).toVec2();
			glm::vec2 currentPos = UnitSphere::ToGeodetic2D(hitPoint).toVec2();

			this->_angles.x -= glm::degrees(currentPos.x - lastPos.x);
			this->_angles.y += glm::degrees(currentPos.y - lastPos.y);

			this->checkAngles();
			this->updateCameraRotation();
		}
		else
		{
			glm::quat q = rotationBetweenVectors(this->_lastGlobePos, hitPoint);

			this->_q = this->_q * q;
			this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
		}

	}
}

void SceneViewControllerOriginal::lookAt(float lon, float lat)
{
	if (true == this->_lockPolesAxis)
	{
		double lonDegrees(0.0); //
		double latDegrees(0.0); //
		double lonRad(0.0); //
		double latRad(0.0); //

		lonRad = static_cast<double>(glm::radians(lon));
		latRad = static_cast<double>(glm::radians(lat));

		latRad = glm::clamp(latRad, -M_PI_2, M_PI_2);

		while
            (lonRad < -M_PI - static_cast<double>(std::numeric_limits<float>::epsilon()))
        {
			lonRad += M_PI;
        }
		while
            (lonRad > M_PI + static_cast<double>(std::numeric_limits<float>::epsilon()))
        {
			lonRad -= M_PI;
        }

		lonDegrees = glm::degrees(lonRad);
		this->_angles.x = static_cast<float>(lonDegrees);
		latDegrees = glm::degrees(-1.0 * latRad);
		this->_angles.y = static_cast<float>(latDegrees);
		this->updateCameraRotation();
	}
	else
	{
		this->_q = glm::quat_cast(glm::mat3(1.f));

		this->_q = glm::rotate(this->_q, glm::radians(lat), glm::vec3(0.f, 1.f, 0.f));
		this->_q = glm::rotate(this->_q, -1 * glm::radians(lon), glm::vec3(0.f, 0.f, 1.f));

		this->_sceneView->getCamera()->setModelMatrix(glm::mat4_cast(this->_q));
	}
}

/// ///////////////////////////////////////////////////////////////////////////////////////////////////
/// Events
/// ///////////////////////////////////////////////////////////////////////////////////////////////////

bool SceneViewControllerOriginal::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
	if
		(std::type_index(typeid(ZoomInRequest)) == eventType)
	{
		this->zoomIn();
		return true;
	}
	else if
		(std::type_index(typeid(ZoomOutRequest)) == eventType)
	{
		this->zoomOut();
		return true;
	}
	else if
		(std::type_index(typeid(CameraResetRequest)) == eventType)
	{
		this->resetCamera();
		return true;
	}
	else if
		(std::type_index(typeid(CameraAlignedWithPolesRequest)) == eventType)
	{
// 		this->lockPolesAxis();
// 		this->updateCameraRotation();
		return true;
	}
	else if
		(std::type_index(typeid(CameraPolesLockRequest)) == eventType)
	{
		bool state = static_cast<const CameraPolesLockRequest&>(event).getState();
		this->setLockPolesAxis(state);
		return true;
	}
	else if
		(std::type_index(typeid(LockViewRequest)) == eventType)
	{
		bool locked = static_cast<const LockViewRequest&>(event).getLocked();
		this->_viewLocked = locked;
	}
	else if
		(std::type_index(typeid(ChangeRasterizationModeRequest)) == eventType)
	{
		OGL::RasterizationMode mode = static_cast<const ChangeRasterizationModeRequest&>(event).getMode();
		this->_sceneView->getRenderer()->setRasterizationMode(mode);

		if
			(OGL::fill == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));
		}
		else if
			(OGL::line == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));

		}
		else if
			(OGL::point == mode)
		{
			auto found = this->_possibleActions.find(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE));
			if
				(this->_possibleActions.end() != found)
			{
				this->_possibleActions.erase(found);
			}
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE));
			this->_possibleActions.insert(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE));
		}
		return true;
	}

	return false;
}

void SceneViewControllerOriginal::updatePossibleActions(const std::set<SceneViewControllerOriginalActions>& toEnable,
	const std::set<SceneViewControllerOriginalActions>& toDisable)
{
	ActionType actionType;

	for
		(const auto& it : toEnable)
	{
		actionType = static_cast<ActionType>(it);
		this->_possibleActions.insert(actionType);
	}

	for
		(const auto& it : toDisable)
	{
		actionType = static_cast<ActionType>(it);

		auto found = this->_possibleActions.find(actionType);
		if
			(this->_possibleActions.end() != found)
		{
			this->_possibleActions.erase(found);
		}
	}
}

bool SceneViewControllerOriginal::performAction(const ActionType& actionType,
    const InputEvent& event,
    const InputDevicesState& devicesState)
{
    if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::RESET) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Reset camera position");
        this->resetCamera();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewControllerOriginalActions", "Start moving camera");
        return this->startCameraMovement(event, devicesState);
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Move camera");
        return this->moveCamera(event, devicesState, true, 1.f);
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewControllerOriginalActions", "Stop moving camera");
        return this->stopCameraMovement();
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Zoom-in to");

        this->zoomInTo();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Zoom-in");
        this->zoomIn();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Zoom-out from");

        this->zoomOut();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Zoom-out");
        this->zoomOut();
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Increase zoom speed");
        this->_zoomSpeed = 2.f;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Decrease zoom speed");
        this->_zoomSpeed = 1.f;
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Toggle poles axis locking");
        this->setLockPolesAxis(!this->_lockPolesAxis);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Resize scene viewport");
        this->_sceneView->setViewport(glm::ivec2(event.getWidth(), event.getHeight()));
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::fill);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::line);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE) == actionType)
    {
        ChangeRasterizationModeRequest outEvent(OGL::point);
        this->_appEventsDispatcher.dispatch<ChangeRasterizationModeRequest>(outEvent);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE) == actionType)
    {
        this->cursorMoveEvent(devicesState);
        return true;
    }
    else if
        (static_cast<ActionType>(SceneViewControllerOriginalActions::UPDATE_CURSOR_COORDS) == actionType)
    {
        Log::log(LogLevel::Debug, "SceneViewController", "Update geographical coordinates of the cursor");

        glm::vec2 lonlat;
        glm::vec3 hitpoint;
        bool onWorld = this->_sceneView->getLonLat(devicesState.getMouseX(),
                                                   devicesState.getMouseY(),
                                                   lonlat, hitpoint);
        CursorLocationChangedEvent appEvent(static_cast<std::uint32_t>(devicesState.getMouseX()),
                                            static_cast<std::uint32_t>(devicesState.getMouseY()),
                                            lonlat.x, lonlat.y, onWorld, hitpoint);
        this->_appEventsDispatcher.dispatch<CursorLocationChangedEvent>(appEvent);
        return true;
    }

    return false;
}

void SceneViewControllerOriginal::initEvents()
{
	// Configure actions

	// - Cursor move
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE),
		"Cursor move", "~X|Y");

	// - Camera reset
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::RESET),
		"Reset zoom and location", ">R");  /* keypress: R*/

// - Camera move start
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START),
		"Start camera movement", ">BUTTON0"); /* buttonpress: left */

// - Camera move
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE),
		"Move camera", "}BUTTON0 ~X|Y"); /* buttonpress: left + move */

// - Camera move end
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_END),
		"Conclude camera movement", "<BUTTON0"); /* buttonrelease: left */

// - Zoom-in
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO),
		"Zoom-in to a location", "*Y+"); /* wheel: forward on Y axis */

// - Zoom-in (alternative binding
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN),
		"Zoom-in", "}PAGE_UP"); /* keypress: page-up */

// - Zoom-out
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM),
		"Zoom-out from a location", "*Y-"); /* wheel: backward on Y axis */

// - Zoom-in (alternative binding
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT),
		"Zoom-out", "}PAGE_DOWN"); /* keypress: page-down */

// - Increase zoom speed
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED),
		"Increase zoom speed", ">LEFT_CONTROL"); /* keypress: left ctrl */

// - Decrease zoom speed
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED),
		"Decrease zoom speed", "<LEFT_CONTROL"); /* keyrelease: left ctrl */

// - Toggle poles axis locking
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING),
		"Toggle poles axis locking", "}LEFT_SHIFT >A");

	// - Update cursor coordinates
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::UPDATE_CURSOR_COORDS),
		"Update geographical coordinates of the cursor", "__ ~X|Y");

	// - Resize viewport
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT),
		"Resize scene viewport", "#?");

	// - Use full mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_FULL_MODE),
		"Use full mode", "}LEFT_SHIFT >F");

	// - Use wireframe mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE),
		"Use wireframe mode", "}LEFT_SHIFT >W");

	// - Use dot mode
	this->defineAction(static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE),
		"Use vertex mode", "}LEFT_SHIFT >V");

	this->_possibleActions = { static_cast<ActionType>(SceneViewControllerOriginalActions::RESET),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::CURSOR_MOVE),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::MOVE_START),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN_TO),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_IN),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT_FROM),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::ZOOM_OUT),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::INCREASE_ZOOM_SPEED),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::DECREASE_ZOOM_SPEED),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::TOGGLE_POLES_AXIS_LOCKING),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::UPDATE_CURSOR_COORDS),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::RESIZE_VIEWPORT),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::USE_WIREFRAME_MODE),
							  static_cast<ActionType>(SceneViewControllerOriginalActions::USE_DOT_MODE), };

	this->_appEventsDispatcher.subscribe<ZoomInRequest>(this);
	this->_appEventsDispatcher.subscribe<ZoomOutRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraResetRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraAlignedWithPolesRequest>(this);
	this->_appEventsDispatcher.subscribe<CameraPolesLockRequest>(this);
	this->_appEventsDispatcher.subscribe<LockViewRequest>(this);
	this->_appEventsDispatcher.subscribe<ChangeRasterizationModeRequest>(this);
}

void SceneViewControllerOriginal::cleanEvents()
{
	this->_appEventsDispatcher.unsubscribe<ZoomInRequest>(this);
	this->_appEventsDispatcher.unsubscribe<ZoomOutRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraResetRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraAlignedWithPolesRequest>(this);
	this->_appEventsDispatcher.unsubscribe<CameraPolesLockRequest>(this);
	this->_appEventsDispatcher.unsubscribe<LockViewRequest>(this);
	this->_appEventsDispatcher.unsubscribe<ChangeRasterizationModeRequest>(this);

	this->_sceneView = nullptr; // not owning
}

} // namespace SEAScope
