// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "nanogui/button.h"

#include "tools/side_menu.h"

#include "core/app_event.h"
#include "core/app_events_dispatcher.h"

#include "ui/nanogui/nanogui_side_menu.h"
#include "ui/nanogui/nanogui_main_menu.h"
#include "ui/nanogui/nanogui_about_menu.h"

#include "events/remote_pickle_export_request.h"
#include "events/remote_png_export_request.h"
#include "events/change_year_month_request.h"
#include "events/change_year_request.h"
#include "events/side_menu_resize_request.h"
#include "events/display_catalogue_request.h"
#include "events/granule_info_ready_for_display_event.h"
#include "events/display_extraction_info_request.h"
#include "events/collection_removal_request.h"

using namespace SEAScope;

SideMenu::SideMenu(AppEventsDispatcher& appEventsDispatcher, nanogui::Widget* parent, bool processingToolsAvailable)
: _appEventsDispatcher(appEventsDispatcher),
  _ui(nullptr),
  _aboutMenu(nullptr),
  _mainMenu(nullptr),
  _catalogueButton(nullptr),
  _aboutButton(nullptr)
{
	this->_ui = new NanoguiSideMenu(parent);
	this->_ui->setFixedWidth(320);
	this->_ui->setFixedHeight(parent->height() - 136 /* timeline height */);
	this->_ui->anchor->validCorners.push_back(NanoguiCorner::TOP_LEFT);
	this->_ui->anchor->validCorners.push_back(NanoguiCorner::TOP_RIGHT);
	this->_ui->content = new NanoguiMainMenu(this->_ui);
	this->_ui->content->setPosition(Eigen::Vector2i(0,0));
	this->_ui->anchor->setCorner(NanoguiCorner::TOP_RIGHT);
	this->_ui->setSize(this->_ui->fixedSize());
	this->_ui->setVisible(false);

	// Help menu
	this->_aboutMenu = new NanoguiAboutMenu();
	this->_aboutMenu->init(parent);
#if defined __linux__ || defined __APPLE__
	this->_aboutMenu->window->setFixedSize(Eigen::Vector2i(360, 510));
#else
	this->_aboutMenu->window->setFixedSize(Eigen::Vector2i(360, 430));
#endif

	this->_mainMenu = dynamic_cast<NanoguiMainMenu*>(this->_ui->content);
	this->_mainMenu->setFixedWidth(320);

	// Hide processing options in the extraction dialog if the processor is
	// not enabled
	this->_mainMenu->extractionDialog->processorOptions->setVisible(processingToolsAvailable);

	// Sidemenu header
	this->_ui->header->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
													    nanogui::Alignment::Middle,
													    2, 3));
	this->_catalogueButton = new nanogui::Button(this->_ui->header, "Catalogue",
                                                 ENTYPO_ICON_DATABASE);
    this->_catalogueButton->setFlags(nanogui::Button::ToggleButton);

    nanogui::Widget* spacer = new nanogui::Widget(this->_ui->header);
    spacer->setFixedWidth(20);
    spacer->setWidth(20);

	this->_aboutButton = new nanogui::Button(this->_ui->header, "",
                                            ENTYPO_ICON_HELP_WITH_CIRCLE);
    this->_aboutButton->setBackgroundColor(nanogui::Color(0,0,0,0));
    this->_aboutButton->setFixedSize(Eigen::Vector2i(20, 20));

	this->_mainMenu->collectionsControl->setListChangedCallback([this] () {
        this->_mainMenu->update();
		this->_ui->reflowNeeded = true;
	});
    this->_mainMenu->collectionsControl->setCollectionRemovedCallback([this](const CollectionId& colId) {
        CollectionRemovalRequest event(colId);
        this->_appEventsDispatcher.dispatch<CollectionRemovalRequest>(event);
    });

	this->_aboutMenu->closeCallback = [this]() {
		this->_aboutMenu->window->hide();
	};

	this->_aboutMenu->window->hide();

	this->_mainMenu->granuleInfo->monthCallback = [this] (std::uint16_t year, std::uint8_t month) {
        ChangeYearMonthRequest event(year, month);
        this->_appEventsDispatcher.dispatch<ChangeYearMonthRequest>(event);
		//app.engine.timeline.view->setMonth(year, month);
	};
	this->_mainMenu->granuleInfo->yearCallback = [this] (std::uint16_t year) {
        ChangeYearRequest event(year);
        this->_appEventsDispatcher.dispatch<ChangeYearRequest>(event);
		// app.engine.timeline.view->setYear(year);
	};

	this->_mainMenu->extractionDialog->pickleCallback = [this] () {
        RemotePickleExportRequest event;
        this->_appEventsDispatcher.dispatch<RemotePickleExportRequest>(event);
	};
	this->_mainMenu->extractionDialog->pngCallback = [this] () {
        RemotePNGExportRequest event;
        this->_appEventsDispatcher.dispatch<RemotePNGExportRequest>(event);
	};

	this->_catalogueButton->setChangeCallback([this](bool pushed) {
		this->_mainMenu->collectionsControl->setVisible(pushed);
		this->_catalogueButton->setPushed(pushed);
		this->_mainMenu->update();
	});

	this->_aboutButton->setCallback([this]() {
		if
			(this->_aboutMenu->window->visible())
		{
			this->_aboutMenu->window->setVisible(false);
		}
		else
		{
			Eigen::Vector2i pos(
				std::floor(0.5 * (this->_ui->parent()->width() - this->_aboutMenu->window->width())),
				std::floor(0.5 * (this->_ui->parent()->height() - this->_aboutMenu->window->height())));
			this->_aboutMenu->window->display(pos);
			this->_aboutMenu->window->requestFocus();
		}
	});

    this->_appEventsDispatcher.subscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.subscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.subscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.subscribe<DisplayExtractionInfoRequest>(this);
}

SideMenu::~SideMenu()
{
    this->_appEventsDispatcher.unsubscribe<SideMenuResizeRequest>(this);
    this->_appEventsDispatcher.unsubscribe<GranuleInfoReadyForDisplayEvent>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayCatalogueRequest>(this);
    this->_appEventsDispatcher.unsubscribe<DisplayExtractionInfoRequest>(this);

    delete this->_aboutMenu;
    this->_aboutMenu = nullptr;
}

/**
 *
 */
bool SideMenu::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SideMenuResizeRequest)) == eventType)
    {
        int width = static_cast<const SideMenuResizeRequest&>(event).getWidth();
        int height = static_cast<const SideMenuResizeRequest&>(event).getHeight();

		this->_ui->anchor->update();
        this->_ui->setFixedSize(Eigen::Vector2i(width, height));
        this->_ui->setSize(Eigen::Vector2i(width, height));
        this->_mainMenu->update();
        this->_ui->reflowNeeded = true;

        return true;
    }
    else if
        (std::type_index(typeid(GranuleInfoReadyForDisplayEvent)) == eventType)
    {
        const GranuleInfo& info = static_cast<const GranuleInfoReadyForDisplayEvent&>(event).getGranuleInfo();

		this->_mainMenu->extractionDialog->setVisible(false);
		this->_mainMenu->granuleInfo->displayInfo(info);
		this->_mainMenu->granuleInfo->setVisible(true);
		this->_mainMenu->update();
		this->_ui->setVisible(true);
		this->_ui->reflowNeeded = true;

        return true;
    }
    else if
        (std::type_index(typeid(DisplayCatalogueRequest)) == eventType)
    {
		if
			(false == this->_ui->visible())
		{
			this->_ui->setVisible(true);
		}

		if
			(false == this->_mainMenu->collectionsControl->visible())
		{
			this->_mainMenu->collectionsControl->setVisible(true);
			this->_catalogueButton->setPushed(true);
			this->_mainMenu->update();
		}
    }
    else if
        (std::type_index(typeid(DisplayExtractionInfoRequest)) == eventType)
    {
        bool success(false);

        const std::map<RenderableId, Extraction>& extractionData = static_cast<const DisplayExtractionInfoRequest&>(event).getExtractionData();
        const std::map<RenderableId, std::string>& labels = static_cast<const DisplayExtractionInfoRequest&>(event).getLabels();

        this->_mainMenu->extractionDialog->setExtraction(extractionData, labels, success);
        this->_mainMenu->extractionDialog->setVisible(true);
        this->_mainMenu->granuleInfo->setVisible(false);
        this->_mainMenu->update();
        this->_ui->setVisible(true);
        this->_ui->reflowNeeded = true;

        return true;
    }

    return false;
}

NanoguiMainMenu* SideMenu::getMainMenu() const
{
    return this->_mainMenu;
}

/*
	this->_ui->anchor->updateCallback(0, 0, 0, 0); //Hack required with some window managers (the other workaround is to resize the window)
*/
