// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_annotation_dialog.h"
#include "ui/nanogui/nanogui_properties_editor.h"
#include "ui/nanogui/nanogui_attribute.h"
#include "core/measurements.h"
#include "core/data_node.h"
#include "core/log.h"
#include "core/strptime.h"

#include "ui/nanogui/nanogui_time_coverage_editor.h"
#include "ui/nanogui/nanogui_image_view.h"


using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiAnnotationDialog::NanoguiAnnotationDialog(nanogui::Widget* parent)
: nanogui::Widget(parent),
  collectionAttr(nullptr),
  variableAttr(nullptr),
  spatialCoverageAttr(nullptr),
  _editor(nullptr),
  btns(nullptr),
  extractBtn(nullptr),
  delBtn(nullptr),
  editBtn(nullptr),
  saveBtn(nullptr),
  extractCallback(nullptr),
  deleteCallback(nullptr),
  editCallback(nullptr),
  saveCallback(nullptr),
  _timeCoverageEditor(nullptr),
  _savePropertiesCallback(nullptr),
  _updateTimeCoverageCallback(nullptr)
{
    nanogui::Widget* attrs; //

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Fill,
                                           0, 6));

    attrs = new nanogui::Widget(this);
    attrs->setFixedWidth(360);
	attrs->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Minimum,
										   3, 0));

    this->_timeCoverageEditor = new NanoguiTimeCoverageEditor(attrs);
    this->_timeCoverageEditor->setCallback([this] (const std::string& startStr, const std::string& stopStr) {
        if
            (nullptr == this->_updateTimeCoverageCallback)
        {
            return;
        }

        time_t tStart;	//
        timestamp_t timestampStart;	//
        std::tm tsStart = {};	//

        strptime(startStr.c_str(), "%Y-%m-%d %H:%M:%S", &tsStart);
        tStart = std::mktime(&tsStart);
        timestampStart = static_cast<timestamp_t>(tStart);

        time_t tStop;	//
        timestamp_t timestampStop;	//
        std::tm tsStop = {};	//

        strptime(stopStr.c_str(), "%Y-%m-%d %H:%M:%S", &tsStop);
        tStop = std::mktime(&tsStop);
        timestampStop = static_cast<timestamp_t>(tStop);

        this->_updateTimeCoverageCallback(this->target,
                                          1000 * static_cast<double>(timestampStart),
                                          1000 * static_cast<double>(timestampStop));
    });

    this->spatialCoverageAttr = new NanoguiAttribute(attrs);
    this->spatialCoverageAttr->setName("Spatial coverage");
    this->spatialCoverageAttr->setValue("_");
	this->collectionAttr = new NanoguiAttribute(attrs);
	this->collectionAttr->nameLbl->setFixedWidth(354);
	this->collectionAttr->valueLbl->setFixedWidth(344);
	this->collectionAttr->setName("Collection");
	this->variableAttr = new NanoguiAttribute(attrs);
	this->variableAttr->nameLbl->setFixedWidth(354);
	this->variableAttr->valueLbl->setFixedWidth(344);
	this->variableAttr->setName("Variable");

    this->_editor = new NanoguiPropertiesEditor(this);
    this->_editor->resize(Eigen::Vector2i(360, 200));
    this->_editor->setSaveCallback([this] (const DictNode& properties) {
        if
            (nullptr != this->_savePropertiesCallback)
        {
            this->_savePropertiesCallback(this->target, properties);
        }
    });

    this->btns = new nanogui::Widget(this);
    this->btns->setFixedWidth(360);
    this->btns->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                 nanogui::Alignment::Middle,
                                                 2, 4));

	this->extractBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_PUBLISH);
	this->extractBtn->setFixedSize(Eigen::Vector2i(20, 20));
	this->extractBtn->setTooltip("Extract granule data intersecting this shape.");

	this->delBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_TRASH);
	this->delBtn->setFixedSize(Eigen::Vector2i(20, 20));
	this->delBtn->setTooltip("Delete this shape.");

    this->editBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_EDIT);
    this->editBtn->setFlags(nanogui::Button::ToggleButton);
    this->editBtn->setFixedSize(Eigen::Vector2i(20, 20));
    this->editBtn->setTooltip("Edit this shape.");

    this->saveBtn = new nanogui::Button(this->btns, "", ENTYPO_ICON_SAVE);
    this->saveBtn->setFixedSize(Eigen::Vector2i(20, 20));
    this->saveBtn->setTooltip("Save this shape as JSON.");

	this->extractBtn->setCallback([this]() {
		if
			(nullptr != this->extractCallback)
		{
			this->extractCallback(this->target);
		}
	});

	this->delBtn->setCallback([this]() {
		if
			(nullptr != this->deleteCallback)
		{
			this->deleteCallback(this->target);
		}
	});

    this->editBtn->setChangeCallback([this] (bool pushed) {
        this->enableEdition(pushed, false);
    });

    this->saveBtn->setCallback([this]() {
        if
            (nullptr != this->saveCallback)
        {
            this->saveCallback(this->target);
        }
    });

    this->setVisible(false);
};

/**
 * Destructor.
 */
NanoguiAnnotationDialog::~NanoguiAnnotationDialog()
{
};

/**
 *
 */
void NanoguiAnnotationDialog::enableEdition(bool enabled, bool silent)
{
    this->editBtn->setPushed(enabled);
    this->extractBtn->setEnabled(!enabled);
    this->delBtn->setEnabled(!enabled);
    this->_editor->setPreventImageViewer(enabled);
    if
        ((false == silent) && (nullptr != this->editCallback))
    {
        this->editCallback(this->target, enabled);
    }
}

/**
 *
 */
bool NanoguiAnnotationDialog::enable(const RenderableId& rId, const GranuleInfo& info, DictNode&& properties, double start, double stop)
{
    Log::log(LogLevel::Debug, "Annotation dialog", "enable");

    if
        (  (rId.granule.first == this->target.granule.first)
        && (rId.granule.second == this->target.granule.second)
        && true == this->visible())
    {
        return false;
    }

    // Collection
	if
		(0 < info.collection.label.size())
	{
		this->collectionAttr->setValue(info.collection.label);
		this->collectionAttr->setVisible(true);
	}
	else
	{
		this->collectionAttr->setVisible(false);
	}

    // Variable
	if
		(0 < info.variable.label.size())
	{
		this->variableAttr->setValue(info.variable.label);
		this->variableAttr->setVisible(true);
	}
	else
	{
		this->variableAttr->setVisible(false);
	}

    // Time
    this->updateTemporalCoverage(start, stop);
    this->_timeCoverageEditor->setVisible(true);

    // Space
    this->updateSpatialCoverage(static_cast<double>(info.area),
                                static_cast<double>(info.perimeter),
                                static_cast<double>(info.lon),
                                static_cast<double>(info.lat));

    auto found = properties.children.find("properties");
    if
        (properties.children.end() != found)
    {
        this->_editor->setProperties(std::move(properties));
    }
    else
    {
        DictNode emptyProperties;
        this->_editor->setProperties(std::move(emptyProperties));
    }

    this->extractBtn->setEnabled(true);
    this->delBtn->setEnabled(true);
    this->editBtn->setEnabled(true);
    this->editBtn->setPushed(false);
    this->saveBtn->setEnabled(true);
    this->setVisible(true);

    this->target = rId;

    return true;
};

void NanoguiAnnotationDialog::updateSpatialCoverage(double area, double perimeter, double lon, double lat)
{
    std::ostringstream buffer("");  //
	std::string tmpStr;	//
	bool ok;	//
    bool areaOk;    //

    areaOk = SEAScope::formatArea(area, tmpStr);
    if
        (true == areaOk)
    {
        buffer << "Area: " << tmpStr;
    }

    ok = SEAScope::formatDistance(perimeter, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        if
            (true == areaOk)
        {
            buffer << "Perimeter: " << tmpStr;
        }
        else
        {
            buffer  << "Length: " << tmpStr;
        }
    }

    ok = SEAScope::formatLongitude(lon, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        buffer << "Longitude: " << tmpStr;
    }

    ok = SEAScope::formatLatitude(lat, tmpStr);
    if
        (true == ok)
    {
        if
            (0 < buffer.tellp())
        {
            buffer << "\n";
        }

        buffer << "Latitude: " << tmpStr;
    }

    this->spatialCoverageAttr->setValue(buffer.str());
}

/**
 *
 */
void NanoguiAnnotationDialog::updateTemporalCoverage(double start, double stop)
{
	char startStr[20];	//
	char stopStr[20];	//
	time_t timestampStart;	//
	time_t timestampStop;	//
	std::tm* ltimeStart;	//
	std::tm* ltimeStop;	//

    timestampStart = static_cast<long>(.001 * start);
    ltimeStart = std::localtime(&timestampStart);
    strftime(startStr, sizeof(startStr), "%Y-%m-%d %H:%M:%S", ltimeStart);

    timestampStop = static_cast<long>(.001 * stop);
    ltimeStop = std::localtime(&timestampStop);
    strftime(stopStr, sizeof(stopStr), "%Y-%m-%d %H:%M:%S", ltimeStop);

    this->_timeCoverageEditor->setValue(startStr, stopStr);
}

/**
 *
 */
bool NanoguiAnnotationDialog::disable()
{
    Log::log(LogLevel::Debug, "Annotation dialog", "disable");
    if
        (true == this->visible())
    {
        this->_editor->hideImageViewer();
        this->extractBtn->setEnabled(false);
        this->delBtn->setEnabled(false);
        this->editBtn->setEnabled(false);
        this->editBtn->setPushed(false);
        if
            (nullptr != this->editCallback)
        {
            this->editCallback(this->target, false);
        }
        this->saveBtn->setEnabled(false);
	    this->setVisible(false);
        this->setFocused(false);
        return true;
    }
	return false;
};

/**
 *
 */
void NanoguiAnnotationDialog::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float btnsPosX; //
    float btnsPosY; //
    float btnsSizeX;    //
    float btnsSizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());

    btnsPosX = static_cast<float>(this->btns->position().x());
    btnsPosY = static_cast<float>(this->btns->position().y());
    btnsSizeX = static_cast<float>(this->btns->width());
    btnsSizeY = static_cast<float>(this->btns->height());

	nvgBeginPath(ctx);
	nvgMoveTo( ctx
		     , posX + btnsPosX
		     , posY + btnsPosY);
	nvgLineTo( ctx
		     , posX + btnsPosX + btnsSizeX
		     , posY + btnsPosY);
	nvgStrokeColor(ctx, nanogui::Color(50,50,50,50));
	nvgStroke(ctx);

	nvgBeginPath(ctx);
	nvgRect( ctx
		   , posX + btnsPosX
		   , posY + btnsPosY
		   , btnsSizeX, btnsSizeY);
	nvgFillColor(ctx, nanogui::Color(255,255,255,10));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiAnnotationDialog::setSavePropertiesCallback(const std::function<void(const RenderableId&, const DictNode&)>& callback)
{
    if
        (nullptr != this->_savePropertiesCallback)
    {
        return false;
    }

    this->_savePropertiesCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiAnnotationDialog::setUpdateTimeCoverageCallback(const std::function<void(const RenderableId&, double, double)>& callback)
{
    if
        (nullptr != this->_updateTimeCoverageCallback)
    {
        return false;
    }

    this->_updateTimeCoverageCallback = callback;
    return true;
}
