// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iomanip>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_camera_control.h"
#include "ui/nanogui/nanogui_icon.h"

#include "core/app_events_dispatcher.h"
#include "core/input_events_dispatcher.h"
#include "3d/scene_view.h"

#include "events/scene_view_updated_event.h"
#include "events/cursor_location_changed_event.h"

using namespace SEAScope;

/**
 *
 */
NanoguiCameraControl::NanoguiCameraControl(nanogui::Widget* parent, AppEventsDispatcher& appEventsDispatcher, const InputEventsDispatcher& inputEventsDispatcher)
: nanogui::Widget(parent), NanoguiComponent(parent), location(nullptr),
  altitude(nullptr), scale(nullptr),
  _appEventsDispatcher(appEventsDispatcher), _inputEventsDispatcher(inputEventsDispatcher)
{
	nanogui::Widget* spacer;	//
	NanoguiIcon* icon;	//

	this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
										   nanogui::Alignment::Minimum,
										   3, 0));

	/* Coordinates */
	icon = new NanoguiIcon(this, ENTYPO_ICON_LOCATION_PIN);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Location: longitude and latitude under the mouse cursor");

	this->location = new nanogui::Label(this, "", "sans-bold");
	this->location->setFixedSize(Eigen::Vector2i(98, 25));

	/* Zoom */
	icon = new NanoguiIcon(this, ENTYPO_ICON_TRIANGLE_UP);
	icon->setFixedSize(Eigen::Vector2i(20, 20));
	icon->setFontSize(18);
	icon->setColor(nanogui::Color(255,255,255,255));
	icon->setTooltip("Altitude of the camera (above sea level)");

	this->altitude = new nanogui::Label(this, "00000m", "sans-bold");
	this->altitude->setFixedSize(Eigen::Vector2i(62, 20));

	spacer = new nanogui::Widget(this);
	spacer->setFixedSize(Eigen::Vector2i(18, 24));

	/* Scale */
	this->scale = new NanoguiScale(this);
    this->scale->setFixedSize(Eigen::Vector2i(65, 24));

    this->_appEventsDispatcher.subscribe<SceneViewUpdatedEvent>(this);
    this->_appEventsDispatcher.subscribe<CursorLocationChangedEvent>(this);
};

/**
 *
 */
NanoguiCameraControl::~NanoguiCameraControl()
{
    this->_appEventsDispatcher.unsubscribe<SceneViewUpdatedEvent>(this);
    this->_appEventsDispatcher.unsubscribe<CursorLocationChangedEvent>(this);
};

/**
 *
 */
bool NanoguiCameraControl::setSize(int w, int h)
{
	this->setFixedSize(Eigen::Vector2i(w, h));
	return true;
};

/**
 *
 */
void NanoguiCameraControl::draw(NVGcontext* ctx)
{
	nvgBeginPath(ctx);
	nvgRoundedRect(ctx,
					static_cast<float>(this->mPos.x()), static_cast<float>(this->mPos.y()),
					static_cast<float>(this->mSize.x()), static_cast<float>(this->mSize.y()), 12);
	nvgFillColor(ctx, nanogui::Color(255,255,255,10));
	nvgFill(ctx);
	nanogui::Widget::draw(ctx);
};

/**
 *
 */
bool NanoguiCameraControl::handleAppEvent(std::type_index eventType, const AppEvent& event)
{
    if
        (std::type_index(typeid(SceneViewUpdatedEvent)) == eventType)
    {
        // Since camera changed we need to update cursor location

        std::ostringstream buffer;	//

        glm::vec2 lonlat;
        glm::vec3 hitpoint;
        const SceneView* sceneViewTmp = static_cast<const SceneViewUpdatedEvent&>(event).getSceneView();
        bool onWorld = sceneViewTmp->getLonLat(this->_inputEventsDispatcher.getInputDevicesState().getMouseX(),
                                               this->_inputEventsDispatcher.getInputDevicesState().getMouseY(),
                                               lonlat, hitpoint);
        float resolution;	//
        resolution = sceneViewTmp->getPixelSize();

        float altitudeValue;	//
        altitudeValue = sceneViewTmp->getAltitude();

        sceneViewTmp = nullptr;

        if
            (false == onWorld)
        {
            this->location->setCaption("Lost in space");
        }
        else
        {
            buffer << std::fixed << std::setprecision(2) << lonlat.x << "°, ";
            buffer << std::fixed << std::setprecision(2) << lonlat.y << "°";
            this->location->setCaption(buffer.str());
        }

        this->scale->setResolution(resolution);

        buffer.str("");
        if
            (10000 > altitudeValue)
        {
            buffer << std::fixed << std::setprecision(0) << altitudeValue << "m";
        }
        else
        {
            altitudeValue = 0.001f * altitudeValue;
            buffer << std::fixed << std::setprecision(1) << altitudeValue << "km";
        }
        this->altitude->setCaption(buffer.str());

        return true;
    }
    else if
        (std::type_index(typeid(CursorLocationChangedEvent)) == eventType)
    {
        std::ostringstream buffer;	//
        const CursorLocationChangedEvent& typedEvent = static_cast<const CursorLocationChangedEvent &>(event);
        if
            (false == typedEvent.getOnWorld())
        {
            this->location->setCaption("Lost in space");
        }
        else
        {
            buffer << std::fixed << std::setprecision(2) << typedEvent.getLon() << "°, ";
            buffer << std::fixed << std::setprecision(2) << typedEvent.getLat() << "°";
            this->location->setCaption(buffer.str());
        }

        return true;
    }

    return false;
}
