// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_contextual_menu.h"
#include "ui/nanogui/nanogui_draggable_pane.h"
#include "ui/nanogui/nanogui_annotation_dialog.h"
#include "ui/nanogui/nanogui_granule_dialog.h"
#include "ui/nanogui/nanogui_path_selector.h"
#include "core/data_node.h"
#include "core/path.h"
#include "core/log.h"

using namespace SEAScope;

/**
 *
 */
NanoguiContextualMenu::NanoguiContextualMenu()
: _window(nullptr),
  _title(nullptr),
  _closeBtn(nullptr),
  _annotationDialog(nullptr),
  _granuleDialog(nullptr),
  _pathBlock(nullptr),
  _pathSelector(nullptr),
  _saveButtonsBlock(nullptr),
  _saveDataButton(nullptr),
  _saveCallback(nullptr),
  _closeCallback(nullptr),
  _renderableType(RenderableType::GRANULE)
{
};

/**
 *
 */
NanoguiContextualMenu::~NanoguiContextualMenu()
{
  // Deleted by nanogui
  this->_window = nullptr;
  this->_title = nullptr;
  this->_closeBtn = nullptr;
  this->_annotationDialog = nullptr;
  this->_granuleDialog = nullptr;
  this->_pathBlock = nullptr;
  this->_pathSelector = nullptr;
  this->_saveButtonsBlock = nullptr;
  this->_saveDataButton = nullptr;

  // reset callbacks;
  this->_saveCallback = nullptr;
  this->_closeCallback = nullptr;
};

/**
 *
 */
NanoguiAnnotationDialog* NanoguiContextualMenu::getAnnotationDialog() const
{
    return this->_annotationDialog;
}

/**
 *
 */
NanoguiGranuleDialog* NanoguiContextualMenu::getGranuleDialog() const
{
    return this->_granuleDialog;
}

/**
 *
 */
bool NanoguiContextualMenu::setSaveCallback(const std::function<void(const RenderableId&, const std::string&)>& callback)
{
    if
        (nullptr != this->_saveCallback)
    {
        return false;
    }

    this->_saveCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiContextualMenu::setCloseCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_closeCallback)
    {
        return false;
    }

    this->_closeCallback = callback;
    return true;
}

/**
 *
 */
bool NanoguiContextualMenu::init(nanogui::Widget* screen)
{
	nanogui::Widget* headerbar;	//
	std::int32_t headerHeight;	//

	if
		(nullptr != this->_window)
	{
		/* Already initialized */
		return false;
	}

	headerHeight = 20;

	this->_window = new NanoguiDraggablePane(screen);
	this->_window->headerHeight = static_cast<std::uint32_t>(headerHeight);

	headerbar = new nanogui::Widget(this->_window);
	headerbar->setFixedSize(Eigen::Vector2i(360, headerHeight));
	headerbar->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal
											   , nanogui::Alignment::Middle
											   , 2, 0));

	this->_title = new nanogui::Label(headerbar, "", "sans-bold");
	this->_title->setFixedWidth(340);

	this->_closeBtn = new nanogui::Button(headerbar, "", ENTYPO_ICON_CROSS);
	this->_closeBtn->setFixedSize(Eigen::Vector2i(16, 16));
	this->_closeBtn->setCallback([this]() {
		if
			(nullptr != this->_closeCallback)
		{
			this->_closeCallback();
		}
	});

    this->_window->keyCallback = [this] (int key, int /*scancode*/, int action, int /*modifiers*/) -> bool {
        if
            (action == GLFW_PRESS || action == GLFW_REPEAT)
        {
            if
                (GLFW_KEY_ESCAPE == key
                && nullptr != this->_closeCallback)
            {
                this->_closeCallback();
                return true;
            }
        }

        return false;
    };

    this->_annotationDialog = new NanoguiAnnotationDialog(this->_window);
    this->_annotationDialog->saveCallback = [this](const RenderableId& /*rId*/) {
        this->_pathBlock->setVisible(true);
        this->_saveButtonsBlock->setVisible(true);
        this->_pathSelector->show();
        this->_window->updateLayout();
    };

    this->_granuleDialog = new NanoguiGranuleDialog(this->_window);

    this->_pathBlock = new nanogui::Widget(this->_window);
    this->_pathBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                       nanogui::Alignment::Fill,
                                                       2, 3));

    nanogui::Widget* pathTitleBlock(nullptr);
    pathTitleBlock = new nanogui::Widget(this->_pathBlock);
    pathTitleBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                     nanogui::Alignment::Middle,
                                                     0, 0));
    new nanogui::Label(pathTitleBlock, "Select file path", "sans-bold", 16);

    nanogui::Widget* pathContainer = new nanogui::Widget(this->_pathBlock);
    pathContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                    nanogui::Alignment::Middle,
                                                    0, 0));
    this->_pathSelector = new NanoguiPathSelector(pathContainer);
    this->_pathSelector->resize(Eigen::Vector2i(356, 40));
    this->_pathSelector->setChangeCallback([this] (const std::string&) {
        const std::string& path = this->_pathSelector->getPath();
        this->_saveDataButton->setEnabled( (0 < path.size()) && (false == Path::dirExists(path)));
    });

    this->_saveButtonsBlock = new nanogui::Widget(this->_window);
    this->_saveButtonsBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                              nanogui::Alignment::Maximum,
                                                              0, 0));
    this->_saveButtonsBlock->setFixedSize(Eigen::Vector2i(356, 20));

    nanogui::Widget* saveButtons = new nanogui::Widget(this->_saveButtonsBlock);
    saveButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                  nanogui::Alignment::Middle,
                                                  3, 2));
    this->_saveDataButton = new nanogui::Button(saveButtons, "Save");
    this->_saveDataButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_saveDataButton->setFontSize(16);
    this->_saveDataButton->setEnabled(false);
    this->_saveDataButton->setCallback([this] () {
        const std::string& path = this->_pathSelector->getPath();

        if
            (nullptr != this->_saveCallback)
        {
            this->_saveCallback(this->_renderableId, path);
        }

        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });

    nanogui::Button* saveNokButton(nullptr);
    saveNokButton = new nanogui::Button(saveButtons, "Cancel");
    saveNokButton->setFixedSize(Eigen::Vector2i(60, 18));
    saveNokButton->setFontSize(16);
    saveNokButton->setCallback([this] () {
        this->_pathSelector->hide();
        this->_pathBlock->setVisible(false);
        this->_saveButtonsBlock->setVisible(false);
        this->_window->updateLayout();
    });

    this->_pathSelector->hide();
    this->_pathBlock->setVisible(false);
    this->_saveButtonsBlock->setVisible(false);

    return true;
}

/**
 *
 */
bool NanoguiContextualMenu::display(const Eigen::Vector2i& cursorPos)
{
    bool ok;    //
    Eigen::Vector2i pos;    //

    ok = this->adaptPosition(cursorPos, pos);
    this->_window->display(pos);
    if
        (ok)
    {
        this->_window->requestFocus();
		if
			(RenderableType::GRANULE == this->_renderableType)
		{
			this->_granuleDialog->requestFocus();
        }
        else
        {
            this->_annotationDialog->requestFocus();
        }
    }
    return ok;
}

/**
 *
 */
void NanoguiContextualMenu::hide()
{
    this->_annotationDialog->disable();
    this->_granuleDialog->disable();

    this->_pathSelector->hide();
    this->_pathBlock->setVisible(false);
    this->_saveButtonsBlock->setVisible(false);

    this->_window->hide();
}

/**
 *
 */
bool NanoguiContextualMenu::adaptPosition(const Eigen::Vector2i& cursorPosIn, Eigen::Vector2i& result) const
{
    float scaleFactor;  //
    bool fitFound;  //
    bool fits;  //
    std::int32_t x;   //
    std::int32_t y;   //
    Eigen::Vector2i cursorPos;  //

    /* Apply scale-factor on the cursor position */
    scaleFactor = this->_window->screen()->pixelRatio();
    cursorPos = Eigen::Vector2i(std::floor(static_cast<float>(cursorPosIn.x()) / scaleFactor),
                                std::floor(static_cast<float>(cursorPosIn.y()) / scaleFactor));

    fitFound = true;
    const Eigen::Vector2i& parentSize = this->_window->parent()->size();
    const Eigen::Vector2i& selfSize = this->_window->size();

    /* By default, center on the cursor horizontal position */
    x = static_cast<std::int32_t>(std::floor(cursorPos.x() - 0.5 * selfSize.x()));
    if
        (0 > x)
    {
        /* Move the menu to the right of the cursor with a small margin */
        x = cursorPos.x() + 20;
        if
            (parentSize.x() < x + selfSize.x())
        {
            /* the screen estate is too small, move menu at the horizontal
               center of the screen */
            x = static_cast<std::int32_t>(std::floor(0.5 * (parentSize.x() - selfSize.x())));
            fits = (0<x && parentSize.x() < (x + selfSize.x()));
            fitFound = fitFound && fits;
        }
    }

    if
        (parentSize.x() < x + selfSize.x())
    {
        /* Move the menu to the left of the cursor with a small margin */
        x = cursorPos.x() - selfSize.x() - 20;
        if
            (0 > x)
        {
            /* the screen estate is too small, move menu at the horizontal
               center of the screen */
            x = static_cast<std::int32_t>(std::floor(0.5 * (parentSize.x() - selfSize.x())));
            fits = (0<x && parentSize.x() < (x + selfSize.x()));
            fitFound = fitFound && fits;
        }
    }

    /* The menu is displayed below the cursor by default */
    y = cursorPos.y() + 20;
    if
        (parentSize.y() < y + selfSize.y())
    {
        /* Move the menu above cursor */
        y = cursorPos.y() - 20 - selfSize.y();
        if
            (0 > y)
        {
            y = static_cast<std::int32_t>(std::floor(0.5 * (parentSize.y() - selfSize.y())));
            fits = (0<y && parentSize.y() < (y + selfSize.y()));
            fitFound = fitFound && fits;
        }
    }

    result = Eigen::Vector2i(x,y);
    return fitFound;
}

/**
 *
 */
bool NanoguiContextualMenu::setTarget(const RenderableId& rId, const RenderableType& rType, const GranuleInfo& info, DictNode&& properties, double start, double stop)
{
    this->_pathSelector->hide();
    this->_pathBlock->setVisible(false);
    this->_saveButtonsBlock->setVisible(false);

    this->_renderableId = rId;
	this->_renderableType = rType;
    if
        (RenderableType::GRANULE == rType)
    {
        this->_title->setCaption("Selected granule");
        this->_annotationDialog->disable();
        this->_granuleDialog->enable(rId, info);
    }
    else
    {
        this->_title->setCaption("Selected annotation");
        this->_granuleDialog->disable();
        this->_annotationDialog->enable(rId, info, std::move(properties), start, stop);
    }

    this->_window->updateLayout();
    return true;
}
