// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <sstream>

#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_granule_dialog.h"
#include "core/measurements.h"
#include "core/log.h"

using namespace SEAScope;

/**
 * Constructor.
 */
NanoguiGranuleDialog::NanoguiGranuleDialog(nanogui::Widget* parent):
nanogui::Widget(parent), btns(nullptr), collectionAttr(nullptr),
granuleAttr(nullptr), variableAttr(nullptr), timeCoverageAttr(nullptr),
spatialCoverageAttr(nullptr), infoCallback(nullptr),
centerTimeCallback(nullptr), extractCallback(nullptr)
{
	nanogui::Widget* attrs;	//
	nanogui::Button* btn;	//

	this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Fill,
										   0, 6));

	attrs = new nanogui::Widget(this);
	attrs->setFixedWidth(360);
	attrs->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										    nanogui::Alignment::Minimum,
										    3, 0));

	this->btns = new nanogui::Widget(this);
	this->btns->setFixedWidth(360);
	this->btns->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
										   nanogui::Alignment::Middle,
										   2, 4));

	this->granuleAttr = new NanoguiAttribute(attrs);
	this->granuleAttr->nameLbl->setFixedWidth(354);
	this->granuleAttr->valueLbl->setFixedWidth(344);
	this->granuleAttr->setName("Granule");
	this->timeCoverageAttr = new NanoguiAttribute(attrs);
	this->timeCoverageAttr->nameLbl->setFixedWidth(354);
	this->timeCoverageAttr->valueLbl->setFixedWidth(344);
	this->timeCoverageAttr->setName("Time coverage");
	this->spatialCoverageAttr = new NanoguiAttribute(attrs);
	this->spatialCoverageAttr->nameLbl->setFixedWidth(354);
	this->spatialCoverageAttr->valueLbl->setFixedWidth(344);
	this->spatialCoverageAttr->setName("Spatial coverage");
	this->collectionAttr = new NanoguiAttribute(attrs);
	this->collectionAttr->nameLbl->setFixedWidth(354);
	this->collectionAttr->valueLbl->setFixedWidth(344);
	this->collectionAttr->setName("Collection");
	this->variableAttr = new NanoguiAttribute(attrs);
	this->variableAttr->nameLbl->setFixedWidth(354);
	this->variableAttr->valueLbl->setFixedWidth(344);
	this->variableAttr->setName("Variable");

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_PUBLISH);
	btn->setTooltip("Extract data that ovrlap the footprint of this granule");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->extractCallback)
		{
			this->extractCallback(this->target);
		}
	});

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_INFO);
	btn->setTooltip("Info");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->infoCallback)
		{
			this->infoCallback(this->target);
		}
	});

	btn = new nanogui::Button(this->btns, "", ENTYPO_ICON_CLOCK);
	btn->setTooltip("Center the timeline on this granule's datetime");
	btn->setFixedSize(Eigen::Vector2i(20, 20));
	btn->setCallback([this]() {
		if
			(nullptr != this->centerTimeCallback)
		{
			this->centerTimeCallback(this->target);
		}
	});

};

/**
 * Destructor.
 */
NanoguiGranuleDialog::~NanoguiGranuleDialog()
{
};

/**
 *
 */
bool NanoguiGranuleDialog::enable(const RenderableId& rId, const GranuleInfo& info)
{
	std::ostringstream buffer;	//
	char startStr[20];	//
	char stopStr[20];	//
	time_t timestampStart;	//
	time_t timestampStop;	//
	std::tm* ltimeStart;	//
	std::tm* ltimeStop;	//
    std::string tmpStr; //
    bool hasLength; //
    bool hasArea;   //
    bool ok;    //

	this->target = rId;

	Log::log(LogLevel::Debug, "Granule dialog", "enable");

	if
		(0 < info.collection.label.size())
	{
		this->collectionAttr->setValue(info.collection.label);
		this->collectionAttr->setVisible(true);
	}
	else
	{
		this->collectionAttr->setVisible(false);
	}

	if
		(nullptr != info.metadata)
	{
		/* start */
		timestampStart = static_cast<long>(.001 * static_cast<double>(info.metadata->start));
		ltimeStart = std::localtime(&timestampStart);
		strftime(startStr, sizeof(startStr), "%Y-%m-%d %H:%M:%S", ltimeStart);

		/* stop */
		timestampStop = static_cast<long>(.001 * static_cast<double>(info.metadata->stop));
		ltimeStop = std::localtime(&timestampStop);
		strftime(stopStr, sizeof(stopStr), "%Y-%m-%d %H:%M:%S", ltimeStop);

		buffer << startStr << " ... " << stopStr;
		this->timeCoverageAttr->setValue(buffer.str());
		this->timeCoverageAttr->setVisible(true);

		if
			(0 < info.metadata->dataId.size())
		{
			this->granuleAttr->setValue(info.metadata->dataId);
			this->granuleAttr->setVisible(true);
		}
		else
		{
			this->granuleAttr->setVisible(false);
		}
	}
	else
	{
		this->timeCoverageAttr->setVisible(false);
	}

	if
		(0 < info.variable.label.size())
	{
		this->variableAttr->setValue(info.variable.label);
		this->variableAttr->setVisible(true);
	}
	else
	{
		this->variableAttr->setVisible(false);
	}

	buffer.str("");
    hasArea = (false == std::isnan(info.area));
    if
        (true == hasArea)
    {
        ok = SEAScope::formatArea(static_cast<double>(info.area), tmpStr);
        if
            (true == ok)
        {
            buffer << "Area: " << tmpStr;
        }
    }

    hasLength = (false == std::isnan(info.perimeter));
    if
        (true == hasLength)
    {
        if
            (true == hasArea)
        {
            ok = SEAScope::formatDistance(static_cast<double>(info.perimeter), tmpStr);
            if
                (true == ok)
            {
                buffer << "\nPerimeter: " << tmpStr;
            }
        }
        else
        {
            ok = SEAScope::formatDistance(static_cast<double>(info.perimeter), tmpStr);
            if
                (true == ok)
            {
                buffer << "Length: " << tmpStr;
            }
        }

    }
	this->spatialCoverageAttr->setValue(buffer.str());

	this->setVisible(true);
	return true;
};

/**
 *
 */
bool NanoguiGranuleDialog::disable()
{
	Log::log(LogLevel::Debug, "Granule dialog", "disable");
	if
		(true == this->visible())
	{
		this->setVisible(false);
        this->setFocused(false);
		return true;
	}
	return false;
};

/**
 *
 */
void NanoguiGranuleDialog::draw(NVGcontext* ctx)
{
	nvgBeginPath(ctx);
	nvgMoveTo( ctx
		     , static_cast<float>(this->mPos.x() + this->btns->position().x())
		     , static_cast<float>(this->mPos.y() + this->btns->position().y()));
	nvgLineTo( ctx
		     , static_cast<float>(this->mPos.x() + this->btns->position().x() + this->btns->width())
		     , static_cast<float>(this->mPos.y() + this->btns->position().y()));
	nvgStrokeColor(ctx, nanogui::Color(50,50,50,50));
	nvgStroke(ctx);

	nvgBeginPath(ctx);
	nvgRect( ctx
		   , static_cast<float>(this->mPos.x() + this->btns->position().x())
		   , static_cast<float>(this->mPos.y() + this->btns->position().y())
		   , static_cast<float>(this->btns->width())
           , static_cast<float>(this->btns->height()));
	nvgFillColor(ctx, nanogui::Color(255,255,255,10));
	nvgFill(ctx);

	nanogui::Widget::draw(ctx);
};
