// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "vendors/_nanogui.h"

#include "ui/nanogui/nanogui_main_menu.h"
#include "core/log.h"

using namespace SEAScope;

/**
 * 
 */
NanoguiMainMenu::NanoguiMainMenu(nanogui::Widget* parent)
: nanogui::Widget(parent), granuleInfo(nullptr), extractionDialog(nullptr)
, sourcesControl(nullptr), collectionsControl(nullptr), reflowNeeded(true)
{
	this->setLayout( new nanogui::BoxLayout( nanogui::Orientation::Vertical
										   , nanogui::Alignment::Middle
										   ,  0, 0));

    this->granuleInfo = new NanoguiGranuleShowcase(this);
    this->granuleInfo->setTitle("Granule info");
    this->granuleInfo->setVisible(false);

    this->extractionDialog = new NanoguiExtractionDialog(this);
    this->extractionDialog->setTitle("Extracted data");
    this->extractionDialog->setVisible(false);

    this->sourcesControl = new NanoguiSourcesControl(this);
    this->sourcesControl->setVisible(false);

    this->collectionsControl = new NanoguiCollectionsControl(this);
    this->collectionsControl->setTitle("Catalogue");
    this->collectionsControl->setVisible(false);
};

/**
 *
 */
NanoguiMainMenu::~NanoguiMainMenu()
{
	this->granuleInfo = nullptr;
	this->extractionDialog = nullptr;
	this->sourcesControl = nullptr;
	this->collectionsControl = nullptr;
}

/**
 *
 */
bool NanoguiMainMenu::scrollEvent(const Eigen::Vector2i& pos, const Eigen::Vector2f& delta)
{
	nanogui::Widget::scrollEvent(pos, delta);

	/* return true to prevent events propagation */
	return true;
}

/**
 *
 */
void NanoguiMainMenu::draw(NVGcontext* ctx)
{
	if
		(this->reflowNeeded)
	{
		this->performLayout(ctx);
		this->reflowNeeded = false;
	}
	nanogui::Widget::draw(ctx);
};

/**
 *
 */
void NanoguiMainMenu::update()
{
	NanoguiScrollableWithTitle* resizeable(nullptr);	//
	const std::int32_t minGapSize(12);	//
	std::int32_t childrenCount(0);	//
	std::int32_t childHeight(0);	//
	std::int32_t verticalSpace(0);	//
    double childHeightTmp(0.0); //

	verticalSpace = this->fixedHeight();

	for
		(auto& child : this->mChildren)
	{
		if
			(child->visible())
		{
			childrenCount = childrenCount + 1;
		}
	}

	if
		(0 >= childrenCount)
	{
		return;
	}

	verticalSpace = verticalSpace - (childrenCount - 1) * minGapSize;

	for
		(auto& child : this->mChildren)
	{
		if
			(false == child->visible())
		{
			continue;
		}

		childHeight = child->preferredSize(this->screen()->nvgContext()).y();
		Log::log(LogLevel::Debug, "Main menu", "Child preferred height: {}",
				 childHeight);
		if
			(childHeight > verticalSpace / childrenCount)
		{
			childHeightTmp = std::floor(verticalSpace / childrenCount);
            childHeight = static_cast<int>(childHeightTmp);
		}
		Log::log(LogLevel::Debug, "Main menu", "Child actual height: {}",
				 childHeight);

		resizeable = dynamic_cast<NanoguiScrollableWithTitle*>(child);
		if
			(nullptr != resizeable)
		{
			resizeable->resize(this->fixedWidth(), childHeight);
			resizeable = nullptr;
		}
		else
		{
			child->setFixedHeight(childHeight);
		}

		verticalSpace = verticalSpace - childHeight;
		childrenCount = childrenCount - 1;
	}

	this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
										   nanogui::Alignment::Middle,
										   0, minGapSize));
    this->reflowNeeded = true;
}
