// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/log.h"
#include "ui/nanogui/nanogui_numberbox.h"
#include "num_utils.h"

using namespace SEAScope;

/**
 *
 */
NanoguiNumberBox::NanoguiNumberBox(nanogui::Widget* parent)
: nanogui::FloatBox<double>(parent),
  _delta(0.0),
  _requiresKeyboardValidation(false),
  _hasPendingChanges(false),
  _useDelayedCommit(false),
  _validationFramesDelay(0),
  _elapsedFrames(0),
  _validMin(std::numeric_limits<double>::lowest()),
  _validMax(std::numeric_limits<double>::max())
{
};

/**
 *
 */
NanoguiNumberBox::~NanoguiNumberBox()
{
};

/**
 *
 */
void NanoguiNumberBox::draw(NVGcontext* ctx)
{
    float posX; //
    float posY; //
    float sizeX;    //
    float sizeY;    //

    posX = static_cast<float>(this->mPos.x());
    posY = static_cast<float>(this->mPos.y());
    sizeX = static_cast<float>(this->mSize.x());
    sizeY = static_cast<float>(this->mSize.y());

	nanogui::FloatBox<double>::draw(ctx);

    if
        (true == this->_hasPendingChanges)
    {
        // Draw a line below value to hint user that an action is needed
        nvgBeginPath(ctx);
        for
            (float k = 0.f; k * 5.f + 3.f <= sizeX - 2.f; k = k + 1.F)
        {
            nvgMoveTo(ctx, posX + 1.f + 5.f * k, posY + sizeY - 2.f);
            nvgLineTo(ctx, posX + 1.f + 5.f * k + 3.f, posY + sizeY - 2.f);

        }
        nvgStrokeColor(ctx, nanogui::Color(30, 200, 230, 200));
        nvgStrokeWidth(ctx, 2);
        nvgStroke(ctx);

        if
            (true == this->_useDelayedCommit)
        {
            // Note: having this in the rendering loop is definitely not good
            // design but it is a lot easier than implementing a timer running
            // in a thread to throttle and delay changes
            if
                (this->_elapsedFrames >= this->_validationFramesDelay)
            {
                // commit value and execute callback immediately
                this->mValue = this->mValueTemp;
                this->mCommitted = true;

                if
                    (nullptr != this->mCallback)
                {
                    this->mCallback(this->mValue);
                }

                this->_elapsedFrames = 0;
                this->_hasPendingChanges = false;
                this->_useDelayedCommit = false;
            }
            else
            {
                this->_elapsedFrames = static_cast<uint8_t>(this->_elapsedFrames + 1u);
            }
        }
    }

	if
		(this->mMouseFocus && this->mEnabled && this->mEditable)
	{
		nvgBeginPath(ctx);
		nvgRoundedRect( ctx
					  , posX, posY
					  , sizeX, sizeY
					  , 3);
		nvgStrokeColor(ctx, nanogui::Color(255, 215, 30, 200));
		nvgStroke(ctx);
	}
};

/**
 *
 */
bool NanoguiNumberBox::focusEvent(bool focused)
{
    bool ok(false);

    ok = nanogui::TextBox::focusEvent(focused);

    if
        (false == focused)
    {
        // edited value has been committed by TextBox::focusEvent
        this->_hasPendingChanges = false;
        this->_useDelayedCommit = false;
    }


    return ok;
}

/**
 *
 */
bool NanoguiNumberBox::scrollEvent(const Eigen::Vector2i& p, const Eigen::Vector2f& rel)
{
    double newValue;	//
    double relYTmp;  //

    if
        (false == (this->mEnabled && this->mEditable))
    {
        return false;
    }

	nanogui::FloatBox<double>::scrollEvent(p, rel);

    relYTmp = static_cast<double>(rel.y());
    if
        (true == this->mCommitted)
    {
	    newValue = this->value() + relYTmp * this->_delta;
    }
    else
    {
        newValue = std::stod(this->mValueTemp) + relYTmp * this->_delta;
    }
	Log::log(LogLevel::Debug, "NanoguiNumberBox", "scroll {} {} {}", relYTmp, this->_delta, newValue);

    newValue = std::min(std::max(newValue, this->_validMin), this->_validMax);

    std::ostringstream buffer;
    buffer << newValue;
    this->mValueTemp = buffer.str();
    this->_hasPendingChanges = (this->mValueTemp != this->mValue);
    if
        (true == this->_hasPendingChanges)
    {
        if
            (0 >= this->_validationFramesDelay)
        {
            // commit value and execute callback immediately
            this->mValue = this->mValueTemp;
            this->mCommitted = true;
            this->_hasPendingChanges = false;
            this->_useDelayedCommit = false;

            if
                (nullptr != this->mCallback)
            {
                this->mCallback(this->mValue);
            }
        }
        else
        {
            // Start delay
            this->_elapsedFrames = 0;
            this->_useDelayedCommit = true;

            // required for nanogui to use mValueTemp instead of mValue when
            // drawing the widget
            this->mCommitted = false;
        }
    }

	return true;
};

/**
 *
 */
bool NanoguiNumberBox::keyboardEvent(int key, int scancode, int action, int modifiers)
{
    bool ok(false);
    if
        (false == (this->mEditable && this->mEnabled))
    {
        return false;
    }

    if
        ((GLFW_PRESS == action) && (GLFW_KEY_ESCAPE == key))
    {
        // Restore last committed value
        this->mValueTemp = this->mValue;
        this->mCommitted = true;
        this->_hasPendingChanges = false;
        this->_useDelayedCommit = false;
        this->_elapsedFrames = 0;
        this->setFocused(false);
    }

    ok = nanogui::FloatBox<double>::keyboardEvent(key, scancode, action, modifiers);

    if
        (true == this->_requiresKeyboardValidation)
    {
        if
            ((GLFW_PRESS == action) && (GLFW_KEY_ENTER == key))
        {
            // commit value and execute callback immediately
            this->mValue = this->mValueTemp;

            if
                (nullptr != this->mCallback)
            {
                this->mCallback(this->mValue);
            }

            this->_hasPendingChanges = false;
            this->_useDelayedCommit = false;
        }
        else
        {
            // mValueTemp contains the edited value and it will not be commited
            // until the user presses the Enter key
            this->_hasPendingChanges = (this->mValueTemp != this->mValue);
        }
    }
    else if
        (GLFW_RELEASE == action)
    {
        this->_hasPendingChanges = (this->mValueTemp != this->mValue);

        if
            (true == this->_hasPendingChanges)
        {
            if
                (0 >= this->_validationFramesDelay)
            {
                // commit value and execute callback immediately
                this->mValue = this->mValueTemp;
                this->_hasPendingChanges = false;

                if
                    (nullptr != this->mCallback)
                {
                    this->mCallback(this->mValue);
                }
            }
            else
            {
                // Start delay
                this->_elapsedFrames = 0;
                this->_useDelayedCommit = true;
            }
        }
        else
        {
            this->_useDelayedCommit = false;
        }
    }

    return ok;
}

/**
 *
 */
bool NanoguiNumberBox::setDelta(double value)
{
    if
        (floating_point_equality(value, this->_delta))
    {
        return false;
    }

    this->_delta = value;
    return true;
}

/**
 *
 */
double NanoguiNumberBox::getDelta() const
{
    return this->_delta;
}

/**
 *
 */
bool NanoguiNumberBox::setRequiresKeyboardValidation(bool value)
{
    if
        (value == this->_requiresKeyboardValidation)
    {
        return false;
    }

    this->_requiresKeyboardValidation = value;
    return true;
}

/**
 *
 */
bool NanoguiNumberBox::getRequiresKeyboardValidation() const
{
    return this->_requiresKeyboardValidation;
}

/**
 *
 */
bool NanoguiNumberBox::setValidationFramesDelay(std::uint8_t value)
{
    if
        (value == this->_validationFramesDelay)
    {
        return false;
    }

    this->_validationFramesDelay = value;
    return true;
}

/**
 *
 */
std::uint8_t NanoguiNumberBox::getValidationFramesDelay() const
{
    return this->_validationFramesDelay;
}

/**
 *
 */
bool NanoguiNumberBox::setValidMin(double value)
{
    if
        (floating_point_equality(value, this->_validMin))
    {
        return false;
    }
    this->_validMin = value;
    return true;
}

/**
 *
 */
bool NanoguiNumberBox::setValidMax(double value)
{
    if
        (floating_point_equality(value, this->_validMax))
    {
        return false;
    }
    this->_validMax = value;
    return true;
}
