// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "ui/nanogui/nanogui_path_selector.h"

#include "ui/nanogui/nanogui_list.h"
#include "ui/nanogui/nanogui_clickable_label.h"
#include "ui/nanogui/nanogui_anchored.h"
#include "ui/nanogui/nanogui_textfield.h"

#include "core/path.h"

using namespace SEAScope;

/**
 *
 */
NanoguiPathSelector::NanoguiPathSelector(nanogui::Widget* parent)
: nanogui::Widget(parent),
  _inputText(nullptr),
  _dropdownContainer(nullptr),
  _itemsList(nullptr),
  _browseButton(nullptr),
  _selectButton(nullptr),
  _parentDirButton(nullptr),
  _changeCallback(nullptr),
  _dropdownCallback(nullptr),
  _updateRequested(false),
  _selected("")
{
    std::string separator(1, Path::separator());
    Path::currentWorkingDir(this->_parentPath);
    this->_parentPath = this->_parentPath + separator;

    this->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                           nanogui::Alignment::Minimum,
                                           0, 0));

    nanogui::Widget* labelBlock(nullptr);
    labelBlock = new nanogui::Widget(this);
    labelBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                 nanogui::Alignment::Minimum,
                                                 3, 2));
    nanogui::Widget* inputTextBlock(nullptr);
    inputTextBlock = new nanogui::Widget(labelBlock);
    inputTextBlock->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                     nanogui::Alignment::Maximum,
                                                     0, 0));
    this->_inputText = new NanoguiTextField(inputTextBlock);
    this->_inputText->setAlignment(nanogui::TextBox::Alignment::Center);
    this->_inputText->setFontSize(18);
    this->_inputText->setFixedSize(Eigen::Vector2i(100, 20));
    this->_inputText->setPlaceholder("Please provide a path");
    this->_inputText->setCallback([this] (const std::string& path) -> bool {
        this->updateExplorer(path);

        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(path);
        }
        return 0 < path.size();
    });

    nanogui::Widget* blk(nullptr);
    blk = new nanogui::Widget(labelBlock);
    blk->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                          nanogui::Alignment::Maximum,
                                          1, 0));
    this->_browseButton = new nanogui::Button(blk, "", ENTYPO_ICON_FOLDER);
    this->_browseButton->setFixedSize(Eigen::Vector2i(20, 20));
    this->_browseButton->setFontSize(16);

    this->_dropdownContainer = NanoguiAnchored::create(this->_browseButton);
    this->_dropdownContainer->offset = 0u;
    this->_dropdownContainer->placement = static_cast<std::uint8_t>(NanoguiAnchored::ON_LEFT + NanoguiAnchored::ON_BOTTOM);
    this->_dropdownContainer->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                               nanogui::Alignment::Minimum,
                                                               0, 0));

    nanogui::Widget* explorerButtons = new nanogui::Widget(this->_dropdownContainer);
    explorerButtons->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Horizontal,
                                                      nanogui::Alignment::Middle,
                                                      10, 10));

    this->_selectButton = new nanogui::Button(explorerButtons, "Select");
    this->_selectButton->setFixedSize(Eigen::Vector2i(60, 18));
    this->_selectButton->setFontSize(16);
    this->_selectButton->setEnabled(false);
	this->_selectButton->setTextColor(nanogui::Color(255,255,255,255));
	this->_selectButton->setBackgroundColor(nanogui::Color(48, 87, 139, 255));
    this->_selectButton->setCallback([this] () {
        this->_inputText->setValue(this->_selected);
        this->updateExplorer(this->_selected);
        if
            (nullptr != this->_changeCallback)
        {
            this->_changeCallback(this->_selected);
        }

        this->hideDropDown();
    });

    this->_parentDirButton = new nanogui::Button(explorerButtons, "", ENTYPO_ICON_ARROW_UP);
    this->_parentDirButton->setFixedSize(Eigen::Vector2i(18, 18));
    this->_parentDirButton->setFontSize(16);
    this->_parentDirButton->setCallback([this] () {
        std::string delimiter(1, Path::separator());
        std::size_t lastSeparatorPosition;

        if
            (  (this->_parentPath.size() > delimiter.size())
            && (0 == this->_parentPath.compare(this->_parentPath.size() - delimiter.size(), delimiter.size(), delimiter)))
        {
            lastSeparatorPosition = this->_parentPath.substr(0, this->_parentPath.size() - 1).rfind(delimiter);
        }
        else
        {
            lastSeparatorPosition = this->_parentPath.rfind(delimiter);
        }
        this->_parentPath = this->_parentPath.substr(0, lastSeparatorPosition + 1);
        this->updateExplorer(this->_parentPath);
    });

    this->_itemsList = new NanoguiList(this->_dropdownContainer);
    this->_itemsList->content->setLayout(new nanogui::BoxLayout(nanogui::Orientation::Vertical,
                                                                nanogui::Alignment::Fill,
                                                                1, 1));
    this->_itemsList->content->setBackgroundFillColor(nanogui::Color(80, 80, 80, 255));
    this->_browseButton->setCallback([this]() {
        std::string fullPath;
        this->getFullPath(this->_inputText->value(), fullPath);
        this->_inputText->setValue(fullPath);

        this->toggleDropDown();
    });

    this->updateExplorer(this->_selected);
    this->hideDropDown();
}

/**
 *
 */
NanoguiPathSelector::~NanoguiPathSelector()
{

}

/**
 *
 */
void NanoguiPathSelector::enable()
{
    this->setEnabled(true);
    this->_browseButton->setEnabled(true);
    this->_inputText->setEnabled(true);
}

/**
 *
 */
void NanoguiPathSelector::disable()
{
    this->setEnabled(false);
    this->_browseButton->setEnabled(false);
    this->_inputText->setEnabled(false);
}

/**
 *
 */
void NanoguiPathSelector::show()
{
    this->setVisible(true);
}

/**
 *
 */
void NanoguiPathSelector::hide()
{
    this->hideDropDown();
    this->setVisible(false);
}

/**
 *
 */
const std::string& NanoguiPathSelector::getPath() const
{
    return this->_inputText->value();
}


/**
 *
 */
bool NanoguiPathSelector::setPath(const std::string& path)
{
    if
        (path == this->_inputText->value())
    {
        return false;
    }

    this->_inputText->setValue(path);

    if
        (nullptr != this->_changeCallback)
    {
        this->_changeCallback(path);
    }
    return true;
}

/**
 *
 */
bool NanoguiPathSelector::setChangeCallback(const std::function<void(const std::string&)>& callback)
{
    if
        (nullptr != this->_changeCallback)
    {
        return false;
    }

    this->_changeCallback = callback;
    return true;
}

/**
 *
 */
void NanoguiPathSelector::resize(const Eigen::Vector2i& size)
{
    this->setFixedSize(size);

    this->_inputText->setFixedWidth(size.x() - 26);
    this->_itemsList->setFixedSize(Eigen::Vector2i(size.x() - 26, 180));
}


/**
 *
 */
bool NanoguiPathSelector::setDropDownCallback(const std::function<void()>& callback)
{
    if
        (nullptr != this->_dropdownCallback)
    {
        return false;
    }

    this->_dropdownCallback = callback;
    return true;
}

/**
 *
 */
void NanoguiPathSelector::toggleDropDown()
{
    if
        (this->_dropdownContainer->visible())
    {
        this->hideDropDown();
    }
    else
    {
        this->showDropDown();
    }
}

/**
 *
 */
void NanoguiPathSelector::showDropDown()
{
    this->_dropdownContainer->setVisible(true);
    if
        (nullptr != this->_dropdownCallback)
    {
        this->_dropdownCallback();
    }
}

/**
 *
 */
void NanoguiPathSelector::hideDropDown()
{
    this->_dropdownContainer->setVisible(false);
}

/**
 *
 */
void NanoguiPathSelector::getFullPath(const std::string& path, std::string& fullPath)
{
    std::string parentPath("");
    std::string separator(1, Path::separator());
    bool relative(false);
    std::size_t lastSeparatorPosition = path.rfind(separator);

    relative = (std::string::npos == lastSeparatorPosition);
    if
        (true == relative)
    {
        // relative path
        #ifdef __APPLE__
        Path::homeDir(parentPath);
        #else
        Path::currentWorkingDir(parentPath);
        #endif
        fullPath = parentPath + separator + path;
    }
    else
    {
        fullPath = path;
    }
}

/**
 *
 */
void NanoguiPathSelector::updateExplorer(const std::string& path)
{
    std::string separator(1, Path::separator());
    std::string fullPath("");
    bool relative(false);
    std::forward_list<std::string> dirContent;
    std::size_t lastSeparatorPosition = path.rfind(separator);

    relative = (std::string::npos == lastSeparatorPosition);
    if
        (true == relative)
    {
        // relative path
        #ifdef __APPLE__
        Path::homeDir(this->_parentPath);
        #else
        Path::currentWorkingDir(this->_parentPath);
        #endif
        this->_parentPath = this->_parentPath + separator;
        fullPath = this->_parentPath + path;
    }
    else
    {
        this->_parentPath = path.substr(0, lastSeparatorPosition + 1);
        fullPath = path;
    }

    if
        (  (true == Path::dirExists(path))
        && (path.size() >= separator.size())
        && (0 == path.compare(path.size() - separator.size(), separator.size(), separator)))
    {
        this->_parentPath = path;
    }

    NanoguiClickableLabel* label(nullptr);
    std::size_t availableWidgets = this->_widgets.size();

    // Show content of parent directory
    Path::listFiles(this->_parentPath, dirContent);
    dirContent.sort();
    std::size_t count = 0;
    for
        (const auto& item: dirContent)
    {
        if
            (count < availableWidgets)
        {
            label = this->_widgets.at(count);
        }
        else
        {
            label = new NanoguiClickableLabel(this->_itemsList->content);
            label->setChangeCallback([this] (const std::string& value, bool selected, float duration) {
                if
                    (true == selected)
                {
                    // Find the widgets that corresponds to this value (if any)
                    for
                        (auto widget : this->_widgets)
                    {
                        widget->setSelected((widget->caption() == value), true /*silenced*/);
                    }

                    Path::canonicalPath(this->_parentPath + value, this->_selected);
                }

                if
                    ((false == selected) && (500000000.f > duration))
                {
                    std::string canonicalPath;
                    Path::canonicalPath(this->_parentPath + value, canonicalPath);

                    // Consider this is a double-click
                    if
                        (Path::dirExists(this->_parentPath + value))
                    {
                        std::string delimiter(1, Path::separator());
                        this->_selected = canonicalPath + delimiter;
                        this->_updateRequested = true;
                    }
                    else
                    {
                        this->_selected = canonicalPath;
                        this->hideDropDown();
                    }

                    this->_inputText->setValue(this->_selected);
                    if
                        (nullptr != this->_changeCallback)
                    {
                        this->_changeCallback(this->_selected);
                    }
                }

                this->_selectButton->setEnabled(selected);
            });
            this->_widgets.push_back(label);
        }

        label->setCaption(item.substr(item.rfind(separator) + 1));
        label->setSelected((item == fullPath), true);
        label->setVisible(true);
        label->setEnabled(true);

        count = count + 1;

        if
            (Path::dirExists(item))
        {
            // Assign directory widget
            label->setIcon(ENTYPO_ICON_FOLDER);
        }
        else
        {
            // Assign file widget
            label->setIcon(ENTYPO_ICON_TEXT_DOCUMENT);
        }
    }

    if
        (0 < count)
    {
        int newHeight = static_cast<int>(count) * label->preferredSize(this->screen()->nvgContext()).y() + 2 /*margin*/ + (static_cast<int>(count) - 1) * 1 /*inter spacing*/;
        this->_itemsList->content->setFixedHeight(newHeight);
        this->_itemsList->content->setHeight(newHeight);
    }
    else
    {
        this->_itemsList->content->setFixedHeight(0);
    }

    // Hide all unused widgets
    while
        (count < this->_widgets.size())
    {
        label = this->_widgets.at(count);
        label->setCaption("");
        label->setSelected(false, true);    // unselect by default
        label->setVisible(false);
        label->setEnabled(false);
        count = count + 1;
    }

    this->_itemsList->update(true /*reset scrolling*/);
}

/**
 *
 */
void NanoguiPathSelector::draw(NVGcontext* ctx)
{
    if
        (true == this->_updateRequested)
    {
        this->updateExplorer(this->_inputText->value());
        this->_updateRequested = false;
    }

    nanogui::Widget::draw(ctx);
}

/**
 *
 */
void NanoguiPathSelector::setPlaceholder(const std::string& msg)
{
    this->_inputText->setPlaceholder(msg);
}
