// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>
#include <thread>
#include <chrono>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include "core/log.h"
#include "platform/sockets_monitor_impl.h"

#include "core/perfs.h"
#include "core/sequential_buffer.h"

using namespace SEAScope;

/**
 *
 */
SocketsMonitorImpl::SocketsMonitorImpl(): signalHandler(nullptr), terminationRequested(false), epfd(-1), read_pipe(-1), write_pipe(-1)
{
};

/**
 *
 */
SocketsMonitorImpl::~SocketsMonitorImpl()
{
};

/**
 *
 */
bool SocketsMonitorImpl::handleEvent(const SocketJob& event)
{
	Log::log(LogLevel::Debug, "Unix sockets monitor", "handle event");
	if
		(event.terminate)
	{
		this->terminationRequested = true;
	}
	else if
		(event.subscribe)
	{
		return this->subscribe(event.fileDescriptor, event.handler, event.handlerFactory);
	}
	else
	{
		// Unsubscribe
		this->unsubscribe(event.fileDescriptor);
	}
	return false;
};

/**
 *
 */
bool SocketsMonitorImpl::subscribe(socket_t fd, IProtocolHandler* handler, const ProtocolHandlerFactory& handlerFactory)
{
	SequentialBuffer<SOCKET_CHUNKSIZE>* buffer;	//
	UnixMonitoredSocket monitorEntry;	//
	bool ok;	//
	int rc;	// Return code

	Log::log(LogLevel::Debug, "Unix sockets monitor", "Subscribe {}", fd);
	Log::log(LogLevel::Debug, "Unix sockets monitor", "with handler? {}", nullptr != handler);
	Log::log(LogLevel::Debug, "Unix sockets monitor", "with handler factory? {}", nullptr != handlerFactory);
	Log::log(LogLevel::Debug, "Unix sockets monitor", "available buffers: {}", this->bufferPool.available.size());
	auto found = this->monitored.find(fd);
	if
		(this->monitored.end() != found)
	{
		Log::log( LogLevel::Error, "Unix sockets monitor"
				, "File descriptor {} is already monitored"
				, fd);
		return false;
	}

	ok = false;
	buffer = this->bufferPool.acquire(ok);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Unix sockets monitor"
				, "No buffer available in the pool!");
		return false;
	}

	monitorEntry.ioHandler = handler;
	monitorEntry.clientHandler = handlerFactory;
	monitorEntry.buffer = buffer;
	monitorEntry.epollEvent.data.fd = fd;
	monitorEntry.epollEvent.events = EPOLLIN|EPOLLET;

	rc = epoll_ctl( this->epfd
				  , EPOLL_CTL_ADD
				  , fd
				  , &(monitorEntry.epollEvent));
	if
		(-1 == rc)
	{
		// release buffer
		this->bufferPool.release(buffer);

		return false;
	}

	this->monitored[fd] = monitorEntry;
	return true;
};

/**
 *
 */
bool SocketsMonitorImpl::unsubscribe(socket_t fd)
{
	bool ok;	//
	int rc;	//

	Log::log(LogLevel::Debug, "Unix sockets monitor", "Unsubscribe {}", fd);
	auto found = this->monitored.find(fd);
	if
		(this->monitored.end() == found)
	{
		return false;
	}

	/* Remove fd from epoll watchlist */
	rc = epoll_ctl( this->epfd
				  , EPOLL_CTL_DEL
				  , fd
				  , &(found->second.epollEvent));
	if
		(-1 == rc)
	{
		Log::log( LogLevel::Error, "Unix sockets monitor"
				, "Could not remove epoll event from the watchlist");
		return false;
	}

	/* Release buffer */
	ok = this->bufferPool.release(found->second.buffer);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "Unix sockets monitor"
				, "Could not release buffer");
		return false;
	}

	this->monitored.erase(found);
	return true;
};

/**
 *
 */
bool SocketsMonitorImpl::init(const std::function<void()>& signalHandlerIn, int& eventWakeFd)
{
	this->signalHandler = signalHandlerIn;

	// Create pipe that will wake from epoll_wait
    int pipefds[2] = {};
    pipe(pipefds);
    this->read_pipe = pipefds[0];
    this->write_pipe = pipefds[1];
	eventWakeFd = this->write_pipe;

    // make read-end non-blocking
    int flags = fcntl(this->read_pipe, F_GETFL, 0);
    fcntl(this->write_pipe, F_SETFL, flags|O_NONBLOCK);

	return true;
};

/**
 *
 */
bool SocketsMonitorImpl::monitorSockets()
{
	int i;	//
	int n;	//
	int rc;	//
	std::size_t maxEvents = 64;	//
	bool ok;	//

	struct epoll_event event, current_event;
	// Buffer where events are returned.
	struct epoll_event* events = static_cast<epoll_event*>(calloc(maxEvents, sizeof event));
	this->epfd = epoll_create1(0);

    // add the read end to the epoll
    epoll_event ev = {};
    ev.events = EPOLLIN;
    ev.data.fd = this->read_pipe;
    epoll_ctl(this->epfd, EPOLL_CTL_ADD, this->read_pipe, &ev);


	while
		(true)
	{
		n = epoll_wait(this->epfd, events, static_cast<int>(maxEvents), -1);
		Log::log(LogLevel::Debug, "Unix sockets monitor", "echoing... {}", n);
		for
			(i = 0; i < n; ++i)
		{
			current_event = events[i];

			if ((current_event.events & EPOLLERR) ||
				(current_event.events & EPOLLHUP) ||
				(!(current_event.events & EPOLLIN)))
			{
				// Close fd because there was an error or socket was not ready
				// for reading
				Log::log( LogLevel::Error, "Unix sockets monitor"
						, "An error occured on the socket. Closing the {} file descriptor..."
						, current_event.data.fd);
				this->unsubscribe(current_event.data.fd);
				close(current_event.data.fd);
			}
			else if
				(current_event.events & EPOLLRDHUP)
			{
				Log::log( LogLevel::Warn, "Unix sockets monitor"
						, "Connection reset by peer. Closing the {} file descriptor."
						, current_event.data.fd);
				this->unsubscribe(current_event.data.fd);
				close(current_event.data.fd);
			}
			else if (this->read_pipe == current_event.data.fd)
			{
				std::uint8_t b;
				read(current_event.data.fd, &b, 1);
				// Handle events sent by sockets manager
				if
					(nullptr != this->signalHandler)
				{
					this->signalHandler();
				}
			}
			else
			{
				Log::log(LogLevel::Debug, "Unix sockets monitor", "handling another socket {}", current_event.data.fd);
				auto found = this->monitored.find(current_event.data.fd);
				if
					(this->monitored.end() == found)
				{
					Log::log(LogLevel::Warn, "Unix sockets monitor"
							, "Reacted on an unmonitored socket ({})."
							, current_event.data.fd);
					continue;
				}
				if
					(nullptr != found->second.ioHandler)
				{
					std::uint64_t left;
					std::uint8_t* res = found->second.buffer->getPointer(left);
					std::uint64_t toRead = (SOCKET_CHUNKSIZE > left) ? left:SOCKET_CHUNKSIZE;
					std::int64_t loaded = read(current_event.data.fd, res, toRead);

					std::uint64_t consumed;
					bool msgComplete;

					if
						(0 >= loaded)
					{
						Log::log(LogLevel::Debug, "Unix sockets monitor",
								"Nothing received from client, closing connection");
						this->unsubscribe(current_event.data.fd);
						close(current_event.data.fd);
						continue;
					}

					while
						(0 < loaded)
					{
						Log::log( LogLevel::Debug, "Unix sockets monitor"
								, "Read data from fd {}: {} {} => {}"
								, current_event.data.fd, toRead, loaded, res);

						std::uint64_t cumulatedRead;	//
						cumulatedRead = 0;
						msgComplete = true;
						while
							(cumulatedRead < static_cast<unsigned long>(loaded))
						{
							msgComplete = false;
							msgComplete = found->second.ioHandler->parseStream(current_event.data.fd, res, cumulatedRead, static_cast<unsigned long>(loaded) - cumulatedRead, consumed);
							cumulatedRead += consumed;
							if
								(true == msgComplete)
							{
								Log::log( LogLevel::Debug, "Unix sockets monitor"
										, "Message complete");
							}
							else
							{
								Log::log( LogLevel::Debug, "Unix sockets monitor"
										, "Message still being read");
							}
						}

						left = 0;
						res = found->second.buffer->getPointer(left);
						toRead = (SOCKET_CHUNKSIZE > left) ? left:SOCKET_CHUNKSIZE;
						loaded = read(current_event.data.fd, res, toRead);

						if
							((true == msgComplete) && (0 == loaded))
						{
							Log::log(LogLevel::Debug, "Unix sockets monitor",
                                     "Closing connection after receiving messages");
							this->unsubscribe(current_event.data.fd);
						}
						/*
						if
							(0 > loaded)
						{
							// TODO: try to implement this in a better way
							// because a negative return code from read may
							// indicate a real error

							Log::log(LogLevel::Debug, "UnixSocketsMonitor", "Nothing more to load");
							loaded = 0;
							break;
						}
						*/
					}
				}
				else if
					(nullptr != found->second.clientHandler)
				{
					Log::log( LogLevel::Debug, "Unix sockets monitor"
							, "Should create a new socket !!!");
					while
						(true)
					{
						struct sockaddr client_addr;
						socklen_t client_len;
						socket_t clientFd;
						client_len = sizeof client_addr;
						IProtocolHandler* pHandler;

						// accept4 makes clientFd non-blocking
						clientFd = accept4( current_event.data.fd, &client_addr
										  , &client_len
										  , SOCK_NONBLOCK | SOCK_CLOEXEC);
						if
							(-1 == clientFd)
						{
							if
								((EAGAIN == errno) || (EWOULDBLOCK == errno))
							{
								break;  // We have processed all incoming connections.
							}
							else
							{
								perror("accept");
								break;
							}
						}

						pHandler = found->second.clientHandler(ok);
						if
							(false == ok)
						{
							Log::log( LogLevel::Error, "Unix sockets monitor"
									, "Could not get protocol handler for the client connection");
							rc = close(clientFd);
							if
								(-1 == rc)
							{
								Log::log( LogLevel::Error, "Unix socket Monitor"
										, "Could not close client socket");
							}
							continue;
						}

						// Monitor communication from client
						this->subscribe(clientFd, pHandler, nullptr);
					} // All incoming connections handled
				} // accept client connection
			}
		}

		if
			(this->terminationRequested)
		{
			Log::log(LogLevel::Debug, "Unix sockets monitor", "Termination requested, stopping sockets monitoring");
			break;
		}
	}

	for
		(auto& it : this->monitored)
	{
		close(it.first);
	}
	this->monitored.clear();
	close(this->read_pipe);
	close(this->write_pipe);
    close(this->epfd);

	free(events);

	return false;
};

/**
 *
 */
bool SocketsMonitorImpl::triggerEventsProcessing()
{
	std::uint8_t b = '0';
	Log::log(LogLevel::Debug, "Unix sockets monitor", "Trigger events processing");

	// Just generate some activity on the file descriptor for epoll to wake up
	write(this->write_pipe, &b, 1);
	return true;
};

/**
 *
 */
bool SocketsMonitorImpl::sendData(socket_t sock, std::uint8_t* buffer, std::uint64_t dataSize)
{
	std::uint64_t remaining;	//
	std::int64_t sent;	//

	// Send data
	remaining = dataSize;
	while
		(0 < remaining)
	{
		sent = send(sock, buffer, remaining, MSG_NOSIGNAL);
		if
			(-1 == sent)
		{
			if
				((errno == EWOULDBLOCK) || (errno == EAGAIN))
			{
				continue;
			}

			/* Something wrong happened */
			break;
		}

		/* Prepare next write */
		remaining -= static_cast<std::uint64_t>(sent);
		buffer += sent;
	}

	return 0 == remaining;
};
