// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "platform/path_impl.h"

#include <sstream>
#include <Shlwapi.h>
#include <windows.h>
#include <strsafe.h>
#include <iostream>
#include <shlobj_core.h>

using namespace SEAScope;

time_t filetime_to_timet(FILETIME const& ft)
{
	ULARGE_INTEGER ull;
	ull.LowPart = ft.dwLowDateTime;
	ull.HighPart = ft.dwHighDateTime;
	return static_cast<time_t>(ull.QuadPart / 10000000ULL - 11644473600ULL);
}

void PrintLastError()
{
	// Retrieve the system error message for the last-error code

	LPVOID lpMsgBuf;
	DWORD dw = GetLastError();

	FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER |
		FORMAT_MESSAGE_FROM_SYSTEM |
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dw,
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPTSTR)&lpMsgBuf,
		0, NULL);

	// Display the error message and exit the process

	std::cout << (const char*)lpMsgBuf << std::endl;

	LocalFree(lpMsgBuf);
}

std::wstring toWStringUnlockMaxPath(const std::string& str)
{
	size_t numCharConverted = 0;
	wchar_t* dstStr = new wchar_t[str.size() + 10];

	mbstowcs_s(&numCharConverted, dstStr, str.size() + 10, str.c_str(), _TRUNCATE);

	std::wstring wstr = dstStr;

	delete[] dstStr;

	return wstr;
}

std::wstring toWString(const std::string& str)
{
	size_t numCharConverted = 0;
	wchar_t* dstStr = new wchar_t[str.size() + 10];

	mbstowcs_s(&numCharConverted, dstStr, str.size() + 10, str.c_str(), _TRUNCATE);

	std::wstring wstr = dstStr;

	delete[] dstStr;

	return wstr;
}

std::string toString(const std::wstring& wstr)
{
	size_t numCharConverted = 0;
	char* dstStr = new char[wstr.size() + 10];

	wcstombs_s(&numCharConverted, dstStr, wstr.size() + 10, wstr.c_str(), _TRUNCATE);

	std::string str = dstStr;

	delete[] dstStr;

	return str;
}

/**
*
*/
PathImpl::PathImpl()
{
};

/**
*
*/
PathImpl::~PathImpl()
{
};

/**
* Check if a file exists.
*/
bool PathImpl::fileExists(const std::string& path)
{
	HANDLE hFile;
	bool ret = true;

	std::wstring wstr = toWStringUnlockMaxPath(path);

	hFile = CreateFileW(wstr.c_str(),               // file to open
		GENERIC_READ,          // open for reading
		FILE_SHARE_READ,       // share for reading
		NULL,                  // default security
		OPEN_EXISTING,         // existing file only
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED, // normal file
		NULL);                 // no attr. template

	if (hFile == INVALID_HANDLE_VALUE) // file doesn't exist
	{
		ret = false;
	}
	else
	{
		// file exists
		CloseHandle(hFile);
	}

	return ret;
};

/**
* Check that it is possible to create or modify the file
*/
bool PathImpl::writeFile(const std::string& path)
{
	HANDLE hFile;
	bool ret = true;

	std::wstring wstr = toWStringUnlockMaxPath(path);

	hFile = CreateFileW(wstr.c_str(),   // file to open
		GENERIC_WRITE,					// open for reading
		0,								// no share 
		NULL,							// default security
		OPEN_ALWAYS,     // existing file or new one
		FILE_ATTRIBUTE_NORMAL,			// normal file
		NULL);							// no attr. template

	if (hFile == INVALID_HANDLE_VALUE)  // Unable to open or create the file
	{
		ret = false;
	}
	else
	{
		CloseHandle(hFile);
	}

	return ret;
};

/**
*
*/
bool PathImpl::fileSize(const std::string& path, std::int64_t& size)
{
	HANDLE hFile;
	LARGE_INTEGER fileSize;
	bool ret = true;

	std::wstring wstr = toWStringUnlockMaxPath(path);

	hFile = CreateFileW(wstr.c_str(),					// file to open
		GENERIC_READ,									// open for reading
		FILE_SHARE_READ,								// share for reading
		NULL,											// default security
		OPEN_EXISTING,									// existing file only
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,	// normal file
		NULL);											// no attr. template

	if (hFile == INVALID_HANDLE_VALUE) // file doesn't exist
	{
		ret = false;
	}
	else
	{
		BOOL fileSizeRet = GetFileSizeEx(hFile, &fileSize);

		if (fileSizeRet == TRUE)
		{
			size = fileSize.QuadPart;
		}
		else
		{
			ret = false;
		}

		CloseHandle(hFile);
	}

	return ret;
};

/**
* Check if a directory exists.
*/
bool PathImpl::dirExists(const std::string& path)
{
	std::wstring wstr = toWStringUnlockMaxPath(path);

	DWORD dw = GetFileAttributesW(wstr.c_str());

	if (INVALID_FILE_ATTRIBUTES != dw &&
		dw & FILE_ATTRIBUTE_DIRECTORY)
	{
		return true;
	}

	return false;
}

/**
* Create directory
*/
bool PathImpl::createDir(const std::string& path, mode_t)
{
	std::wstring wstr = toWStringUnlockMaxPath(path);

	BOOL dw = CreateDirectoryW(wstr.c_str(), nullptr);

	if (dw == 0)
	{
		return false;
	}

	return true;
};


/**
* Create directory tree.
*/
bool PathImpl::createDirs(const std::string& path, mode_t mode)
{
	char* searchStart;
	char* slashPointer;
	char* cpath;
	bool ok;

	cpath = const_cast<char*>(path.c_str());
	ok = true;
	searchStart = cpath;
	slashPointer = std::strchr(searchStart, '/');
	while
		(true == ok
			&& 0 != slashPointer)
	{
		if (slashPointer != searchStart)
		{
			// Neither root nor double slash in path
			*slashPointer = '\0'; /* artificially truncate the path at first slash found */
			ok = this->createDir(cpath, mode);
			*slashPointer = '/'; /* restore truncated path */
		}
		searchStart = slashPointer + 1;
		slashPointer = std::strchr(searchStart, '/');
	}
	if
		(true == ok)
	{
		ok = this->createDir(path, mode);
	}
	return ok;
};

/**
*
*/
void PathImpl::splitPaths(const std::string& paths, const std::string& delim, std::forward_list<std::string>& result)
{
	auto start = (size_t)0;
	auto end = paths.find(delim);
	auto it = result.before_begin();

	while
		(end != std::string::npos)
	{
		it = result.insert_after(it, paths.substr(start, end - start));
		start = end + delim.length();
		end = paths.find(delim, start);
	}
	result.insert_after(it, paths.substr(start, end));
};

void SEAScope::PathImpl::joinPaths(const std::vector<std::string>& paths, std::string& result)
{
	std::ostringstream buffer;	//

	for
		(const auto& path : paths)
	{
		buffer << path << std::string("\\");
	}
	result = buffer.str();
	result.pop_back(); // Remove trailing slash
}

char SEAScope::PathImpl::separator() const
{
	return '\\';
}

bool PathImpl::canonicalPath(const std::string& inputPath, std::string& outputPath)
{
	std::string delimiter(1, this->separator());
	std::forward_list<std::string> pathElements;
	std::uint16_t ignoredCount(0u);

	outputPath = "";

	this->splitPaths(inputPath, delimiter, pathElements);
	pathElements.reverse();
	for
		(const auto& pathElement : pathElements)
	{
		if
			(".." == pathElement)
		{
			ignoredCount = ignoredCount + 1u;
		}
		else if
			(0 < ignoredCount)
		{
			ignoredCount = ignoredCount - 1u;
		}
		else if
			("." == pathElement)
		{
			// noop
		}
		else if
			(0 < outputPath.size())
		{
			outputPath = pathElement + delimiter + outputPath;
		}
		else
		{
			outputPath = pathElement;
		}
	}

	return true;
}

bool PathImpl::currentWorkingDir(std::string& path)
{
	// Get path length
	DWORD len = GetCurrentDirectory(0, nullptr);

	// Resize string buffer
	path.resize(len);

	// Get the path directly in the string buffer
	len = GetCurrentDirectory(len, const_cast<LPSTR>(path.data()));

	return true;
}

/**
*
*/
bool PathImpl::listFiles(const std::string& path, std::forward_list<std::string>& result)
{
	HANDLE hFind = INVALID_HANDLE_VALUE;
	bool ret = true;
	WIN32_FIND_DATAW ffd;
	DWORD dwError;

	if (true == dirExists(path))
	{
		std::wstring wpath = toWStringUnlockMaxPath(path) + std::wstring(L"\\*");
		std::wstring wpath2 = toWString(path);

		hFind = FindFirstFileW(wpath.c_str(), &ffd);

		if (INVALID_HANDLE_VALUE == hFind)
		{
			return false;
		}

		std::wstring str;

		do
		{
			str = ffd.cFileName;

			if (str != L".." && str != L".")
			{
				std::wstring wpathfull = wpath2 + L"\\" + ffd.cFileName;

				result.push_front(toString(wpathfull));

			}
		} while (FindNextFileW(hFind, &ffd) != 0);

		dwError = GetLastError();

		if (dwError != ERROR_NO_MORE_FILES)
		{
			ret = false;
		}

		FindClose(hFind);
	}

	return ret;
};

/**
*
*/
bool PathImpl::listFilesRecursive(const std::string& path, std::forward_list<std::string>& result)
{
	bool ok;
	std::forward_list<std::string> tmpList;

	ok = listFiles(path, tmpList);

	if (true == ok)
	{
		for (auto& str : tmpList)
		{
			if (true == PathImpl::dirExists(str))
			{
				PathImpl::listFilesRecursive(str, result);
			}
			else if (true == PathImpl::fileExists(str))
			{
				result.push_front(str);
			}
		}

	}

	return true;
};

/**
*
*/
bool PathImpl::deleteFile(const std::string& path)
{
	bool ok;	//
	int rc;	// Return code

	ok = this->fileExists(path);
	if
		(false == ok)
	{
		return false;
	}

	rc = remove(path.c_str());
	ok &= (0 == rc);
	return ok;
}

/**
 *
 */
void PathImpl::fixPlatformPath(std::string& str)
{
	size_t pos = 0;
	while (pos != std::string::npos)
	{
		pos = str.find('/', pos);

		if (pos != std::string::npos)
		{
			str[pos] = '\\';
		}
	}

	DWORD len = GetFullPathName(str.c_str(), 0, nullptr, nullptr);

	char* buf = new char[len + 1];

	GetFullPathName(str.c_str(), len + 1, buf, nullptr);

	str = std::string("\\\\?\\") + buf;

	delete[] buf;
}

/**
 *
 */
bool PathImpl::lastModified(const std::string& path, std::time_t& result)
{
	FILETIME last;
	HANDLE hFile;
	bool ret = false;

	hFile = CreateFile(
		path.c_str(),          // file to open
		GENERIC_READ,          // open for reading
		FILE_SHARE_READ,       // share for reading
		NULL,                  // default security
		OPEN_EXISTING,         // existing file only
		FILE_FLAG_BACKUP_SEMANTICS, // file or directory
		NULL);                 // no attr. template

	if (INVALID_HANDLE_VALUE != hFile)
	{
		BOOL res = GetFileTime(hFile, NULL, NULL, &last);

		if (TRUE == res)
		{
			result = filetime_to_timet(last);

			CloseHandle(hFile);

			ret = true;
		}
	}

	return ret;
}

/**
 *
 */
bool PathImpl::homeDir(std::string& outputPath)
{
	PWSTR p = NULL;

	HRESULT hRes = SHGetKnownFolderPath(FOLDERID_Profile, 0, nullptr, &p);

	if
		(S_OK != hRes)
	{
		CoTaskMemFree(p);
		return false;
	}

	outputPath = toString(std::wstring(p));
	CoTaskMemFree(p);

	return true;
}
