#!/bin/bash

# =============================================================================
# @author: <sylvain.herledan@oceandatlab.com>
# @date: 2016-11-16
#
# Build functions to generate the static libraries required to build
# the SEAScope client.
# =============================================================================

set -o nounset
set -o errexit
set -o pipefail

build_zlib()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="http://zlib.net/current/zlib-1.2.12.tar.gz";

	# Download source if needed
	if [ ! -f "${src_dir}/zlib-1.2.12.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/zlib-1.2.12.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/zlib-1.2.12" ];
	then
		tar xzf "${src_dir}/zlib-1.2.12.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libz.a" ];
	then
		cd "${build_dir}/zlib-1.2.12"
		./configure --prefix="${rootfs}" \
                --libdir="${rootfs}/lib" \
					      --static \
					      --64
		make -j ${CPUs}
		make install
	fi
	
	# Back to root	
	cd "${root_dir}"
}

build_libpng()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="http://downloads.sourceforge.net/sourceforge/libpng/libpng-1.6.24.tar.xz";

	# Download source if needed
	if [ ! -f "${src_dir}/libpng-1.6.24.tar.xz" ];
	then
		wget "${uri}" -O "${src_dir}/libpng-1.6.24.tar.xz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/libpng-1.6.24" ];
	then
		tar xJf "${src_dir}/libpng-1.6.24.tar.xz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libpng16.a" ];
	then
    cd "${build_dir}/libpng-1.6.24"
    LDFLAGS="-L${rootfs}/lib ${LDFLAGS}" \
    CPPFLAGS="-I${rootfs}/include ${CPPFLAGS}" \
    ./configure --prefix="${rootfs}" \
                --libdir="${rootfs}/lib" \
                --with-sysroot="${rootfs}" \
                --disable-shared
    make -j${CPUs}
    make install
  fi
	
	# Back to root	
	cd "${root_dir}"
}


build_sqlite3()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://sqlite.org/2016/sqlite-src-3110000.zip";

	# Download source if needed
	if [ ! -f "${src_dir}/sqlite-src-3110000.zip" ];
	then
		wget "${uri}" -O "${src_dir}/sqlite-src-3110000.zip"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/sqlite-src-3110000" ];
	then
		unzip "${src_dir}/sqlite-src-3110000.zip" -d "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libsqlite3.a" ];
	then
		cd "${build_dir}/sqlite-src-3110000"

        set +o nounset
        local flags="${CPPFLAGS} -DSQLITE_ENABLE_COLUMN_METADATA=1 \
								 -DSQLITE_ENABLE_UNLOCK_NOTIFY \
								 -DSQLITE_ENABLE_DBSTAT_VTAB=1 \
								 -DSQLITE_SECURE_DELETE"
        set -o nounset

	    rm -f sqlite3.h
        CPPFLAGS="${flags}" \
        ./configure --prefix="${rootfs}" \
                    --libdir="${rootfs}/lib" \
                    --disable-amalgamation \
                    --disable-tcl \
                    --disable-readline \
                    --enable-fts3 \
                    --enable-fts4 \
                    --enable-fts5 \
                    --enable-rtree \
                    --enable-json1
		make # json extension breaks parallel build
		make install
	fi
	
	# Back to root	
	cd "${root_dir}"
}

build_glfw()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://github.com/glfw/glfw.git";

	# Download source if needed
	if [ ! -d "${src_dir}/glfw" ];
	then
		git clone "${uri}" "${src_dir}/glfw"

		cd "${src_dir}/glfw"
		git checkout c32dc3a0856e339d8de582582fcbedec430e14b4
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libglfw3.a" ];
	then
		cd "${src_dir}/glfw"
	    cmake -DCMAKE_INSTALL_PREFIX="${rootfs}" \
            -DCMAKE_INSTALL_LIBDIR="${rootfs}/lib" \
            -DBUILD_SHARED_LIBS=OFF \
            .
		make -j${CPUs}
		make install
	fi
	
	# Back to root	
	cd "${root_dir}"
}

build_ossp_uuid()
{
	local root_dir;
	local src_dir;
	local build_dir;
	local rootfs;
	local uri;

  root_dir="${1}"; shift;
  src_dir="${root_dir}/sources";
  build_dir="${root_dir}/build";
  rootfs="${root_dir}/rootfs";
  uri="https://mirrors.ocf.berkeley.edu/debian/pool/main/o/ossp-uuid/ossp-uuid_1.6.2.orig.tar.gz";

	# Download source if needed
	if [ ! -f "${src_dir}/uuid-1.6.2.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/uuid-1.6.2.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/uuid-1.6.2" ];
	then
		tar xzf "${src_dir}/uuid-1.6.2.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libuuid.a" ];
	then
		cd "${build_dir}/uuid-1.6.2"
		./configure --prefix="${rootfs}" \
					--libdir="${rootfs}/lib" \
					--without-perl \
					--without-php \
					--without-pgsql
		make -j ${CPUs}
		make install
	fi

	# Back to root
	cd "${root_dir}"
}

