# =============================================================================
# @author: <ziad.khoury.hanna@oceandatalab.com>
# @date: 2019-03-27
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#
# Wrapper that initializes configuration, chooses a version of SEAScope which
# is compatible with the host system and starts the application with the
# options that provide the best performances.
# =============================================================================

property onUnsupportedBundle : "Alert" -- "Alert" | "Ignore"
property onCollectionExists : "Merge" -- "Ask" | "Replace" | "Merge" | "Stop"
property onConfigExists : "Keep both" -- "Ask" | "Replace" | "Keep both" | "Stop"
property onGranuleExists : "Skip" -- "Ask" | "Replace" | "Skip" | "Stop"

global seascopeApp
global successful
global changed
global totalBundlesCount
global lastExecutedScript
global supportsProgress

on run
	try
		set successful to false
		set changed to false
		set totalBundlesCount to 0

		set onUnsupportedBundle to getDefaults of "onUnsupportedBundle" given default:onUnsupportedBundle
		set onCollectionExists to getDefaults of "onCollectionExists" given default:onCollectionExists
		set onConfigExists to getDefaults of "onConfigExists" given default:onConfigExists
		set onGranuleExists to getDefaults of "onGranuleExists" given default:onGranuleExists
		repeat
			set prefs to {¬
				"When dragged-on item is unsupported:" & tab & tab & tab & "-> " & onUnsupportedBundle,¬
				"When a collection being imported already exists:" & tab & "-> " & onCollectionExists,¬
				"When there's a confilict between a collection's config and the one being imported:" & tab & tab & tab & tab & tab & "-> " & onConfigExists,¬
				"When a granule being imported already exists:" & tab & "-> " & onGranuleExists¬
			}
			display dialog¬
				"This droplet will import data bundles dragged onto its icon into SEAScope's workspace." & linefeed &¬
				linefeed &¬
				"There are a user-settable preferences for the actions to take:" & return & "• " & joinList(prefs, return & "• ")¬
				buttons {"Change Preferences", "Done"}¬
				default button "Done"¬
				with title "Import to SEAScope Droplet"

			if the button returned of the result is "Change Preferences" then
				set title to "Import to SEAScope Droplet Preferences"

				choose from list prefs¬
					with title title¬
					with prompt "Pick the preference you wish to change:"¬
					OK button name "Change"¬
					without multiple selections allowed and empty selection allowed

				set choice to getChoiceFromChooseResult(the result)
				repeat with i from 1 to length of prefs
					if item i of prefs is choice then
						set prompt to questionFromPerfDesction(choice)
						if i is 1 then
							set optionsList to {"Alert", "Ignore"}
							set defaultOption to onUnsupportedBundle
						else if i is 2 then
							set optionsList to {"Ask", "Replace", "Merge", "Stop"}
							set defaultOption to onCollectionExists
						else if i is 3 then
							set optionsList to {"Ask", "Replace", "Keep both", "Stop"}
							set defaultOption to onConfigExists
						else if i is 4 then
							set optionsList to {"Ask", "Replace", "Skip", "Stop"}
							set defaultOption to onGranuleExists
						end if

						choose from list optionsList¬
							with title title¬
							with prompt prompt¬
							default items {defaultOption}¬
							OK button name "Change"¬
							without multiple selections allowed and empty selection allowed

						set option to getChoiceFromChooseResult(the result)
						if option is false then
							exit repeat
						end if

						if i is 1 then
							set onUnsupportedBundle to option
							setDefaults of "onUnsupportedBundle" to option
						else if i is 2 then
							set onCollectionExists to option
							setDefaults of "onCollectionExists" to option
						else if i is 3 then
							set onConfigExists to option
							setDefaults of "onConfigExists" to option
						else if i is 4 then
							set onGranuleExists to option
							setDefaults of "onGranuleExists" to option
						end if

						exit repeat
					end if
				end repeat
			else
				return
			end if
		end repeat
	on error errMsg number errorNumber
		display alert ("An unknown error occurred:  " & errorNumber as text) message errMsg as critical
	end try
end run

on getChoiceFromChooseResult(r)
	if r is false then
		return false
	else
		return item 1 of r
	end if
end getChoiceFromChooseResult

on questionFromPerfDesction(description)
	set UPPERCASE to "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	set LOWERCASE to "abcdefghijklmnopqrstuvwxyz"

	set firstChar to first character of description
	considering case
		set firstCharOffset to offset of firstChar in UPPERCASE
	end considering
	if firstCharOffset > 0 then
		set firstChar to (character firstCharOffset of LOWERCASE)
	end if

	return "What should this droplet do " & firstChar & (text 2 thru ((offset of ":" in description) - 1) of description) & "?"
end questionFromPerfDesction

on getDefaults of key given default:default
	try
		return execution of {"defaults", "read", id of me, key} with output
	on error number 1
		return default
	end try
end getDefaults

on setDefaults of key to value
	execution of {"defaults", "write", id of me, key, value} without output
end setDefaults

-- This droplet processes files dropped onto the applet
on open bundles
	try
		-- on open might be called multiple times before on quit is called
		-- so we check if changed wasn't defined we set it to false
		-- otherwise we leave it untouched
		try
			changed
		on error number -2753
			-- first time on open is called
			set changed to false
			set totalBundlesCount to 0
		end try

		try
			progress total steps
			set supportsProgress to true
		on error
			set supportsProgress to false
		end try

		set successful to true
		set bundlesCount to length of bundles
		set totalBundlesCount to totalBundlesCount + bundlesCount

		if supportsProgress then
			set progress total steps to bundlesCount * 100
			set progress completed steps to 0
			set progress description to "Processing bundles..."
			set progress additional description to "Looking for SEAScope application..."
		end if

		set seascopeApp to getSEAScope()
		set dataPath to getDataPath(POSIX path of seascopeApp)

		repeat with i from 1 to bundlesCount
			set bundleDescription to "bundle " & i & " of " & bundlesCount
			repeat 1 times
				if supportsProgress then
					set progress additional description to "Processing " & bundleDescription
				end if

				set bundle to item i of bundles
				if isFile(bundle) is true then
					try
						if supportsProgress then
							set progress additional description to "Trying to extract " & bundleDescription
							delay 0.1 -- small delay to update the UI with the new progress description
						end if
						set bundle to extractArchive(bundle)
					on error number 1
						if onUnsupportedBundle is "Alert" then
							set the bundle_info to info for bundle
							set bundle_name to the name of the bundle_info
							display alert "Unsupported bundle type" message "The file “" & bundle_name & "” cannot be imported." buttons {"Stop", "Continue"} default button 2 cancel button "Stop" as warning
						end if
						exit repeat
					end try
				end if

				if isFolder(bundle) is true then
					set changed to importCollections(dataPath, bundle, true, bundleDescription) or changed
				end if

				if supportsProgress then set progress completed steps to i * 100
			end repeat
		end repeat

		if supportsProgress then
			-- Reset the progress information
			set progress total steps to 0
			set progress completed steps to 0
			set progress description to ""
			set progress additional description to ""
		end if
	on error errMsg number errorNumber
		set successful to false
		if errorNumber is -128 then
		else if errorNumber is 1 then
			display alert errMsg message lastExecutedScript as critical
		else
			display alert ("An unknown error occurred:  " & errorNumber as text) message errMsg as critical
		end if
		quit
	end try
end open

on quit
	try
		if successful is false then
			continue quit
			return
		end if

		if changed is true then
			set bundlesMessage to totalBundlesCount & " bundle"
			if totalBundlesCount > 1 then
				set bundlesMessage to bundlesMessage & "s"
			end if
			set headerMessage to "Successfully import " & bundlesMessage & " into SEAScope's workspace."
			if application seascopeApp is running then
				display alert¬
					"Import successful" message headerMessage & return & return &¬
					"To see the imported data, you need to restart SEAScope" buttons {"Restart SEAScope", "Dissmis"} default button "Dissmis"
				if the button returned of the result is "Restart SEAScope" then
					try
						tell application seascopeApp to quit
					end try
					delay 1
					try
						tell application seascopeApp to launch
					end try
				end if
			else
				display alert¬
					"Import successful" message headerMessage buttons {"Dissmis"} default button "Dissmis"
			end if
		else
			display alert¬
				"Nothing was imported" message "You probably already had them." buttons {"Dissmis"} default button "Dissmis"
		end if
	on error errMsg number errorNumber
		display alert ("An unknown error occurred:  " & errorNumber as text) message errMsg as critical
	end try

	continue quit -- allows the script to quit
end quit

on importCollections(dataPath, dir, keepOriginals, bundleDescription)
	local changed
	set changed to false

	if supportsProgress then
		set progress additional description to "Importing collections from " & bundleDescription
	end if

	set dir to removeTrailingSlashOfPOSIXPath(POSIX path of dir)
	set configs to find(dir, {"-maxdepth", 4, "-type", "f", "-name", "config.ini"})
	set configsCount to length of configs
	repeat with i from 1 to configsCount
		if supportsProgress then
			set collectionDescription to "collection " & i & " of " & configsCount & " from " & bundleDescription
			set progress additional description to "Importing " & collectionDescription
		end if

		set changed to importCollection(dataPath, item i of configs, keepOriginals) or changed

		if supportsProgress then
			set progress completed steps to progress completed steps + 100 / configsCount
		end if
	end repeat

	return changed
end importCollections

on importCollection(dataPath, collectionConfigPath, keepOriginals)
	local changed

	set collectionPath to getParentDirPath(collectionConfigPath)
	set collectionName to getBaseName(collectionPath)
	set dataCollectionPath to dataPath & "/" & collectionName
	set dataConfigPath to dataCollectionPath & "/config.ini"

	if POSIX file collectionPath is POSIX file dataCollectionPath then
		-- Skip collections in the workspace's data dir
		return false
	end if

	if isFolder(dataCollectionPath) is true then
		if onCollectionExists is "Ask" then
			display dialog "A collection named “" & collectionName & "” already exists in your workspace. Do you want to replace it with the one being imported?" buttons {"Stop", "Replace", "Merge"} default button "Merge" cancel button "Stop"
			set collectionAction to the button returned of the result
		else
			set collectionAction to onCollectionExists
		end if
	else
		set collectionAction to "Replace"
	end if

	if collectionAction is "Stop" then
		error -128
	else if collectionAction is "Replace" then
		moveFolder(collectionPath, dataCollectionPath, keepOriginals)
		return true
	else if collectionAction is "Merge" then
	end if

	# By default, no change to the target collection
	set changed to false

	# Collection configuration
	# *************************************************************************
	# In case a config file exists in both the new collection and the target,
	# create a backup of the target version (ie. if it differs from the version)
	# and use the new version instead.

	if isFile(dataConfigPath) is true and areTheSame(collectionConfigPath, dataConfigPath) is false then
		if onConfigExists is "Ask" then
			display dialog "A collection named “" & collectionName & "” in your workspace already has a config that differs from the one being imported. Do you want to replace it?" buttons {"Stop", "Replace", "Keep both"} default button "Keep both" cancel button "Stop"
			set configAction to the button returned of the result
		else
			set configAction to onConfigExists
		end if
	else
		set configAction to "Replace"
	end if

	if configAction is "Stop" then
		error -128
	else if configAction is "Replace" then
	else if configAction is "Keep both" then
		backupFile(dataConfigPath)
		set changed to true
	end if
	moveFile(collectionConfigPath, dataConfigPath, keepOriginals)

	# Collection granules
	# *************************************************************************
	# If a granule directory exists in both the new collection and the target,
	# then the new version will be discarded.

	set granules to find(collectionPath, {"-mindepth", 1, "-maxdepth", 1, "-type", "d"})
	repeat with a from 1 to length of granules
		set granulePath to item a of granules
		set granuleName to getBaseName(granulePath)
		set dataGranulePath to dataCollectionPath & "/" & granuleName

		if isFolder(dataGranulePath) is true then
			if onGranuleExists is "Ask" then
				display dialog "A granule named “" & granuleName & "” in a collection named “" & collectionName & "” already exists in your workspace. Do you want to replace it with the one being imported?" buttons {"Stop", "Replace", "Skip"} default button "Skip" cancel button "Stop"
				set granuleAction to the button returned of the result
			else
				set granuleAction to onGranuleExists
			end if
		else
			set granuleAction to "Replace"
		end if

		if granuleAction is "Stop" then
			error -128
		else if granuleAction is "Replace" then
			moveFolder(granulePath, dataGranulePath, keepOriginals)
			set changed to true
		else if granuleAction is "Skip" then
		end if
	end repeat

	return changed
end importCollection

on joinList(theList, theDelimiter)
	set AppleScript's text item delimiters to theDelimiter
	set theString to theList as string
	set AppleScript's text item delimiters to ""
	return theString
end joinList

on splitString(theString, theDelimiter)
	set AppleScript's text item delimiters to theDelimiter
	set theList to every text item of theString
	set AppleScript's text item delimiters to ""
	return theList
end splitString

on pathExists(aPath)
	try
		POSIX file aPath as alias
		return true
	on error number -1700
		return false
	end try
end pathExists

on isFile(f)
	try
		execution of {"test", "-f", f} without output
		return true
	on error number 1
		return false
	end try
end isFile

on isFolder(f)
	try
		execution of {"test", "-d", f} without output
		return true
	on error number 1
		return false
	end try
end isFolder

on areTheSame(file1, file2)
	try
		execution of {"diff", file1, file2} without output
		return true
	on error number 1
		return false
	end try
end areTheSame

on moveFile(source, destination, keepOriginals)
	if keepOriginals is true then
		execution of {"cp", "-af", source, destination} without output
	else
		execution of {"mv", source, destination} without output
	end if
end moveFile

on moveFolder(source, destination, keepOriginals)
	if keepOriginals is true then
		execution of {"cp", "-af", source, destination} without output
	else
		execution of {"rm", "-rf", destination} without output
		execution of {"mv", source, destination} without output
	end if
end moveFolder

on extractArchive(archive)
	set extractionDir to (POSIX path of (path to temporary items as string)) & "extract"
	execution of {"rm", "-rf", extractionDir} without output
	execution of {"mkdir", "-p", extractionDir} without output
	execution of {"tar", "xf", archive, "-C", extractionDir} without output
	return extractionDir
end extractArchive

on find(path, options)
	set results to execution of ({"find", path} & options) with output
	set results to splitString(results, return)
	return results
end find

on getCurrentDatestamp()
	set {year:y, month:m, day:d, hours:h, minutes:m, seconds:s} to current date
	return "" & y & "" & pad2(m) & "" & pad2(d) & "" & pad2(h) & "" & pad2(m) & "" & pad2(s)
end getCurrentDatestamp

on pad2(num)
	return text -2 thru -1 of ((100 + num) as string)
end pad2

on backupFile(filePath)
	if class of filePath is alias or class of filePath is file then
		set filePath to POSIX path of filePath
	end if

	set datestamp to getCurrentDatestamp()
	set pathParts to splitString(filePath, "/")
	set dirPath to joinList(items 1 thru -2 of pathParts, "/")
	set baseName to item -1 of pathParts
	set baseNameParts to splitString(baseName, ".")

	if length of baseNameParts > 1 then
		set baseNameWithoutExtension to joinList(items 1 thru -2 of baseNameParts, ".")
		set extension to "." & item -1 of baseNameParts
	else
		set baseNameWithoutExtension to baseName
		set extension to ""
	end if

	set counter to 1
	set backupPath to dirPath & "/" & baseNameWithoutExtension & "-" & datestamp & extension
	repeat while pathExists(backupPath)
		set counter to counter + 1
		set backupPath to dirPath & "/" & baseNameWithoutExtension & "-" & datestamp & "-" & counter & extension
	end repeat

	moveFile(filePath, backupPath, false)
end backupFile

on removeTrailingSlashOfPOSIXPath(path)
	if last item of path is "/" then
		return text 1 thru -2 of path
	end if
	return path
end removeTrailingSlashOfPOSIXPath

on getParentDirPath(aPath)
	if class of aPath is alias or class of aPath is file then
		set aPath to POSIX path of aPath
	end if
	return joinList(items 1 thru -2 of splitString(aPath, "/"), "/")
end getParentDirPath

on getBaseName(aPath)
	if class of aPath is alias or class of aPath is file then
		set aPath to POSIX path of aPath
	end if
	return item -1 of splitString(aPath, "/")
end getBaseName

on getAppPath(appName)
	try
		return execution of {"/usr/bin/python", "-c", "from AppKit import NSWorkspace; print NSWorkspace.sharedWorkspace().fullPathForApplication_('" & appName & "') or exit(1)"} with output
	on error number 1
	end try
	set appPath to POSIX path of (choose application with prompt "Where is " & appName & "?" as alias)
end getAppPath

on getSEAScope()
	set appPath to ""
	repeat until pathExists(appPath & "/Contents/MacOS/SEAScope")
		set appPath to getAppPath("SEAScope")
		if pathExists(appPath & "/Contents/MacOS/SEAScope") is true then exit repeat
	end repeat
	return appPath
end getSEAScope

on getDataPath(appPath)
	return execution of {appPath & "/Contents/MacOS/SEAScope", "--print-data-path"} with output
end getDataPath

to getScript of command given output:withOutput
	set theScript to ""
	repeat with i from 1 to length of command
		set arg to item i of command

		if class of arg is alias or class of arg is file or class of arg is «class bmrk» then
			set arg to POSIX path of arg
		end if
		set theScript to theScript & " " & (quoted form of (arg as string))
	end repeat

	if length of theScript > 0 and withOutput is not true then
		set theScript to theScript & " &>/dev/null"
	end if
	return theScript
end getScript

on execution of command given output:withOutput
	set theScript to getScript of command given output:withOutput
	set lastExecutedScript to theScript
	--display dialog theScript
	return do shell script theScript
end execution

