#!/bin/bash

# =============================================================================
# @author: <sylvain.herledan@oceandatalab.com>
# @date: 2017-06-22
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#
# Wrapper that initializes configuration, chooses a version of SEAScope which
# is compatible with the host system and starts the application with the
# options that provide the best performances.
# =============================================================================


set -o errexit
set -o nounset
set -o pipefail

PROGNAME="$(basename "${0}")";
PROGDIR="$(readlink -f "$(dirname "${0}")")";
readonly PROGNAME;
readonly PROGDIR;
readonly ARGS=( "${@}" );

has()
{
	local command_name;
  command_name="${1}"; shift;

	set +o errexit
	hash "${command_name}" > /dev/null 2>&1;
  # shellcheck disable=SC2181
	if [ 0 -eq ${?} ];
	then
		set -o errexit
		return 0;
	fi

	set -o errexit
	return 1;
}

initialize_workspace()
{
  local config_path;
  local workspace_dir;

  config_path="${1}"; shift;
  workspace_dir="$(dirname "${config_path}")";

  mkdir -p "${workspace_dir}/data"

	cat > "${config_path}" <<EOF
[paths]
plugins=
colormaps=
operators=
data=${workspace_dir}/data
state=${workspace_dir}/state.db
index=${workspace_dir}/index.fb
annotations=${workspace_dir}/annotations.db

[rendering]
glRenderer=OpenGL3
glContextMajor=3
glContextMinor=3
windowWidth=800
windowHeight=800
antialiasing=false
antialiasingType=msaa
antialiasingFactor=8
fullscreen=false
vsync=true
minGranuleScreenEstate=0.008

[timeline]
minYear=1988
maxYear=
timespans=6h|6 hours timespan|10800|10800,1d|1 day timespan|43200|43200,3d|3 days timespan|129600|129600,1w|1 week timespan|302400|302400
timesteps=6h|Move by 6 hours|21600,Day|Move by 24 hours|86400,Week|Move by 7 days|604800,Event|Move to previous/next available data|0
defaultTimespan=1d
defaultTimestep=Event
useNearestMode=true
useNearestInDayMode=true
advancedScrolling=false
initialDatetime=

[misc]
worldMap=lowres
skipIndexCheck=false
searchRequestMinDelay=0.1

[processor]
enabled = true
address = 127.0.0.1
port = 53450
EOF
}

usage()
{
	echo -e "
-------------------------------------------------------------------------------
SEAScope viewer -- Geospatial data visualisation and processing tool
  developed by OceanDataLab: https://oceandatalab.com

Documentation available on the SEAScope website:
  https://seascope.oceandatalab.com

For questions, requests and bug reports, please send an email to:
  seascope@oceandatalab.com
-------------------------------------------------------------------------------

Usage: ${PROGNAME} [-h] [-d] [-c CONFIG_FILE] [-l HOST:PORT]
	

where:
    -h  show this help text.
    -l  define the IP address and port the application will listen to for
        incoming commands (default: 127.0.0.1:11155).
    -d  debug mode: note that this will print a lot of text and slow down the
        application. Should only be used by developers.
    -c  path of the configuration file."
}

main()
{
  local _install_dir;
  local install_dir;
  local config_path;
	local listen_opt;
	local debug_opt;
	local seascope;
  local version_opt;

  _install_dir="${1}"; shift;
  install_dir="$(readlink -f "${_install_dir}")";
  config_path="$(readlink -f "$(pwd)")/config.ini";
  listen_opt="127.0.0.1:11155";
  debug_opt="info";
  version_opt="";
  seascope=;

	# Parse arguments
	while [ "$#" -gt 0 ]; do
	  case "$1" in
		-l) listen_opt="$2"; shift 2;;
		-h) usage; exit 0;;
		-d) debug_opt="debug"; shift 1;;
    -c) config_path="${2}"; shift 2;;
    -v) version_opt="-v"; shift 1;;

		--listen) echo "$1 requires an argument" >&2; exit 1;;
		--listen=*) listen_opt="${1#*=}"; shift 1;;
		--help) usage; exit 0;;
		--debug) debug_opt="debug"; shift 1;;
    --config) echo "${1} requires an argument" >&2; exit 1;;
    --config=*) config_path="${1#*=}"; shift 1;;
    --version) version_opt="--version"; shift 1;;

		-*) echo "unknown option: $1" >&2; exit 1;;
		*) echo "unhandled argument: $1"; exit 1;;
	  esac
	done

	if [ ! -f "${config_path}" ];
	then
		initialize_workspace "${config_path}"
	fi

	# GPU-related tests
	local gl_version;
	local mesa_version;
	local video_driver;

  gl_version="$("${install_dir}/check_gl_version")";
  set +o pipefail
  mesa_version="$(echo "${gl_version}" \
                 | grep -i 'mesa' \
                 | sed -r 's/^([^\s\t ]+).*$/\1/')";
  set -o pipefail
  video_driver="unknown";

	if has "lspci";
	then
      set +o pipefail
	  video_driver="$(lspci -nnk 2>/dev/null \
							| grep -i vga -A3 \
							| grep 'in use' \
							| sed -r 's/.*in use:\s*(.+)/\1/' \
							| sed -r 's/\s//g')";
      set -o pipefail
	fi
	
	echo -e "
.:| SYSTEM INFORMATION |:.
--------------------------"
  echo -e "· Install directory: ${install_dir}";
  echo -e "· Config path: ${config_path}";
  echo -e "· Debug options: ${debug_opt}";
	echo -e "· OpenGL version: ${gl_version}";
	echo -e "· Video driver: ${video_driver}";
	[ -n "${mesa_version}" ] && echo -e "· MESA version: ${mesa_version}";
	echo -e "· Listening for external requests on ${listen_opt}";
	echo -e "\n\n";

  seascope="${install_dir}/seascope_glad";

	# Run the command using special launchers if they are available
	if has "optirun";
	then
		optirun "${seascope}" \
      -w "${install_dir}" \
			-c "${config_path}" \
			-l "${listen_opt}" \
			-f "${debug_opt}" \
      "${version_opt}"
	elif has "primusrun";
	then
		primusrun "${seascope}" \
      -w "${install_dir}" \
			-c "${config_path}" \
			-l "${listen_opt}" \
			-f "${debug_opt}" \
      "${version_opt}"
	else
		if [ "3.0" == "${mesa_version}" ];
		then
			# Try to fix issues with MESA
			MESA_GL_VERSION_OVERRIDE=3.3 \
			MESA_GLSL_VERSION_OVERRIDE=330 \
			"${seascope}" \
        -w "${install_dir}" \
				-c "${config_path}" \
				-l "${listen_opt}" \
				-f "${debug_opt}" \
        "${version_opt}"
		else
			"${seascope}" \
        -w "${install_dir}" \
				-c "${config_path}" \
				-l "${listen_opt}" \
				-f "${debug_opt}" \
        "${version_opt}"
		fi
	fi
}

main "${PROGDIR}" "${ARGS[@]}"
exit 0
