#!/bin/bash

# =============================================================================
# @author: <ziad.khoury.hanna@oceandatalab.com>
# @date: 2018-07-16
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#
# Wrapper that initializes configuration, chooses a version of SEAScope which
# is compatible with the host system and starts the application with the
# options that provide the best performances.
# =============================================================================


set -o errexit
set -o nounset
set -o pipefail

readonly PROGNAME="$(basename "${0}")"
readonly PROGDIR="$(cd "$(dirname "${0}")"; pwd -P)"

readonly PRODUCT_NAME='SEAScope'
readonly PRODUCT_BUNDLE_IDENTIFIER="com.oceandatalab.${PRODUCT_NAME}"
readonly APP_SUPPORT_DIR="${HOME}/Library/Application Support/${PRODUCT_BUNDLE_IDENTIFIER}"
readonly LOG_DIR="${HOME}/Library/Logs/${PRODUCT_BUNDLE_IDENTIFIER}"

default_config()
{
  local workspace_dir;

  workspace_dir="${1}"; shift;

  cat <<EOF
[paths]
plugins=
colormaps=${workspace_dir}/colormaps
operators=
data=${workspace_dir}/data
state=${workspace_dir}/state.db
index=${workspace_dir}/index.fb
annotations=${workspace_dir}/annotations.db

[rendering]
glRenderer=OpenGL3
glContextMajor=3
glContextMinor=3
windowWidth=800
windowHeight=800
antialiasing=false
antialiasingType=msaa
antialiasingFactor=8
fullscreen=false
vsync=true
minGranuleScreenEstate=0.008

[timeline]
minYear=1988
maxYear=
timespans=6h|6 hours timespan|10800|10800,1d|1 day timespan|43200|43200,3d|3 days timespan|129600|129600,1w|1 week timespan|302400|302400
timesteps=6h|Move by 6 hours|21600,Day|Move by 24 hours|86400,Week|Move by 7 days|604800,Event|Move to previous/next available data|0
defaultTimespan=1d
defaultTimestep=Event
useNearestMode=true
useNearestInDayMode=true
advancedScrolling=false
initialDatetime=

[misc]
worldMap=lowres
skipIndexCheck=false
searchRequestMinDelay=0.1

[processor]
enabled = true
address = 127.0.0.1
port = 53450
EOF
}

usage()
{
	echo -e "
-------------------------------------------------------------------------------
SEAScope viewer -- Geospatial data visualisation and processing tool
  developed by OceanDataLab: https://oceandatalab.com

Documentation available on the SEAScope website:
  https://seascope.oceandatalab.com

For questions, requests and bug reports, please send an email to:
  seascope@oceandatalab.com
-------------------------------------------------------------------------------

Usage: ${PROGNAME} [-h] [-d] [-c CONFIG_FILE] [-l HOST:PORT]


where:
    -h  show this help text.
    -l  define the IP address and port the application will listen to for
        incoming commands (default: 127.0.0.1:11155).
    -d  debug mode: note that this will print a lot of text and slow down the
        application. Should only be used by developers.
    -c  path of the configuration file.

    --print-data-path
        print the data dir defined in the configuration file currently in use
        and exits."
}

get_config_paths()
{
    local config_path;
    local requested_keys;
    local section;
    local key;
    local value;

    config_path="${1}"; shift;
    requested_keys="$@";

    while IFS='= ' read key value
    do
        if [[ ${key} == \;* ]]; then
            : # Skip comments
        elif [[ ${key} == \[*] ]]; then
            section="${key}";
        elif [[ \
            ${key} && \
            "${section}" == "[paths]" && \
            " ${requested_keys} " =~ " ${key} " \
        ]]; then
            eval "config_${key}=${value}";
        fi
    done < "${config_path}"
}

abspath_alias()
{
    osascript 2>/dev/null << EOS
tell application "Finder" to get the POSIX path of ¬
    (original item of (POSIX file "$1" as alias) as text)
EOS
}

abspath()
{
    local file;

    file="${1}"; shift;

    if [ -L "${file}" ]; then
        cd "$(dirname "${file}")";
        file="$(basename "${file}")";

        # Iterate down a (possible) chain of symlinks
        while [ -L "${file}" ]; do
            file="$(readlink "${file}")";
            if cd "$(dirname "${file}")" 2> /dev/null; then
                file="$(basename "${file}")";
            fi
        done

        if [[ "${file}" != /* ]]; then
            # Compute the canonicalized name by finding the physical path
            # for the directory we're in and appending the target file.
            file="$(pwd -P)/${file}";
        fi
    fi

    # Use applescript to ask Finder to resolve the alias.
    # But requires the user to give SEAScope permission to control Finder,
    # which isn't ideal.
    #file="$(abspath_alias "${file}" || echo "${file}")";

    if [[ "${file}" != /* ]]; then
        cd "$(dirname "${file}")";
        file="$(pwd -P)/$(basename "${file}")";
    fi

    echo "${file}";
}

main()
{
    local _install_dir;
    local install_dir;
    local app_dir;
    local config_path;
    local listen_opt;
    local debug_opt;
    local print_data_path;
    local version_opt;

    _install_dir="${1}"; shift;
    install_dir="$(cd "${_install_dir}"; pwd -P)";
    app_dir="$(dirname "$(dirname "${install_dir}")")";
    config_path="";
    listen_opt="127.0.0.1:11155";
    debug_opt="info";
    print_data_path="false";
    version_opt="";

    # Parse arguments
    while [ "$#" -gt 0 ]; do
        case "$1" in
            -l) listen_opt="$2"; shift 2;;
            -h) usage; exit 0;;
            -d) debug_opt="debug"; shift 1;;
            -c) config_path="${2}"; shift 2;;
            -v) version_opt="-v"; shift 1;;

            --listen) echo "$1 requires an argument" >&2; exit 1;;
            --listen=*) listen_opt="${1#*=}"; shift 1;;
            --help) usage; exit 0;;
            --debug) debug_opt="debug"; shift 1;;
            --config) echo "${1} requires an argument" >&2; exit 1;;
            --config=*) config_path="${1#*=}"; shift 1;;
            --version) version_opt="--version"; shift 1;;

            --print-data-path) print_data_path="true"; shift 1; break;;

            # macOS's Process Serial Number - passed when launched via GUI
            -psn_*) shift 1;;

            -*) echo "unknown option: $1" >&2; exit 1;;
            *) echo "unhandled argument: $1" >&2; exit 1;;
        esac
    done

    if [ -z "${config_path}" ]; then
        config_path="${APP_SUPPORT_DIR}/config.ini";
        if [ ! -e "${config_path}" -a ! -h "${config_path}" ]; then
            mkdir -p "${APP_SUPPORT_DIR}";
            ln -s "${HOME}/${PRODUCT_NAME}-workspace/config.ini" "${config_path}";
        fi
    fi

    config_path="$(abspath "${config_path}")";

    if [ ! -f "${config_path}" ]; then
        >&2 echo "Config file not found, creating it with defaults";
        mkdir -p "$(dirname "${config_path}")";
        default_config "$(dirname "${config_path}")" > "${config_path}";
    fi

    local config_colormaps;
    local config_data;
    local config_index;
    get_config_paths "${config_path}" colormaps data index;

    mkdir -p "${config_data}";
    if [ ! -d "${config_colormaps}" ]; then
        cp -r "${app_dir}/Contents/Resources/colormaps" "${config_colormaps}";
    fi

    if [ "${print_data_path}" == "true" ]; then
        echo -n "${config_data}";
        exit 0;
    fi

    echo "SEAScope will be listening for external requests on ${listen_opt}";
    echo;
    echo "Running from: ${install_dir}";
    echo "Using config: ${config_path}";
    echo "Debug option: ${debug_opt}";
    echo;
    echo;

    exec "${app_dir}/Contents/Resources/seascope" \
        -w "${app_dir}/Contents/Resources" \
        -c "${config_path}" \
        -l "${listen_opt}" \
        -f "${debug_opt}" \
        "${version_opt}"
}

if [ "3" == "$(lsof -p $$ -a -d '0-2' -F n | grep -c 'n/dev/null')" ]; then
    # Launched by Launch Services
    mkdir -p "${LOG_DIR}"
    exec 1> "${LOG_DIR}/seascope_out.log"
    exec 2> "${LOG_DIR}/seascope_err.log"
fi
main "${PROGDIR}" "$@"
exit 0
