#!/bin/bash

# =============================================================================
# author: <sylvain.herledan@oceandatalab.com>
# date: 2017-07-12
#
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================

set -o nounset
set -o errexit
set -o pipefail

readonly PROGNAME="$(basename "${0}")";
readonly PROGDIR="$(cd "$(dirname "${0}")"; pwd -P)"

# Settings
# -----------------------------------------------------------------------------
readonly BASE_URL='https://seascope.oceandatalab.com/data';

set +o nounset
if [ -n "${CMDER_ROOT}" ];
then
  FIND='/usr/bin/find'
else
  FIND="$(which find)";
fi
set -o nounset

is_macos()
{
  case $OSTYPE in
    darwin*) return 0;;
    *) return 1;;
  esac
}

has()
{
  local command_name;

  command_name="${1}"; shift;

  set +o errexit
  hash "${command_name}" > /dev/null 2>&1;
  if [ 0 -eq ${?} ];
  then
    set -o errexit
    return 0;
  fi

  set -o errexit
  return 1;
}

nice_display()
{
  local input_path;
  input_path="${1}"; shift;

  while read -r line
  do
    echo -e "${line}"
  done < "${input_path}"
}

usage()
{
  echo -e "
-------------------------------------------------------------------------------
SEAScope case studies helper script

For questions, requests and bug reports, please send an email to:
  seascope@oceandatalab.com
-------------------------------------------------------------------------------

Usage:
  Display this help message
    ${PROGNAME} -h

  List case studies available on the SEAScope website
    ${PROGNAME} [-t TMP_PATH] list

  where:
      * TMP_PATH is a writeable directory where the listing will be downloaded
        (and destroyed once it has been viewed).

  Retrieve a case study
    ${PROGNAME} [-t TMP_PATH] get CASE_STUDY_NAME

    where:
      * CASE_STUDY_NAME should be one of the case study identifiers returned by the
        'list' command.
      * TMP_PATH is a writeable directory where data will be temporarily
        stored until the downloaded file integrity is checked.

  Retrieve and install a case study
    ${PROGNAME} [-t TMP_PATH] get CASE_STUDY_NAME TARGET_DIR

    where:
      * CASE_STUDY_NAME should be either one of the case study identifiers returned
        by the 'list' command or the path of a local case study archive.
      * TARGET_DIR should be the data directory of the SEAScope viewer.
      * TMP_PATH is a writeable directory where data will be temporarily
        stored and unpacked until the installation is complete.

    Note: If CASE_STUDY_NAME is the path of a file in the local filesystem, the
    script will install the contents of this file instead of retrieving the
    case study from the SEAScope website."

  if is_macos;
  then
    echo "
    Note: With a default install on macOS TARGET_DIR will be:
    ~/SEAScope-workspace/data"
  fi
}

download_error()
{
  >&2 echo -e "\033[1;31mX\033[0m \033[1;37mERROR\033[0m
This script only supports a few commands to manage downloads but none of them
is available on your system.

Supported download commands:
 - wget
 - curl

If you use another command to handle downloads and can not install one of the
supported ones, please make a support request by sending an email to 
seascope@oceandatalab.com."
  exit 1
}

list_cmd()
{
  local _tmp;
  local url;
  local list_path;

  _tmp="${1}"; shift;
  url="${BASE_URL}/list.txt";
  list_path="";
  
  echo "${_tmp}" "${url}"
  download "${_tmp}" "${url}" "list_path";

  nice_display "${list_path}"
  rm "${list_path}"
}

download()
{
  local _tmp;
  local url;
  local tmp_path;

  _tmp="${1}"; shift;
  url="${1}"; shift;

  mkdir -p "${_tmp}"
  tmp_path="$(mktemp -p "${_tmp}" "${PROGNAME}.XXX")";

  if has "wget";
  then
    local rc;
    set +o errexit
    wget -q --show-progress "${url}" -O "${tmp_path}"
    rc=${?};
    set -o errexit

    if [ ${rc} != 0 ]; then
      >&2 echo -e "\033[1;31mX\033[0m \033[1;37mERROR\033[0m"
      >&2 echo "Failed to download ${url} with wget."
      >&2 echo "Please check the case study identifier."
      exit 1
    fi
    eval "${1}"="${tmp_path}"
  elif has "curl";
  then
    local rc;
    set +o errexit
    curl -XGET --progress-bar "${url}" > "${tmp_path}"
    rc=${?};
    set -o errexit

    if [ ${rc} != 0 ]; then
      >&2 echo -e "\033[1;31mX\033[0m \033[1;37mERROR\033[0m"
      >&2 echo "Failed to download ${url} with curl."
      >&2 echo "Please check the case study identifier."
      exit 1
    fi
    eval "${1}"="${tmp_path}"
  else
    download_error
  fi
}

get_cmd()
{
  local tmp_dir;
  local hotspot_id;
  local tar_path;
  local filename;
  local url;
  local md5sum;

  tmp_dir="${1}"; shift;
  hotspot_id="${1}"; shift;
  tar_path="${hotspot_id}.tar";
  filename="$(basename "${tar_path}")";
  url="${BASE_URL}/${filename}";

  echo -e "
\033[1;34m↳\033[0m \033[1;37mRetrieving the \"${hotspot_id}\" hotspot...\033[0m
"

  if [ -f "${hotspot_id}" ];
  then
    echo "Using provided local case study archive."
    filename="${hotspot_id}";
  elif [ -f "${filename}" ];
  then
    echo "Case study archive already downloaded."
  else
    if has "md5sum";
    then
      md5sum='md5sum';
    elif has "md5";
    then
      md5sum='md5 -r';
    else
      >&2 echo -e "\033[1;31mX\033[0m \033[1;37mERROR\033[0m"
      >&2 echo "The md5sum command is required to check the integrity of the"
      >&2 echo "downloaded case study archive. Alas it does not seem to be "
      >&2 echo "available on your system, please install it using your "
      >&2 echo "preferred method and rerun this script."
      exit 1
    fi

    local hash_path;
    hash_path=""
    download "${tmp_dir}" "${url}.md5" hash_path

    local dl_path;
    dl_path=""
    download "${tmp_dir}" "${url}" dl_path;

    # Check md5 hash
    local download_hash;
    download_hash="$( ${md5sum} "${dl_path}" \
                          | cut -d\  -f1)";
    local ref_hash;
    ref_hash="$(cat "${hash_path}")";
    if [ "${download_hash}" != "${ref_hash}" ]; then
      >&2 echo -e "\033[1;31mX\033[0m \033[1;37mERROR\033[0m"
      >&2 echo "MD5 hash verification failed."
      >&2 echo "There must have been an issue during the download, please"
      >&2 echo "check that you have enough disk space and try again."
      exit 1
    fi

    # Move result to target
    mv "${dl_path}" "${filename}"

    # Cleanup
    rm -f "${hash_path}"
  fi
  
  if [ "${#}" -gt 0 ];
  then
    local target_dir;
    target_dir="${1}"; shift;
    install_hotspot "${tmp_dir}" "${filename}" "${target_dir}"
  fi
}

install_hotspot()
{
  local _tmp;
  local tar_path;
  local data_dir;
  local tmp_dir;
  local index_update_needed;

  _tmp="${1}"; shift;
  tar_path="${1}"; shift;
  data_dir="${1}"; shift;
  mkdir -p "${_tmp}"
  tmp_dir="$(mktemp -d -p "${_tmp}" "${PROGNAME}.XXX")";
  index_update_needed=0;

  echo -e "
\033[1;34m↳\033[0m \033[1;37mInstalling the case study...\033[0m
"

  mkdir -p "${data_dir}"
  tar xf "${tar_path}" -C "${tmp_dir}"

  local hotspot_collections;
  hotspot_collections=$("${FIND}" "${tmp_dir}" \
                                -mindepth 2 \
                                -maxdepth 2 \
                                -type d);

  for hotspot_collection_path in ${hotspot_collections};
  do
    local collection_name;
    local data_collection_path;

    collection_name="$(basename "${hotspot_collection_path}")";
    data_collection_path="${data_dir}/${collection_name}";
    if [ ! -d "${data_collection_path}" ];
    then
      echo "Collection ${collection_name} did not exist in the data directory"
      mv "${hotspot_collection_path}" "${data_collection_path}"
      index_update_needed=1
      continue
    fi

    # Collection configuration
    # *************************************************************************
    # In case a config file exists in the case study and in the target, create
    # a backup of the target version (ie. if it differs from the case study
    # version) and use the case study version instead.
    local hotspot_config_path;
    local data_config_path;
    hotspot_config_path="${hotspot_collection_path}/config.ini";
    data_config_path="${data_collection_path}/config.ini";
    if [ -f "${data_config_path}" ];
    then
      local rc;
      local backup_path;
      set +o errexit
      diff "${hotspot_config_path}" "${data_config_path}" >/dev/null 2>&1
      rc="${?}";
      set -o errexit
      backup_path="${data_config_path}.$(date '+%Y%m%d%H%M%S').bak";
      [ "0" != "${rc}" ] \
        && echo "Backup existing configuration for ${collection_name}" \
        && mv "${data_config_path}" "${backup_path}"
    fi
    mv "${hotspot_config_path}" "${data_config_path}"

    # Collection granules
    # *************************************************************************
    # If a granule directory exists in the case study and in the target, then
    # the case study version will be discarded.
    local hotspot_granules;
    hotspot_granules=$("${FIND}" "${hotspot_collection_path}" \
                              -mindepth 1 \
                              -maxdepth 1 \
                              -type d);
    for hotspot_granule_path in ${hotspot_granules};
    do
      local granule_name;
      local data_granule_path;
      granule_name="$(basename "${hotspot_granule_path}")";
      data_granule_path="${data_collection_path}/${granule_name}";
      if [ -d "${data_granule_path}" ];
      then
        echo "Granule ${granule_name} already in the data directory. Skipped."
        rm -rf "${hotspot_granule_path}"
      else
        mv "${hotspot_granule_path}" "${data_granule_path}"
        index_update_needed=1
      fi
    done
  done

  echo -e "
\033[1;32m✓\033[0m \033[1;37mSUCCESS\033[0m
"

  if [ ${index_update_needed} -gt 0 ];
  then
    echo -e "\033[1;33m
Data have been successfully installed but they will only appear in the SEAScope
viewer after you regenerate the index file.

Remove the index.fb file in your SEAScope directory and restart the viewer
(regenerating the index might take a while, so please be patient).
\033[0m"
  fi
}

main()
{
  [ "${#}" == 0 ] && usage && exit 1

  local _tmp;
  _tmp="/tmp";

  while [ $# -gt 0 ];
  do
    local cmd;
    cmd="${1}"; shift;
    case "${cmd}" in
      "list") list_cmd "${_tmp}" "$@"; exit 0;;
      "get") get_cmd "${_tmp}" "$@"; exit 0;;
      "-t") _tmp="${1}"; shift;;
      "-h"|"--help") usage; exit 0;;
      *) usage; exit 1;;
    esac
  done
}

main "$@"
exit 0
