// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cctype>
#include <cstdlib>
#include <algorithm>
#include <sstream>

#include "vendors/_ini_parser.h"
#include "vendors/_fmt.h"

#include "core/log.h"
#include "local/collection_ini_parser.h"
#include "debug.h"

using namespace SEAScope;

static RenderMethod renderMethods(const std::string name, bool& ok)
{
    if
        ("RASTER" == name)
    {
        ok = true;
        return RenderMethod::RASTER;
    }
    else if
        ("ARROWS" == name)
    {
        ok = true;
        return RenderMethod::ARROWS;
    }
    else if
        ("BARBS" == name)
    {
        ok = true;
        return RenderMethod::BARBS;
    }
    else if
        ("TRAJECTORIES" == name)
    {
	    ok = true;
        return RenderMethod::TRAJECTORIES;
    }
    else if
        ("DOTSCLOUD" == name)
    {
        ok = true;
        return  RenderMethod::DOTSCLOUD;
    }
    else if
        ("STREAMLINES" == name)
    {
        ok = true;
        return RenderMethod::STREAMLINES;
    }
    else if
        ("RAWRGB" == name)
    {
        ok = true;
        return RenderMethod::RAWRGB;
    }

    ok = false;
    return RenderMethod::RASTER;
};


static FilterMode filterModes(const std::string& name, bool& ok)
{
    if
        ("NEAREST" == name)
    {
        ok = true;
        return FilterMode::NEAREST;
    }
    else if
        ("BILINEAR" == name)
    {
        ok = true;
        return FilterMode::BILINEAR;
    }

    ok = false;
    return FilterMode::NEAREST;
};

/**
 *
 */
CollectionINIParser::CollectionINIParser()
{

};

/**
 *
 */
CollectionINIParser::~CollectionINIParser()
{
};

/**
 *
 */
bool CollectionINIParser::extractTags(const std::string& input, std::map<std::string, std::string>& result)
{
	std::vector<std::string> keyvals;	//
	std::vector<std::string> tagValue;	//

	splitString(input, ',', keyvals);
	for
		(const auto& keyval: keyvals)
	{
		tagValue.clear();
		splitString(keyval, ':', tagValue);
		if
			(2 == tagValue.size())
		{
			result[tagValue[0]] = tagValue[1];
		}
		else
		{
			Log::log( LogLevel::Error, "CollectionINIParser"
					, "Tag does not match the expected KEY:VALUE format: {}"
					, keyval);
			return false;
		}
	}

	return true;
};

/**
 *
 */
std::string CollectionINIParser::trim(const std::string& input)
{
	auto  wsfront=std::find_if_not(input.begin(),input.end(),[](int c){return std::isspace(c);});
	return std::string(wsfront,std::find_if_not(input.rbegin(),std::string::const_reverse_iterator(wsfront),[](int c){return std::isspace(c);}).base());
};

/**
 *
 */
std::vector<std::string>& CollectionINIParser::splitString(const std::string& input, char delim, std::vector<std::string>& result)
{
	std::stringstream ss(input);	//
	std::string item;	//
	while
		(std::getline(ss, item, delim))
	{
		item = trim(item);
		if
			(false == item.empty())
		{
			result.push_back(item);
		}
	}

	return result;
};

/**
 *
 */
bool CollectionINIParser::validate(const std::string& uri, INI::Level& root) const
{
	std::vector<std::string> missing;	//
	std::vector<std::string> variables;	//
	const std::array<std::string, 6> requiredGeneral = {"label",
													    "xSeamless",
													    "ySeamless",
													    "mustBeCurrent",
													    "tags",
													    "variables"};
	const std::array<std::string, 15> requiredVar = {"label",
												     "units",
												     "fields",
												     "color",
												     "colormap",
												     "defaultRendering",
												     "tags",
												     "filterMode",
												     "logscale",
												     "min",
												     "max",
												     "opacity",
												     "zindex",
												     "particlesCount",
												     "streamlineLength"};

	try
	{
		INI::Parser p(uri.c_str());
		root = p.top();
	}
	catch
		(std::runtime_error& e)
	{
		Log::log( LogLevel::Error, "CollectionINIParser"
				, "Error during parsing of {}"
				, uri);
		Log::log( LogLevel::Debug, "CollectionINIParser"
				, "{}"
				, e.what());
		return false;
	}

	if
		(root.sections.end() == root.sections.find("general"))
	{
		Log::log(LogLevel::Error, "CollectionINIParser",
                 "[general] section missing in {}", uri);
		return false;
	}

	// Check fields from [general] section
	INI::Level& general = root("general");
	for
		(const auto& field : requiredGeneral)
	{
		if
			(general.values.end() == general.values.find(field))
		{
			missing.push_back(fmt::format("Field: [general] {}", field));
		}
	}
	if
		(general.values.end() == general.values.find("variables"))
	{
		// Further validation impossible without the variables list
		return false;
	}

	splitString(general["variables"], ',', variables);
	for
		(const auto& varname : variables)
	{
		if
			(root.sections.end() == root.sections.find(varname))
		{
			missing.push_back(fmt::format("Section: [{}]", varname));

			// Go on with the next variable
			continue;
		}

		INI::Level& varSection = root(varname);
		for
			(const auto& field : requiredVar)
		{
			if
				(varSection.values.end() == varSection.values.find(field))
			{
				missing.push_back(fmt::format("Field: [{}] {}", varname, field));
			}
		}
	}

	// Check fields for all variables
	if
		(0 < missing.size())
	{
		Log::log(LogLevel::Error, "CollectionINIParser",
				 "Configuration file {} is incomplete, the following elements have not been found:", uri);
		for
			(const auto& msg : missing)
		{
			Log::log(LogLevel::Error, "CollectionINIParser", msg);
		}
		return false;
	}

	return true;
}

/**
 *
 */
bool CollectionINIParser::load(uint32_t colId, const std::string& uri, CollectionCfg& colCfg, std::vector<RenderingCfg*>& rCfgs) const
{
	std::vector<std::string> colorComponents;	//
	std::vector<std::string> variables;	//
	std::vector<std::string> fields;	//
	std::vector<std::string> tag;	//
	VariableId defaultVariable;	//
	RenderingCfg* rCfg;	//
	std::uint32_t varsCount;	//
	std::uint8_t r;	//
	std::uint8_t g;	//
	std::uint8_t b;	//
	INI::Level root;	//
	bool ok;	//

	defaultVariable = 0;

	Log::log( LogLevel::Debug, "CollectionINIParser"
			, "Parsing file {}"
			, uri);

	ok = this->validate(uri, root);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "CollectionINIParser",
                 "Validation of the {} configuration file failed", uri);
		return false;
	}

	// Delete objects before vector clear
	for (RenderingCfg* rCfgTmp : rCfgs)
	{
		delete rCfgTmp;
	}

	rCfgs.clear();
	INI::Level& general = root("general");

	// Extract lists
	splitString(general["variables"], ',', variables);
	Log::log( LogLevel::Debug, "CollectionINIParser"
			, "Variables: {}"
			, general["variables"]);

	colCfg.label = general["label"];
	colCfg.xSeamless = "true" == general["xSeamless"];
	colCfg.ySeamless = "true" == general["ySeamless"];
	colCfg.NEWSAligned = false;
	colCfg.mustBeCurrent = "true" == general["mustBeCurrent"];
	extractTags(general["tags"], colCfg.tags);

	Log::log( LogLevel::Debug, "CollectionINIParser"
			, "Number of variables: {}"
			, variables.size());

	varsCount = 0;
	for
		(const auto& varName : variables)
	{
		VariableCfg vCfg;
		INI::Level& vInfo = root(varName);
		vCfg.label  = vInfo["label"];
		vCfg.units = vInfo["units"];

		fields.clear();
		splitString(vInfo["fields"], ',', fields);
		vCfg.fields.insert(fields.begin(), fields.end());

		/* Parse color parameter */
		colorComponents.clear();
		splitString(vInfo["color"], ',', colorComponents);
		if
			(3 != colorComponents.size())
		{
			Log::log( LogLevel::Error, "CollectionINIParser"
					, "Invalid format for the \"color\" parameter: {} (expected: color = r,g,b)"
					, vInfo["color"]);
			return false;
		}
		r = static_cast<std::uint8_t>(std::atoi(colorComponents[0].c_str()));
		g = static_cast<std::uint8_t>(std::atoi(colorComponents[1].c_str()));
		b = static_cast<std::uint8_t>(std::atoi(colorComponents[2].c_str()));

        RenderMethod rMethod = renderMethods(vInfo["defaultRendering"], ok);
        if
            (false == ok)
		{
			Log::log( LogLevel::Error, "CollectionINIParser"
					, "Rendering method \"{}\" is not supported"
					, vInfo["defaultRendering"]);
			return false;
		}
		vCfg.defaultRendering = rMethod;

		extractTags(vInfo["tags"], vCfg.tags);

		colCfg.variables[varsCount] = vCfg;

		rCfg = new RenderingCfg();
		rCfg->_renderMethod = rMethod;
		rCfg->_logScale = "true" == vInfo["logscale"];
		rCfg->_min = static_cast<float>(std::atof(vInfo["min"].c_str()));
		rCfg->_max = static_cast<float>(std::atof(vInfo["max"].c_str()));
		rCfg->_opacity = static_cast<float>(std::atof(vInfo["opacity"].c_str()));
		rCfg->_zindex = static_cast<float>(std::atof(vInfo["zindex"].c_str()));
		rCfg->_particlesCount = static_cast<std::uint32_t>(std::atoi(vInfo["particlesCount"].c_str()));
		rCfg->_particleTTL = 0; // No longer used
		rCfg->_streamlineLength = static_cast<std::uint32_t>(std::atoi(vInfo["streamlineLength"].c_str()));
		rCfg->_streamlineSpeed = rCfg->_streamlineSpeed = static_cast<float>(std::atof(vInfo["streamlineSpeed"].c_str()));
		if
			(rCfg->_streamlineSpeed == 0.f)
		{
			rCfg->_streamlineSpeed = 1.f;
		}

		rCfg->_colormap = vInfo["colormap"];
		rCfg->_color = {{ r, g, b }};
		rCfg->target.granuleLevel = false;
		rCfg->target.collection = std::make_pair(0u, colId);
		rCfg->target.variable = varsCount;

        FilterMode fMode = filterModes(vInfo["filterMode"], ok);
        if
            (false == ok)
		{
			Log::log( LogLevel::Error, "CollectionINIParser"
					, "Filter mode \"{}\" is not supported"
					, vInfo["filterMode"]);
			return false;
		}
		rCfg->_filterMode = fMode;

		if
			(varName == general["defaultVariable"])
		{
			defaultVariable = varsCount;
		}

		rCfgs.push_back(rCfg);
		++varsCount;
	}
	colCfg.defaultVariable = defaultVariable;

	Debug::displayCollection(colCfg);
	return true;
};
