// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <sstream>
#include <iostream>
#include "local/collections_explorer.h"
#include "local/index_serializer.h"
#include "local/idf_reader.h"
#include "core/path.h"
#include "core/log.h"
#include "index_dialog.h"
#include "debug.h"

using namespace SEAScope;

/**
 *
 */
CollectionsExplorer::CollectionsExplorer()
: dataPath(""), detectOutdatedIndex(true)
{
}

/**
 *
 */
CollectionsExplorer::~CollectionsExplorer()
{
}

/**
 *
 */
bool CollectionsExplorer::init(const std::string& path, const std::string& idx,
							   bool ignoreIndex)
{
	bool ok;	//
	bool buildIndex;	//
	bool indexLoaded;	//
	std::vector<std::string> collections;	//

	this->indexPath = idx;

	ok = Path::dirExists(path);
	if
		(false == ok)
	{
		return false;
	}

	this->dataPath = path;

	ok = this->listCollections(this->dataPath, collections);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "CollectionsExplorer",
				 "Could not find any collection. Skiping index creation");
		return false;
	}

	indexLoaded = false;
	if
		(false == ignoreIndex)
	{
		indexLoaded = this->loadIndex(collections);
	}

	buildIndex = (  false == indexLoaded
				 && false == ignoreIndex);

	ok = this->registerData(collections, buildIndex);

	return ok;
}

/**
 *
 */
bool CollectionsExplorer::listCollections(const std::string& dataPathIn,
										  std::vector<std::string>& result) const
{
	std::forward_list<std::string> files;	//
	std::ostringstream buffer;	//
	std::string basename;	//
	std::size_t slashPos;	//
	bool isCollectionDir;	//
	bool isDir;	//
	bool ok;	//

	ok = Path::listFiles(dataPathIn, files);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& subpath : files)
	{
		slashPos = subpath.find_last_of(SEAScope::Path::separator());
		basename = subpath.substr(slashPos+1);
		isDir = Path::dirExists(subpath);
		if
			(  0 < subpath.size()
			&& true == isDir
			&& 0 != basename.compare(0, 1, "."))
		{
			buffer.str("");
			buffer << subpath << SEAScope::Path::separator() << "config.ini";
			isCollectionDir = Path::fileExists(buffer.str());
			if
				(true == isCollectionDir)
			{
				result.push_back(subpath);
			}
		}
	}
	return true;
}

/**
 *
 */
bool CollectionsExplorer::isIndexUpdateNeeded(const std::string& indexPathIn,
											  const std::string& dataPathIn,
											  const std::vector<std::string>& collections) const
{
	bool ok;	//
	std::time_t indexTimestamp;	//
	std::time_t dataTimestamp;  //
	std::time_t collectionTimestamp;	//

	ok = Path::lastModified(indexPathIn, indexTimestamp);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "CollectionsExplorer",
                 "Could not get last modification time for {}", indexPathIn);
		return false;
	}

	ok = Path::lastModified(dataPathIn, dataTimestamp);
	if
		(false == ok)
	{
		Log::log(LogLevel::Error, "CollectionsExplorer",
				 "Could not get last modification time for {}", dataPathIn);
		return false;
	}

	if
		(dataTimestamp > indexTimestamp)
	{
		/* Index might be outdated */
		return true;
	}

	for
		(const auto& collectionPath : collections)
	{
		ok = Path::lastModified(collectionPath, collectionTimestamp);
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "CollectionsExplorer",
					 "Could not get last modification time for {}",
					 collectionPath);
			return false;
		}

		if
			(collectionTimestamp > indexTimestamp)
		{
			/* Index might be outdated */
			return true;
		}
	}

	return false;
}

/**
 *
 */
bool CollectionsExplorer::loadIndex(const std::vector<std::string>& collections)
{
    IndexSerializer serializer; //
	bool indexLoaded;	//
	bool useExistingIndex;	//
	bool modificationsDetected;	//

	useExistingIndex = true;
	indexLoaded = false;

    if
        (  true == Path::fileExists(this->indexPath)
		&& true == this->detectOutdatedIndex)
    {
		modificationsDetected = this->isIndexUpdateNeeded(this->indexPath,
														  this->dataPath,
														  collections);
		if
			(true == modificationsDetected)
		{
			useExistingIndex = !SEAScope::askAboutIndexUpdate("It appears some of the files in your data directory are newer than the index file, which could therefore be outdated. Do you want to rebuild the index?");
			if
				(false == useExistingIndex)
			{
				/* User chose to regenerate index, delete old one */
				Path::deleteFile(this->indexPath);
			}
		}
	}

	if
		(true == useExistingIndex
        && true == Path::fileExists(this->indexPath))
	{
		/* Use existing index */
		indexLoaded = serializer.load(this->index, this->indexPath);
	}
	else
	{
		std::cout << "*******************************************************************************" << std::endl;
		std::cout << "SEAScope will now build an index of the data granules" << std::endl;
		std::cout << "This operation will take several minutes as each granule file must be opened to" << std::endl;
		std::cout << " extract information." << std::endl;
		std::cout << "The resulting index will be saved so this step will not be necessary anymore" << std::endl;
		std::cout << "(until you add/remove/modify granules) and SEAScope will start up much faster." << std::endl;
		std::cout << "*******************************************************************************" << std::endl;
	}

	return indexLoaded;
}

/**
 *
 */
bool CollectionsExplorer::registerData(const std::vector<std::string>& collections, bool buildIndex)
{
    IndexSerializer serializer; //
	std::uint32_t colId;
	bool ok;	//

	Log::log( LogLevel::Debug, "CollectionsExplorer", "registerData");

	for
		(const auto& collectionPath : collections)
	{
		ok = this->registerCollection(collectionPath, colId);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "CollectionsExplorer"
					, "Failed to register collection");
			return false;
		}
		if
			(true == buildIndex)
		{
			this->findGranules(collectionPath, colId);
		}
	}

    if
        (  true == buildIndex
        && false == Path::fileExists(this->indexPath)
		&& true == Path::writeFile(this->indexPath))
    {
        ok = serializer.save(this->index, this->indexPath);
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "CollectionsExplorer", "Failed to save index");
			return false;
		}
	    Log::log(LogLevel::Debug, "CollectionsExplorer", "Index saved");
    }

	return true;
}

/**
 *
 */
bool CollectionsExplorer::findGranules(const std::string& path, std::uint32_t colId)
{
	std::cout << " -> Indexing granules in " << path << std::endl;
	Log::log( LogLevel::Debug, "CollectionsExplorer"
			, "findGranules({})"
			, colId);
	std::forward_list<std::string> files;	//
	bool ok;	//

	ok = Path::listFilesRecursive(path, files);
	if
		(false == ok)
	{
		return false;
	}

	for
		(const auto& filePath : files)
	{
		if
			(  3 < filePath.size()
			&& ".nc" == filePath.substr(filePath.size() - 3))
		{
			this->indexGranule(filePath, colId);
		}
	}

	return true;
};

/**
 *
 */
bool CollectionsExplorer::indexGranule(const std::string& filePath, std::uint32_t colId)
{
	GranuleMetadata gInfo;	//
	bool alreadyIndexed;	//
	bool ok;	//

	alreadyIndexed = this->index.hasGranule(filePath);
	if
		(true == alreadyIndexed)
	{
		/* Granule already indexed */
		Log::log( LogLevel::Debug, "CollectionsExplorer"
				, "Granule already indexed: {}"
				, filePath);
		return false;
	}

	//ok = IDFReader::extractGranuleInfo(filePath, gInfo);
	ok = this->reader.extractGranuleInfo(filePath, gInfo);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "CollectionsExplorer"
				, "Could not extract granule information from {}"
				, filePath);
		return false;
	}

	/* TODO: include resolution information, required by the index */

	gInfo.collectionId = colId;

	ok = this->index.add(gInfo, false);

	return ok;
};

/**
 *
 */
bool CollectionsExplorer::registerCollection(const std::string& path, std::uint32_t& colId)
{
	CollectionCfg colCfg;	//
	std::vector<RenderingCfg*> rCfgs;	//
	std::stringstream buffer;	//
	bool alreadyRegistered;	//
	bool ok;	//

	alreadyRegistered = this->catalogue.hasCollection(path);
	if
		(true == alreadyRegistered)
	{
		Log::log( LogLevel::Debug, "CollectionsExplorer"
				, "Collection already registered: {}"
				, path);
		return false;
	}

	colId = this->catalogue.getID();

	buffer << path << SEAScope::Path::separator() << "config.ini";
	ok = this->parser.load(colId, buffer.str(), colCfg, rCfgs);
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "CollectionsExplorer"
				, "Could not parse collection configuration found in {}"
				, path);
		return false;
	}

	Log::log( LogLevel::Debug, "CollectionsExplorer"
			, "Registering collection {}"
			, path);
	Debug::displayCollection(colCfg);

	ok = this->catalogue.addCollection(colId, colCfg, rCfgs, buffer.str());
	if
		(false == ok)
	{
		Log::log( LogLevel::Error, "CollectionsExplorer"
				, "Failed to register collection {} in the catalogue"
				, colId);
		return false;
	}

	return ok;
};
