// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>

#include "local/embedded_catalogue.h"
#include "filter_logger.h"
#include "core/log.h"
#include "core/path.h"
#include "debug.h"
#include "platform/path_impl.h"

using namespace SEAScope;

/**
 *
 */
EmbeddedCatalogue::EmbeddedCatalogue()
{
};

/**
 *
 */
EmbeddedCatalogue::~EmbeddedCatalogue()
{
};

/**
 *
 */
bool EmbeddedCatalogue::init(const Settings& s)
{
    bool ok(false);

	if
		(nullptr == Log::implementation)
	{
        Log::implementation = new FilterLogger();
        ok = static_cast<FilterLogger*>(Log::implementation)->init(s.logFilter);
        if
            (false == ok)
        {
            delete Log::implementation;
            Log::implementation = nullptr;

            std::cerr << "Failed to initialize logger with '" << s.logFilter << "'" << std::endl;

            return false;
        }
	}

	if
		(nullptr == Path::implementation)
	{
		Path::implementation = new SEAScope::PathImpl();
	}

#if defined __linux__ || defined __APPLE__
	setenv("TZ", "UTC", 1);
#elif _WIN32
	_putenv_s("TZ", "UTC");
#endif

	this->explorer.detectOutdatedIndex = false; // this will be handled by the index plugin
	return this->explorer.init(s.dataPath, s.indexPath, true /*ignore index */);
};

/**
 *
 */
bool EmbeddedCatalogue::reset()
{
	if
		(nullptr != Path::implementation)
	{
		delete Path::implementation;
		Path::implementation = nullptr;
	}

	if
		(nullptr != Log::implementation)
	{
		delete Log::implementation;
		Log::implementation = nullptr;
	}

    return true;
}

/**
 *
 */
bool EmbeddedCatalogue::listCollections(std::set<std::uint32_t>& results) const
{
	return this->explorer.catalogue.listCollections(results);
};

/**
 *
 */
bool EmbeddedCatalogue::getCollectionCfg(std::uint32_t id, CollectionCfg& cfg, std::vector<RenderingCfg*>& renderings) const
{
	bool ok;	//
	ok = this->explorer.catalogue.getCollection(id, cfg, renderings);
	Log::log( LogLevel::Debug, "EmbeddedCatalogue"
			, "getCollectionCfg: {}"
			, id);
	Debug::displayCollection(cfg);
	return ok;
};

/**
 *
 */
bool EmbeddedCatalogue::canHandle(const std::string& uri) const
{
    return "embedded" == uri;
};

/**
 *
 */
EmbeddedCatalogue* EmbeddedCatalogue::create()
{
	return new EmbeddedCatalogue();
};

std::string EmbeddedCatalogue::getPath(const CollectionId & id) const
{
	auto it = explorer.catalogue.ids.begin();
	
	for (; it != explorer.catalogue.ids.end(); ++it)
	{
		if (id.second == it->second)
		{
			return it->first;
		}
	}
	
	return std::string();
}
