// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <iostream>

#include "local/embedded_index.h"
#include "filter_logger.h"
#include "core/log.h"
#include "platform/path_impl.h"
#include "core/path.h"

using namespace SEAScope;

/**
 *
 */
EmbeddedIndex::EmbeddedIndex()
{
};

/**
 *
 */
EmbeddedIndex::~EmbeddedIndex()
{
};

/**
 *
 */
bool EmbeddedIndex::init(const Settings& s)
{
    bool ok(false);

	if
		(nullptr == Log::implementation)
	{
        FilterLogger * logger = new FilterLogger();
        ok = logger->init(s.logFilter);
        if
            (false == ok)
        {
            delete logger;
            logger = nullptr;

            std::cerr << "Failed to initialize logger with '" << s.logFilter << "'" << std::endl;

            return false;
        }
		Log::implementation = logger;
        logger = nullptr;
	}

	if
		(nullptr == Path::implementation)
	{
		Path::implementation = new SEAScope::PathImpl;
	}

#if defined __linux__ || defined __APPLE__
	setenv("TZ", "UTC", 1);
#elif _WIN32
	_putenv_s("TZ", "UTC");
#endif

	this->explorer.detectOutdatedIndex = !s.skipIndexCheck;
	return this->explorer.init(s.dataPath, s.indexPath, false /*ignore index*/);
};

/**
 *
 */
bool EmbeddedIndex::reset()
{
	if
		(nullptr != Path::implementation)
	{
		delete Path::implementation;
		Path::implementation = nullptr;
	}

	if
		(nullptr != Log::implementation)
	{
		delete Log::implementation;
		Log::implementation = nullptr;
	}

    return true;
}

/**
 *
 */
bool EmbeddedIndex::searchGranules( const SearchCriteria& criteria
						   , std::vector<GranuleMetadata>& result) const
{
	bool ok;	//
	ok = this->explorer.index.findGranules(criteria, result);
	return ok;
};

/**
 *
 */
std::array<bool, 2> EmbeddedIndex::searchPrevNext( timestamp_t current
						    , const std::set<CollectionId>& collections
							, timestamp_t& prev
							, timestamp_t& next
							, timestamp_t minDelta) const
{
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata prevGranule;	//
	GranuleMetadata nextGranule;	//

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "searchPrevNext: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "Local collections list built");

	auto oks = this->explorer.index.adjacentGranules(1000 * current, 1000 * minDelta, localCollections, prevGranule, nextGranule);

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "Adjacent granules ok");
	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "{} {}"
			, oks[0], oks[1]);
	if
		(true == oks[0])
	{
		prev = static_cast<timestamp_t>(0.0005 * static_cast<double>(prevGranule.start)  + 0.0005 * static_cast<double>(prevGranule.stop));
	}
	if
		(true == oks[1])
	{
		next = static_cast<timestamp_t>(0.0005 * static_cast<double>(nextGranule.start) + 0.0005 * static_cast<double>(nextGranule.stop));
	}

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "Returning from searchPrevNext");

    return oks;
};

/**
 *
 */
bool EmbeddedIndex::searchClosestBound( timestamp_t current
						    , const std::set<CollectionId>& collections
							, timestamp_t& closest) const
{
	timestamp_t delta;	//
	timestamp_t closestTmp;	//
	std::set<std::uint32_t> localCollections;	//
	GranuleMetadata closestGranule;	//
	bool ok;	//

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "searchClosestBoundt: {}"
			, current);

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	Log::log( LogLevel::Debug, "EmbeddedIndex"
			, "Local collections list built");

	ok = this->explorer.index.closestGranuleBound(1000 * current, localCollections, closestGranule);

	if
		(true == ok)
	{
		Log::log(LogLevel::Debug, "EmbeddedIndex", "closest granule found: {} {} {}", closestGranule.uris[0].uri, closestGranule.start, closestGranule.stop);
		if
			(  closestGranule.start <= 1000 * current
			&& closestGranule.stop >= 1000 * current)
		{
			/* A granule contains the current datetime, no bounds required */
			closest = current;
			if
				(closestGranule.stop ==  1000 * current)
			{
				/* if granule end, remove one second or the granule will not be displayed */
				closest -= 1;
			}
			return true;
		}
		closestTmp = current;
		delta = std::numeric_limits<uint64_t>::max();
		if
			(  closestGranule.start < current
			&& delta > current - closestGranule.start)
		{
			delta = current - closestGranule.start;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.start));
		}
		else if
			(  closestGranule.start > current
			&& delta > closestGranule.start - current)
		{
			delta = closestGranule.start - current;
			closestTmp = static_cast<timestamp_t>(0.001 * static_cast<double>(closestGranule.start));
		}

		/* Remove one second when using the end of the granule as a result
		   or the granule will no be displayed */
		if
			(  closestGranule.stop < current
			&& delta > current - closestGranule.stop)
		{
			delta = current - closestGranule.stop;
			closestTmp = static_cast<timestamp_t>(-1.0 + 0.001 * static_cast<double>(closestGranule.stop));
		}
		else if
			(  closestGranule.stop > current
			&& delta > closestGranule.stop - current)
		{
			delta = closestGranule.stop - current;
			closestTmp = static_cast<timestamp_t>(-1.0 + 0.001 * static_cast<double>(closestGranule.stop));
		}

		closest = closestTmp;
	}

	Log::log(LogLevel::Debug, "EmbeddedIndex", "closest bound: {}", closest);
    return ok;
};

/**
 *
 */
bool EmbeddedIndex::getDataCoverage(const std::set<CollectionId>& collections, EventsCoverage& result) const
{
	std::set<std::uint32_t> localCollections;	//
	bool ok;	//

	/* Build list of local collections identifiers */
	for
		(const auto& itCol : collections)
	{
		localCollections.insert(itCol.second);
	}

	ok = this->explorer.index.getCoverage(localCollections, result);
	return ok;
};

/**
 *
 */
bool EmbeddedIndex::getDataCoverage(const CollectionId& collectionId, EventsCoverage& result) const
{
	bool ok;	//

	ok = this->explorer.index.getCoverage(collectionId.second, result);
	return ok;
};

/**
 *
 */
EmbeddedIndex* EmbeddedIndex::create()
{
    return new EmbeddedIndex();
};

/**
 *
 */
bool EmbeddedIndex::canHandle(const std::string& uri) const
{
	return "embedded" == uri;
};
