// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <cmath>
#include <cstddef>
#include <sstream>
#include <iomanip>
#include <algorithm>
#include "core/log.h"
#include "data/image.h"
#include "local/idf_reader.h"

#ifdef _WIN32
	#include "core/strptime.h"
#endif

using namespace SEAScope;

/**
 *
 */
std::time_t IDFReader::str2timestamp(const std::string& timeStr)
{
	std::size_t microsecondsPosition;	//
	std::int32_t microseconds;	//
	std::string withoutMicro;	//
	std::string microsecondsStr;	//
	std::tm t = {};	//
	std::time_t timestamp;	//
	std::time_t result;	//

	microseconds = 0;
	microsecondsPosition = timeStr.find_first_of(".");
	withoutMicro = timeStr.substr(0, microsecondsPosition);
	if
		(std::string::npos != microsecondsPosition)
	{
		microsecondsStr = timeStr.substr( 1 + microsecondsPosition
										, timeStr.length() - 1);
		microseconds = std::atoi(microsecondsStr.c_str());
	}
	strptime(withoutMicro.c_str(), "%Y-%m-%dT%H:%M:%S", &t);
	timestamp = std::mktime(&t);
	/* result in milliseconds */
	result = 1000 * timestamp + static_cast<time_t>(0.001 * static_cast<double>(microseconds));
	Log::log( LogLevel::Debug, "IDFReader"
			, "Time coverage extremum: {} {} ### 1000 * {} + .001 * {} = {}"
			, timeStr, withoutMicro, timestamp, microseconds, result);
    return result;
};


/**
 *
 */
bool IDFReader::extractGranuleInfo( const std::string& path
								  , GranuleMetadata& info) const
{
	netCDF::NcVar lat;	//
	netCDF::NcVar lon;	//
	netCDF::NcVar x;	//
	netCDF::NcVar y;	//
	netCDF::NcGroupAtt attr;	//
	std::size_t gcpCount;	//
	std::size_t i;	//
	std::size_t j;	//
	float lon0;	//
	float* lats;	//
	float* lons;	//
	std::int32_t* xs;	//
	std::int32_t* ys;	//
	float resolution;	//
	std::string resolutionUnits;	//
	std::string timeCoverageStartStr;	//
	std::string timeCoverageStopStr;	//
    IDFDescriptor descriptor;   //

	Log::log(LogLevel::Debug, "IDFReader", "extractGranuleInfo({})", path);
    netCDF::NcFile fileHandler(path, netCDF::NcFile::read);
	if
		(true == fileHandler.isNull())
	{
		Log::log(LogLevel::Error, "IDFReader", "Could not open file {}", path);
		return false;
	}
    descriptor.uri = path;

	/* -------------------- Extract mandatory attributes --------------------*/

	// idf_granule_id
    attr = fileHandler.getAtt("idf_granule_id");
	if
		(true == attr.isNull())
	{
		Log::log(LogLevel::Error, "IDFReader"
				, "Global attribute idf_granule_id missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
    attr.getValues(info.dataId);

    // time_coverage_start
	attr = fileHandler.getAtt("time_coverage_start");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_start missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
	attr.getValues(timeCoverageStartStr);
    info.start = static_cast<timestamp_t>(IDFReader::str2timestamp(timeCoverageStartStr));
	//Log::log(LogLevel::Debug, "IDFReader", "Time coverage start: {} {}", timeCoverageStartStr, info.start);

    // time_coverage_end
	attr = fileHandler.getAtt("time_coverage_end");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute time_coverage_end missing from {}"
				, path);
		fileHandler.close();
		return false;
	}
	attr.getValues(timeCoverageStopStr);
	info.stop = static_cast<timestamp_t>(IDFReader::str2timestamp(timeCoverageStopStr));
	//Log::log(LogLevel::Debug, "IDFReader", "Time coverage stop: {} {}", timeCoverageStopStr, info.stop);

	// resolution
    attr = fileHandler.getAtt("idf_spatial_resolution");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute idf_spatial_resolution missing from {}"
				, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
    attr.getValues(&resolution);

	// resolution units
	attr = fileHandler.getAtt("idf_spatial_resolution_units");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
			, "Global attribute idf_spatial_resolution_units missing from {}"
			, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
	attr.getValues(resolutionUnits);

	if
		("m" != resolutionUnits)
	{
		/* TODO: convert resolution in meters */
	}
	descriptor.resolution = resolution;

	// subsampling factor
	attr = fileHandler.getAtt("idf_subsampling_factor");
	if
		(true == attr.isNull())
	{
		Log::log( LogLevel::Error, "IDFReader"
				, "Global attribute idf_subsampling_factor missing from {}"
				, path);
		/* TODO: return false instead of warning */
		Log::log(LogLevel::Warn, "IDFReader", "Error ignored (for now)");
	}
	attr.getValues(&descriptor.subsamplingFactor);

	/* ------------------- Extract optional attributes ----------------------*/

	info.availableAttributes.reset(); /* attributes all flagged unavailable */

	// title
	attr = fileHandler.getAtt("title");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::TITLE), 1);
		attr.getValues(info.title);
	}

	// institution
	attr = fileHandler.getAtt("institution");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::INSTITUTION), 1);
		attr.getValues(info.institution);
	}

	// comment
	attr = fileHandler.getAtt("comment");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::COMMENT), 1);
		attr.getValues(info.comment);
	}

	// [product_]id
	attr = fileHandler.getAtt("id");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::FILE_ID), 1);
		attr.getValues(info.fileId);
	}

	// product_version
	attr = fileHandler.getAtt("product_version");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::PRODUCT_VERSION), 1);
		attr.getValues(info.productVersion);
	}

	// geospatial_lat_min
	attr = fileHandler.getAtt("geospatial_lat_min");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::LAT_MIN), 1);
		attr.getValues(&info.latMin);
	}

	// geospatial_lat_max
	attr = fileHandler.getAtt("geospatial_lat_max");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::LAT_MAX), 1);
		attr.getValues(&info.latMax);
	}

	// geospatial_lon_min
	attr = fileHandler.getAtt("geospatial_lon_min");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::LON_MIN), 1);
		attr.getValues(&info.lonMin);
	}

	// geospatial_lon_max
	attr = fileHandler.getAtt("geospatial_lon_max");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::LON_MAX), 1);
		attr.getValues(&info.lonMax);
	}

	// creator_email
	attr = fileHandler.getAtt("creator_email");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::CREATOR_EMAIL), 1);
		attr.getValues(info.creatorEmail);
	}

	// station_id
	attr = fileHandler.getAtt("station_id");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::STATION_ID), 1);
		attr.getValues(info.stationId);
	}

	// platform
	attr = fileHandler.getAtt("platform");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::PLATFORM), 1);
		attr.getValues(info.platform);
	}

	// sensor
	attr = fileHandler.getAtt("sensor");
	if
		(false == attr.isNull())
	{
		info.availableAttributes.set(
			static_cast<std::size_t>(IDFAttributes::SENSOR), 1);
		attr.getValues(info.sensor);
	}

	/* Find out which model is used */
	std::vector<std::string> modelRowCell;	//
	modelRowCell.push_back("time");
	modelRowCell.push_back("cell");
	modelRowCell.push_back("row");

	std::vector<std::string> modelLatLon;	//
	modelLatLon.push_back("time");
	modelLatLon.push_back("lat");
	modelLatLon.push_back("lon");

	std::vector<std::string> modelYX;	//
	modelYX.push_back("time");
	modelYX.push_back("y");
	modelYX.push_back("x");

	std::vector<std::string> modelNjNi;	//
	modelNjNi.push_back("time");
	modelNjNi.push_back("nj");
	modelNjNi.push_back("ni");

	std::vector<std::string> modelStation;	//
	modelStation.push_back("time");
	modelStation.push_back("station");

	std::vector<std::string> modelTraj;	//
	modelTraj.push_back("time");

	auto dims = fileHandler.getDims();
	for
		(const auto& it : dims)
	{
		auto foundLatLon = std::find(modelLatLon.begin(), modelLatLon.end(), it.first);
		if
			(modelLatLon.end() != foundLatLon)
		{
			modelLatLon.erase(foundLatLon);
		}

		auto foundRowCell = std::find(modelRowCell.begin(), modelRowCell.end(), it.first);
		if
			(modelRowCell.end() != foundRowCell)
		{
			modelRowCell.erase(foundRowCell);
		}

		auto foundYX = std::find(modelYX.begin(), modelYX.end(), it.first);
		if
			(modelYX.end() != foundYX)
		{
			modelYX.erase(foundYX);
		}

		auto foundNjNi = std::find(modelNjNi.begin(), modelNjNi.end(), it.first);
		if
			(modelNjNi.end() != foundNjNi)
		{
			modelNjNi.erase(foundNjNi);
		}

		auto foundStation = std::find(modelStation.begin(), modelStation.end(), it.first);
		if
			(modelStation.end() != foundStation)
		{
			modelStation.erase(foundStation);
		}

		auto foundTraj = std::find(modelTraj.begin(), modelTraj.end(), it.first);
		if
			(modelTraj.end() != foundTraj)
		{
			modelTraj.erase(foundTraj);
		}
		Log::log(LogLevel::Debug, "IDFReader", "Dimension {}: {}", it.first, it.second.getSize());
	}

	gcpCount = 0;
	if
		(0 >= modelLatLon.size())
	{
		info.dataModel = DataModel::LAT_LON;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from lat/lon model");
		descriptor.shape.yArity = static_cast<std::uint16_t>(fileHandler.getDim("lat_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("lon_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("lat").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("lon").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_lat_gcp");
		x = fileHandler.getVar("index_lon_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lons = new float[descriptor.shape.xArity];
		lats = new float[descriptor.shape.yArity];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
				descriptor.shape.gcps.emplace_back( lats[j]
												  , static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0))
												  , static_cast<uint16_t>(xs[i]), static_cast<uint16_t>(ys[j]));
				//Log::log(LogLevel::Error, "IDFReader", "longitude: {} -> {}", lons[i], std::fmod(lons[i] -lon0, 360.0) + lon0);
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelRowCell.size())
	{
		info.dataModel = DataModel::ROW_CELL;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from row/cell model");
		descriptor.shape.yArity = static_cast<std::uint16_t>(fileHandler.getDim("row_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("cell_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("row").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("cell").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_row_gcp");
		x = fileHandler.getVar("index_cell_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
				descriptor.shape.gcps.emplace_back( lats[j * descriptor.shape.xArity + i]
												  , static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0))
												  , static_cast<uint16_t>(xs[i]), static_cast<uint16_t>(ys[j]));
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelYX.size())
	{
		info.dataModel = DataModel::Y_X;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from y/x model");
		descriptor.shape.yArity = static_cast<std::uint16_t>(fileHandler.getDim("y_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("x_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("y").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("x").getSize());

		Log::log(LogLevel::Debug, "IDFReader", "Get variables");
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		y = fileHandler.getVar("index_y_gcp");
		x = fileHandler.getVar("index_x_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get values");
		gcpCount = static_cast<size_t>(descriptor.shape.xArity * descriptor.shape.yArity);
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];
		ys = new int32_t[descriptor.shape.yArity];

		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);
		y.getVar(ys);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

		j = 0;
		for
			(;j<descriptor.shape.yArity; ++j)
		{
			i = 0;
			for
				(; i < descriptor.shape.xArity; ++i)
			{
				descriptor.shape.gcps.emplace_back( lats[j * descriptor.shape.xArity + i]
												  , static_cast<float>(std::fmod(lons[j * descriptor.shape.xArity + i] - lon0, 360.0) + static_cast<double>(lon0))
												  , static_cast<uint16_t>(xs[i]), static_cast<uint16_t>(ys[j]));
			}
		}

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
		delete [] ys;
		ys = nullptr;
	}
	else if
		(0 >= modelNjNi.size())
	{
		info.dataModel = DataModel::NJ_NI;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from nj/ni model");
		descriptor.shape.yArity = static_cast<std::uint16_t>(fileHandler.getDim("nj_gcp").getSize());
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("ni_gcp").getSize());
		descriptor.yArity = static_cast<std::uint32_t>(fileHandler.getDim("nj").getSize());
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("ni").getSize());
		y = fileHandler.getVar("index_nj_gcp");
		x = fileHandler.getVar("index_ni_gcp");
	}
	else if
		(0 >= modelStation.size())
	{
		info.dataModel = DataModel::TIME_STATION;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from station model");
		descriptor.shape.yArity = 1;
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("station_gcp").getSize());
		descriptor.yArity = 1;
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
		x = fileHandler.getVar("index_station_gcp");
	}
	else if
		(0 >= modelTraj.size())
	{
		info.dataModel = DataModel::TIME;
		Log::log(LogLevel::Debug, "IDFReader", "Get shape from trajectory model");
		descriptor.shape.yArity = 1;
		descriptor.shape.xArity = static_cast<std::uint16_t>(fileHandler.getDim("time_gcp").getSize());
		descriptor.yArity = 1;
		descriptor.xArity = static_cast<std::uint32_t>(fileHandler.getDim("time").getSize());
		lat = fileHandler.getVar("lat_gcp");
		lon = fileHandler.getVar("lon_gcp");
		x = fileHandler.getVar("index_time_gcp");

		Log::log(LogLevel::Debug, "IDFReader", "Get {} values", descriptor.shape.xArity);
		gcpCount = descriptor.shape.xArity;
		lats = new float[gcpCount];
		lons = new float[gcpCount];
		xs = new int32_t[descriptor.shape.xArity];

		lat.getVar(lats);
		lon.getVar(lons);
		x.getVar(xs);

		//lon0 = std::fmod(lons[0] + 180.0, 360.0) - 180.0;
		lon0 = -180.0;

        i = 0;
        for
            (; i < descriptor.shape.xArity; ++i)
        {
            descriptor.shape.gcps.emplace_back( lats[i]
                                              , static_cast<float>(std::fmod(lons[i] - lon0, 360.0) + static_cast<double>(lon0))
                                              , static_cast<uint16_t>(xs[i]), static_cast<uint16_t>(0));
        }

		delete [] lons;
		lons = nullptr;
		delete [] lats;
		lats = nullptr;
		delete [] xs;
		xs = nullptr;
	}

	Log::log(LogLevel::Debug, "IDFReader", "Rebuild GCPs list");

    info.uris.push_back(descriptor);

	Log::log( LogLevel::Debug, "IDFReader"
			, "GranuleInfo extracted!");

	lons = nullptr;
	lats = nullptr;
	xs = nullptr;
	ys = nullptr;

	fileHandler.close();

	return true;
};

/**
 *
 */
bool IDFReader::extractGranuleData(const std::string& path, const std::set<std::string>& fields, DataModel dataModel, GranuleData& data) const
{
	netCDF::NcVar var;	//
	netCDF::NcVarAtt offsetAtt;	//
	netCDF::NcVarAtt scaleAtt;	//
	netCDF::NcVarAtt fillValueAtt;	//
	float offset;	//
	float scaleFactor;	//
	float fillValue;	//
	std::uint8_t* result;	//

	bool ok;	//

	result = nullptr;

    netCDF::NcFile fileHandler(path.c_str(), netCDF::NcFile::read);
    if
        (true == fileHandler.isNull())
    {
        Log::log( LogLevel::Error, "IDFReader", "Could not open netCDF file {}"
                , path);
        return false;
    }

	for
		(const auto& field : fields)
	{
		Log::log(LogLevel::Debug, "IDFReader", "Extracting field: {}", field);
		var = fileHandler.getVar(field);

        data[field];
        DataInfo& info = data[field].info;

		if
			(DataModel::LAT_LON == dataModel)
		{
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("lon").getSize());
			info.yArity = static_cast<std::uint16_t>(fileHandler.getDim("lat").getSize());
		}
		else if
			(DataModel::Y_X == dataModel)
		{
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("x").getSize());
			info.yArity = static_cast<std::uint16_t>(fileHandler.getDim("y").getSize());
		}
		else if
			(DataModel::NJ_NI == dataModel)
		{
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("ni").getSize());
			info.yArity = static_cast<std::uint16_t>(fileHandler.getDim("nj").getSize());
		}
		else if
			(DataModel::ROW_CELL == dataModel)
		{
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("cell").getSize());
			info.yArity = static_cast<std::uint16_t>(fileHandler.getDim("row").getSize());
		}
		else if
			(DataModel::TIME_STATION == dataModel)
		{
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("time").getSize());
			info.yArity = static_cast<std::uint16_t>(fileHandler.getDim("station").getSize());
		}
		else if
			(DataModel::TIME == dataModel)
		{
            Log::log(LogLevel::Debug, "IDFReader", "extract data from trajectory");
			info.xArity = static_cast<std::uint16_t>(fileHandler.getDim("time").getSize());
			info.yArity = 1;
		}

		if
			(  "time" == field
			&& DataModel::TIME == dataModel)
		{
			// Time is stored as double precision but SEAScope only handles
			// floats.
			double* timeBuffer = new double[info.xArity];
			double minTime;
			double maxTime;
			double timeExtent;
			int n;
			Log::log(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);

			var.getVar(timeBuffer);

			minTime = std::numeric_limits<double>::max();
			maxTime = std::numeric_limits<double>::min();
			n = info.xArity;
			while
				(--n > -1)
			{
				if
					(timeBuffer[n] > maxTime)
				{
					maxTime = timeBuffer[n];
				}
				if
					(timeBuffer[n] < minTime)
				{
					minTime = timeBuffer[n];
				}
			}
			Log::log(LogLevel::Debug, "IDFReader", "Time range: [{}, {}]", minTime, maxTime);

			timeExtent = maxTime - minTime;
			Log::log(LogLevel::Debug, "IDFReader", "Time extent: {}", timeExtent / 86400);
			if
				(timeExtent > double(std::numeric_limits<float>::max()))
			{
				Log::log(LogLevel::Error, "IDFReader", "Overflow: the time range is too wide to be stored as a float value");
				fileHandler.close();
				return false;
			}
			offset = float(minTime);
			scaleFactor = 1.f;

			info.channels = 1;
			info.dataType = OGL::float32;
			info.offsets.push_back(offset);
			info.scaleFactors.push_back(scaleFactor);
			info.fillValues.push_back(-1.f);

			data[field].buffer.allocate(info.xArity, info.yArity, info.channels,
										info.dataType, nullptr);
			float* floatResult = static_cast<float*>(data[field].buffer.data());
			n = info.xArity;
			while
				(--n > -1)
			{
				float tmpValue = static_cast<float>(timeBuffer[n] - static_cast<double>(offset)) * scaleFactor;
				Log::log(LogLevel::Debug, "IDFReader", "n = {}, tmpValue = {}" , n, tmpValue);
				floatResult[n] = tmpValue;
			}

			n = info.xArity;
			for
				(int i = 0; i < n; ++i)
			{
				Log::log(LogLevel::Debug, "IDFReader", "result: {} + ({} * {})", offset, floatResult[i], scaleFactor);
			}

			delete[] timeBuffer;
		}
		else
		{
			info.channels = 1;
			info.dataType = OGL::unsigned_byte;

			/* Get offset */
			offsetAtt = this->readAttribute(var, "add_offset", ok);
			offset = 0.f;
			if
				(true == ok)
			{
				offsetAtt.getValues(&offset);
			}
			info.offsets.push_back(offset);

			/* Get scale */
			scaleAtt = this->readAttribute(var, "scale_factor", ok);
			scaleFactor = 1.f;
			if
				(true == ok)
			{
				scaleAtt.getValues(&scaleFactor);
			}
			info.scaleFactors.push_back(scaleFactor);

			/* Get fill value */
			fillValueAtt = this->readAttribute(var, "_FillValue", ok);
			if
				(true == ok)
			{
				fillValueAtt.getValues(&fillValue);
			}
			else
			{
				Log::log(LogLevel::Error, "IDFReader", "Cannot find fill value for {}", field);
				fileHandler.close();
				return false;
			}
			info.fillValues.push_back(fillValue);

			data[field].buffer.allocate( info.xArity, info.yArity, info.channels
									   , info.dataType, nullptr);
			result = static_cast<std::uint8_t*>(data[field].buffer.data());
			Log::log(LogLevel::Debug, "IDFReader", "Extracting values {}: {}x{}x{}", field, info.xArity, info.yArity, info.channels);
			var.getVar(result);
		}
	}

	fileHandler.close();
	result = nullptr;
	return true;
};

/**
 *
 */
netCDF::NcVarAtt IDFReader::readAttribute(netCDF::NcVar& var, const std::string& attrName, bool& found) const
{
	netCDF::NcVarAtt attribute;	//
	auto attrMap = var.getAtts();
	auto it = attrMap.find(attrName);

	if
		(it != attrMap.end())
	{
		found = true;
		return it->second;
	}
	found = false;
	return attribute;
};
