// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstdint>
#include <vector>
#include <iostream>
#include <fstream>
#include "core/log.h"
#include "core/api_in.h"
#include "core/api_out.h"
#include "api/api_generated.h"
#include "local/index_serializer.h"

using namespace SEAScope;

/**
 *
 */
IndexSerializer::IndexSerializer()
{
};

/**
 *
 */
IndexSerializer::~IndexSerializer()
{
};

/**
 *
 */
bool IndexSerializer::load(GranulesIndex& target, const std::string& path)
{
    flatbuffers::FlatBufferBuilder builder; //
    std::uint64_t localId;  //
    std::streamsize headerSize; //
    std::streampos totalSize;   //
    std::streamsize bufferSize;   //
    char* memblock; //

    std::ifstream fileHandler(path, std::ios::in|std::ios::binary|std::ios::ate);
    if
        (false == fileHandler.is_open())
    {
        Log::log( LogLevel::Error, "IndexSerializer"
                , "Could not open serialized index {}"
                , path);
        return false;
    }

    totalSize = fileHandler.tellg();
    headerSize = sizeof(std::uint64_t);

    fileHandler.seekg(0, std::ios::beg);
    fileHandler.read (reinterpret_cast<char*>(&bufferSize), headerSize);

    Log::log(LogLevel::Debug, "IndexSerializer", "File size: {}", totalSize);
    Log::log(LogLevel::Debug, "IndexSerializer", "Header size: {}", headerSize);
    Log::log(LogLevel::Debug, "IndexSerializer", "Buffer size: {}", bufferSize);

    if
        (totalSize < headerSize + bufferSize)
    {
        Log::log( LogLevel::Error, "IndexSerializer"
                , "Serialized index file {} does not contain a complete buffer"
                , path);
        fileHandler.close();
        return false;
    }
    memblock = new char [static_cast<size_t>(bufferSize)];

    fileHandler.read(memblock, bufferSize);
    fileHandler.close();

    auto cmd = flatbuffers::GetRoot<SEAScope::API::SearchGranulesResponse>(memblock);
    auto granules = cmd->granules();
    auto granulesCount = granules->Length();

    Log::log(LogLevel::Debug, "IndexSerializer", "Loaded {} granules from the serialized index", granulesCount);
    for
        (flatbuffers::uoffset_t i = 0; i < granulesCount; ++i)
    {
        auto granule = granules->Get(i);
        localId = granule->granuleId();
        target.data[localId];
        APIIn::granuleMetadata(granule, target.data[localId]);
        target.ids[target.data[localId].dataId] = localId;
    }
    target.idGenerator = 1 + granulesCount;

    delete[] memblock;
    memblock = nullptr;

	return true;
}

/**
 *
 */
bool IndexSerializer::save(const GranulesIndex& target, const std::string& path)
{
    std::uint8_t* resultBuffer; //
    std::streamsize resultSize;   //
    flatbuffers::FlatBufferBuilder builder; //
    std::vector<flatbuffers::Offset<SEAScope::API::GranuleMetadata>> granulesVector;    //
	bool ok;	//

	ok = false;

    for
        (const auto& it : target.data)
    {
        const GranuleMetadata& itGranule = it.second;
		auto granule = APIOut::granuleMetadata(builder, itGranule, ok);
		if
			(false == ok)
		{
			Log::log(LogLevel::Error, "IndexSerializer", "Failed to serialize granule metadata");
			return false;
		}
		granulesVector.push_back(granule);
    }

    auto granules = builder.CreateVector(granulesVector);
    auto result = SEAScope::API::CreateSearchGranulesResponse(builder, granules);
    builder.Finish(result);

    Log::log(LogLevel::Debug, "IndexSerializer", "Result built");

    resultBuffer = builder.GetBufferPointer();
    resultSize = builder.GetSize();

    std::ofstream fileHandler(path, std::ios::out|std::ios::binary|std::ios::trunc);
    if
        (false == fileHandler.is_open())
    {
        Log::log( LogLevel::Error, "IndexSerializer"
                , "Could not open serialized index {}"
                , path);
        return false;
    }
    fileHandler.write(reinterpret_cast<const char*>(&resultSize), sizeof(std::uint64_t));
    fileHandler.write(reinterpret_cast<const char*>(resultBuffer), resultSize);
    fileHandler.close();

    Log::log(LogLevel::Debug, "IndexSerializer", "Saved index: {} bytes", resultSize);
    builder.ReleaseBufferPointer();

	return true;
};
