// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2022 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <cstdio>
#include <vector>

#include "simplelogger.h"
#include "core/log.h"
#include "platform/path_impl.h"
#include "core/path.h"
#include "core/structs.h"
#include "core/rendering_cfg.h"
#include "local/collection_ini_parser.h"
#include "local/collections_explorer.h"

using namespace SEAScope;

void displayCollection(const CollectionCfg& colCfg)
{
	Log::log( LogLevel::Debug, "Main"
			, "Collection: {} x:{} y:{} c:{} #v:{}"
			, colCfg.label, colCfg.xSeamless, colCfg.ySeamless, colCfg.mustBeCurrent, colCfg.variables.size());

	/* Display collection tags */
	for
		(const auto& tag : colCfg.tags)
	{
		Log::log( LogLevel::Debug, "Main"
				, " #{} {}"
				, tag.first, tag.second);
	}

	/* Display variables info */
	for
		(const auto& it : colCfg.variables)
	{
		Log::log( LogLevel::Debug, "Main"
				, " * Variable[{}]: {} [{}]"
				, it.first, it.second.label, (int)it.second.defaultRendering);
		for
			(const auto& field : it.second.fields)
		{
			Log::log( LogLevel::Debug, "Main"
					, "   - {}"
					, field);
		}

		for
			(const auto& tag : it.second.tags)
		{
			Log::log( LogLevel::Debug, "Main"
					, "   #{} {}"
					, tag.first, tag.second);
		}
	}

};

void displayGranules(const GranulesIndex& index)
{
	for
		(const auto& it : index.data)
	{
		Log::log( LogLevel::Debug, "Main"
				, "{} {} {}"
				, it.second.uris[0].uri, it.second.start, it.second.stop);
	}
};

int main(int argc, char** argv)
{
	CollectionsExplorer explorer;	//
	CollectionCfg colCfg;	//
	std::vector<RenderingCfg*> renderings;	//	
	CollectionINIParser parser;	//
	std::string dataPath;	//
	std::string indexPath;	//
	bool ok;	//

	setenv("TZ", "UTC", 1);

	Log::implementation = new SimpleLogger();

	if
		(3 > argc)
	{
		Log::log( LogLevel::Error, "Main"
				, "You must provide a root directory for collections data and the path of the index file");
		return 1;
	}

	Path::implementation = new PathImpl();

	dataPath = argv[1];
	indexPath = argv[2];
	ok = explorer.init(dataPath, indexPath, false);

	std::set<std::uint32_t> collections;
	explorer.catalogue.listCollections(collections);
	for
		(const auto& it: collections)
	{
		Log::log(LogLevel::Debug, "Main", "Collection {}", it);
		explorer.catalogue.getCollection(it, colCfg, renderings);
		displayCollection(colCfg);
	}

	Log::log( LogLevel::Debug, "Main"
			, "Granules index contains {} values"
			, explorer.index.data.size());

	displayGranules(explorer.index);

	return 0;
}
