#!/bin/bash

# =============================================================================
# This file is part of SEAScope, a 3D visualisation and analysis application
# for satellite, in-situ and numerical model data.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# SEAScope is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
# =============================================================================
# @author: <sylvain.herledan@oceandatlab.com>
# @date: 2016-10-14
#
# Build functions to generate the static libraries required to build
# SEAScope plugins for local index, catalogue and datastream.
# =============================================================================

set -o nounset
set -o errexit

readonly CPUs=$(if [ -f /proc/cpuinfo ]; then grep -c ^processor /proc/cpuinfo; else sysctl -n hw.ncpu; fi);

function build_zlib()
{
	local root_dir="${1}";
	local src_dir="${root_dir}/sources";
	local build_dir="${root_dir}/build";
	local rootfs="${root_dir}/rootfs";
	local uri="http://zlib.net/current/zlib-1.2.12.tar.gz";

	# Download source if needed
	if [ ! -f "${src_dir}/zlib-1.2.12.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/zlib-1.2.12.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/zlib-1.2.12" ];
	then
		tar xzf "${src_dir}/zlib-1.2.12.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libz.a" ];
	then
		cd "${build_dir}/zlib-1.2.12"
		./configure --prefix="${rootfs}" \
					--static \
					--64
		make -j ${CPUs}
		make install
	fi
	
	# Back to root	
	cd "${root_dir}"
}

function build_hdf5()
{
	local root_dir="${1}";
	local src_dir="${root_dir}/sources";
	local build_dir="${root_dir}/build";
	local rootfs="${root_dir}/rootfs";
	local uri="https://support.hdfgroup.org/ftp/HDF5/releases/hdf5-1.10";
	uri="${uri}/hdf5-1.10.1/src/hdf5-1.10.1.tar.bz2"

	# Download source
	if [ ! -f "${src_dir}/hdf5-1.10.1.tar.bz2" ];
	then
		wget "${uri}" -O "${src_dir}/hdf5-1.10.1.tar.bz2"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/hdf5-1.10.1" ];
	then
		tar xjf "${src_dir}/hdf5-1.10.1.tar.bz2" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libhdf5.a" ];
	then
		cd "${build_dir}/hdf5-1.10.1"

		./configure \
				--prefix="${rootfs}" \
				--enable-hl \
				--enable-cxx \
				--disable-fortran \
				--disable-java \
				--enable-unsupported \
				--enable-threadsafe \
				--enable-build-mode=production \
				--enable-static \
				--disable-shared \
				--disable-sharedlib-rpath \
				--with-zlib="${rootfs}" \
				--with-pthread="/usr" \
				--with-pic
			make -j ${CPUs}
			make install
	fi

	# Back to root
	cd "${root_dir}"
}

function build_netcdf_c()
{
	local root_dir="${1}";
	local src_dir="${root_dir}/sources";
	local build_dir="${root_dir}/build";
	local rootfs="${root_dir}/rootfs";
	local uri="https://github.com/Unidata/netcdf-c/archive/v4.4.1.1.tar.gz";
	local static_libs="";
	local dynamic_libs="-lpthread -ldl"; 

	# Download source
	if [ ! -f "${src_dir}/netcdf-c-4.4.1.1.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/netcdf-c-4.4.1.1.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/netcdf-c-4.4.1.1" ];
	then
		tar xzf "${src_dir}/netcdf-c-4.4.1.1.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libnetcdf.a" ];
	then
		cd "${build_dir}/netcdf-c-4.4.1.1"
		./configure \
			CPPFLAGS="-I${rootfs}/include ${CPPFLAGS}" \
			LDFLAGS="-L${rootfs}/lib ${LDFLAGS}" \
			LIBS="${static_libs} ${dynamic_libs} " \
			--prefix="${rootfs}" \
			--disable-doxygen \
			--disable-examples \
			--disable-utilities \
			--enable-static \
			--disable-shared \
			--enable-fsync \
			--enable-mmap \
			--disable-dap \
			--with-pic
		make -j ${CPUs}
		make install
	fi

	# Back to root
	cd "${root_dir}"
}

function build_netcdf_cxx()
{
	local root_dir="${1}";
	local src_dir="${root_dir}/sources";
	local build_dir="${root_dir}/build";
	local rootfs="${root_dir}/rootfs";
	local uri="https://github.com/Unidata/netcdf-cxx4/archive/v4.3.0.tar.gz";
	local static_libs="-lnetcdf -lhdf5_hl -lhdf5 -lz";
	local dynamic_libs="-lpthread -lm -ldl";

	# Download source
	if [ ! -f "${src_dir}/netcdf-cxx-4.3.0.tar.gz" ];
	then
		wget "${uri}" -O "${src_dir}/netcdf-cxx-4.3.0.tar.gz"
	fi

	# Extract archive
	if [ ! -d "${build_dir}/netcdf-cxx4-4.3.0" ];
	then
		tar xzf "${src_dir}/netcdf-cxx-4.3.0.tar.gz" -C "${build_dir}"
	fi

	# Build
	if [ ! -f "${rootfs}/lib/libnetcdf_c++4.a" ];
	then
		cd "${build_dir}/netcdf-cxx4-4.3.0"
		./configure \
			LIBS="${static_libs} ${dynamic_libs} " \
			CPPFLAGS="-I${rootfs}/include ${CPPFLAGS}" \
			LDFLAGS="-L${rootfs}/lib ${LDFLAGS}" \
			--prefix="${rootfs}" \
			--enable-static \
			--disable-shared \
			--disable-doxygen \
			--with-pic
		make -j ${CPUs}
		make install
	fi

	# Back to root
	cd "${root_dir}"
}
